Package: BAMBI
Type: Package
Title: Bivariate Angular Mixture Models
Version: 2.3.5
Date: 2023-03-08
Author: Saptarshi Chakraborty,
        Samuel W.K. Wong
Maintainer: Saptarshi Chakraborty <chakra.saptarshi@gmail.com>
Description: Fit (using Bayesian methods) and simulate mixtures of univariate and bivariate angular distributions. Chakraborty and Wong (2021) <doi:10.18637/jss.v099.i11>.
License: GPL-3
LazyData: TRUE
RoxygenNote: 7.2.3
LinkingTo: Rcpp, RcppArmadillo
Imports: stats, stats4, graphics, lattice, grDevices, Rcpp, tcltk,
        qrng, mvtnorm, gtools, parallel, label.switching, methods,
        coda, future.apply, loo (>= 2.4.1), RColorBrewer,
        bridgesampling, scales, numDeriv
Suggests: future, gridExtra
Depends: R (>= 3.2.0)
URL: https://doi.org/10.18637/jss.v099.i11
BugReports: https://github.com/c7rishi/BAMBI/issues
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2023-03-08 22:28:25 UTC; chakrab2
Repository: CRAN
Date/Publication: 2023-03-08 23:10:05 UTC
Built: R 4.2.0; aarch64-apple-darwin20; 2023-07-11 00:42:26 UTC; unix
Archs: BAMBI.so.dSYM

Package: BAMBI
Type: Package
Title: Bivariate Angular Mixture Models
Version: 2.3.7
Date: 2026-05-05
Authors@R: c(person(given = "Saptarshi",
                      family = "Chakraborty",
                      role = c("aut", "cre"),
                      email = "chakra.saptarshi@gmail.com"),
               person(given = c("Samuel", "W.K."),
                      family = "Wong",
                      role = "aut"),
               person(given = "Florence",
                      family = "Bockting",
                      role = "ctb", 
                      comment = "Patch for compatibility with {loo}"))
Maintainer: Saptarshi Chakraborty <chakra.saptarshi@gmail.com>
Description: Fit (using Bayesian methods) and simulate mixtures of univariate and bivariate angular distributions. Chakraborty and Wong (2021) <doi:10.18637/jss.v099.i11>.
License: GPL-3
LazyData: TRUE
RoxygenNote: 7.3.3
LinkingTo: Rcpp, RcppArmadillo
Imports: stats, stats4, graphics, lattice, grDevices, Rcpp, tcltk,
        qrng, mvtnorm, gtools, parallel, label.switching, methods,
        coda, future.apply, loo (>= 2.4.1), RColorBrewer,
        bridgesampling, scales, numDeriv
Suggests: future, gridExtra
Depends: R (>= 3.2.0)
URL: https://doi.org/10.18637/jss.v099.i11
BugReports: https://github.com/c7rishi/BAMBI/issues
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-05-08 20:25:41 UTC; chakrab2
Author: Saptarshi Chakraborty [aut, cre],
  Samuel W.K. Wong [aut],
  Florence Bockting [ctb] (Patch for compatibility with {loo})
Repository: CRAN
Date/Publication: 2026-05-09 06:02:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-09 06:27:57 UTC; unix
Archs: BAMBI.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.mcmc.list,angmcmc)
S3method(bridge_sampler,angmcmc)
S3method(contour,angmcmc)
S3method(densityplot,angmcmc)
S3method(logLik,angmcmc)
S3method(loo,angmcmc)
S3method(plot,angmcmc)
S3method(print,angmcmc)
S3method(print,stepfit)
S3method(print,summary_angmcmc)
S3method(quantile,angmcmc)
S3method(summary,angmcmc)
S3method(waic,angmcmc)
export(DIC)
export(add_burnin_thin)
export(angmcmc)
export(bestcriterion)
export(bestmodel)
export(circ_cor)
export(circ_varcor_model)
export(contour_model)
export(d_fitted)
export(dvm)
export(dvmcos)
export(dvmcosmix)
export(dvmmix)
export(dvmsin)
export(dvmsinmix)
export(dwnorm)
export(dwnorm2)
export(dwnorm2mix)
export(dwnormmix)
export(extractsamples)
export(fit_angmix)
export(fit_incremental_angmix)
export(fit_vmcosmix)
export(fit_vmmix)
export(fit_vmsinmix)
export(fit_wnorm2mix)
export(fit_wnormmix)
export(fix_label)
export(is.angmcmc)
export(latent_allocation)
export(lpdtrace)
export(minuspi_to_pi)
export(paramtrace)
export(pointest)
export(r_fitted)
export(rvm)
export(rvmcos)
export(rvmcosmix)
export(rvmmix)
export(rvmsin)
export(rvmsinmix)
export(rwnorm)
export(rwnorm2)
export(rwnorm2mix)
export(rwnormmix)
export(select_chains)
export(surface_model)
export(vm2_mle)
export(zero_to_2pi)
import(graphics)
import(stats)
importFrom(RColorBrewer,brewer.pal)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(bridgesampling,bridge_sampler)
importFrom(coda,as.mcmc.list)
importFrom(future.apply,future_lapply)
importFrom(grDevices,colorRampPalette)
importFrom(gtools,rdirichlet)
importFrom(lattice,densityplot)
importFrom(lattice,wireframe)
importFrom(loo,loo)
importFrom(loo,waic)
importFrom(methods,is)
importFrom(mvtnorm,rmvnorm)
importFrom(parallel,detectCores)
importFrom(qrng,sobol)
importFrom(stats4,mle)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
useDynLib(BAMBI, .registration = TRUE)

# Bambi

[![PyPi version](https://badge.fury.io/py/bambi.svg)](https://badge.fury.io/py/bambi)
[![Build Status](https://github.com/bambinos/bambi/actions/workflows/test.yml/badge.svg)](https://github.com/bambinos/bambi/actions/workflows/test.yml)
[![codecov](https://codecov.io/gh/bambinos/bambi/branch/master/graph/badge.svg?token=ZqH0KCLKAE)](https://codecov.io/gh/bambinos/bambi)
[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)

BAyesian Model-Building Interface in Python

## Overview

Bambi is a high-level Bayesian model-building interface written in Python. It's built on top of the [PyMC3](https://github.com/pymc-devs/pymc3) probabilistic programming framework, and is designed to make it extremely easy to fit mixed-effects models common in social sciences settings using a Bayesian approach.

## Installation

Bambi requires a working Python interpreter (3.7+). We recommend installing Python and key numerical libraries using the [Anaconda Distribution](https://www.anaconda.com/products/individual#Downloads), which has one-click installers available on all major platforms.

Assuming a standard Python environment is installed on your machine (including pip), Bambi itself can be installed in one line using pip:

    pip install bambi

Alternatively, if you want the bleeding edge version of the package you can install from GitHub:

    pip install git+https://github.com/bambinos/bambi.git

### Dependencies

Bambi requires working versions of ArviZ, formulae, NumPy, pandas, PyMC3 and statsmodels. Dependencies are listed in `requirements.txt`, and should all be installed by the Bambi installer; no further action should be required.

## Example

In the following two examples we assume the following basic setup

```python
import bambi as bmb
import numpy as np
import pandas as pd

data = pd.DataFrame({
    "y": np.random.normal(size=50),
    "g": np.random.choice(["Yes", "No"], size=50),
    "x1": np.random.normal(size=50),
    "x2": np.random.normal(size=50)
})
```

### Linear regression

```python
model = bmb.Model("y ~ x1 + x2", data)
fitted = model.fit()
```

In the first line we create and build a Bambi `Model`. The second line tells the sampler to start
running and it returns an `InferenceData` object, which can be passed to several ArviZ functions
such as `az.summary()` to get a summary of the parameters distribution and sample diagnostics or
 `az.plot_traces()` to visualize them.


### Logistic regression

Here we just add the `family` argument set to `"bernoulli"` to tell Bambi we are modelling a binary
response. By default, it uses a logit link. We can also use some syntax sugar to specify which event
we want to model. We just say `g['Yes']` and Bambi will understand we want to model the probability
of a `"Yes"` response. But this notation is not mandatory. If we use `"g ~ x1 + x2"`, Bambi will
pick one of the events to model and will inform us which one it picked.


```python
model = bmb.Model("g['Yes'] ~ x1 + x2", data, family="bernoulli")
fitted = model.fit()
```

## Documentation

The Bambi documentation can be found in the [official docs](https://bambinos.github.io/bambi/index.html)

## Citation

If you use Bambi and want to cite it please use [![arXiv](https://img.shields.io/badge/arXiv-2012.10754-b31b1b.svg)](https://arxiv.org/abs/2012.10754)

Here is the citation in BibTeX format

```
@misc{capretto2020,
      title={Bambi: A simple interface for fitting Bayesian linear models in Python},
      author={Tomás Capretto and Camen Piho and Ravin Kumar and Jacob Westfall and Tal Yarkoni and Osvaldo A. Martin},
      year={2020},
      eprint={2012.10754},
      archivePrefix={arXiv},
      primaryClass={stat.CO}
}
```

## Contributions

Bambi is a community project and welcomes contributions. Additional information can be found in the [Contributing](https://github.com/bambinos/bambi/blob/master/CONTRIBUTING.md) Readme.

For a list of contributors see the [GitHub contributor](https://github.com/bambinos/bambi/graphs/contributors) page

## Code of Conduct

Bambi wishes to maintain a positive community. Additional details can be found in the [Code of Conduct](https://github.com/bambinos/bambi/blob/master/CODE_OF_CONDUCT.md)

## License

[MIT License](https://github.com/bambinos/bambi/blob/master/LICENSE)

#' The bivariate von Mises sine model
#'
#' @param n number of observations. Ignored if at least one of the other parameters have length k > 1, in which
#' case, all the parameters are recycled to length k to produce k random variates.
#' @param x  bivariate vector  or a two-column matrix with each row being a bivariate vector of angles
#' (in radians)  where the densities are to be evaluated.
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0}.
#' @param log logical. Should the log density be returned instead?
#' @param method Rejection sampling method to be used. Available choices are \code{"naive"} (default) or \code{"vmprop"}. See details.
#' @details
#' The bivariate von Mises sine model density at the point \eqn{x = (x_1, x_2)} is given by
#' \deqn{f(x) = C_s (\kappa_1, \kappa_2, \kappa_3) \exp(\kappa_1 \cos(T_1) + \kappa_2 \cos(T_2) + \kappa_3 \sin(T_1) \sin(T_2))}
#' where
#' \deqn{T_1 = x_1 - \mu_1;  T_2 = x_2 - \mu_2}
#' and \eqn{C_s (\kappa_1, \kappa_2, \kappa_3)} denotes the normalizing constant for the sine model.
#'
#' Two different rejection sampling methods are implemented for random generation. If \code{method = "vmprop"}, then first the y-marginal
#' is drawn from the associated marginal density, and then x is generated from the conditional distributio of x given y. The marginal generation of
#' y is implemented in a rejection sampling scheme with proposal being either von Mises (if the target marginal density is unimodal), or a mixture of
#' von Mises (if bimodal), with optimally chosen concentration. This the method suggested in Mardia et al. (2007). On the other hand, when
#' \code{method = "naive"} (default) a (naive) bivariate rejection sampling scheme with (bivariate) uniform propsoal is used.
#'
#' Note that although method = \code{"vmprop"} may provide better efficiency when the density is highly concentrated, it does have
#' an (often substantial) overhead due to the optimziation step required to find a reasonable proposal concentration parameter.
#' This can compensate the efficiency gains of this method, especially when \code{n} is not large.
#'
#'
#'
#'
#' @return \code{dvmsin} gives the density  and \code{rvmsin} generates random deviates.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # when x is a bivariate vector and parameters are all scalars,
#' # dvmsin returns single density
#' dvmsin(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when x is a two column matrix and parameters are all scalars,
#' # dmvsin returns a vector of densities calculated at the rows of
#' # x with the same parameters
#' dvmsin(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # if x is a bivariate vector and at least one of the parameters is
#' # a vector, all parameters are recycled to the same length, and
#' # dvmsin returns a vector of with ith element being the density
#' # evaluated at x with parameter values kappa1[i], kappa2[i],
#' # kappa3[i], mu1[i] and mu2[i]
#' dvmsin(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # if x is a two column matrix and at least one of the parameters is
#' # a vector, rows of x and the parameters are recycled to the same
#' # length, and dvmsin returns a vector of with ith element being the
#' # density evaluated at ith row of x with parameter values kappa1[i],
#' # kappa2[i], # kappa3[i], mu1[i] and mu2[i]
#' dvmsin(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # when parameters are all scalars, number of observations generated
#' # by rvmsin is n
#' rvmsin(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when at least one of the parameters is a vector, all parameters are
#' # recycled to the same length, n is ignored, and the number of
#' # observations generated by rvmsin is the same as the length of the
#' # recycled vectors
#' rvmsin(n, kappa1, kappa2, kappa3, mu1, mu2)
#'
#' @export


rvmsin <- function(n, kappa1=1, kappa2=1,
                   kappa3=0, mu1=0, mu2=0, method="naive")
{
  if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
  if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
  if(n < 0) stop("invalid n")


  # if (is.null(method)) {
  #   if (n > 1e5) method <- "vmprop"
  #   else method <- "naive"
  # }

  if (!method %in% c("naive", "vmprop"))
    stop("method must be either \'naive\' or \'vmprop\'")

  if(max(length(kappa1), length(kappa2), length(kappa3),
         length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    k1 <- expanded[[1]]; k2 <- expanded[[2]]; k3 <- expanded[[3]]
    mu1 <- expanded[[4]]; mu2 <- expanded[[5]]

    t(vapply(1:length(k1),
             function(j) rvmsin_1par(1, k1[j], k2[j], k3[j],
                                            mu1[j], mu2[j], "naive"), c(0, 0)))
  } else {
    # if (is.null(method) & max(kappa1, kappa2, abs(kappa3)) < 0.1)
    #   method <- "vmprop"

    rvmsin_1par(n, kappa1, kappa2, kappa3, mu1, mu2, method)
  }

}

# rvmsin_naive <- function(n, kappa1=1, kappa2=1,
#                    kappa3=0, mu1=0, mu2=0) {
#
#   opt_obj <- function(k1=1, k2=1, k3=0, mu1=0, mu2=0) {
#     # for numerical stability, if k3 < 0, fabs(k1+k3) < 1e-5 or fabs(k2+k3) < 1e-5
#     # make k3 = k3 * (1+1e-5)
#     if (k3 < 0) {
#       while (abs(k1 + k3) < 1e-5 || abs(k2 + k3) < 1e-5) {
#         k3 = k3*(1+1e-5)
#       }
#     }
#     obj <- optim(c(0,0), fn = function(x) -(k1*cos(x[1]-mu1)+k2*cos(x[2]-mu2)+k3*sin(x[1]-mu1)*sin(x[2]-mu2)),
#                  gr = function(x)  -c(-k1*sin(x[1]-mu1)+k3*cos(x[1]-mu1)*sin(x[2]-mu2),
#                                       -k2*sin(x[2]-mu2)+k3*sin(x[1]-mu1)*cos(x[2]-mu2)))
#     -obj$value
#
#   }
#
#   if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
#     expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
#     k1 <- expanded[[1]]; k2 <- expanded[[2]]; k3 <- expanded[[3]]
#     mu1 <- expanded[[4]]; mu2 <- expanded[[5]]
#     upper_bd_all <- vapply(1:length(k1),
#                            function(h) opt_obj(k1[h], k2[h], k3[h], mu1[h], mu2[h]),
#                            0)
#     rsin_manypar(k1, k2, k3, mu1, mu2, upper_bd_all)
#   } else {
#     upper_bd <- opt_obj(kappa1, kappa2, kappa3, mu1, mu2)
#     rsin_onepar(n, kappa1, kappa2, kappa3, mu1, mu2, upper_bd)
#   }
# }

#' @rdname rvmsin
#' @export
dvmsin <- function(x, kappa1=1, kappa2=1, kappa3=0, mu1=0, mu2=0, log = FALSE)
{
  if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be non-negative")
  if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
  if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
     || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")

  if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    kappa1 <- expanded[[1]]
    kappa2 <- expanded[[2]]
    kappa3 <- expanded[[3]]
    mu1 <- expanded[[4]]
    mu2 <- expanded[[5]]
  }

  par.mat <- rbind(kappa1, kappa2, kappa3, mu1, mu2)
  n_par <- ncol(par.mat)
  if (length(x) == 2) x <- matrix(x, nrow = 1)
  n_x <- nrow(x)


  if (n_par == 1) {
    log_den <- c(ldsin_manyx_onepar(x, kappa1, kappa2, kappa3,
                                    mu1, mu2))

  } else if (n_x == 1) {
    log_den <- c(ldsin_onex_manypar(c(x), kappa1, kappa2, kappa3,
                                    mu1, mu2))
  } else {
    x_set <- 1:nrow(x)
    par_set <- 1:n_par
    expndn_set <- expand_args(x_set, par_set)
    x_set <- expndn_set[[1]]
    par_set <- expndn_set[[2]]
    log_den <- c(ldsin_manyx_manypar(x[x_set, ], kappa1[par_set],
                                     kappa2[par_set], kappa3[par_set],
                                     mu1[par_set], mu2[par_set]))
  }

  if (log) log_den
  else exp(log_den)
}

#'
# dvmsin <- function(x, kappa1=1, kappa2=1, kappa3=0, mu1=0, mu2=0)
# {
#   if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be non-negative")
#   if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
#   if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
#   if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
#      || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")
#
#   if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
#     expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
#     kappa1 <- expanded[[1]]
#     kappa2 <- expanded[[2]]
#     kappa3 <- expanded[[3]]
#     mu1 <- expanded[[4]]
#     mu2 <- expanded[[5]]
#   }
#
#   par.mat <- rbind(kappa1, kappa2, kappa3, mu1, mu2)
#   n_par <- ncol(par.mat)
#   if (length(x) == 2) x <- matrix(x, nrow = 1)
#   n_x <- nrow(x)
#
#
#   if (n_par == 1) {
#     log_den <- c(ldsin_manyx_onepar(x, kappa1, kappa2, kappa3,
#                                     mu1, mu2))
#
#   } else if (n_x == 1) {
#     log_den <- c(ldsin_onex_manypar(c(x), kappa1, kappa2, kappa3,
#                                     mu1, mu2))
#   } else {
#     x_set <- 1:nrow(x)
#     par_set <- 1:n_par
#     expndn_set <- expand_args(x_set, par_set)
#     x_set <- expndn_set[[1]]
#     par_set <- expndn_set[[2]]
#     log_den <- c(ldsin_manyx_manypar(x[x_set, ], kappa1[par_set],
#                                      kappa2[par_set], kappa3[par_set],
#                                      mu1[par_set], mu2[par_set]))
#   }
#
#   if (log) log_den
#   else exp(log_den)
# }

#' The bivariate von Mises sine model mixtures
#' @inheritParams rvmsin
#' @param n number of observations.
#' @param x matrix of angles (in radians) where the density is to be evaluated, with each row being a
#' single bivariate vector of angles.
#'
#' @param pmix vector of mixture proportions.
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0} for each component.
#'
#' @details All the argument vectors \code{pmix, kappa1, kappa2, kappa3, mu1} and \code{mu2} must be of
#' the same length ( = component size of the mixture model), with \eqn{j}-th element corresponding to the
#' \eqn{j}-th component of the mixture distribution.
#' @details The bivariate von Mises sine model mixture distribution with component size \code{K = \link{length}(p.mix)} has density
#' \deqn{g(x) = \sum p[j] * f(x; \kappa_1[j], \kappa_2[j], \kappa_3[j], \mu_1[j], \mu_2[j])}
#' where the sum extends over \eqn{j}; \eqn{p[j]; \kappa_1[j], \kappa_2[j], \kappa_3[j]}; and \eqn{\mu_1[j], \mu_2[j]} respectively denote the mixing proportion,
#' the three concentration parameters and the two mean parameter for the \eqn{j}-th component, \eqn{j = 1, ..., K},
#' and \eqn{f(. ; \kappa_1, \kappa_2, \kappa_3, \mu_1, \mu_2)} denotes the density function of the von Mises sine model
#' with concentration parameters \eqn{\kappa_1, \kappa_2, \kappa_3} and  mean parameters \eqn{\mu_1, \mu_2}.
#'
#' @return \code{dvmsinmix} computes the density (vector if x is a two column matrix with more than one row)
#' and \code{rvmsinmix} generates random deviates from the mixture density.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' pmix <- c(0.3, 0.4, 0.3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # mixture densities calculated at the rows of x
#' dvmsinmix(x, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' # number of observations generated from the mixture distribution is n
#' rvmsinmix(n, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' @export

rvmsinmix <- function(n, kappa1, kappa2, kappa3, mu1, mu2, pmix, method="naive")
{
  allpar <- list(kappa1=kappa1, kappa2=kappa2, kappa3=kappa3,
                 mu1=mu1, mu2=mu2, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(c(allpar$kappa1, allpar$kappa2) < 0)) stop("kappa1 and kappa2 must be non-negative")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)

  out <- matrix(0, n, 2)
  ncomp <- allpar_len[1] # number of components
  comp_ind <- cID(tcrossprod(rep(1, n), allpar$pmix), ncomp, runif(n))
  # n samples from multinom(ncomp, pmix)
  for(j in seq_len(ncomp)) {
    obs_ind_j <- which(comp_ind == j)
    n_j <- length(obs_ind_j)
    if(n_j > 0) {
      out[obs_ind_j, ] <- rvmsin(n_j, kappa1[j], kappa2[j],
                                 kappa3[j], mu1[j], mu2[j], method)
    }
  }

  out
}


#' @rdname rvmsinmix
#' @export
dvmsinmix <- function(x, kappa1, kappa2, kappa3,
                      mu1, mu2, pmix, log=FALSE)
{
  allpar <- list("kappa1"=kappa1, "kappa2"=kappa2, "kappa3"=kappa3,
                 "mu1"=mu1, "mu2"=mu2, "pmix"=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len)) stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(c(allpar$kappa1, allpar$kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)
  if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
     || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")


  ncomp <- length(kappa1)

  if (length(x) == 2) x <- matrix(x, nrow=1)

  allcompden <- vapply(1:ncomp,
                       function(j) dvmsin(x, kappa1[j], kappa2[j],
                                          kappa3[j], mu1[j], mu2[j], FALSE),
                       rep(0, nrow(x)))

  mixden <- c(allcompden %*% pmix)

  if (log) {
    log(mixden)
  } else {
    mixden
  }

}


#' Fitting bivariate von Mises sine model mixtures using MCMC
#' @param ... arguments (other than \code{model}) passed to \link{fit_angmix}
#'
#' @details Wrapper for \link{fit_angmix} with \code{model = "vmsin"}
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' fit.vmsin.20
#'
#' @export

fit_vmsinmix <- function(...)
{
  fit_angmix(model = "vmsin", ...)
}

vmsin_var_cor_singlepar_numeric <- function(kappa1, kappa2, kappa3) {
  fn_vmsin_const <- function(pars) {
    const_vmsin(pars[1], pars[2], pars[3])
  }

  const <- fn_vmsin_const(c(kappa1, kappa2, kappa3))
  grad <- numDeriv::grad(fn_vmsin_const, c(kappa1, kappa2, kappa3))
  names(grad) <- c("k1", "k2", "k3")
  hess <- numDeriv::hessian(fn_vmsin_const, c(kappa1, kappa2, kappa3))
  dimnames(hess) <- list(c("k1", "k2", "k3"), c("k1", "k2", "k3"))

  rho_fl <- unname(
    (grad["k3"] * hess["k1", "k2"]) /
    sqrt(
      hess["k1", "k1"] * (const - hess["k1", "k1"])
      * hess["k2", "k2"] * (const - hess["k2", "k2"])
    )
  )

  rho_js <- unname(
    grad["k3"] /
    sqrt((const - hess["k1", "k1"]) * (const - hess["k2", "k2"]))
  )

  var1 <- unname(1 - grad["k1"]/const)
  var2 <- unname(1 - grad["k2"]/const)

  # if (rho_js <= -1) {
  #   rho_js <- -1
  # } else if (rho_js >= 1) {
  #   rho_js <- 1
  # }
  #
  # dat <- rvmsin(N, kappa1, kappa2, kappa3, 0, 0)
  #
  # ave_sin1sin2 <- sum(sin(dat[, 1]) * sin(dat[, 2]))/N
  # ave_cos1cos2 <- sum(cos(dat[, 1]) * cos(dat[, 2]))/N
  #
  # ave_sin1sq <- sum(sin(dat[, 1])^2)/N
  # ave_cos1sq <- 1-ave_sin1sq
  # ave_cos1 <- sum(cos(dat[, 1]))/N
  #
  # ave_sin2sq <- sum(sin(dat[, 2])^2)/N
  # ave_cos2sq <- 1-ave_sin2sq
  # ave_cos2 <- sum(cos(dat[, 2]))/N
  #
  # rho_js <- ave_sin1sin2/sqrt(ave_sin1sq * ave_sin2sq)
  # # ifelse(ave_sin1sin2 >= 0, 1, -1) *
  # # min(abs(ave_sin1sin2)/sqrt(ave_sin1sq * ave_sin2sq), 1)
  #
  # rho_fl <- rho_js *
  #   ave_cos1cos2/sqrt(ave_cos1sq * ave_cos2sq)
  # # ifelse(ave_cos1cos2 >= 0, 1, -1) *
  # # min(abs(ave_cos1cos2)/sqrt(ave_cos1sq * ave_cos2sq), 1)
  #
  # var1 <- min(1 - ave_cos1, 1)
  # var2 <- min(1 - ave_cos2, 1)

  list(var1 = var1, var2 = var2, rho_fl = rho_fl, rho_js = rho_js)
}


vmsin_var_cor_singlepar <- function(kappa1, kappa2, kappa3, ...) {
  if (max(kappa1, kappa2, abs(kappa3) > 150)) {
    out <- vmsin_var_cor_singlepar_numeric(kappa1, kappa2, kappa3)
  } else {
    out <- vmsin_var_corr_anltc(kappa1, kappa2, kappa3)
  }

  for (rho in c("rho_js", "rho_fl")) {
    if (out[[rho]] <= -1) {
      out[[rho]] <- -1
    } else if (out[[rho]] >= 1) {
      out[[rho]] <- 1
    }
  }

  for (var in c("var1", "var2")) {
    if (out[[var]] <= 0) {
      out[[var]] <- 0
    } else if (out[[var]] >= 1) {
      out[[var]] <- 1
    }
  }

  out
}





rvmsin_1par <- function(n=1, kappa1=1, kappa2=1, kappa3=1,
                               mu1=0, mu2=0, method = "vmprop")
{
  if (abs(abs(kappa3) - kappa1) < 1e-8 |
      abs(abs(kappa3) - kappa2) < 1e-8)
    kappa3 <- kappa3*(1+1e-6)

  if (method == "vmprop") {

    log_const_vmsin <- log(const_vmsin(kappa1, kappa2, kappa3))
    log_2pi <- log(2*pi)

    unimodal_x <- TRUE
    phistar <- NULL

    # first check if the joint density is bimodal
    if (kappa3^2 >=  kappa1*kappa2) {
      # now check if the x marginal is bimodal
      if (A_bessel(kappa2) > kappa1*kappa2/kappa3^2) {
        unimodal_x <- FALSE
        phistar_eqn <- function(phi) {
          a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi)^2)
          cos(phi)*A_bessel(a_phi)/a_phi - kappa1/kappa3^2
        }
        # find the modes in (-pi, pi)
        find_root <- uniroot.all(phistar_eqn, c(-pi, pi))
        # if no root found, use the naive two dimensional rejection sampler
        if (is.null(find_root)) {
          method <- "naive"
        } else {
          phistar <- find_root[1]
        }

      }

    }


    if (method=="vmprop" & unimodal_x) {

      grid_0_2pi <- seq(0, 2*pi, length.out = 100)

      # do minimax to find optimum kappa and K
      # abs difference between the target log marginal den
      # and log proposal den
      obj_kappa_abs <- function(kappa) {
        obj_kappa_phi <- function(phi) {
          cos_phi_mu1 <- cos(phi-mu1)
          a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi-mu1)^2)
          log_I_a_phi <- log(besselI(a_phi, 0))
          log_target_marginal <- -log_const_vmsin + log_2pi + log_I_a_phi + kappa2*cos_phi_mu1
          log_proposal <- -log_2pi - log(besselI(kappa, 0)) + kappa*cos_phi_mu1
          abs(log_target_marginal - log_proposal)
        }
        optimize(obj_kappa_phi, c(0, 2*pi), maximum=TRUE)$objective
      }

      # minimize obj_kappa_abs wrt kappa
      kappa_opt <- optimize(obj_kappa_abs, c(0, max(50, kappa2*2)))$minimum


      log_prop_den <- function(kappa, phi) {
        cos_phi <- cos(phi)
        -log_2pi - log(besselI(kappa, 0)) +
          kappa*cos_phi
      }

      # find max of log_prop_den
      max_log_prop_den <- max(log_prop_den(kappa_opt, grid_0_2pi))
      # any point with density < exp(max_log_prop_den)*exp(-10)
      # are less likely to appear, so make exp(max_log_prop_den)*exp(-10)
      # as the lower bound for propsal density to avoid instability


      # browser()

      # now maximize the log ratio of the two densities
      # w.r.t phi, given kappa = kappa_opt
      obj_kappa_phi <- function(kappa, phi) {
        cos_phi_mu1 <- cos(phi-mu1)
        a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi-mu1)^2)
        log_I_a_phi <- log(besselI(a_phi, 0))
        log_target_marginal <- -log_const_vmsin + log_2pi + log_I_a_phi + kappa2*cos_phi_mu1
        log_proposal <- pmax(-log_2pi - log(besselI(kappa, 0)) +
                               kappa*cos_phi_mu1, max_log_prop_den-10)
        # log_proposals are >= max_log_prop_den -10 to avoid instability
        # also, proposal deviates with log density less than this bound are unlikely
        log_target_marginal - log_proposal
      }

      # maximize the log-ratio over a grid
      (logK <- max(obj_kappa_phi(kappa_opt, seq(0, 2*pi, length.out = 200))))


      # cat(exp(-logK))
      # rcos_unimodal(1, kappa1, kappa2, kappa3, mu1, mu2,
      #               kappa_opt, log(besselI(kappa_opt, 0, TRUE)) + kappa_opt,
      #               logK, log_const_vmsin)

      rsin_unimodal(n, kappa1, kappa2, kappa3, mu1, mu2,
                    kappa_opt, log(BESSI0_C(kappa_opt)), logK,
                    log_const_vmsin)

    }

    else if (method=="vmprop" & !unimodal_x) {

      # browser()

      # do all otimizations is (-pi, pi) then revert back to [0, 2*pi]


      # change the modes into [0, 2*pi]
      mode_1 <- prncp_reg(mu1 + phistar)
      mode_2 <- prncp_reg(mu1 - phistar)
      # sin_phistar <- sin(phistar)
      unifpropn <- 1e-10
      vmpropn <- (1-1e-10)/2
      grid_0_2pi <- seq(0, 2*pi, length.out = 100)


      # proposal for y marginal = vmpropn-vmpropn-unifpropn mix
      # of vm(kappa, mu2 +- phistar) and unif(0, 2*pi)
      # and unif(0, 2*pi) (to avoid overflow of the ratio where
      # the densities are flat)
      # do minimax to find optimum kappa and K
      # don't worry about mu1 mu2 while optimizing!
      # abs difference between the target log marginal den
      # and log proposal den
      obj_kappa_abs <- function(kappa) {
        obj_kappa_phi <- function(phi) {
          cos_phi_mu1 <- cos(phi-mu1)
          a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi-mu1)^2)
          log_I_a_phi <- log(besselI(a_phi, 0))
          log_target_marginal <- -log_const_vmsin + log_2pi + log_I_a_phi + kappa2*cos_phi_mu1
          log_proposal <-
            -log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                                  kappa*cos(phi-mode_1) +
                                  log(1+exp(kappa*(cos(phi-mode_2)
                                                   - cos(phi-mode_1))))
            ) + unifpropn)
          # simplified mixture density
          abs(log_target_marginal - log_proposal)
        }
        # optimize(obj_kappa_phi, c(0, 2*pi), maximum=TRUE)$objective
        max(obj_kappa_phi(grid_0_2pi))
      }


      # minimize obj_kappa_abs wrt kappa
      (kappa_opt <- optimize(obj_kappa_abs,
                             c(0, max(kappa1, kappa2,
                                      abs(kappa3))))$minimum)


      # browser()
      #
      #       target_den <- function(phi) {
      #         cos_phi_mu1 <- cos(phi-mu1)
      #         a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi-mu1)^2)
      #         log_I_a_phi <- log(besselI(a_phi, 0))
      #         log_target_marginal <- -log_const_vmsin + log_2pi + log_I_a_phi + kappa2*cos_phi_mu1
      #         exp(log_target_marginal)
      #       }

      log_prop_den <- function(kappa, phi) {
        cos_phi <- cos(phi)
        -log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                              kappa*cos(phi-mode_1) +
                              log(1+exp(kappa*(cos(phi-mode_2)
                                               - cos(phi-mode_1))))
        ) + unifpropn)

      }


      # find max of log_prop_den
      max_log_prop_den <- max(log_prop_den(kappa_opt, grid_0_2pi))
      # any point with density < exp(max_log_prop_den)*exp(-10)
      # are less likely to appear, so make exp(max_log_prop_den)*exp(-10)
      # as the lower bound for propsal density to avoid instability


      # now maximize the log ratio of the two densities
      # w.r.t phi, given kappa = kappa_opt
      obj_kappa_phi <- function(kappa, phi) {
        cos_phi_mu1 <- cos(phi-mu1)
        a_phi <- sqrt(kappa2^2 + kappa3^2*sin(phi-mu1)^2)
        log_I_a_phi <- log(besselI(a_phi, 0))
        log_target_marginal <- -log_const_vmsin + log_2pi + log_I_a_phi + kappa2*cos_phi_mu1
        log_proposal <-
          pmax(
            -log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                                  kappa*cos(phi-mode_1) +
                                  log(1+exp(kappa*(cos(phi-mode_2)
                                                   - cos(phi-mode_1))))
            ) + unifpropn),
            max_log_prop_den-10)
        # log_proposals are >= max_log_prop_den -10 to avoid instability
        # also, proposal deviates with log density less than this bound are unlikely
        log_target_marginal - log_proposal
      }


      #
      #       prop_den <- function(kappa, phi) {
      #         exp(log_prop_den(kappa, phi))
      #       }


      #
      # pp <- function(kappa, phi) {
      #   cos_phi <- cos(phi)
      #   log_proposal <-
      #     pmax(-log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
      #                                kappa*cos(phi-mode_1) +
      #                                log(1+exp(kappa*(cos(phi-mode_2) - cos(phi-mode_1)))))
      #                          + unifpropn), max_log_prop_den-10)
      #   exp(log_proposal)
      # }
      # #
      # integrate(target_den, 0, 2*pi)
      # integrate(function(x) prop_den(kappa_opt, x), 0, 2*pi)

      # maximize the log-ratio over a grid
      (logK <- max(obj_kappa_phi(kappa_opt, seq(0, 2*pi, length.out = 200))))

      exp(logK)


      # poin <- seq(0, 2*pi, length.out = 100)
      # # ggp <- target_den(poin)
      # # ppp <- prop_den(kappa_opt, poin)
      # # #
      # # plot(poin, ggp, type="l", ylim=range(ggp, ppp, ggp/ppp))
      # # points(poin, ppp, type="l", col = "blue")
      # # points(poin, ggp/ppp, type="l")
      #
      #
      # lggp <- log(target_den(poin))
      # lppp <- log(prop_den(kappa_opt, poin))
      # difp <- obj_kappa_phi(kappa_opt, poin)
      # #
      # plot(poin, lggp, type="l", ylim=range(lggp, lppp, lggp-lppp))
      # points(poin, lppp, type="l", col = "blue")
      # points(poin, lggp-lppp, type="l")
      # points(poin, difp, type="l", col = "red")
      # abline(v=c(mode_1, mode_2))
      # #
      # #
      # #
      # # browser()
      # cat(exp(-logK))

      rsin_bimodal(n, kappa1, kappa2, kappa3, mu1, mu2,
                   kappa_opt, log(BESSI0_C(kappa_opt)), logK,
                   log_const_vmsin, mode_1, mode_2,
                   vmpropn, unifpropn)

    }
  }

  else if (method == "naive") {

    opt_obj <- function(k1=1, k2=1, k3=0, mu1=0, mu2=0) {
      # for numerical stability, if k3 < 0, fabs(k1+k3) < 1e-5 or fabs(k2+k3) < 1e-5
      # make k3 = k3 * (1+1e-5)
      if (k3 < 0) {
        while (abs(k1 + k3) < 1e-5 || abs(k2 + k3) < 1e-5) {
          k3 = k3*(1+1e-5)
        }
      }
      obj <- optim(c(0,0), fn = function(x) -(k1*cos(x[1]-mu1)+k2*cos(x[2]-mu2)+k3*sin(x[1]-mu1)*sin(x[2]-mu2)),
                   gr = function(x)  -c(-k1*sin(x[1]-mu1)+k3*cos(x[1]-mu1)*sin(x[2]-mu2),
                                        -k2*sin(x[2]-mu2)+k3*sin(x[1]-mu1)*cos(x[2]-mu2)))
      -obj$value

    }
      upper_bd <- opt_obj(kappa1, kappa2, kappa3, mu1, mu2)
      rsin_onepar(n, kappa1, kappa2, kappa3, mu1, mu2, upper_bd)

  }
}

#' @keywords internal
Ainv <- function(x) {
  if(x < 0.53) (2*x + x^3 + 5/6*x^5)
  else if(x >= 0.86 && x < 0.95) ((9 - 8*x + 3*x^2) / (8 * (1 - x)))
  else ((1.28 - 0.53*x^2) * tan(x*pi/2))
}


#' @keywords internal
A_bessel <- function(x){
  besselI(x, 1, TRUE)/besselI(x, 0, TRUE)
}


#' @keywords internal
start_par_vm <- function(data.sub) {
  x1 <- data.sub
  Sbar <- mean(sin(x1))
  Cbar <- mean(cos(x1))
  muhat <- atan(Sbar/Cbar) + pi * (Cbar < 0)
  Rbar <- sqrt(Sbar^2 + Cbar^2)
  k <- Ainv(Rbar)
  c(k, prncp_reg(muhat))
}  #starting parameters from  a dataset

#' @keywords internal
start_clus_kmeans_vm <- function(data.full, comp = 2, nstart = 10){
  data.full.cart <- t(sapply(data.full, function(x) c(cos(x), sin(x))))
  data.kmean <- kmeans(data.full.cart, centers = comp, nstart = nstart)
  ids <- data.kmean$cluster
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vm(data.full[clust.ind[[m]]]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2, function(x) sum_sq(x[1:3]))))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #kmeans, then start_par for each cluster

#' @keywords internal
start_clus_rand_vm <- function(data.full, comp = 2, nstart = 10) {
  rand.pi  <- runif(comp, 1/(2*comp), 2/comp)
  rand.pi <- rand.pi/sum(rand.pi)

  rand.multinom <- t(rmultinom(length(data.full), 1, rand.pi))
  ids <- apply(rand.multinom, 1, which.max)
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vm(data.full[clust.ind[[m]]]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2, function(x) sum_sq(x[1:3]))))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #random groups, then start_par for each cluster




#' @keywords internal
start_par_vmsin <- function(data.sub) {
  x1 <- data.sub[,1]; y1 <- data.sub[,2]
  Sbarphi <- mean(sin(x1))
  Cbarphi <- mean(cos(x1))
  phibar <- atan(Sbarphi/Cbarphi) + pi * (Cbarphi < 0)
  Rbarphi <- sqrt(Sbarphi^2 + Cbarphi^2)
  k1 <- Ainv(Rbarphi)

  Sbarpsi <- mean(sin(y1))
  Cbarpsi <- mean(cos(y1))
  psibar <- atan(Sbarpsi/Cbarpsi) + pi * (Cbarpsi < 0)
  Rbarpsi <- sqrt(Sbarpsi^2 + Cbarpsi^2)
  k2 <- Ainv(Rbarpsi)


  sindiffphi <- sin(outer(x1, x1, "-"))
  sindiffpsi <- sin(outer(y1, y1, "-"))
  rho <- sum(sindiffphi*sindiffpsi)/sum(sindiffphi^2)/sum(sindiffpsi^2)

  c(k1, k2, rho*sqrt(k1*k2), prncp_reg(phibar), prncp_reg(psibar))
}  #starting parameters from  a dataset

#' @keywords internal
start_clus_kmeans_vmsin <- function(data.full, comp = 2, nstart = 10){
  data.full.cart <- t(apply(data.full, 1, sph2cart))
  data.kmean <- kmeans(data.full.cart, centers = comp, nstart = nstart)
  ids <- data.kmean$cluster
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vmsin(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2, function(x) sum_sq(x[1:3]))))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #kmeans, then start_par for each cluster

#' @keywords internal
start_clus_rand_vmsin <- function(data.full, comp = 2, nstart = 10) {
  rand.pi  <- runif(comp, 1/(2*comp), 2/comp)
  rand.pi <- rand.pi/sum(rand.pi)

  rand.multinom <- t(rmultinom(nrow(data.full), 1, rand.pi))
  ids <- apply(rand.multinom, 1, which.max)
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vmsin(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2, function(x) sum_sq(x[1:3]))))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #random groups, then start_par for each cluster






#' @keywords internal
start_par_vmcos <- function(data.sub) {
  x1 <- data.sub[,1];
  y1 <- data.sub[,2]
  Sbarphi <- mean(sin(x1))
  Cbarphi <- mean(cos(x1))
  phibar <- atan(Sbarphi/Cbarphi) + pi * (Cbarphi < 0)
  Rbarphi <- sqrt(Sbarphi^2 + Cbarphi^2)
  k1 <- Ainv(Rbarphi)
  #k1 <- min(15, Ainv(Rbarphi))

  Sbarpsi <- mean(sin(y1))
  Cbarpsi <- mean(cos(y1))
  psibar <- atan(Sbarpsi/Cbarpsi) + pi * (Cbarpsi < 0)
  Rbarpsi <- sqrt(Sbarpsi^2 + Cbarpsi^2)
  k2 <- Ainv(Rbarpsi)
  #k2 <- min(15, Ainv(Rbarpsi))

  mu <- prncp_reg(phibar)
  nu <- prncp_reg(psibar)

  Sbarphi_psi <- mean(sin(x1-y1))
  Cbarphi_psi <- mean(cos(x1-y1))
  phi_psibar <- atan(Sbarphi_psi/Cbarphi_psi) + pi * (Cbarphi_psi < 0)
  Rbarphi_psi <- sqrt(Sbarphi_psi^2 + Cbarphi_psi^2)
  k3.unsgn <- Ainv(Rbarphi_psi)


  sindiffphi <- sin(outer(x1, x1, "-"))
  sindiffpsi <- sin(outer(y1, y1, "-"))
  rho <- sum(sindiffphi*sindiffpsi)/sum(sindiffphi^2)/sum(sindiffpsi^2)

  # Sbarphi_psi <- mean(sin(x1-y1+mu-nu))
  # Cbarphi_psi <- mean(cos(x1-y1+mu-nu))
  # Rbarphi_psi <- sqrt(Sbarphi_psi^2 + Cbarphi_psi^2)
  k3 <- sign(rho)*k3.unsgn



  c(k1, k2, k3, mu, nu)
}  #starting parameters from  a dataset

#' @keywords internal
start_clus_kmeans_vmcos <- function(data.full, comp = 2, nstart = 10){
  data.full.cart <- t(apply(data.full, 1, sph2cart))
  data.kmean <- kmeans(data.full.cart, centers = comp, nstart = nstart)
  ids <- data.kmean$cluster
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vmcos(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2, function(x) sum_sq(x[1:3]))))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #kmeans, then start_par for each cluster

#' @keywords internal
start_clus_rand_vmcos <- function(data.full, comp = 2, nstart = 10){
  rand.pi  <- runif(comp, 1/(2*comp), 2/comp)
  rand.pi <- rand.pi/sum(rand.pi)

  rand.multinom <- t(rmultinom(nrow(data.full), 1, rand.pi))
  ids <- apply(rand.multinom, 1, which.max)
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_vmcos(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(order(apply(par, 2,  sum_sq)))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #random groups, then start_par for each cluster

basic_surfaceplot <- function(xpoints, ypoints, denmat,
                              xlab, ylab, zlab,
                              main = main, ...) {
  inargs <- list(...)
  inargs$x <- denmat~x*y
  inargs$data <- data.frame(x = xpoints,
                            y=rep(ypoints, each=length(xpoints)),
                            denmat=denmat)
  if (is.null(inargs$outerbox)) inargs$outerbox <- FALSE
  if (is.null(inargs$par.settings)) inargs$par.settings <- list(axis.line = list(col = 'transparent'))
  if (is.null(inargs$xlab)) inargs$xlab <- xlab
  if (is.null(inargs$ylab)) inargs$ylab <- ylab
  if (is.null(inargs$colorkey)) inargs$colorkey <- FALSE
  if (is.null(inargs$main)) inargs$main <- main
  if (is.null(inargs$neval)) inargs$neval <- 100
  if (is.null(inargs$aspect)) inargs$aspect <- c(61/87, 0.4)
  if (is.null(inargs$zlab)) inargs$zlab <- list("Density", rot=90)
  if (is.null(inargs$screen)) inargs$screen <- list(z=-30, x=-60)
  if (is.null(inargs$colorkey)) inargs$colorkey <- FALSE
  if (is.null(inargs$scales))
    inargs$scales <- list(arrows=FALSE, col=1)
  if (is.null(inargs$drape)) inargs$drape <- TRUE
  if (is.null(inargs$light.source))
    inargs$light.source <- c(10,0,10)
  if (is.null(inargs$col.regions))
    inargs$col.regions <- grDevices::colorRampPalette(c("steelblue", "green",
                                                        "yellow", "orange", "red"))(60)
  if (is.null(inargs$par.settings))
    inargs$par.settings <-
    list(axis.line = list(col = 'transparent'
    ),
    layout.heights = list(
      top.padding = 0,
      main.key.padding = 0,
      key.axis.padding = 0,
      axis.xlab.padding = 0,
      xlab.key.padding = 0,
      key.sub.padding = 0,
      bottom.padding = 0
    ),
    layout.widths = list(
      left.padding = 0,
      key.ylab.padding = 0,
      ylab.axis.padding = 0,
      axis.key.padding = 0,
      right.padding = 0
    ))

  if (is.null(inargs$zoom))
    inargs$zoom <- 0.85


  do.call(lattice::wireframe, inargs)
}

# list of functions


#' The bivariate von Mises cosine model
#' @inheritParams rvmsin
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0}.
#' @param ... additional arguments to be passed to dvmcos. See details.
#'
#' @details
#' The bivariate von Mises cosine model density at the point \eqn{x = (x_1, x_2)} is given by
#' \deqn{f(x) = C_c (\kappa_1, \kappa_2, \kappa_3) \exp(\kappa_1 \cos(T_1) + \kappa_2 \cos(T_2) + \kappa_3 \cos(T_1 - T_2))}
#' where
#' \deqn{T_1 = x_1 - \mu_1;  T_2 = x_2 - \mu_2}
#' and \eqn{C_c (\kappa_1, \kappa_2, \kappa_3)} denotes the normalizing constant for the cosine model.
#'
#' Because \eqn{C_c} involves an infinite  alternating series with product of Bessel functions,
#' if \code{kappa3 < -5} or \code{max(kappa1, kappa2, abs(kappa3)) > 50}, \eqn{C_c} is evaluated
#' numerically via (quasi) Monte carlo method for
#' numerical stability. These (quasi) random numbers can be provided through the
#' argument \code{qrnd}, which must be a two column matrix, with each element being
#' a  (quasi) random number between 0 and 1. Alternatively, if \code{n_qrnd} is
#' provided (and \code{qrnd} is missing), a two dimensional sobol sequence of size \code{n_qrnd} is
#' generated via the function \link[qrng]{sobol} from the R package \code{qrng}. If none of \code{qrnd}
#' or \code{n_qrnd} is available, a two dimensional sobol sequence of size 1e4 is used. By default Monte
#' Carlo approximation is used only if \code{kappa3 < -5} or \code{max(kappa1, kappa2, abs(kappa3)) > 50}.
#' However, a forced Monte Carlo approximation can be made (irrespective of the choice of \code{kappa1, kappa2} and
#' \code{kappa3}) by setting \code{force_approx_const = TRUE}. See examples.
#'
#' @return \code{dvmcos} gives the density  and \code{rvmcos} generates random deviates.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # when x is a bivariate vector and parameters are all scalars,
#' # dvmcos returns single density
#' dvmcos(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when x is a two column matrix and parameters are all scalars,
#' # dmvsin returns a vector of densities calculated at the rows of
#' # x with the same parameters
#' dvmcos(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # if x is a bivariate vector and at least one of the parameters is
#' # a vector, all parameters are recycled to the same length, and
#' # dvmcos returns a vector with ith element being the density
#' # evaluated at x with parameter values kappa1[i], kappa2[i],
#' # kappa3[i], mu1[i] and mu2[i]
#' dvmcos(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # if x is a two column matrix and at least one of the parameters is
#' # a vector, rows of x and the parameters are recycled to the same
#' # length, and dvmcos returns a vector with ith element being the
#' # density evaluated at ith row of x with parameter values kappa1[i],
#' # kappa2[i], # kappa3[i], mu1[i] and mu2[i]
#' dvmcos(x, kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # when parameters are all scalars, number of observations generated
#' # by rvmcos is n
#' rvmcos(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when at least one of the parameters is a vector, all parameters are
#' # recycled to the same length, n is ignored, and the number of
#' # observations generated by rvmcos is the same as the length of the
#' # recycled vectors
#' rvmcos(n, kappa1, kappa2, kappa3, mu1, mu2)
#'
#'
#' \donttest{
#' ## Visualizing (quasi) Monte Carlo based approximations of
#' ## the normalizing constant through density evaluations.
#'
#' # "good" setup, where the analytic formula for C_c can be
#' # calculated without numerical issues
#' # kappa1 = 1, kappa2 = 1, kappa3 = -2, mu1 = pi, mu2 = pi
#'
#' n_qrnd <-  (1:500)*20
#' # analytic
#' good.a <- dvmcos(c(3,3), 1, 1, -2, pi, pi, log=TRUE)
#' # using quasi Monte Carlo
#' good.q <- sapply(n_qrnd,
#'                  function(j)
#'                    dvmcos(c(3,3), 1, 1, -2, pi, pi,
#'                           log=TRUE, n_qrnd = j,
#'                           force_approx_const = TRUE))
#' # using ordinary Monte Carlo
#' set.seed(1)
#' good.r <- sapply(n_qrnd,
#'                  function(j)
#'                    dvmcos(c(3,3), 1, 1, -2, pi, pi,
#'                           log=TRUE,
#'                           qrnd = matrix(runif(2*j), ncol = 2),
#'                           force_approx_const = TRUE))
#'
#'
#' plot(n_qrnd, good.q, ylim = range(good.a, good.q, good.r),
#'      col = "orange", type = "l",
#'      ylab = "",
#'      main = "dvmcos(c(3,3), 1, 1, -2, pi, pi, log = TRUE)")
#' points(n_qrnd, good.r, col = "skyblue", type = "l")
#' abline(h = good.a, lty = 2, col = "grey")
#' legend("topright",
#'        legend = c("Sobol", "Random", "Analytic"),
#'        col = c("orange", "skyblue", "grey"),
#'        lty = c(1, 1, 2))
#'
#'
#' # "bad" setup, where the calculating C_c
#' # numerically using the analytic formula is problematic
#' # kappa1 = 100, kappa2 = 100, kappa3 = -200, mu1 = pi, mu2 = pi
#'
#' n_qrnd <-  (1:500)*20
#'
#' # using quasi Monte Carlo
#' bad.q <- sapply(n_qrnd,
#'                 function(j)
#'                   dvmcos(c(3,3), 100, 100, -200, pi, pi,
#'                          log=TRUE, n_qrnd = j,
#'                          force_approx_const = TRUE))
#' # using ordinary Monte Carlo
#' set.seed(1)
#' bad.r <- sapply(n_qrnd,
#'                 function(j)
#'                   dvmcos(c(3,3), 100, 100, -200, pi, pi,
#'                          log=TRUE,
#'                          qrnd = matrix(runif(2*j), ncol = 2),
#'                          force_approx_const = TRUE))
#'
#'
#' plot(n_qrnd, bad.q, ylim = range(bad.q, bad.r),
#'      col = "orange", type = "l",
#'      ylab = "",
#'      main = "dvmcos(c(3,3), 100, 100, -200, pi, pi, log = TRUE)")
#' points(n_qrnd, bad.r, col = "skyblue", type = "l")
#' legend("topright",
#'        legend = c("Sobol", "Random"),
#'        col = c("orange", "skyblue"), lty = 1)
#'}
#'
#' @export
rvmcos <- function(n, kappa1=1, kappa2=1,
                   kappa3=0, mu1=0, mu2=0,
                   method="naive")
{
  if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
  if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
  if(n < 0) stop("invalid n")


  # if (is.null(method)) {
  #   if (n > 1e5) method <- "vmprop"
  #   else method <- "naive"
  # }

  if (!method %in% c("naive", "vmprop"))
    stop("method must be either \'naive\' or \'vmprop\'")

  if(max(length(kappa1), length(kappa2), length(kappa3),
         length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    k1 <- expanded[[1]]; k2 <- expanded[[2]]; k3 <- expanded[[3]]
    mu1 <- expanded[[4]]; mu2 <- expanded[[5]]

    t(vapply(1:length(k1),
             function(j) rvmcos_1par(1, k1[j], k2[j], k3[j],
                                     mu1[j], mu2[j], "naive"), c(0, 0)))
  } else {
    # if (is.null(method) & max(kappa1, kappa2, abs(kappa3)) < 0.1)
    #   method <- "vmprop"

    rvmcos_1par(n, kappa1, kappa2, kappa3, mu1, mu2, method)
  }

}


# rvmcos <- function(n, kappa1=1, kappa2=1, kappa3=0,
#                    mu1=0, mu2=0) {
#   if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
#   if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
#   if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
#
#   opt_obj <- function(k1=1, k2=1, k3=0, mu1=0, mu2=0) {
#     # for numerical stability, if k3 < 0, fabs(k1+k3) < 1e-5 or fabs(k2+k3) < 1e-5
#     # make k3 = k3 * (1+1e-5)
#     if (k3 < 0) {
#       while (abs(k1 + k3) < 1e-5 || abs(k2 + k3) < 1e-5) {
#         k3 = k3*(1+1e-5)
#       }
#     }
#     obj <- optim(c(0,0), fn = function(x) -(k1*cos(x[1]-mu1)+k2*cos(x[2]-mu2)+k3*cos(x[1]-x[2]-mu1+mu2)),
#                  gr = function(x)  -c(-k1*sin(x[1]-mu1)-k3*sin(x[1]-x[2]-mu1+mu2),
#                                       -k2*sin(x[2]-mu2)+k3*sin(x[1]-x[2]-mu1+mu2)))
#     -obj$value
#
#   }
#
#   if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
#     expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
#     k1 <- expanded[[1]]; k2 <- expanded[[2]]; k3 <- expanded[[3]]
#     mu1 <- expanded[[4]]; mu2 <- expanded[[5]]
#     upper_bd_all <- vapply(1:length(k1),
#                            function(h) opt_obj(k1[h], k2[h], k3[h], mu1[h], mu2[h]),
#                            0)
#     rcos_manypar(k1, k2, k3, mu1, mu2, upper_bd_all)
#   } else {
#     upper_bd <- opt_obj(kappa1, kappa2, kappa3, mu1, mu2)
#     qrnd_grid <- sobol(1e4, 2, FALSE)
#     cat(exp(log(const_vmcos(kappa1, kappa2, kappa3,
#                             qrnd_grid))
#             - upper_bd)/(4*pi^2))
#     rcos_onepar(n, kappa1, kappa2, kappa3, mu1, mu2, upper_bd)
#   }
# }




#' @rdname rvmcos
#' @importFrom qrng sobol
#' @export

dvmcos <- function(x, kappa1=1, kappa2=1, kappa3=0, mu1=0,
                   mu2=0, log=FALSE, ...) {

  if (any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
  if (any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if (any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
  if ((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
      || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")

  ell <- list(...)

  if (!is.null(ell$qrnd)) {
    qrnd_grid <- ell$qrnd
    dim_qrnd <- dim(qrnd_grid)
    if (!is.matrix(qrnd_grid) | is.null(dim_qrnd) |
        dim_qrnd[2] != 2)
      stop("\'qrnd\' must be a two column matrix")
    n_qrnd <- dim_qrnd[1]
  } else if (!is.null(ell$n_qrnd)){
    n_qrnd <- round(ell$n_qrnd)
    if (n_qrnd < 1)
      stop("n_qrnd must be a positive integer")
    qrnd_grid <- sobol(n_qrnd, 2, FALSE)
  } else {
    n_qrnd <- 1e4
    qrnd_grid <- sobol(n_qrnd, 2, FALSE)
  }


  if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    kappa1 <- expanded[[1]]
    kappa2 <- expanded[[2]]
    kappa3 <- expanded[[3]]
    mu1 <- expanded[[4]]
    mu2 <- expanded[[5]]
  }

  par.mat <- rbind(kappa1, kappa2, kappa3, mu1, mu2)
  n_par <- ncol(par.mat)
  if (length(x) == 2) x <- matrix(x, nrow = 1)
  n_x <- nrow(x)



  if (is.null(ell$force_approx_const)) {
    force_approx_const <- FALSE
  } else if (!is.logical(ell$force_approx_const)) {
    stop("\'force_approx_const\' must be logical.")
  } else {
    force_approx_const <- ell$force_approx_const
  }


  l_const_all <- rep(0, n_par)

  for(j in 1:n_par) {
    if (all(!force_approx_const,
            ((kappa3[j] >= -5 & max(kappa1[j], kappa2[j],
                                    abs(kappa3[j])) <= 50) |
             abs(kappa3[j]) < 1e-4))) {

      l_const_all[j] <- log(const_vmcos_anltc(kappa1[j], kappa2[j],
                                              kappa3[j]))
    } else {
      l_const_all[j] <- log(const_vmcos_mc(kappa1[j], kappa2[j],
                                           kappa3[j], qrnd_grid));
    }
  }

  if (n_par == 1) {
    log_den <- c(ldcos_manyx_onepar(x, kappa1, kappa2, kappa3,
                                    mu1, mu2, l_const_all))

  } else if (n_x == 1) {
    log_den <- c(ldcos_onex_manypar(c(x), kappa1, kappa2, kappa3,
                                    mu1, mu2, l_const_all))
  } else {
    x_set <- 1:nrow(x)
    par_set <- 1:n_par
    expndn_set <- expand_args(x_set, par_set)
    x_set <- expndn_set[[1]]
    par_set <- expndn_set[[2]]
    log_den <- c(ldcos_manyx_manypar(x[x_set, ], kappa1[par_set],
                                     kappa2[par_set], kappa3[par_set],
                                     mu1[par_set], mu2[par_set],
                                     l_const_all[par_set]))
  }

  if (log) log_den
  else exp(log_den)

}

# dvmcos <- function(x, kappa1=1, kappa2=1, kappa3=0, mu1=0, mu2=0) {
#
#   if(any(c(kappa1, kappa2) < 0)) stop("kappa1 and kappa2 must be nonnegative")
#   if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
#   if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
#   if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
#      || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")
#
#
#   if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
#     expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
#     k1 <- expanded[[1]]; k2 <- expanded[[2]]; k3 <- expanded[[3]]
#     mu1 <- expanded[[4]]; mu2 <- expanded[[5]]
#     sobol_grid <- sobol_2d_1e4_from_seed_1
#     par.mat <- rbind(k1,k2,k3,mu1,mu2)
#     l_const_all <- log_const_vmcos_all(par.mat, sobol_grid, ncores = 1)
#     if(length(x) != 2) {
#       x_set <- 1:nrow(x)
#       par_set <- 1:length(kappa1)
#       expndn_set <- expand_args(x_set, par_set)
#       x_set <- expndn_set[[1]]
#       par_set <- expndn_set[[2]]
#       as.vector(dcos_manyx_manypar(x[x_set, ], k1[par_set], k2[par_set], k3[par_set], mu1[par_set], mu2[par_set], l_const_all[par_set]))
#     } else{
#       as.vector(dcos_onex_manypar(x, k1, k2, k3, mu1, mu2, l_const_all))
#     }
#
#   } else {
#     sobol_grid <- sobol_2d_1e4_from_seed_1
#     const.vmcos <- const_vmcos(kappa1, kappa2, kappa3, sobol_grid, ncores = 1)
#     if(length(x) != 2){
#       as.vector(dcos_manyx_onepar(x, kappa1, kappa2, kappa3, mu1, mu2, log(const.vmcos)))
#     } else{
#       exp(ldcosnum(x[1], x[2], c(kappa1, kappa2, kappa3, mu1, mu2)))/const.vmcos
#     }
#   }
#
#
#
# }

#' The bivariate von Mises cosine model mixtures
#' @inheritParams rvmsinmix
#' @inheritParams rvmcos
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0} for each component.
#'
#' @details All the argument vectors \code{pmix, kappa1, kappa2, kappa3, mu1} and \code{mu2} must be of
#' the same length ( = component size of the mixture model), with \eqn{j}-th element corresponding to the
#' \eqn{j}-th component of the mixture distribution.
#' @details The bivariate von Mises cosine model mixture distribution with component size \code{K = \link{length}(pmix)} has density
#' \deqn{g(x) = \sum p[j] * f(x; \kappa_1[j], \kappa_2[j], \kappa_3[j], \mu_1[j], \mu_2[j])}
#' where the sum extends over \eqn{j}; \eqn{p[j]; \kappa_1[j], \kappa_2[j], \kappa_3[j]}; and \eqn{\mu_1[j], \mu_2[j]} respectively denote the mixing proportion,
#' the three concentration parameters and the two mean parameter for the \eqn{j}-th cluster, \eqn{j = 1, ..., K},
#' and \eqn{f(. ; \kappa_1, \kappa_2, \kappa_3, \mu_1, \mu_2)} denotes the density function of the von Mises cosine model
#' with concentration parameters \eqn{\kappa_1, \kappa_2, \kappa_3} and  mean parameters \eqn{\mu_1, \mu_2}.
#'
#' @return \code{dvmcosmix} computes the density  and \code{rvmcosmix} generates random deviates from the mixture density.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' pmix <- c(0.3, 0.4, 0.3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # mixture densities calculated at the rows of x
#' dvmcosmix(x, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' # number of observations generated from the mixture distribution is n
#' rvmcosmix(n, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' @export

rvmcosmix <- function(n, kappa1, kappa2, kappa3,
                      mu1, mu2, pmix, method = "naive", ...)
{
  allpar <- list(kappa1=kappa1, kappa2=kappa2, kappa3=kappa3,
                 mu1=mu1, mu2=mu2, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(c(allpar$kappa1, allpar$kappa2) < 0)) stop("kappa1 and kappa2 must be non-negative")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)

  # opt_I_k13 <- function(k1, k2, k3, mu1, mu2) {
  #   I_k13 <- function(y) BESSI0_C(sqrt(k1 * k1 + k3 * k3 + 2 * k1 * k3 * cos(y - mu2)))
  #   optimize(I_k13, c(0, 2*pi), maximum = TRUE)$maximum
  # }
  #
  # upper_bd_all <- vapply(1:length(kappa1),
  #                        function(h) opt_I_k13(kappa1[h], kappa2[h], kappa3[h],
  #                                              mu1[h], mu2[h]),
  #                        0)
  #
  # clus_label <- cID(t(replicate(allpar$pmix, n = n)), length(allpar$pmix), runif(n))
  # rcos_manypar(allpar$kappa1[clus_label], allpar$kappa2[clus_label], allpar$kappa3[clus_label],
  #              allpar$mu1[clus_label], allpar$mu2[clus_label], upper_bd_all[clus_label])

  out <- matrix(0, n, 2)
  ncomp <- allpar_len[1] # number of components
  comp_ind <- cID(tcrossprod(rep(1, n), allpar$pmix), ncomp, runif(n))
  # n samples from multinom(ncomp, pmix)
  for(j in seq_len(ncomp)) {
    obs_ind_j <- which(comp_ind == j)
    n_j <- length(obs_ind_j)
    if(n_j > 0) {
      out[obs_ind_j, ] <- rvmcos(n_j, kappa1[j], kappa2[j],
                                 kappa3[j], mu1[j], mu2[j], method)
    }
  }

  out

}


#' @rdname rvmcosmix
#' @export
dvmcosmix <- function(x, kappa1, kappa2, kappa3,
                      mu1, mu2, pmix, log=FALSE, ...)
{
  allpar <- list("kappa1"=kappa1, "kappa2"=kappa2, "kappa3"=kappa3,
                 "mu1"=mu1, "mu2"=mu2, "pmix"=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len)) stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(c(allpar$kappa1, allpar$kappa2) <= 0)) stop("kappa1 and kappa2 must be positive")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)
  if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
     || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")

  ncomp <- length(kappa1)

  if (length(x) == 2) x <- matrix(x, nrow=1)

  allcompden <- vapply(1:ncomp,
                       function(j) dvmcos(x, kappa1[j], kappa2[j],
                                          kappa3[j], mu1[j], mu2[j], FALSE, ...),
                       rep(0, nrow(x)))

  mixden <- c(allcompden %*% pmix)

  if (log) {
    log(mixden)
  } else {
    mixden
  }
}


#' Fitting bivariate von Mises cosine model mixtures using MCMC
#' @inheritParams fit_vmsinmix
#'
#' @details
#' Wrapper for \link{fit_angmix} with \code{model = "vmcos"}.
#'
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmcos.10 <- fit_vmcosmix(tim8, ncomp = 3, n.iter =  10,
#'                              n.chains = 1)
#' fit.vmcos.10
#'
#' @export

fit_vmcosmix <- function(...)
{
  fit_angmix(model="vmcos", ...)
}




vmcos_var_cor_singlepar_numeric <- function(kappa1, kappa2, kappa3, qrnd_grid) {
  # N <- 1e4

  # browser()

  fn_log_vmcos_const <- function(pars) {
    const_vmcos(pars[1], pars[2], pars[3], uni_rand = qrnd_grid, return_log = TRUE)
  }

  # const <- fn_log_vmcos_const(c(kappa1, kappa2, kappa3))
  grad_over_const <- numDeriv::grad(fn_log_vmcos_const, c(kappa1, kappa2, kappa3))
  names(grad_over_const) <- c("k1", "k2", "k3")
  hess_over_const <- numDeriv::hessian(fn_log_vmcos_const, c(kappa1, kappa2, kappa3)) +
    tcrossprod(grad_over_const)
  dimnames(hess_over_const) <- list(c("k1", "k2", "k3"), c("k1", "k2", "k3"))

  rho_fl <- unname(
    ((grad_over_const["k3"] - hess_over_const["k1", "k2"]) *
       hess_over_const["k1", "k2"]) /
      sqrt(
        hess_over_const["k1", "k1"] * (1 - hess_over_const["k1", "k1"])
        * hess_over_const["k2", "k2"] * (1 - hess_over_const["k2", "k2"])
      )
  )

  rho_js <- unname(
    (grad_over_const["k3"] - hess_over_const["k1", "k2"]) /
      sqrt((1 - hess_over_const["k1", "k1"]) *
             (1 - hess_over_const["k2", "k2"]))
  )

  var1 <- unname(1 - grad_over_const["k1"])
  var2 <- unname(1 - grad_over_const["k2"])
  # dat <- rvmcos(N, kappa1, kappa2, kappa3, 0, 0)
  #
  # ave_sin1sin2 <- sum(sin(dat[, 1]) * sin(dat[, 2]))/N
  # ave_cos1cos2 <- sum(cos(dat[, 1]) * cos(dat[, 2]))/N
  #
  # ave_sin1sq <- sum(sin(dat[, 1])^2)/N
  # ave_cos1sq <- 1-ave_sin1sq
  # ave_cos1 <- sum(cos(dat[, 1]))/N
  #
  # ave_sin2sq <- sum(sin(dat[, 2])^2)/N
  # ave_cos2sq <- 1-ave_sin2sq
  # ave_cos2 <- sum(cos(dat[, 2]))/N
  #
  # rho_js <- ave_sin1sin2/sqrt(ave_sin1sq * ave_sin2sq)
  # # ifelse(ave_sin1sin2 >= 0, 1, -1) *
  # # min(abs(ave_sin1sin2)/sqrt(ave_sin1sq * ave_sin2sq), 1)
  #
  # rho_fl <- rho_js *
  #   ave_cos1cos2/sqrt(ave_cos1sq * ave_cos2sq)
  # # ifelse(ave_cos1cos2 >= 0, 1, -1) *
  # # min(abs(ave_cos1cos2)/sqrt(ave_cos1sq * ave_cos2sq), 1)
  #
  # var1 <- min(1 - ave_cos1, 1)
  # var2 <- min(1 - ave_cos2, 1)

  list(var1 = var1, var2 = var2, rho_fl = rho_fl, rho_js = rho_js)
}


vmcos_var_cor_singlepar <- function(kappa1, kappa2, kappa3,
                                    qrnd_grid) {
  if (max(kappa1, kappa2, abs(kappa3)) > 50 |
      kappa3 < 0) {
    out <- vmcos_var_cor_singlepar_numeric(kappa1, kappa2,
                                           kappa3, qrnd_grid)
    # } else if(kappa3 < -1 | max(kappa1, kappa2, abs(kappa3)) > 50) {
    #   vmcos_var_corr_mc(kappa1, kappa2, kappa3, qrnd_grid)
  } else {
    out <- vmcos_var_corr_anltc(kappa1, kappa2, kappa3)
  }

  for (rho in c("rho_js", "rho_fl")) {
    if (out[[rho]] <= -1) {
      out[[rho]] <- -1
    } else if (out[[rho]] >= 1) {
      out[[rho]] <- 1
    }
  }

  for (var in c("var1", "var2")) {
    if (out[[var]] <= 0) {
      out[[var]] <- 0
    } else if (out[[var]] >= 1) {
      out[[var]] <- 1
    }
  }

  out
}



rvmcos_1par <- function(n=1, kappa1, kappa2, kappa3, mu1, mu2, method)
{

  if (abs(abs(kappa3) - kappa1) < 1e-8 |
      abs(abs(kappa3) - kappa2) < 1e-8)
    kappa3 <- kappa3*(1+1e-6)


  if (method == "vmprop") {
    qrnd_grid <- sobol(1e4, 2, FALSE)
    log_const_vmcos <- log(const_vmcos(kappa1, kappa2, kappa3, qrnd_grid))
    log_2pi <- log(2*pi)

    unimodal_y <- TRUE
    phistar <- NULL
    method <- "vmprop"


    # first check if the joint density is bimodal
    if (kappa3 <  - kappa1*kappa2/(kappa1+kappa2+1e-16)) {
      # now check if the y marginal is bimodal
      if (A_bessel(abs(kappa1+kappa3)) >
          -abs(kappa1+kappa3)*kappa2/(kappa1*kappa3 + 1e-16)) {
        unimodal_y <- FALSE
        phistar_eqn <- function(phi) {
          kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos(phi))
          -kappa1*kappa3*A_bessel(kappa13)/kappa13 - kappa2
        }
        find_root <- uniroot.all(phistar_eqn, c(0, 2*pi))
        # if no root found, use the naive two dimensional rejection sampler
        if (is.null(find_root)) {
          method <- "naive"
        } else {
          phistar <- find_root[1]
        }

      }

    }

    # browser()

    if (method == "vmprop" & unimodal_y) {

      grid_0_2pi <- seq(0, 2*pi, length.out = 100)

      # do minimax to find optimum kappa and K
      # don't worry about mu1 mu2 while optimizing!
      # abs difference between the target log marginal den
      # and log proposal den
      obj_kappa_abs <- function(kappa) {
        obj_kappa_phi <- function(phi) {
          cos_phi <- cos(phi)
          kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos_phi)
          log_I_k_13 <- log(besselI(kappa13, 0))
          log_target_marginal <- -log_const_vmcos + log_2pi + log_I_k_13 + kappa2*cos_phi
          log_proposal <- -log_2pi - log(besselI(kappa, 0)) + kappa*cos_phi
          abs(log_target_marginal - log_proposal)
        }
        optimize(obj_kappa_phi, c(0, 2*pi), maximum=TRUE)$objective
      }

      # minimize obj_kappa_abs wrt kappa
      kappa_opt <- optimize(obj_kappa_abs, c(0, max(50, kappa2*2)))$minimum


      log_prop_den <- function(kappa, phi) {
        cos_phi <- cos(phi)
        -log_2pi - log(besselI(kappa, 0)) +
          kappa*cos_phi
      }

      # find max of log_prop_den
      max_log_prop_den <- max(log_prop_den(kappa_opt, grid_0_2pi))
      # any point with density < exp(max_log_prop_den)*exp(-10)
      # are less likely to appear, so make exp(max_log_prop_den)*exp(-10)
      # as the lower bound for propsal density to avoid instability


      # browser()

      # now maximize the log ratio of the two densities
      # w.r.t phi, given kappa = kappa_opt
      obj_kappa_phi <- function(kappa, phi) {
        cos_phi <- cos(phi)
        kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos_phi)
        log_I_k_13 <- log(besselI(kappa13, 0))
        log_target_marginal <- -log_const_vmcos + log_2pi + log_I_k_13 + kappa2*cos_phi
        log_proposal <- pmax(-log_2pi - log(besselI(kappa, 0)) +
                               kappa*cos_phi, max_log_prop_den-10)
        # log_proposals are >= max_log_prop_den -10 to avoid instability
        # also, proposal deviates with log density less than this bound are unlikely
        log_target_marginal - log_proposal
      }

      # maximize the log-ratio over a grid
      (logK <- max(obj_kappa_phi(kappa_opt, seq(0, 2*pi, length.out = 200))))


      # browser()
      # cat(exp(-logK))
      # rcos_unimodal(1, kappa1, kappa2, kappa3, mu1, mu2,
      #               kappa_opt, log(besselI(kappa_opt, 0, TRUE)) + kappa_opt,
      #               logK, log_const_vmcos)

      rcos_unimodal(n, kappa1, kappa2, kappa3, mu1, mu2,
                    kappa_opt, log(BESSI0_C(kappa_opt)), logK,
                    log_const_vmcos)

    }

    else if (method == "vmprop" & !unimodal_y) {

      # change the modes into [0, 2*pi]
      mode_1 <- prncp_reg(prncp_reg.minuspi.pi(mu2)+phistar)
      mode_2 <- prncp_reg(prncp_reg.minuspi.pi(mu2)-phistar)
      sin_phistar <- sin(phistar)
      unifpropn <- 1e-10
      vmpropn <- (1-1e-10)/2
      grid_0_2pi <- seq(0, 2*pi, length.out = 100)

      # proposal for y marginal = vmpropn-vmpropn-unifpropn mix
      # of vm(kappa, mu2 +- phistar) and unif(0, 2*pi)
      # and unif(0, 2*pi) (to avoid overflow of the ratio where
      # the densities are flat)
      # do minimax to find optimum kappa and K
      # abs difference between the target log marginal den
      # and log proposal den
      obj_kappa_abs <- function(kappa) {
        obj_kappa_phi <- function(phi) {
          cos_phi <- cos(phi-mu2)
          kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos_phi)
          log_I_k_13 <- log(besselI(kappa13, 0))
          log_target_marginal <- -log_const_vmcos + log_2pi + log_I_k_13 + kappa2*cos_phi
          log_proposal <-
            -log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                                  kappa*cos(phi-mode_1) +
                                  log(1+exp(kappa*(cos(phi-mode_2)
                                                   - cos(phi-mode_1))))
            ) + unifpropn) # simplified mixture density
          abs(log_target_marginal - log_proposal)
        }
        max(obj_kappa_phi(grid_0_2pi))
      }


      # minimize obj_kappa_abs wrt kappa
      (kappa_opt <- optimize(obj_kappa_abs,
                             c(0, max(kappa1, kappa2,
                                      abs(kappa3))))$minimum)


      log_prop_den <- function(kappa, phi) {
        -log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                              kappa*cos(phi-mode_1) +
                              log(1+exp(kappa*(cos(phi-mode_2) - cos(phi-mode_1)))))
                        + unifpropn)
      }


      # find max of log_prop_den
      max_log_prop_den <- max(log_prop_den(kappa_opt, grid_0_2pi))
      # any point with density < exp(max_log_prop_den)*exp(-10)
      # are less likely to appear, so make exp(max_log_prop_den)*exp(-10)
      # as the lower bound for propsal density to avoid instability


      # now maximize the log ratio of the two densities
      # w.r.t phi, given kappa = kappa_opt
      obj_kappa_phi <- function(kappa, phi) {
        cos_phi <- cos(phi-mu2)
        kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos_phi)
        log_I_k_13 <- log(besselI(kappa13, 0))
        log_target_marginal <- -log_const_vmcos + log_2pi + log_I_k_13 + kappa2*cos_phi
        log_proposal <-
          pmax(-log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
                                     kappa*cos(phi-mode_1) +
                                     log(1+exp(kappa*(cos(phi-mode_2) - cos(phi-mode_1)))))
                               + unifpropn), max_log_prop_den-10)
        # log_proposals are >= max_log_prop_den -10 to avoid instability
        # also, proposal deviates with log density less than this bound are unlikely
        log_target_marginal - log_proposal
      }

      # browser()
      #
      # target_den <- function(phi) {
      #   cos_phi <- cos(phi-mu2)
      #   kappa13 <- sqrt(kappa1^2 + kappa3^2 + 2*kappa1*kappa3*cos_phi)
      #   log_I_k_13 <- log(besselI(kappa13, 0))
      #   log_target_marginal <- -log_const_vmcos + log_2pi + log_I_k_13 + kappa2*cos_phi
      #   exp(log_target_marginal)
      # }
      #
      # prop_den <- function(kappa, phi) {
      #   cos_phi <- cos(phi)
      #   log_proposal <-
      #     pmax(-log_2pi +  log(exp(log(vmpropn) - log(besselI(kappa, 0)) +
      #                                kappa*cos(phi-mode_1) +
      #                                log(1+exp(kappa*(cos(phi-mode_2) - cos(phi-mode_1)))))
      #                          + unifpropn), max_log_prop_den-30)
      #   exp(log_proposal)
      # }
      # #
      # # integrate(gg, 0, 2*pi)
      # # integrate(function(x) pp(kappa_opt, x), 0, 2*pi)

      # maximize the log-ratio over a grid
      (logK <- max(obj_kappa_phi(kappa_opt, seq(0, 2*pi, length.out = 200))))

      exp(logK)


      # poin <- seq(0, 2*pi, length.out = 100)
      # lggp <- log(target_den(poin))
      # lppp <- log(prop_den(kappa_opt, poin))
      # # difp <- obj_kappa_phi(kappa_opt, poin)
      # #
      # plot(poin, lggp, type="l", ylim=range(lggp, lppp, lggp-lppp))
      # points(poin, lppp, type="l", col = "blue")
      # # points(poin, lggp-lppp, type="l")
      # # points(poin, difp, type="l", col = "red")
      # abline(v=c(prncp_reg(mu2+phistar), prncp_reg(mu2-phistar)))


      rcos_bimodal(n, kappa1, kappa2, kappa3, mu1, mu2,
                   kappa_opt, log(BESSI0_C(kappa_opt)), logK,
                   log_const_vmcos, mode_1, mode_2,
                   vmpropn, unifpropn)

      # rcos_unimodal(n, kappa1, kappa2, kappa3, mu1, mu2,
      #               kappa_opt, log(BESSI0_C(kappa_opt)), logK,
      #               log_const_vmcos)

      # rcos_unimodal_R(kappa1, kappa2, kappa3, mu1, mu2,
      #                 kappa_opt, log(BESSI0_C(kappa_opt)), logK,
      #                 log_const_vmcos)
    }

  }

  else if (method == "naive") {
    opt_obj <- function(k1=1, k2=1, k3=0, mu1=0, mu2=0) {
      obj <- optim(c(0,0), fn = function(x) -(k1*cos(x[1]-mu1)+k2*cos(x[2]-mu2)+k3*cos(x[1]-x[2]-mu1+mu2)),
                   gr = function(x)  -c(-k1*sin(x[1]-mu1)-k3*sin(x[1]-x[2]-mu1+mu2),
                                        -k2*sin(x[2]-mu2)+k3*sin(x[1]-x[2]-mu1+mu2)))
      -obj$value

    }
    upper_bd <- opt_obj(kappa1, kappa2, kappa3, mu1, mu2)
    # qrnd_grid <- sobol(1e4, 2, FALSE)
    # cat(exp(log(const_vmcos(kappa1, kappa2, kappa3,
    #                         qrnd_grid))
    #         - upper_bd)/(4*pi^2))
    rcos_onepar(n, kappa1, kappa2, kappa3, mu1, mu2, upper_bd)
  }

}

process_startpar <- function(start_par, data,
                             ncomp, model,
                             rand_start) {
  if (is.null(start_par) & !rand_start) {
    starting <-
      do.call(paste0("start_clus_kmeans_", model),
              list(data.full=data, comp=ncomp))
  }

  else if (is.null(start_par) & rand_start) {
    starting <-
      do.call(paste0("start_clus_rand_", model),
              list(data.full=data, comp=ncomp,nstart=5))
  }

  else {

    if (model %in% c("vmsin", "vmcos", "wnorm2")) {
      allpar <- start_par
      if (any(is.null(allpar$kappa1), is.null(allpar$kappa2), is.null(allpar$kappa3),
              is.null(allpar$mu1), is.null(allpar$mu2)) ) {
        stop("too few elements in start_par, with no default")
      }

      allpar1 <- list(allpar$kappa1, allpar$kappa2, allpar$kappa3, allpar$mu1, allpar$mu2)
      allpar_len <- listLen(allpar1)
      if (min(allpar_len) != max(allpar_len)){
        stop("component size mismatch: number of components of in the starting parameter vectors differ")
      }

      starting <- list(par.mat = rbind(start_par$kappa1, start_par$kappa2,
                                       start_par$kappa3,
                                       start_par$mu1,
                                       start_par$mu2))
    }
    else if (model %in% c("vm", "wnorm")) {
      allpar <- start_par
      if (any(is.null(allpar$kappa), is.null(allpar$mu)) ) {
        stop("too few elements in start_par, with no default")
      }

      allpar1 <- list(allpar$kappa, allpar$mu)
      allpar_len <- listLen(allpar1)
      if (min(allpar_len) != max(allpar_len)){
        stop("component size mismatch: number of components of in the starting parameter vectors differ")
      }

      starting <- list(par.mat = rbind(start_par$kappa,
                                       start_par$mu))
    }

    if (ncomp == 1) {
      starting$pi.mix <- 1
    }
    else {
      starting$pi.mix <- start_par$pmix
    }
  }

  starting
}

#' Contourplot for bivariate angular mixture model densities
#' @param model bivariate angular model whose mixture is of interest. Must be one of
#' "vmsin", "vmcos" and "wnorm2".
#' @param kappa1,kappa2,kappa3,mu1,mu2,pmix model parameters and mixing
#' proportions. See the respective mixture model densities (\link{dvmsinmix}, \link{dvmcosmix},
#' \link{dwnorm2mix}) for more details.
#' @param levels numeric vector of levels at which to draw contour lines;
#' passed to the \link{contour} function in graphics.
#' @param nlevels	number of contour levels desired \strong{if} levels is not supplied;
#' passed to the \link{contour} function in graphics.
#' @param xpoints Points on the first (x-) coordinate where the density is to be evaluated.
#' Default to seq(0, 2*pi, length.out=100).
#' @param ypoints Points on the first (x-) coordinate where the density is to be evaluated.
#' Default to seq(0, 2*pi, length.out=100).
#' @param ... additional model specific argment
#' @param xlab,ylab,col,lty,main graphical parameters passed to \link{contour}.
#' @examples
#' contour_model('vmsin', 1, 1, 1.5, pi, pi)
#' contour_model('vmcos', 1, 1, 1.5, pi, pi)
#'
#'
#' @export

contour_model <- function(model = "vmsin", kappa1, kappa2, kappa3, mu1, mu2,
                          pmix = rep(1/length(kappa1), length(kappa1)),
                          xpoints = seq(0, 2*pi, length.out = 100),
                          ypoints = seq(0, 2*pi, length.out = 100),
                          levels, nlevels = 20,
                          xlab="x", ylab="y", col="black",
                          lty=1, main, ...)

{
  if (!model %in% c("vmsin", "vmcos", "wnorm2"))
    stop("model must be one of \"vmsin\", \"vmcos\" or \"wnorm2\"")

  if (missing(levels)) {
    levels <- exp(seq(-20,2, length.out = nlevels))
  }

  if (missing(main))
    main <- paste("Contour plot for", length(kappa1), "component",
                  model, "mixture density")

  coords <- as.matrix(expand.grid(xpoints, ypoints))
  inargs <- list(x = coords, kappa1 = kappa1, kappa2 = kappa2, kappa3 = kappa3,
                 mu1 = mu1, mu2 = mu2, pmix = pmix, ...)
  dens <- do.call(paste0("d", model, "mix"), inargs)

  contour(xpoints, ypoints, matrix(dens, nrow=length(xpoints)), levels=levels,
          xlab=xlab, ylab=ylab, col=col, lty=lty, main=main)

}



#' Surface for bivariate angular mixture model densities
#' @inheritParams contour_model
#' @param kappa1,kappa2,kappa3,mu1,mu2,pmix model parameters and mixing
#' proportions. See the respective mixture model densities (\link{dvmsinmix}, \link{dvmcosmix},
#' \link{dwnorm2mix}) for more details.
#' @param log.density logical. Should log density be used for the plot?
#' @param xlab,ylab,zlab,main graphical parameters passed to \code{lattice::wireframe}
#' @param ... additional arguments passed to \code{lattice::wireframe}
#' @examples
#' surface_model('vmsin', 1, 1, 1.5, pi, pi)
#' surface_model('vmcos', 1, 1, 1.5, pi, pi)
#'
#' @export

surface_model <- function(model = "vmsin", kappa1, kappa2, kappa3, mu1, mu2,
                          pmix = rep(1/length(kappa1), length(kappa1)),
                          xpoints = seq(0, 2*pi, length.out = 30),
                          ypoints = seq(0, 2*pi, length.out = 30),
                          log.density = FALSE, xlab="x", ylab="y",
                          zlab = ifelse(log.density, "Log Density", "Density"),
                          main, ...)
{
  if (!model %in% c("vmsin", "vmcos", "wnorm2"))
    stop("model must be one of \"vmsin\", \"vmcos\" or \"wnorm2\"")

  if (missing(main))
    main <- paste("Density surface for", length(kappa1), "component",
                  model, "mixture density")


  coords <- as.matrix(expand.grid(xpoints, ypoints))
  inargs <- list(x = coords, kappa1 = kappa1, kappa2 = kappa2,
                 kappa3 = kappa3,
                 mu1 = mu1, mu2 = mu2, pmix = pmix)
  dens <- do.call(paste0("d", model, "mix"), inargs)

  denmat <- matrix(dens, nrow = length(xpoints))

  if(log.density) {
    denmat <- log(denmat)
  }

  print(basic_surfaceplot(xpoints = xpoints, ypoints = ypoints,
                          denmat = denmat, xlab = xlab, ylab = ylab,
                          main = main, zlab = zlab, ...))

  # nrden <- nrow(denmat)
  # ncden <- ncol(denmat)
  #
  # denfacet <- denmat[-1, -1] + denmat[-1, -ncden] +
  #   denmat[-nrden, -1] + denmat[-nrden, -ncden]
  #
  # # Create a function interpolating colors in the range of specified colors
  # jet.colors <- grDevices::colorRampPalette( c("blue", "green",
  #                                              "yellow", "orange", "red") )
  # # Generate the desired number of colors from this palette
  # nbcol <- 500
  # color <- jet.colors(nbcol)
  #
  # denfacet <- denmat[-1, -1] + denmat[-1, -ncden] +
  #   denmat[-nrden, -1] + denmat[-nrden, -ncden]
  # # Recode facet z-values into color indices
  # facetcol <- cut(denfacet, nbcol)
  #
  # persp(x=xpoints, y=ypoints, z=denmat, theta = theta, phi = phi,
  #       expand = expand, col = color[facetcol],
  #       ltheta = 120, shade = shade, ticktype = "detailed",
  #       xlab = xlab, ylab = ylab, zlab = zlab,
  #       main = main, ...) -> res

  # wireframe(x=denmat~x*y,
  #           data=data.frame(x=xpoints,
  #                           y=rep(ypoints, each=length(xpoints)),
  #                           denmat=denmat),
  #           xlab=xlab, ylab=ylab, zlab=zlab,
  #           main=main, colorkey=colorkey, outerbox=FALSE,
  #           par.settings=list(axis.line=list(col = 'transparent')),
  #           neval=length(xpoints), aspect = c(61/87, 0.4),
  #           drape=TRUE, light.source=c(10,0,10),
  #           col.regions=colorRampPalette(c("blue", "green",
  #                                          "yellow", "orange",
  #                                          "red"))(100),
  #           scales=list(arrows=FALSE, col=1)
  # )
  # do.call(wireframe, inargs)
  # inargs$outerbox <- FALSE
  # inargs$par.settings <- list(axis.line = list(col = 'transparent'))
  # if (is.null(inargs$xlab)) inargs$xlab <- expression(Theta)
  # if (is.null(inargs$ylab)) inargs$ylab <- expression(Phi)
  # if (is.null(inargs$colorkey)) inargs$colorkey <- FALSE
  # if (is.null(inargs$main)) inargs$main <- list(maintext, font=2, cex=1.5)
  # if (is.null(inargs$neval)) inargs$neval <- 100
  # if (is.null(inargs$aspect)) inargs$aspect <- c(61/87, 0.4)
  # if (is.null(inargs$zlab)) inargs$zlab <- list("Density", rot=90)
  # if (is.null(inargs$screen)) inargs$screen <- list(z=45,x=-45)
  # if (is.null(inargs$colorkey)) inargs$colorkey <- TRUE
  # if (is.null(inargs$scales))
  #   inargs$scales <- list(
  #     arrows = FALSE,
  #     x=list(at = c(-pi, -pi/2, 0, pi, pi/2),
  #            labels = c(expression(-pi), expression(-pi/2), 0,
  #                       expression(pi), expression(pi/2))),
  #     y=list(at = c(-pi, -pi/2, 0, pi, pi/2),
  #            labels = c(expression(-pi), expression(-pi/2), 0,
  #                       expression(pi), expression(pi/2))),
  #     col = "black", font = 1, tck=1)
  # if (is.null(inargs$drape)) inargs$drape <- TRUE
  # if (is.null(inargs$light.source))
  #   inargs$light.source <- c(10,0,10)
  # if (is.null(inargs$col.regions))
  #   inargs$col.regions <- colorRampPalette(c("blue", "green",
  #                                            "yellow", "orange", "red"))(100)
  # if (is.null(inargs$par.settings))
  #   inargs$par.settings <- list(top.padding = 0,
  #                               bottom.padding = 0,
  #                               left.padding=0,
  #                               right.padding=0)
  #
  # inargs$par <- c(k1,k2,k3,0,0)
  #
  #
}

#' Wrap angles into \code{[-pi, pi]} or \code{[0, 2*pi]}
#' @param x numeric vector or matrix or data.frame.
#' @details
#' \code{minuspi_to_pi} wraps \code{x} into \code{[-pi, pi]},
#' while \code{zero_to_pi} wraps \code{x} into \code{[0, 2*pi]}.
#'
#' @examples
#' dat <- matrix(runif(100, -pi, pi), ncol=2)
#' dat1 <- zero_to_2pi(dat)
#' dat2 <- minuspi_to_pi(dat1)
#' all.equal(dat, dat2)
#' @export

zero_to_2pi <- function(x)
{
  x %% (2*pi)
}


#' @rdname zero_to_2pi
#' @export

minuspi_to_pi <- function(x)
{
  prncp_reg.minuspi.pi(x)
}

#' Contour plot for angmcmc objects with bivariate data
#'
#' @inheritParams pointest
#' @param type Passed to \link{d_fitted}. Possible choices are "point-est" and "post-pred".
#' @param x angular MCMC object (with bivariate data).
#' @param show.data logical. Should the data points be added to the contour plot? Ignored if `object` is NOT supplied.
#' @param cex,col,pch graphical parameters passed to \link{points} from graphics for plotting the data points.
#' Ignored if `show.data == FALSE`.
#' @param alpha color transparency for the data points, implemented via \link[scales]{alpha} from package `scales`.
#' Ignored if `show.data == FALSE`.
#' @inheritParams contour_model
#' @param ... additional arguments to be passed to the function \link{contour}.
#' @md
#'
#' @details
#' `contour.angmcmc` is an S3 function for `angmcmc` objects that calls \link{contour} from graphics.
#'
#' To estimate the mixture density required to construct the contour plot, first the parameter vector \eqn{\eta} is estimated
#' by applying \code{fn} on the MCMC samples, yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}. Then the mixture density
#' \eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by \eqn{f(x|\hat{\eta})}.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # now create a contour plot
#' contour(fit.vmsin.20)
#'
#' @export

contour.angmcmc <-  function(x, fn = "MAP", type = "point-est", show.data = TRUE,
                             xpoints = seq(0, 2*pi, length.out = 100),
                             ypoints = seq(0, 2*pi, length.out = 100),
                             levels, nlevels = 20,
                             cex = 1, col = "red", alpha = 0.4,
                             pch = 19, ...)
{
  object <- x

  if (!is.angmcmc(object))
    stop("\'x\' must be an angmcmc object")

  if(x$type != "bi") stop("\"x\" is not a bivariate angmcmc object")

  if(missing(levels)) {
    levels <- exp(seq(-20,2, length.out = nlevels))
  }

  dots <- list(...)

  colnames_data <- colnames(x$data)

  if(is.null(colnames_data)) {
    xlab <- ylab <- ""
  } else {
    xlab <- colnames_data[1]
    ylab <- colnames_data[2]
  }

  if(x$ncomp > 1) {
    main <- paste("Contour plot for fitted", x$ncomp, "component", x$model, "mixtures")
  } else {
    main <- paste("Contour plot for fitted (single component)", x$model)
  }

  coords <- as.matrix(expand.grid(xpoints, ypoints))
  dens <- d_fitted(coords, x, fn = fn, type = type)

  contour_in <- c(
    list(
      x = xpoints,
      y = ypoints,
      z = matrix(dens, nrow=length(xpoints)),
      levels = levels
    ),
    dots
  )

  if (is.null(dots$xlab)) contour_in$xlab <- xlab
  if (is.null(dots$ylab)) contour_in$ylab <- ylab
  if (is.null(dots$main)) contour_in$main <- main

  do.call(contour, contour_in)

  if(show.data) points(x$data, col = scales::alpha(col, alpha),
                       cex = cex, pch = pch)
}


#
# panel_wireframe_cloud <- function(x, y, z, x2, y2, z2,...) {
#   panel.wireframe(x, y, z,...)
#   panel.cloud(x2, y2, z2,...)
# }


#' Density plots for angmcmc objects
#' @inheritParams contour.angmcmc
#' @description Plot fitted angular mixture model density surfaces or curves.
#' @inheritParams pointest
#' @param x angmcmc object.
#' @param data unused. The parameter is already filled with results from fitted angular model. It is kept
#' to ensure compatibility with the lattice S3 generic \code{densityplot}.
#' @param plot logical. Should the density surface (if the fitted data is bivariate) or the density
#' curve (if univariate) be plotted?
#' @param log.density logical. Should log density be used for the plot?
#' @param ... additional arguments passed to \code{lattice::wireframe} if
#' fitted data is bivariate, or to \link{hist} (if (\code{show.hist == TRUE})), if the fitted data is univariate
#' @param show.hist logical. Should a histogram for the data
#' points be added to the plot, if the fitted data is univariate? Ignored if data is
#' bivariate.
#' @param xlab,ylab,zlab,main graphical parameters passed to \code{lattice::wireframe} (if
#' bivariate) or \link{plot} (if univariate). If the data is univariate, \code{zlab} and \code{ylab} can be
#' used interchangeably (both correspond to the density).
#' @param xpoints,ypoints Points on the  x and y coordinates (if bivariate) or only x coordinate
#' (if univariate) where the density is to be evaluated. Each defaults to seq(0, 2*pi, length.out=100).
#'
#' @details
#' When \code{plot==TRUE}, \code{densityplot.angmcmc} calls \code{lattice::wireframe} or
#' \link{plot} from graphics to draw the surface or curve.
#'
#' To estimate the mixture density, first the parameter vector \eqn{\eta} is estimated
#' by applying \code{fn} on the MCMC samples, yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}. Then the mixture density
#' \eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by \eqn{f(x|\hat{\eta})}.
#'
#'
#' Note that \code{densityplot.angmcmc} \strong{does not} plot the kernel densitie estimates
#' of the MCMC parameters. (These plots can be obtained by first converting an \code{angmcmc}
#' object to an \code{mcmc} object via \link[coda]{as.mcmc.list}, and then
#' by using \code{densplot} from package coda on the resulting \code{mcmc.list} object. Instead,
#' \code{densityplot.angmcmc} returns the surface (if 2-D) or the curve (if 1-D)
#' of the fitted model density evaluated at the estimated parameter vector (obtain through \link{pointest}).
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # now create density surface with the default first 1/3 as burn-in and thin = 1
#' library(lattice)
#' densityplot(fit.vmsin.20)
#' # the viewing angles can be changed through the argument 'screen'
#' # (passed to lattice::wireframe)
#' densityplot(fit.vmsin.20, screen = list(z=-30, x=-60))
#' densityplot(fit.vmsin.20, screen = list(z=30, x=-60))
#' # the colors can be changed through 'col.regions'
#' cols <- grDevices::colorRampPalette(c("blue", "green",
#'                                       "yellow", "orange", "red"))(100)
#' densityplot(fit.vmsin.20, col.regions = cols)
#'
#' # Now fit a vm mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vm.20 <- fit_vmmix(wind$angle, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' densityplot(fit.vm.20)
#'
#' @importFrom lattice densityplot wireframe
#'
#' @export

densityplot.angmcmc <- function(x,
                                data = NULL,
                                fn = mean, type = "point-est", log.density = FALSE,
                                xpoints=seq(0, 2*pi, length.out=35),
                                ypoints=seq(0, 2*pi, length.out=35),
                                plot=TRUE,
                                show.hist=ifelse(log.density, FALSE, TRUE),
                                xlab, ylab,
                                zlab = ifelse(log.density, "Log Density", "Density"),
                                main,
                                ...)
{
  if(!is.angmcmc(x))
    stop("\"x\" must be an angmcmc object")

  object <- x

  if (object$type == "bi") {

    colnames_data <- colnames(object$data)

    if (any(missing(xlab), missing(ylab))) {
      if (is.null(colnames_data)) {
        xlab <- ylab <- ""
      } else {
        xlab <- colnames_data[1]
        ylab <- colnames_data[2]
      }
    }

    if (missing(main)) {
      if(x$ncomp > 1) {
        main <- paste("Contour plot for fitted", x$ncomp, "component", x$model, "mixtures")
      } else {
        main <- paste("Contour plot for fitted (single component)", x$model)
      }
    }

    coords <- as.matrix(expand.grid(xpoints, ypoints))
    den <- d_fitted(coords, object, fn = fn, type = type)


    denmat <- matrix(den, nrow=length(xpoints))

    if(log.density) {
      denmat <- log(denmat)
    }


    out <- list(x=xpoints, y=ypoints, density=denmat)

    if (plot) {
      nrden <- nrow(denmat)
      ncden <- ncol(denmat)

      if(object$ncomp > 1) {
        main <- paste("Density surface for fitted", object$ncomp, "component",
                      object$model, "mixtures")
      } else {
        main <- paste("Density surface for fitted (single component)", object$model)
      }

      # # Create a function interpolating colors in the range of specified colors
      # jet.colors <- grDevices::colorRampPalette( c("blue", "green",
      #                                              "yellow", "orange", "red") )
      # # Generate the desired number of colors from this palette
      # nbcol <- 500
      # color <- jet.colors(nbcol)

      # denfacet <- denmat[-1, -1] + denmat[-1, -ncden] +
      #   denmat[-nrden, -1] + denmat[-nrden, -ncden]
      # Recode facet z-values into color indices
      # facetcol <- cut(denfacet, nbcol)

      print(basic_surfaceplot(xpoints = xpoints, ypoints = ypoints,
                              denmat = denmat, xlab = xlab, ylab = ylab,
                              main = main, zlab = zlab, ...))

      # persp(x=xpoints, y=ypoints, z=denmat, theta = theta, phi = phi, expand = expand, col = color[facetcol],
      #       ltheta = 120, shade = shade, ticktype = "detailed",
      #       xlab = xlab, ylab = ylab, zlab = zlab,
      #       main = main, ...) -> res



      # inargs <- list(...)
      # inargs$x <- denmat~x*y
      # inargs$data <- data.frame(x = xpoints,
      #                           y=rep(ypoints, each=length(xpoints)),
      #                           denmat=denmat)
      # inargs$outerbox <- FALSE
      # inargs$par.settings <- list(axis.line = list(col = 'transparent'))
      # if (is.null(inargs$xlab)) inargs$xlab <- xlab
      # if (is.null(inargs$ylab)) inargs$ylab <- ylab
      # if (is.null(inargs$colorkey)) inargs$colorkey <- FALSE
      # if (is.null(inargs$main)) inargs$main <- main
      # if (is.null(inargs$neval)) inargs$neval <- 100
      # if (is.null(inargs$aspect)) inargs$aspect <- c(61/87, 0.4)
      # if (is.null(inargs$zlab)) inargs$zlab <- list("Density", rot=90)
      # if (is.null(inargs$screen)) inargs$screen <- list(z=45, x=-45)
      # if (is.null(inargs$colorkey)) inargs$colorkey <- FALSE
      # if (is.null(inargs$scales))
      #   inargs$scales <- list(arrows=FALSE, col=1)
      # if (is.null(inargs$drape)) inargs$drape <- TRUE
      # if (is.null(inargs$light.source))
      #   inargs$light.source <- c(10,0,10)
      # if (is.null(inargs$col.regions))
      #   inargs$col.regions <- colorRampPalette(c("blue", "green",
      #                                            "yellow", "orange", "red"))(100)
      # if (is.null(inargs$par.settings))
      #   inargs$par.settings <- list(top.padding = 0,
      #                               bottom.padding = 0,
      #                               left.padding=0,
      #                               right.padding=0,
      #                               axis.line=list(col = 'transparent'))
      # do.call(wireframe, inargs)
    }

    invisible(out)
  }
  else {

    den <- d_fitted(xpoints, object, fn = fn, type = type)

    if (log.density) den <- log(den)

    out <- list(x=xpoints, density=den)

    if (plot) {
      if(show.hist){
        histplot <-  hist(object$data, plot = FALSE, ...)
      } else {
        histplot <- NULL
      }

      if (missing(main)) {
        if(object$ncomp > 1) {
          main <- paste("Density plot for fitted", object$ncomp, "component", object$model, "mixtures")
        } else {
          main <- paste("Density plot for fitted (single component)", object$model)
        }
      }

      if (missing(xlab))
        xlab <- "Angles in radians"

      if (missing(ylab)) {
        ylab <- zlab
      }

      y_max <- 1.1* max(den, histplot$density)
      plot(NULL, xlim=range(xpoints), ylim=c(0, y_max), xlab = xlab,
           ylab=ylab, main=main)
      points(xpoints, den, type = "l")

      if(show.hist) plot(histplot, freq = FALSE, add = TRUE, ...)


      title(main = main)
    }
  }
  invisible(out)
}

extract_digits <- function(x) {
  tmp <- regmatches(
    x,
    gregexpr("[[:digit:]]+", x)
  )[[1]]

  out <- if (length(tmp) > 0) {
    strsplit(
      tmp,
      ""
    )[[1]]
  } else ""

  out
}

# get_unicode <- function(x) {
#   subscript_char <- extract_digits(x)
#   subscript_char_unicode <- if (subscript_char != "") {
#     paste0(
#       sapply(
#         subscript_char,
#         function(this_digit) {
#           unicode_num <- 2080 + as.numeric(this_digit)
#           unicode_num_char <- eval(parse(text = paste0("'\\u", unicode_num, "'")))
#         }
#       ),
#       collapse = ""
#     )
#   } else ""
#
#   parameter_greek <- if (grepl("mu", x)) {
#     "\u03BC"
#   } else if (grepl("kappa", x)) {
#     "\u03BA"
#   } else if (grepl("pmix", x)) {
#     "p"
#   }
#
#   out <- paste0(parameter_greek, subscript_char_unicode)
#
#   if (grepl("pmix", x)) {
#     # append superscript 'mix'
#     out <- paste0(out, "\u1D50", "\u2071", "\u02E3")
#   }
#
#   out
# }

get_bquote_expr <- function(x) {
  subscript_char <- extract_digits(x)
  param_char <- gsub(subscript_char, "", x)
  out <- if (grepl("pmix", x)) {
    "p['mix']"
  } else {
    paste0(param_char, "[", subscript_char, "]")
  }
  out
}

parse_text <- function(x) {
  eval(parse(text = x))
}

parse_text_bquote <- function(x) {
  tmp <- paste0("bquote(", x, ")")
  parse_text(tmp)
}

#' Trace plot for parameters from an angmcmc object
#' @inheritParams pointest
#' @param object angular MCMC object.
#' @param par parameter for which trace plot is to be created.
# #' @param press.enter logical. Should the next plot in the series
# #' be shown after you press "Enter"? Ignored if only a single plot
# #' is to be created.
#' @param ... unused
#' @return
#' Returns a single plot if a single \code{par} and a single \code{comp.label} is supplied.
#' Otherwise, a series of plots is produced.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # trace plot for kappa1 in component 1
#' paramtrace(fit.vmsin.20, "kappa1", 1)
#' # for kappa1 in all components
#' paramtrace(fit.vmsin.20, "kappa1")
#' # for all parameters in component 1
#' paramtrace(fit.vmsin.20, comp.label = 1)
#'
#' @importFrom RColorBrewer brewer.pal
#'
#' @export

paramtrace <- function(object, par.name, comp.label, chain.no,
                       ...)
{
  if(!class(object) %in% "angmcmc") stop("\'object\' must be an angmcmc object")

  ell <- list(...)

  if (!is.null(ell$press.enter))
    warning("\'press.enter\' is deprecated. Instead use par(ask=TRUE) before calling paramtrace.")
  if (!is.null(ell$burnin))
    warning("Use of burnin is deprecated in postprocessing. Use \'burnin.prop\' during original MCMC run instead.")
  if (!is.null(ell$thin))
    warning("Use of thin is deprecated in postprocessing. Use \'thin\' during original MCMC run instead.")

  if (missing(par.name)) {
    par.name <- object$par.name
  }  else if (any(!par.name %in% object$par.name)) {
    stop("invalid par.name")
  }

  if (missing(comp.label)) {
    comp.label <- 1:object$ncomp
  } else if (any(!comp.label %in% 1:object$ncomp)) {
    stop("invalid component label")
  }

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }


  if(any(c(length(par.name), length(comp.label)) > 1)) {
    singleplot <- FALSE
  } else {
    singleplot <- TRUE
  }

  n.chains <- length(chain.no)
  col_brew <- brewer.pal(n = max(n.chains, 3), name = "Pastel2")

  if (singleplot) {

    val <- extractsamples(object, par.name, comp.label, chain.no, drop = FALSE)
    plot(NULL, xlim = c(1, object$n.iter.final), ylim = range(val),  ylab="", xlab = "Iteration")
    for(ch in 1:n.chains)
      points(val[, , , ch], type = "l", col = col_brew[ch])
    legend("bottomright", legend = paste("Chain", chain.no), col = col_brew,
           lty = 1)

    par.name.1 <- sapply(par.name, get_bquote_expr)
    if(object$ncomp > 1) {
      ylab <- paste0(par.name.1, "~'for component ", comp.label, "'")
      main <- paste0("'Traceplot for'~", ylab, "~'in ",
                     object$ncomp, "-component ", object$model, " mixtures'")
    } else {
      ylab <- par.name.1
      main <- paste0("Traceplot for'~", ylab, "~'in (single component) ", object$model, "'")
    }
    title(main = parse_text_bquote(main), ylab = parse_text_bquote(ylab))
  }

  # not singleplot
  else {
    nplots <- length(par.name) * length(comp.label)
    currplotno <- 1L
    for(par.curr in par.name) {
      for(comp.label.curr in comp.label) {
        val <- extractsamples(object, par.curr, comp.label.curr, chain.no, drop = FALSE)
        plot(NULL, xlim = c(1, object$n.iter.final), ylim = range(val),  ylab="", xlab = "Iteration")
        for(ch in 1:n.chains)
          points(val[, , , ch], type = "l", col = col_brew[ch])

        legend("bottomright", legend = paste("Chain", chain.no), col = col_brew,
               lty = 1)

        par.curr.1 <- sapply(par.curr, get_bquote_expr)
        if(object$ncomp > 1) {
          ylab <- paste(par.curr.1, "~' for component", comp.label.curr, "'")
          main <- paste0("'Traceplot for'~", ylab, "*' in ", object$ncomp, "-component ", object$model, " mixtures'")
        } else {
          ylab <- par.curr.1
          main <- paste0("'Traceplot for'~", ylab, "*' in  (single component)", object$model, "'")
        }

        title(main = parse_text_bquote(main), ylab = parse_text_bquote(ylab))

        # --not required--
        # if(currplotno < nplots) {
        #   if(press.enter) {
        #     press_enter()
        #     frame()
        #   }
        # }

        currplotno <- currplotno + 1
      }
    }
  }
}


#' Trace and autocorrelation plots of log posterior density or log likelihood from an angmcmc object
#' @inheritParams paramtrace
#' @param object angular MCMC object.
#' @param use.llik logical. Should log likelihood be plotted instead of log posterior? Set
#' to \code{FALSE} by default.
#' @param plot.autocor logical. Should the autocorrelations be plotted as well?
#' @param lag.max maximum lag for autocorrelation.  Passed to \link{acf}. Ignored if
#' \code{plot.autocor = FALSE}.
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # log posterior density trace
#' lpdtrace(fit.vmsin.20)
#' # log likelihood trace
#' lpdtrace(fit.vmsin.20, use.llik = TRUE)
#'
#' @export

lpdtrace <- function(object, chain.no, use.llik = FALSE,
                     plot.autocor = FALSE,
                     # press.enter = TRUE,
                     lag.max = NULL, ...)
{
  if (!is.angmcmc(object)) stop("lpdtrace can only be used for \'angmcmc\' objects")

  ell <- list(...)

  if (!is.null(ell$press.enter))
    warning("\'press.enter\' is deprecated. Instead use par(ask=TRUE) before calling lpdtrace.")
  if (!is.null(ell$burnin))
    warning("Use of burnin is deprecated in postprocessing. Use \'burnin.prop\' during original MCMC run instead.")
  if (!is.null(ell$thin))
    warning("Use of thin is deprecated in postprocessing. Use \'thin\' during original MCMC run instead.")


  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }


  n.chains <- length(chain.no)
  final_iter_set <- object$final_iter
  col_brew <- brewer.pal(n = max(n.chains, 3), name = "Paired")

  if (use.llik) {
    val <- object$llik[final_iter_set, chain.no, drop = FALSE]

    plot(NULL, xlim = c(1, object$n.iter.final), ylim = range(val),
         ylab="Log Likelihood", xlab = "Iteration")
    for(ch in 1:n.chains)
      points(val[, ch], type = "l", col = col_brew[ch])
    legend("bottomright", legend = paste("Chain", chain.no), col = col_brew,
           lty = 1)

    if(object$ncomp > 1) {
      main <- paste("Log likelihood traceplot for ", object$ncomp, "component", object$model, "mixtures")
    } else {
      main <- paste("Log likelihood traceplot for (single component)", object$model)
    }

    title(main = main)



    # if (press.enter & plot.autocor) {
    #   press_enter()
    #   frame()
    # }

    if (plot.autocor) {
      all_autocors <- lapply(1:n.chains,
                             function(ch) acf(val[, ch], lag.max = lag.max, plot = FALSE))

      range_y <- range(unlist(lapply(all_autocors, function(j) j$acf)))
      for(ch in 1:n.chains) {
        acr <- all_autocors[[ch]]
        if (ch  == 1) {
          plot(acr$acf, type = "b", col = col_brew[ch],
               ylab="Log likelihood autocorrelation",
               xlab = "Lag", pch = 16, ylim = range_y)
        } else {
          points(acr$acf, type = "b", col = col_brew[ch], pch = 16)
        }
      }

      abline(h = 0, lty = 2)
      legend("topright", legend = paste("Chain", chain.no), col = col_brew,
             pch = 16, lty = 1)

      if(object$ncomp > 1) {
        main <- paste("Log likelihood autocorrelation plot for ", object$ncomp, "component", object$model, "mixtures")
      } else {
        main <- paste("Log likelihood autocorrelation plot for (single component)", object$model)
      }

      title(main = main)

    }



  } else {

    val <- object$lpd[final_iter_set, chain.no, drop = FALSE]

    plot(NULL, xlim = c(1, object$n.iter.final), ylim = range(val),
         ylab="Log Posterior Density", xlab = "Iteration")
    for(ch in 1:n.chains)
      points(val[, ch], type = "l", col = col_brew[ch])
    legend("bottomright", legend = paste("Chain", chain.no), col = col_brew,
           lty = 1)

    if(object$ncomp > 1) {
      main <- paste("Log Posterior Density traceplot for", object$ncomp, "component", object$model, "mixtures")
    } else {
      main <- paste("Log Posterior Density traceplot fitted (single component)", object$model)
    }

    title(main = main)


    # if (press.enter & plot.autocor) {
    #   press_enter()
    #   frame()
    # }

    if (plot.autocor) {

      all_autocors <- lapply(1:n.chains,
                             function(ch) acf(val[, ch], lag.max = lag.max, plot = FALSE))

      range_y <- range(unlist(lapply(all_autocors, function(j) j$acf)))
      for(ch in 1:n.chains) {
        acr <- all_autocors[[ch]]
        if (ch  == 1) {
          plot(acr$acf, type = "b", col = col_brew[ch],
               ylab="Log posterior autocorrelation",
               xlab = "Lag", pch = 16, ylim = range_y)
        } else {
          points(acr$acf, type = "b", col = col_brew[ch], pch = 16)
        }
      }

      abline(h = 0, lty = 2)
      legend("topright", legend = paste("Chain", chain.no), col = col_brew,
             pch = 16, lty = 1)

      if(object$ncomp > 1) {
        main <- paste("Log posterior autocorrelation plot for ", object$ncomp, "component", object$model, "mixtures")
      } else {
        main <- paste("Log posterior autocorrelation plot for (single component)", object$model)
      }

      title(main = main)

    }

  }



}


#' Summary plots for angmcmc objects
#' @inheritParams paramtrace
#' @inheritParams lpdtrace
#' @param do.paramtrace logical. Should the trace(s) for the
#' parameter(s) be plotted?
#' @param do.lpdtrace logical. Should the log posterior trace
#' be plotted?
#' @param use.llik logical. Should the log likelihood be plotted
#' instead? Ignored if \code{do.lpdtrace == FALSE}.
#' @param x angmcmc object
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' plot(fit.vmsin.20)
#' @export

plot.angmcmc <- function(x, par.name, comp.label, chain.no,
                         do.paramtrace = TRUE,
                         do.lpdtrace = TRUE, use.llik = FALSE,
                         ...)
{
  if (!is.angmcmc(x))
    stop("\'x\' must be an angmcmc object")

  ell <- list(...)
  if (!is.null(ell$press.enter))
    warning("\'press.enter\' is deprecated. Instead use par(ask=TRUE) before calling plot.angmcmc.")

  if (do.paramtrace)
    paramtrace(x, par.name, comp.label, chain.no, ...)

  # if (press.enter) {
  #   press_enter()
  #   frame()
  # }

  if (do.lpdtrace)
    lpdtrace(x, chain.no, use.llik)
}

#' Maximum likelihood estimation of bivariate von Mises parameters
#' @inheritParams fit_angmix
#' @param model Bivariate von Mises model. One of "vmsin", "vmcos" or "indep".
#' @param ... Additional arguments. See details.
#' @details The parameters \code{kappa1} and \code{kappa2} are optimized
#' in log scales. The method of optimization used (passed to \link{optim})
#' can be specified through \code{method} in \code{...}
#' (defaults to \code{"L-BFGS-B"}). Note, however, that
#' lower (0)  and upper (2*pi) bounds for \code{mu1} and \code{mu2}
#' are specified; so not all methods implemented in \link{optim} will work.
#' @return An object of class \link{mle-class}.
#' @examples
#' pars <- list(kappa1 = 3, kappa2 = 2, kappa3 = 1.5, mu1 = 0.5, mu2 = 1.5)
#' nsamp <- 2000
#' model <- "vmsin"
#' set.seed(100)
#' dat_gen <- do.call(paste0("r", model), c(list(n = nsamp), pars))
#'
#' est <- vm2_mle(dat_gen, model = model)
#' library(stats4)
#' coef(est)
#' vcov(est)
#' @export
vm2_mle <- function(data, model = c("vmsin", "vmcos"), ...) {

  model <- model[1]
  dots <- list(...)
  data <- data.matrix(data)
  call <- match.call()

  if (is.null(dots$method)) {
    dots$method <- "L-BFGS-B"
  }
  method <- dots$method

  if (model == "vmsin") {

    # in log scale
    lpd_grad_model_indep_1comp <-  function(par_vec_lscale) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      lpd_grad <- matrix(NA, 6, 1)

      lpd_grad <- signif(
        suppressWarnings(grad_llik_vmsin_C(data, par_vec))*
          c(par_vec[1:2], rep(1, 4)),
        8
      )
      list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    }

    start_par_gen <- start_par_vmsin

    hessian_fn <- function(par_vec) {
      numDeriv::hessian(
        func = function(par_vec) {
          -grad_llik_vmsin_C(data, par_vec)[6]
        },
        x = par_vec
      )
    }

  }

  else if (model == "vmcos") {

    ell <- dots[c("qrnd_grid", "n_qrnd")] #list(qrnd_grid = qrnd, n_qrnd = n_qrnd)

    if (!is.null(ell$qrnd)) {
      qrnd_grid <- ell$qrnd
      dim_qrnd <- dim(qrnd_grid)
      if (!is.matrix(qrnd_grid) | is.null(dim_qrnd) |
          dim_qrnd[2] != 2)
        stop("qrnd_grid must be a two column matrix")
      n_qrnd <- dim_qrnd[1]
    } else if (!is.null(ell$n_qrnd)){
      n_qrnd <- round(ell$n_qrnd)
      if (n_qrnd < 1)
        stop("n_qrnd must be a positive integer")
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    } else {
      n_qrnd <- 1e4
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    }

    # in log scale
    lpd_grad_model_indep_1comp <- function(par_vec_lscale) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      lpd_grad <- matrix(NA, 6, 1)
      lpd_grad[] <- signif(
        suppressWarnings(grad_llik_vmcos_C(data, par_vec, qrnd_grid)) *
          c(par_vec[1:2], rep(1, 4)),
        8
      )

      list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    }

    start_par_gen <- start_par_vmcos

    hessian_fn <- function(par_vec) {
      numDeriv::hessian(
        func = function(par_vec) {
          -grad_llik_vmcos_C(data, par_vec, qrnd_grid)[6]
        },
        x = par_vec
      )
    }

  }


  else if (model == "indep") {
    # in log scale
    lpd_grad_model_indep_1comp <-  function(par_vec_lscale) {
      par_vec <- c(exp(par_vec_lscale[1:2]), 0, par_vec_lscale[4:5])

      lpd_grad_parts <- lapply(
        1:2,
        function(j) {
          signif(
            suppressWarnings(grad_llik_univm_C(data[, j], par_vec[c(j, 3+j)]))*
              c(par_vec[j], 1, 1),
            8
          )
        }
      )

      lpr <- sum(sapply(lpd_grad_parts, "[", 3))
      grad <- rep(0, 5)
      for (j in 1:2) {
        grad[c(j, 3+j)] <- lpd_grad_parts[[j]][1:2]
      }
      list(lpr = lpr, grad = grad)
    }

    start_par_gen <- function(dat) {
      pars_by_dim <- lapply(1:2, function(j) start_par_vm(dat[, j]))
      pars <- numeric(5)
      for (j in 1:2) {
        pars[c(j, 3+j)] <- pars_by_dim[[j]][1:2]
      }
      pars
    }

    hessian_fn <- function(par_vec) {
      numDeriv::hessian(
        func = function(par_vec) {
          -sum(
            sapply(
              1:2,
              function(j) grad_llik_univm_C(data[, j], par_vec[c(j, 3+j)])[3]
            )
          )
        },
        x = par_vec
      )
    }
  }

  start <- start_par_gen(data)
  names(start) <- c("log_kappa1", "log_kappa2", "kappa3", "mu1", "mu2")

  start_lscale <- start
  start_lscale[c("log_kappa1", "log_kappa2")] <-
    log(start[c("log_kappa1", "log_kappa2")])

  opt <- optim(
    par = start_lscale,
    fn = function(par_lscale) {
      -lpd_grad_model_indep_1comp(par_lscale)$lpr
    },
    gr = function(par_lscale) {
      -lpd_grad_model_indep_1comp(par_lscale)$grad
    },
    lower = c(rep(-Inf, 3), 0, 0),
    upper = c(rep(Inf, 3), 2*pi, 2*pi),
    method = method
    # hessian = TRUE
  )

  est_par <- opt$par
  names(est_par)[1:2] <- c("kappa1", "kappa2")
  est_par[c("kappa1", "kappa2")] <- exp(est_par[c("kappa1", "kappa2")])

  hess <- hessian_fn(par_vec = est_par)
  dimnames(hess) <- list(names(est_par), names(est_par))

  if (model == "indep") {
    vcov <- matrix(0, 5, 5)
    dimnames(vcov) <- dimnames(hess)
    vcov[-3, -3] <- solve(hess[-3, -3])
  } else {
    vcov <- solve(hess)
  }

  res <- methods::new(
    "mle",
    call = call,
    coef = est_par,
    fullcoef = unlist(est_par),
    vcov = vcov,
    min = opt$value,
    details = opt,
    minuslogl = function(kappa1, kappa2, kappa3, mu1, mu2) {
      par_lscale <- c(log(kappa1), log(kappa2), kappa3, mu1, mu2)
      -lpd_grad_model_indep_1comp(par_lscale)$lpr
    },
    nobs = nrow(data),
    method = method
  )

  res
}

start_par_wnorm <- function(data.sub) {
  x1 <- prncp_reg(data.sub)

  Sbar <- mean(sin(x1))
  Cbar <- mean(cos(x1))
  muhat <- prncp_reg(atan(Sbar/Cbar))
  Rbar <- sqrt(Sbar^2 + Cbar^2)

  c(1/(1-Rbar), muhat)
}  #starting parameters from  a dataset

start_clus_kmeans_wnorm <- function(data.full, comp = 2, nstart = 5){
  data.full.cart <- t(sapply(data.full, function(x) c(cos(x), sin(x))))
  data.kmean <- kmeans(data.full.cart, centers = comp, nstart = nstart)
  ids <- data.kmean$cluster
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_wnorm(data.full[clust.ind[[m]]]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(colSums(par^2))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #kmeans, then start_par for each cluster



start_clus_rand_wnorm <- function(data.full, comp = 2){
  rand.pi  <- runif(comp, 1/(2*comp), 2/comp)
  rand.pi <- rand.pi/sum(rand.pi)
  rand.multinom <- t(rmultinom(length(data.full), 1, rand.pi))
  ids <- apply(rand.multinom, 1, which.max)
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_wnorm(data.full[clust.ind[[m]]]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(colSums(par^2))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #random groups, then start_par for each cluster


start_par_wnorm2 <- function(data.sub) {
  data.sub.0.2pi <- prncp_reg(data.sub)
  mu <- prncp_reg(colMeans(data.sub.0.2pi))
  sigma.inv.vec <- solve(cov(data.sub.0.2pi))[c(1,4,2)]
  c(sigma.inv.vec, mu)
}  #starting parameters from  a dataset


start_clus_kmeans_wnorm2 <- function(data.full, comp = 2, nstart = 5){
  data.full.cart <- t(apply(data.full, 1, sph2cart))
  data.kmean <- kmeans(data.full.cart, centers = comp, nstart = nstart)
  ids <- data.kmean$cluster
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_wnorm2(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(colSums(par^2))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #kmeans, then start_par for each cluster


start_clus_rand_wnorm2 <- function(data.full, comp = 2, nstart = 5){
  rand.pi  <- runif(comp, 1/(2*comp), 2/comp)
  rand.pi <- rand.pi/sum(rand.pi)
  rand.multinom <- t(rmultinom(nrow(data.full), 1, rand.pi))
  ids <- apply(rand.multinom, 1, which.max)
  clust.ind <- lapply(1:comp, function(i) which(ids == i))
  par <- sapply(1:length(clust.ind), function(m) start_par_wnorm2(data.full[clust.ind[[m]],]))
  pi.mix <- listLen(clust.ind)/length(unlist(clust.ind))
  order.conc <- order(colSums(par^2))
  list("par.mat" = par[,order.conc], "pi.mix" = pi.mix[order.conc], "clust.ind" = clust.ind[order.conc], "id" = ids)
}  #random groups, then start_par for each cluster

#' Add (extra) burnin and thin to angmcmc object after original run
#' @param object angmcmc object
#' @inheritParams fit_angmix
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' lpdtrace(fit.vmsin.20)
#' # Now add extra burn-in
#' fit.vmsin.20.burn <- add_burnin_thin(fit.vmsin.20, 0.3)
#' lpdtrace(fit.vmsin.20.burn)
#' @export
add_burnin_thin <- function(object, burnin.prop=0, thin=1)
{
  if (!is.angmcmc(object))
    stop("object must be an \'angmcmc\' object")
  if(burnin.prop < 0 || burnin.prop >= 1)
    stop("\"burnin.prop\" must be in [0, 1)")
  if(thin < 1)
    stop("\"thin\" must be a positive integer")


  # browser()
  final_iter.orig <- object$final_iter
  final_iter.burn <- setdiff(final_iter.orig,
                             final_iter.orig[seq_len(ceiling(length(final_iter.orig)*burnin.prop))])

  thin <- round(thin)
  thin.final <- object$thin*thin
  final_iter.thin <- final_iter.burn[c(TRUE, rep(FALSE, thin.final-1))]

  object$n.burnin <- final_iter.thin[1] - 1
  object$thin <- thin.final
  object$n.iter.final <- length(final_iter.thin)
  object$final_iter <- final_iter.thin

  object
}

#' Select chains from angmcmc objects
#' @inheritParams pointest
#' @param chain.no labels of chains to be retained in the final sample. If missing,
#' all chains are used.
#' @param ... unused
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              L = c(10, 12), chains_parallel = FALSE,
#'                              n.chains = 2)
#' fit.vmsin.20
#' fit.vmsin.20.1 <- select_chains(fit.vmsin.20, 1)
#' fit.vmsin.20.1
#'
#' @return Returns another angmcmc object with only selected chains passed through \code{chain.no}
#' @export

select_chains <- function(object, chain.no, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }

  object$par.value <- object$par.value[, , , chain.no, drop=FALSE]
  object$clus.ind <- object$clus.ind[, , chain.no, drop=FALSE]
  object$llik.contri <- object$llik.contri[, , chain.no, drop=FALSE]
  object$mem.prob <- object$mem.prob[, , , chain.no, drop=FALSE]
  object$llik <- object$llik[,  chain.no, drop=FALSE]
  object$lpd <- object$lpd[, chain.no, drop=FALSE]
  object$lprior <- object$lprior[, chain.no, drop=FALSE]
  object$accpt.modelpar <- object$accpt.modelpar[, , chain.no, drop=FALSE]
  if (object$return_tune_param)
    object$tune_param <- object$tune_param[, , chain.no, drop=FALSE]
  object$epsilon <- object$epsilon[, chain.no, drop=FALSE]
  object$L <- object$L[chain.no]
  object$propscale <- object$propscale[, chain.no, drop=FALSE]
  object$n.chains <- length(chain.no)

  object
}

#' Create an mcmc.list object from an angmcmc object
#' @param x angmcmc object
#' @param ... unused
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#'
#' # now convert it to mcmc.list
#' library(coda)
#' fit.vmsin.20.mcmc <- as.mcmc.list(fit.vmsin.20)
#' @importFrom coda as.mcmc.list
#' @method as.mcmc.list angmcmc
#'
#' @export
as.mcmc.list.angmcmc <- function(x, ...)
{
  if (!is.angmcmc(x))
    stop("\'x\' must be an angmcmc object")

  object <- x
  npars <- length(object$par.name)
  ncomp <- object$ncomp
  n.iter <- object$n.iter
  n.data <- object$n.data
  n.chains <- object$n.chains

  if (object$ncomp == 1) {
    par_names_long <- object$par.name
  } else {
    par_names_long <- apply(as.matrix(expand.grid(object$par.name, 1:object$ncomp)),
                            1, function(x) paste0(x[1], "[", x[2], "]"))
  }


  start <- object$final_iter[1]
  end <- object$n.iter
  thin <- object$thin

  out <- vector("list", n.chains)

  for(ii in 1:n.chains) {
    par.vals <- object$par.value[, , , ii, drop=FALSE]
    x <- t(matrix(c(par.vals), ncomp*npars, n.iter))
    colnames(x) <- par_names_long
    if (object$ncomp == 1) x <- x[, -1, drop = FALSE]
    out[[ii]] <- coda::mcmc(x, start = start, end = end, thin = thin)
  }

  coda::as.mcmc.list(out)
}




#' Fix label switching in angmcmc objects
#' @inheritParams pointest
#'
#' @param ... arguments other than \code{z, K, complete, mcmc, p}
#' and \code{data} passed to \link[label.switching]{label.switching}. See details.
#'
#' @details \code{fix_label} is a wrapper for  \link[label.switching]{label.switching} from
#' package \code{label.switching} for \code{angmcmc} objects. The arguments
#' \code{z, K, complete, mcmc, p} and \code{data} are appropriately filled in
#' from \code{object}. The \code{label.switching} argument \code{method} can
#' be a scalar or vector; for this wrapper it defaults to \code{"STEPHENS"} if the \code{angmcmc} was
#' created with permutation sampling (by setting perm_sampling = TRUE in
#' \link{fit_angmix}), and to \code{"DATA-BASED"} otherwise.
#'
#' @return Returns a single \code{angmcmc} object or a list of \code{angmcmc} objects (according as whether
#' the argument \code{method} is a scalar or vector) with label switchings corrected (after burn-in and thin)
#' according to the resulting permutation from \link[label.switching]{label.switching}.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # now apply fix_label
#' fit.vmsin.20.fix <- fix_label(fit.vmsin.20)
#'
#' @export

fix_label <- function(object, ...) {

  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  if (object$ncomp == 1)
    stop("Fitted model has only one component")


  # first collect all necessary arguments from all chains,
  # and collapse into a wide matrix matrix
  K <- ncomp <- object$ncomp
  n.data <- object$n.data
  n.iter.final <- object$n.iter.final
  n.chains <- object$n.chains

  final_iter <- object$final_iter

  npar <- length(object$par.name)

  ell <- list(...)

  if (is.null(ell$method)) {
    if (object$perm_sampling)
      ell$method <- "STEPHENS"
    else
      ell$method <- "DATA-BASED"
  }


  method <- ell$method

  iter_no <- matrix(1:(n.chains*n.iter.final), ncol = n.chains)

  z <- matrix(1, n.chains*n.iter.final, n.data)
  p <- array(1, c(n.chains*n.iter.final, n.data, ncomp))

  mcmc_par <- array(1, c(n.chains*n.iter.final, ncomp, npar))

  for(ii in 1:n.chains) {
    z[iter_no[, ii], ] <- t(object$clus.ind[, final_iter, ii])
    p[iter_no[, ii], , ] <- aperm(object$mem.prob[, , final_iter, ii], c(3, 1, 2))
    mcmc_par[iter_no[, ii], , ] <- aperm(object$par.value[, , final_iter, ii], c(3, 2, 1))
  }

  par_long <- aperm(mcmc_par, c(3, 2, 1))
  mem_prob_long <- aperm(p, c(2, 3, 1))
  clus_ind_long <- t(z)

  # browser()

  complete <- function(x, z, pars) {
    # x: data (size = n)
    # z: allocation vector (size = n)
    # pars: K x npars matrix of normal mixture parameters:
    # pars[k, 1] = k-th mixing proportion
    # pars[k, -1] = k-th component parameters
    pars.t <- t(pars)
    rownames(pars.t) <- object$par.name
    K <- ncol(pars.t)

    # pars.t[1, ] : mixing proportions
    # pars.t[-1, ] : model pars

    if (object$type == "bi")
      data_groups <- lapply(1:K, function(j) x[which(z == j), ])
    else
      data_groups <- lapply(1:K, function(j) x[which(z == j)])

    n.groups <- listLen(data_groups)
    res <- 0
    for(j in 1:K) {
      inargs <- list_by_row(pars.t[-1, j, drop = FALSE])
      inargs$x <- data_groups[[j]]
      if (object$model %in% c("wnorm", "wnorm2"))
        inargs$int.displ <- object$int.displ
      else if (object$model == "vmcos")
        inargs$qrnd_grid <- object$qrnd_grid
      res <- res + sum(
        signif(
          log(
            do.call(paste0("d", object$model), inargs)
          ),
          8
        )
      )
    }

    res + sum(n.groups * log(pars.t[1, ]))
  }

  labswitch_in <- addtolist(ell,
                            z = z, K = ncomp,
                            complete = complete,
                            mcmc = mcmc_par,
                            p = p, data = object$data)


  # lab_switch <-
  #   label.switching::label.switching(method = method,
  #                                    z = z, K = ncomp,
  #                                    complete = complete,
  #                                    mcmc = mcmc_par,
  #                                    p = p, data = object$data,
  #                                    ...)

  lab_switch <-
    do.call(label.switching::label.switching, labswitch_in)



  # now rearrage the component labels according to the rearrangement
  # obtained in lab_switch, and revert back to the original (1 higer dim
  # array) form of the outputs

  res <- vector("list", length(method))

  for(j in 1:length(method)) {
    tmp <- object
    method_curr <- method[j]
    perm_mat <- lab_switch[["permutations"]][[j]]

    par.value.mod <- tmp$par.value[, , final_iter, , drop=FALSE]
    mem.prob.mod <- tmp$mem.prob[, , final_iter, , drop=FALSE]
    clus.ind.mod <- tmp$clus.ind[, final_iter, , drop=FALSE]
    if (object$return_tune_param)
      tune_param_mod <- tmp$tune_param[, final_iter, , drop=FALSE]

    # relabel on these subarrays
    for(ii in 1:n.chains) {
      for(iter in 1:n.iter.final) {
        # browser()
        par.value.mod[ , , iter, ii] <- par.value.mod[, perm_mat[iter_no[iter, ii], ], iter, ii]
        mem.prob.mod[, , iter, ii] <- mem.prob.mod[, perm_mat[iter_no[iter, ii], ], iter, ii]
        clus.ind.mod[, iter, ii] <- perm_mat[iter_no[iter, ii], clus.ind.mod[, iter, ii]]
        if (object$return_tune_param) {
          tume_param_temp_mat <- matrix(tune_param_mod[, iter, ii], ncol=ncomp)
          tune_param_mod[, iter, ii] <- c(tume_param_temp_mat[, perm_mat[iter_no[iter, ii], ]])
        }
      }
    }

    tmp$par.value[, , final_iter, ] <- par.value.mod
    tmp$mem.prob[, , final_iter, ] <- mem.prob.mod
    tmp$clus.ind[, final_iter, ] <- clus.ind.mod
    if (object$return_tune_param)
      tmp$tune_param[, final_iter, ] <- tune_param_mod

    tmp$fixed.label <- TRUE
    res[[j]] <- tmp
  }

  if (length(method) == 1)
    res <- res[[1]]

  res
}

#' Point estimates for parameters from an angmcmc object
#' @param object angular MCMC object.
#' @param fn function, or a single character string specifying its name, to evaluate on MCMC samples to estimate
#' parameters.  Defaults to \code{mean}, which computes the estimated posterior mean.
#' Note that if \code{fn = "MODE"} (warning: not \code{"mode"}) or \code{fn = "MAP"}, then the maximum aposteriori estimate (MAP) is
#' calculated.
#' @param par.name vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.
#' @param comp.label vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
#' If \code{NULL}, results for all components are provided.
#' @param chain.no vector of chain numbers whose samples are to be be used.
#' in the estimation. By default all chains are used.
#' @param ... additional arguments to be passed to the function.
#'
#' @return Returns a matrix of point estimates, or vector of point estimates if \code{length(par.name)==1} or \code{length(comp.label)==1}.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # estimate parameters by sample mean
#' (est_mean <- pointest(fit.vmsin.20))
#' # estimate parameters by sample median
#' (est_median <- pointest(fit.vmsin.20, fn = median))
#' # estimate parameters by MAP
#' (est_median <- pointest(fit.vmsin.20, fn = "MODE"))
#' @export

pointest <- function(object, fn = mean, par.name,
                     comp.label, chain.no, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")


  if (is.character(fn)) {
    if (fn == "MODE" | fn == "MAP") {
      do_MAP <- TRUE
    }
    else {
      do_MAP <- FALSE
      fn <- match.fun(fn)
    }
  } else {
    do_MAP <- FALSE
    fn <- match.fun(fn)
  }


  ell <- list(...)
  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (missing(par.name)) {
    par.name <- object$par.name
  }  else if (any(!par.name %in% object$par.name)) {
    stop("invalid par.name")
  }

  if (missing(comp.label)) {
    comp.label <- 1:object$ncomp
  } else if (any(!comp.label %in% 1:object$ncomp)) {
    stop("invalid component label")
  }

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }

  par.posit <- which(object$par.name %in% par.name)

  final_iter <- object$final_iter


  if (!do_MAP) {
    res <- apply(object$par.value[par.posit, comp.label, final_iter, chain.no, drop = FALSE],
                 c(1, 2), fn, drop = FALSE)
  }

  else {
    map_indic <- which(object$lpd[, chain.no, drop=FALSE]
                       == max(object$lpd[final_iter, chain.no]),
                       arr.ind = TRUE)[1, ]  # first entry of max
    res_mat <- as.matrix(object$par.value[ , , map_indic[1], map_indic[2]])

    res <- res_mat[par.posit, comp.label, drop=FALSE]
  }

  rownames(res) <- par.name
  colnames(res) <- comp.label
  res
}


#' Quantile estimates for parameters from an angmcmc object
#' @inheritParams pointest
#' @param x angmcmc object
#' @inheritParams stats::quantile
#' @return Returns a three dimensional array of quantiles, or a matrix (vector) of quantiles
#'  if one (or two) among \code{par.name},  \code{comp.label}, \code{probs} has length 1.
#' @param ... further arguments to pass to \code{quantile}.  In particular, \code{probs = seq(0, 1, 0.25)}
#' is the default vector of quantiles computed for each parameter.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # 0.025th quantiles
#' (quant_025 <- quantile(fit.vmsin.20, prob = 0.025))
#' # 0.975th quantiles
#' (quant_975 <- quantile(fit.vmsin.20, prob = 0.975))
#' # default quantiles
#' (quant_def <- quantile(fit.vmsin.20))
#'
#' @export


quantile.angmcmc <- function(x, par.name, comp.label, chain.no,
                             probs = seq(0, 1, 0.25), ...)
{

  object <- x

  ell <- list(...)

  if (!is.angmcmc(object))
    stop("\'x\' must be an angmcmc object")

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (missing(par.name)) {
    par.name <- object$par.name
  }  else if (any(!par.name %in% object$par.name)) {
    stop("invalid par.name")
  }

  if (missing(comp.label)) {
    comp.label <- 1:object$ncomp
  } else if (any(!comp.label %in% 1:object$ncomp)) {
    stop("invalid component label")
  }

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }

  par.posit <- which(object$par.name %in% par.name)

  final_iter <- object$final_iter

  n_probs <- length(probs)

  res <- list()

  for(jj in 1:n_probs) {
    res[[jj]] <- apply(object$par.value[par.posit, comp.label, final_iter, chain.no, drop = FALSE],
                       c(1, 2),
                       function(y) quantile(y, probs = probs[jj]))
    rownames(res[[jj]]) <- par.name
    colnames(res[[jj]]) <- comp.label
  }

  names(res) <- paste0(round(probs*100), "%")

  res
}


#' Extract MCMC samples for parameters from an angmcmc object
#' @inheritParams pointest
#' @param object angular MCMC object
#' @param drop logical. Should the dimension of the output be dropped, if \code{par.name},
#' \code{comp.label} or \code{chain.no} has a single level?
#' @details The default for both \code{par.name} and \code{comp.label} are the all possible choices
#' available in \code{object}.
#' @return
#' Returns  a four dimensional array with
#'
#' dimension 1 - model parameters and mixing proportions
#' dimention 2 - components
#' dimension 3 - MCMC iterations
#' dimension 4 - chain number
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # extract Markov chain realizations for kappa1 from component 1
#' extr_kappa1_1 <- extractsamples(fit.vmsin.20, "kappa1", 1)
#' # for kappa1 from component from all components
#' extr_kappa1 <- extractsamples(fit.vmsin.20, "kappa1")
#' # for all parameters in component 1
#' extr_1 <- extractsamples(fit.vmsin.20, comp.label = 1)
#'
#' @export

extractsamples <- function(object, par.name, comp.label,
                           chain.no, drop = TRUE, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  ell <- list(...)

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (missing(par.name)) {
    par.name <- object$par.name
  }  else if (any(!par.name %in% object$par.name)) {
    stop("invalid par.name")
  }

  if (missing(comp.label)) {
    comp.label <- 1:object$ncomp
  } else if (any(!comp.label %in% 1:object$ncomp)) {
    stop("invalid component label")
  }

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }

  par.posit <- which(object$par.name %in% par.name)
  final_iter <- object$final_iter
  out <- object$par.value[par.posit, comp.label, final_iter, chain.no, drop = drop]

  # browser()
  if (!drop)
    dimnames(out) <- list(par.name, comp.label, NULL, chain.no)

  out
}


#' Summary statistics for parameters from an angmcmc object
#' @inheritParams base::summary
#' @inheritParams pointest
#' @param object angular MCMC object.
#' @details Computes (after thinning and discarding burn-in) point estimates with 95\% posterior credible sets for all components and all parameters,
#' together with the sample averages of log likelihood and log posterior density.
#' @return Returns a list with elements \code{estimate, lower, upper, llik} and \code{lpd}. \code{estimate}
#' is itself a list with three elements: \code{mean}, \code{median} and \code{mode} providing the
#' sample mean, sample median and (sample) MAP estimates.
#'
#' Note that \code{summary.angmcmc} has its own print method, providing a table the estimated mean and 95\% credible intervals for each parameter
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' summary(fit.vmsin.20)
#'
#' @export

summary.angmcmc <- function(object, par.name, comp.label,
                            chain.no, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  ell <- list(...)

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (missing(par.name)) {
    par.name <- object$par.name
  }  else if (any(!par.name %in% object$par.name)) {
    stop("invalid par.name")
  }

  if (missing(comp.label)) {
    comp.label <- 1:object$ncomp
  } else if (any(!comp.label %in% 1:object$ncomp)) {
    stop("invalid component label")
  }

  if (missing(chain.no)) {
    chain.no <- 1:object$n.chains
  } else if (any(!chain.no %in% 1:object$n.chains)) {
    stop("invalid chain number")
  }


  est <- pointest(object, mean, par.name, comp.label, chain.no)
  map <- pointest(object, "MODE", par.name, comp.label, chain.no)
  quants <- quantile(object, par.name, comp.label, chain.no,
                     probs = c(0.5, 0.025, 0.975))

  med <- quants[[1]]
  low <- quants[[2]]
  up <- quants[[3]]

  llik <- as.numeric(logLik(object))
  lpd <- mean(object$lprior) + llik

  res <- list("estimate " = list(mean = est, median = med, mode = map),  "upper" = up, "lower" = low,
              "llik" = llik, "lpd" = lpd)

  class(res) <- "summary_angmcmc"
  res
}

#' @export
print.summary_angmcmc <- function(x, ...)
{
  res <- x
  print(est_ci(res$estimate$mean, res$lower, res$upper, digits=2))
}



# #' AIC and BIC for angmcmc objects -- not needed
# #' @inheritParams stats::AIC
# #' @inheritParams logLik.angmcmc
# #' @param ... additional argument to be passed to \link{logLik.angmcmc}
# #' @return AIC computes the AIC and BIC computes BIC for \code{angmcmc} objects.
# #'
# #' @details
# #' Note that \code{AIC.angmcmc} and \code{BIC.angmcmc} calls \link{logLik.angmcmc},
# #' which calculates Bayes estimate of the log-likelihood and *not* the maximum
# #' likelihood. As such, care needs to be taken while using theses quantities.
# #'
# #' \eqn{\hat{L}} is estimated by the sample maximum obtained from the MCMC realizations.
# #'
# #' @examples
# #' # illustration only - more iterations needed for convergence
# #' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
# #'                              n.chains = 1)
# #' AIC(fit.vmsin.20)
# #' BIC(fit.vmsin.20)
# #'
# #' @export
# #'
# #' AIC.angmcmc <- function(object, ..., k = 2)
# #' {
# #'   if (!is.angmcmc(object))
# #'     stop("\'object\' must be an angmcmc object")
# #'
# #'   ell <- list(...)
# #'
# #'   if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
# #'     warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")
# #'
# #'   llik <- logLik.angmcmc(object, ...)
# #'   k*attr(llik, "df") - 2*as.numeric(llik)
# #' }
# #'
# #'
# #'
# #' @rdname AIC.angmcmc
# #' @export
# #'
# #' BIC.angmcmc <- function(object, ...)
# #' {
# #'   AIC.angmcmc(object, ..., k=log(object$n.data))
# #' }


#' Deviance Information Criterion (DIC) for angmcmc objects
#' @inheritParams pointest
#' @param ... additional model specific arguments to be passed to \code{DIC}. For example, \code{int.displ}
#' specifies integer dispacement in wnorm and wnorm2 models. See \link{fit_wnormmix} and
#' \link{fit_wnorm2mix} for more details.
#' @param form form of DIC to use. Available choices are 1 and 2 (default). See details.
#' @return Computes the DIC for a given angmcmc object
#' @details Given a deviance function \eqn{D(\theta) = -2 log(p(y|\theta))}, and an estimate
#' \eqn{\theta* = (\sum \theta_i) / N} of the posterior mean
#' \eqn{E(\theta|y)}, where \eqn{y} denote the data, \eqn{\theta} are the unknown
#' parameters of the model, \eqn{\theta_1, ..., \theta_N} are MCMC samples from the posterior
#' distribution of \eqn{\theta} given \eqn{y} and \eqn{p(y|\theta)} is the likelihood function,
#' the (form 1 of) Deviance Infomation Criterion (DIC) is defined as
#' \deqn{DIC = 2 ( (\sum_{s=1}^N D(\theta_s)) / N - D(\theta*) )}
#' The second form for DIC is given by
#' \deqn{DIC = D(\theta*) - 4 \hat{var} \log p(y|\theta_s)}
#' where for \eqn{i = 1, ..., n}, \eqn{\hat{var} \log p(y|\theta)} denotes the estimated variance
#' of the log likelihood based on the realizations \eqn{\theta_1, ..., \theta_N}.
#'
#' Like AIC and BIC, DIC is an asymptotic approximation for large samples, and
#' is only valid when the posterior distribution is approximately normal.
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' DIC(fit.vmsin.20)
#'
#' @export

DIC <- function(object, form = 2, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  ell <- list(...)

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if(!form %in% 1:2)
    stop("form must be either 1 or 2")

  final_iter <- object$final_iter

  all_llik <- object$llik[final_iter, ]
  llik_estpar <-
    as.numeric(logLik.angmcmc(object, fn = mean, method = 1))
  D_all <- -2*all_llik
  Dbar <- mean(D_all)
  Dhat <- -2*llik_estpar

  if (form == 1) {
    pD <- Dbar - Dhat
  } else {
    pD <- 0.5*var(D_all)
  }

  c("pD" = pD, "DIC" = pD + Dbar)
}



#' Watanabe-Akaike Information Criterion (WAIC) for angmcmc objects
#'
#' @param x angmcmc object.
#' @param ... additional model specific arguments to be passed to \link[loo]{waic} from loo. For example, \code{int.displ}
#' specifies integer displacement in wnorm and wnorm2 models. See \link{fit_wnormmix} and
#' \link{fit_wnorm2mix} for more details.
#' @return Computes the WAIC for a given angmcmc object.
#' @details
#' Given a deviance function \eqn{D(\eta) = -2 \log(p(y|\eta))}, and an estimate
#' \eqn{\eta* = (\sum \eta_i) / n} of the posterior mean
#' \eqn{E(\eta|y)}, where \eqn{y = (y_1, ..., y_n)} denote the data, \eqn{\eta} is the unknown
#' parameter vector of the model, \eqn{\eta_1, ..., \eta_N} are MCMC samples from the posterior
#' distribution of \eqn{\eta} given \eqn{y} and \eqn{p(y|\eta)} is the likelihood function,
#' the Watanabe-Akaike Information Criterion (WAIC) is defined as
#' \deqn{WAIC = LPPD - p_W}
#' where
#' \deqn{LPPD  = \sum_{i=1}^n \log (N^{-1} \sum_{s=1}^N p(y_i|\eta_s) )}
#' and (form 1 of)
#' \deqn{p_W =  2 \sum_{i=1}^n [ \log (N^{-1} \sum_{s=1}^N p(y_i|\eta_s) ) - N^{-1} \sum_{s=1}^N \log \:p(y_i|\eta_s) ].}
#' An alternative form (form 2) for \eqn{p_W} is given by
#' \deqn{p_W = \sum_{i=1}^n \hat{var} \log p(y_i|\eta)}
#' where for \eqn{i = 1, ..., n}, \eqn{\hat{var} \log p(y_i|\eta)} denotes the estimated variance
#' of \eqn{\log p(y_i|\eta)} based on the realizations \eqn{\eta_1, ..., \eta_N}.
#'
#' Note that waic.angmcmc calls \link[loo]{waic} for computation. If the likelihood contribution of each data
#' point for each MCMC iteration is available in \code{object} (can be returned by setting \code{return_llik_contri = TRUE})
#' during \link{fit_angmix} call), \code{waic.array} is used; otherwise \code{waic.function} is
#' called. Computation is much faster if the likelihood contributions are available - however, they are very
#' memory intensive, and by default not returned in \link{fit_angmix}.
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1, return_llik_contri = TRUE)
#' library(loo)
#' waic(fit.vmsin.20)
#'
#' @importFrom loo waic loo
#'
#' @export


waic.angmcmc <- function(x, ...)
{
  object <- x

  if (!is.angmcmc(object))
    stop("\'x\' must be an angmcmc object")

  ell <- list(...)

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  final_iter <- object$final_iter


  if (object$return_llik_contri) {
    all_llik_contri <- object$llik.contri
    llik_for_loo <- aperm(all_llik_contri[, final_iter, , drop = FALSE], c(2, 3, 1))
    waic <- loo::waic.array(llik_for_loo,  ...)
  }

  else {
    samples <- object$par.value[, , final_iter, , drop=FALSE]
    chain_id <- unlist(lapply(1:object$n.chains, function(j) rep(j, length(final_iter))))

    dim_samples <- dim(samples)
    samples_long <- array(c(samples),
                          dim = c(dim_samples[1:2], dim_samples[3]*dim_samples[4]))
    dimnames(samples_long)[[1]] <- object$par.name

    calc_llik_contri <- function(data_i, draws) {
      N <- dim(draws)[3]
      out <- rep(NA,  N)
      for (j in 1:N) {
        input_pars <- list_by_row(draws[, , j])
        input_pars$log <- TRUE
        input_pars$x <- c(data_i)
        out[j] <- signif(
          do.call(paste0("d", object$model, "mix"), input_pars),
          8
        )
      }
      out
    }

    data <- as.matrix(object$data)

    waic <- loo::waic.function(calc_llik_contri,
                               data = data, draws = samples_long,
                               ...)


  }

  waic
}


#' Leave-one-out cross-validation (LOO) for angmcmc objects
#' @inheritParams waic.angmcmc
#'
#' @examples
#' \donttest{
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1, return_llik_contri = TRUE)
#' library(loo)
#' loo(fit.vmsin.20)
#' }
#' @export
#'
#' @details
#' Note that loo.angmcmc calls \link[loo]{loo} for computation. If the likelihood contribution of each data
#' point for each MCMC iteration is available in \code{object} (can be returned by setting \code{return_llik_contri = TRUE})
#' during \link{fit_angmix} call), \code{loo.array} is used; otherwise \code{loo.function} is
#' called. Computation is much faster if the likelihood contributions are available - however, they are very
#' memory intensive, and by default not returned in \link{fit_angmix}.
#'
#'

loo.angmcmc <- function(x, ...)
{
  object <- x


  if (!is.angmcmc(object))
    stop("\'x\' must be an angmcmc object")

  ell <- list(...)

  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  final_iter <- object$final_iter

  if (object$return_llik_contri) {
    all_llik_contri <- object$llik.contri
    llik_for_loo <- aperm(all_llik_contri[, final_iter, , drop = FALSE],
                          c(2, 3, 1))
    r_eff <- loo::relative_eff(exp(llik_for_loo), ...)
    looic <- loo::loo.array(llik_for_loo, r_eff = r_eff,
                            ...)
  }

  else {
    samples <- object$par.value[, , final_iter, , drop=FALSE]
    chain_id <- unlist(lapply(1:object$n.chains, function(j) rep(j, length(final_iter))))

    dim_samples <- dim(samples)
    samples_long <- array(c(samples),
                          dim = c(dim_samples[1:2], dim_samples[3]*dim_samples[4]))
    dimnames(samples_long)[[1]] <- object$par.name

    calc_llik_contri <- function(data_i, draws) {
      N <- dim(draws)[3]
      out <- rep(NA,  N)
      for (j in 1:N) {
        input_pars <- list_by_row(draws[, , j])
        input_pars$log <- TRUE
        input_pars$x <- c(data_i)
        out[j] <- signif(
          do.call(paste0("d", object$model, "mix"),
                  input_pars),
          8
        )
      }
      out
    }

    data <- as.matrix(object$data)


    r_eff <- loo::relative_eff(calc_llik_contri,
                               chain_id = chain_id,
                               data=data,
                               draws = samples_long)
    looic <- loo::loo.function(calc_llik_contri, r_eff = r_eff,
                               data = data, draws = samples_long,
                               ...)
  }

  looic
}



#' Density and random deviates from an angmcmc object
#' @inheritParams pointest
#' @inheritParams dvmsin
#' @param type Method of estimating density/generating random deviates. Possible choices are
#' \code{"post-pred"} and \code{"point-est"}. See details. Defaults to \code{"point-est"}.
#' @param object angular MCMC object. The dimension of the model must match with \code{x}.
#' @param x vector, if univariate or a two column matrix, if bivariate, with each row a 2-D vector, (can
#' also be a data frame of similar dimensions) of points where the
#' densities are to be computed.
#' @param n number of observations to be generated.
#' @return \code{d_fitted} gives a vector the densities computed at the given points  and \code{r_fitted}
#' creates a vector (if univariate) or a matrix (if bivariate) with each row being a 2-D point, of random deviates.
#'
#' @details
#'
#' If \code{type = 'point-est'}, density is evaluated/random samples are generated at a point estimate of
#' the parameter values.  To estimate the mixture density, first the parameter vector \eqn{\eta} is estimated
#' by applying \code{fn} on the MCMC samples (using the function \link{pointest}), yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}.
#' Then the mixture density \eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by
#' \eqn{f(x|\hat{\eta})}. The random deviates are generated from the estimated mixture density \eqn{f(x|\hat{\eta})}.
#'
#' If \code{type == 'post-pred'}, posterior predictive samples and densities are returned. That
#' is, the average density \eqn{S^{-1} \sum_{s = 1}^S f(x | \eta_s)} is returned in \code{d_fitted},
#' where \eqn{\eta_1, \dots, \eta_S} is the set posterior MCMC samples obtained from \code{object}. In
#' \code{r_fitted}, first a random sub-sample \eqn{\eta_{(1)}, \dots, \eta_{(n)}} of size \code{n} from the
#' set of posterior samples \eqn{\eta_1, \dots, \eta_S} is drawn (with replacement if \code{n} > S). Then
#' the i-th posterior predictive data point is generated from the mixture density
#' \eqn{f(x|\eta_{(i)})} for i = 1,..., n.
#'
#' @examples
#' set.seed(1)
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' d_fitted(c(0,0), fit.vmsin.20, type = "post-pred")
#' d_fitted(c(0,0), fit.vmsin.20, type = "point-est")
#'
#' r_fitted(10, fit.vmsin.20, type = "post-pred")
#' r_fitted(10, fit.vmsin.20, type = "point-est")
#' @export


d_fitted <- function(x, object, type = "point-est", fn = mean, log=FALSE,
                     chain.no, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  ell <- list(...)
  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (is.data.frame(x))
    x <- as.matrix(x)


  if(object$type == "bi") {
    if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
       || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")
  }

  if (!type %in% c("post-pred", "point-est")) {
    stop ("\'type\' must either be \'post-pred\' or \'point-est\'")
  }

  inargs <- list(x = x)
  if (object$model %in% c("wnorm", "wnorm2"))
    inargs$int.displ <- object$int.displ
  else if (object$model == "vmcos")
    inargs$qrnd_grid <- object$qrnd_grid

  if (type == "point-est") {
    est <- pointest(object, fn = fn)
    inargs <- c(list_by_row(est), inargs)
    inargs$log <- log
    out <- signif(suppressWarnings(do.call(paste0("d", object$model, "mix"), inargs)), 8)
  } else {
    samp <- extractsamples(object, chain.no = chain.no, drop = FALSE)
    dim_samp <- dim(samp)
    nsamp <- dim_samp[3]*dim_samp[4]
    samp_coll <- samp
    dim(samp_coll) <- c(dim_samp[1:2], nsamp)
    dimnames(samp_coll)[1:2] <- dimnames(samp)[1:2]
    inargs$log <- FALSE
    out_list <- lapply(seq_len(nsamp),
                       function(j) {
                         inargs1 <- c(list_by_row(as.matrix(samp_coll[, , j])), inargs)
                         signif(
                           c(suppressWarnings(do.call(paste0("d",
                                                             object$model,
                                                             "mix"), inargs1))),
                           8
                         )
                       })
    out <- Reduce('+', out_list)/nsamp

    if (log) out <- log(out)
  }

  out
}



#' @rdname d_fitted
#' @export
r_fitted <- function(n=1, object, type =  "point-est", fn = mean,
                     chain.no,  ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  ell <- list(...)
  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  if (!type %in% c("post-pred", "point-est")) {
    stop ("\'type\' must either be \'post-pred\' or \'point-est\'")
  }


  inargs <- list()

  if (object$model == "vmcos") {
    inargs$qrnd_grid <- object$qrnd_grid
  }


  if (object$model %in% c("vmcos", "vmsin")) {
    inargs$method <- "naive"
  }


  if (type == "point-est") {
    est <- pointest(object, fn = fn, chain.no = chain.no)
    inargs <- c(list_by_row(as.matrix(est)), inargs)
    inargs$n <- n
    out <- suppressWarnings(do.call(paste0("r", object$model, "mix"), inargs))
  } else {

    samp <- extractsamples(object, chain.no = chain.no, drop = FALSE)
    dim_samp <- dim(samp)
    nsamp <- dim_samp[3]*dim_samp[4]
    samp_coll <- samp
    dim(samp_coll) <- c(dim_samp[1:2], nsamp)
    dimnames(samp_coll)[1:2] <- dimnames(samp)[1:2]
    ids <- sample(seq_len(dim_samp[3]*dim_samp[4]), size = n,
                  replace = n > nsamp)

    if (object$type == "bi") {
      out_row <- numeric(2)
    } else {
      out_row <- numeric(1)
    }

    out <- vapply(ids,
                  function(j) {
                    inargs1 <- c(list_by_row(as.matrix(samp_coll[, , j])), inargs)
                    inargs1$n <- 1
                    c(suppressWarnings(do.call(paste0("r",
                                                      object$model,
                                                      "mix"), inargs1)))
                  },
                  out_row)

    if (object$type == "bi") out <- t(out)
  }

  out
}


#' Extract Log-Likelihood from angmcmc objects
#' @inheritParams pointest
#' @inheritParams stats::logLik
#' @param method interger specifying method of estimating the log likelihood. Must be 1 or 2. Defaults to 1. See details.
#' @param fn function to evaluate on the iteration-wise log-likelihood values obtained during MCMC run if \code{method = 1}; or,
#' if \code{method = 2}, function to evaluate on the MCMC samples for parameter estimation (passed to \link{pointest}).
#' Defaults to \code{max} if \code{method = 1} and \code{mean} if \code{method = 2}.
#'
#' @details
#'
#' There are two ways to estimate the log likelihood from the model. If \code{method = 1},
#' then log likelihood is estimated by applying \code{fn} (defaults to max, if method = 1)
#' direclty on the log likelihood values from observed during the MCMC run.
#' On the other hand, if \code{method == 2}, then  parameter estimates
#' are first computed using \code{pointest} with \code{fn}
#' (defaults to "MODE", if \code{method == 2}) applied on the MCMC samples,
#' and then then log likelihood is evaluated at the parameter estimates.
#'
#'
#' The degrees of the likelihood function is the total number of free parameters estimated in the mixture models,
#' which is equal to \eqn{6K - 1} for bivariate models (vmsin, vmcos and wnorm2), or \eqn{3K - 1} for univariate
#' models (vm and wnorm), where \eqn{K} denotes the number of components in the mixture model.
#'
#' @return Returns an object of class \link{logLik}. This is a number (the estimated log likelihood) with attributes "df"
#' (degrees of freedom) and "nobs" (number of observations).
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' logLik(fit.vmsin.20)
#' @export


logLik.angmcmc <- function(object, method = 1, fn, ...)
{
  if (!is.angmcmc(object))
    stop("\'object\' must be an angmcmc object")

  if (!method %in% c(1, 2))
    stop("method must be either 1 or 2")

  if (missing(fn)) {
    if (method == 1) fn <- "max"
    else fn <- "MODE"
  }

  do_MAP <- FALSE

  if (is.character(fn)) {
    if (fn %in% c("MODE", "MAP")) {
      if (method == 1)
        stop("fn can be \'MODE\' or \'MAP\' only when method = 2")
      else do_MAP <- TRUE
    }
  }

  if (!do_MAP) fn <- match.fun(fn)

  ell <- list(...)
  if (any(!is.null(ell$burnin),  !is.null(ell$thin)))
    warning("Use of burnin and thin are obsolete. Specify \'burnin.prop\' and \'thin\' during original MCMC run, or use \'add_burnin_thin\'.")

  final_iter <- object$final_iter

  if (method == 2) {
    all_den <- d_fitted(object$data, object, fn = fn, log=TRUE)
    llik <- sum(all_den)
  } else {
    llik <- fn(object$llik[final_iter, ])
  }


  if(object$type == "uni") {
    df <- 3*object$ncomp - 1
  } else if (object$cov.restrict == "ZERO") {
    df <- 5*object$ncomp - 1 #the cov terms aren't present
  } else {
    df <- 6*object$ncomp - 1
  }

  n_data <- object$n.data
  result <- llik
  attributes(result) <- list("df" = df, "nobs" = n_data)
  class(result) <- "logLik"
  result

}


#' Log Marginal Likelihood via Bridge Sampling for angmcmc objects
#' @param samples angmcmc object
#' @param ... additional argument passed to \link[bridgesampling]{bridge_sampler}. Note that default for
#' the argument \code{method} is \code{"warp3"}, (instead of \code{"normal"} as used in \code{bridgesampling} package)
#' to account for multi-modality of the posterior density.
#' @param ave_over_chains logical. Separately call \link[bridgesampling]{bridge_sampler} on
#' each chain in the angmcmc object and then take the average? Defaults to \code{TRUE}.
#' See details.
#'
#' @details
#' Marginal likelihood is calculated by first converting the \code{angmcmc} object \code{samples} to an
#' \code{mcmc.list} object, and then by passing the resulting \code{mcmc.list} object to \link[bridgesampling]{bridge_sampler}.
#' If variablity across multiple chains (if any) are very different,
#' then calling \link[bridgesampling]{bridge_sampler} separately for each chain
#' usually provides more stable results; the final log ML is computed by averaging over
#' chain specific MLs.
#'
#' @importFrom bridgesampling bridge_sampler
#'
#' @examples
#' \donttest{
#' library(future)
#' library(parallel)
#' # plan(multisession, gc = TRUE) # parallelize chains
#'
#' set.seed(100)
#' MC.fit <- fit_angmix("vmsin", tim8, ncomp=3, n.iter=5000,
#'                      n.chains = 3)
#'
#'
#' library(bridgesampling)
#' bridge_sampler(MC.fit)
#' }
#'
#' @export

bridge_sampler.angmcmc <- function(samples, ..., ave_over_chains = TRUE)
{

  object <- samples

  if (!is.angmcmc(object))
    stop("\'samples\' must be an angmcmc object")


  ell <- list(...)

  if (is.null(ell$method)) {
    ell$method <- "warp3"
  }

  n.chains <- object$n.chains

  if (object$model == "wnorm2") {
    kappa_tmp <- object$par.value
    sig_tmp <-
      apply(kappa_tmp, 2:4,
            function(x) c(x[1],
                          unname(kappas2sigmas_wnorm2(x[2], x[3], x[4])),
                          x[5:6]))
    names(sig_tmp) <- names(kappa_tmp)
    object$par.value <- sig_tmp
  }


  object_mcmc <- coda::as.mcmc.list(object)

  object_mcmc_matrix <- as.matrix(object_mcmc)

  gam.loc <- object$gam.loc
  gam.rate <- 1/object$gam.scale


  norm.var <- object$norm.var

  if (object$type == "bi" & length(norm.var) == 1) norm.var <- rep(norm.var, 3)


  if (object$ncomp == 1) {

    calc_lpd <- function(par_vec, data) {
      # reparametrize if wnorm2
      if (object$model == "wnorm2") {
        par_vec <- c(sigmas2kappas_wnorm2(par_vec[1], par_vec[2], par_vec[3]),
                     par_vec[4:5])
      }
      allpar_mat <- as.matrix(par_vec)

      rownames(allpar_mat) <- object$par.name[-1]
      inargs <- list_by_row(allpar_mat)
      inargs$x <- data
      if (grepl("wnorm", object$model))
        inargs$int.displ <- object$int.displ
      if (object$model == "vmcos")
        inargs$qrnd_grid <- object$qrnd_grid

      inargs$log <- TRUE
      llik <- sum(signif(do.call(paste0("d", object$model), inargs), 8))

      if (object$type == "bi") {
        lprior <- sum(-0.5*c((log(allpar_mat[1, ]))^2/norm.var[1],
                             (log(allpar_mat[2, ]))^2/norm.var[2],
                             ((allpar_mat[3, ]))^2/norm.var[3]))
      } else {
        lprior <- sum(-0.5*log(allpar_mat[1, ])^2/norm.var)
      }

      unname(llik+lprior)
    }


    # lower and upper limits for reparameterized wnorm2

    if (object$model == "wnorm2") {
      lower <- object$modelpar.lower
      upper <- object$modelpar.upper
      upper[1:2, ] <- 1e5
      lower[3, ] <- -(1-1e-7)
      upper[3, ] <- (1-1e-7)
      lower <- c(lower)
      upper <- c(upper)
    }
    else {
      lower <- c(object$modelpar.lower)
      upper <- c(object$modelpar.upper)
    }

    names(lower) <- names(upper) <- colnames(object_mcmc_matrix)

    if(n.chains == 1 || !ave_over_chains) {
      tmp_mat <- do.call(rbind, lapply(1:n.chains,
                                       function(j)
                                         as.matrix(object_mcmc[[j]])))

      do.call(bridgesampling::bridge_sampler,
              c(list(samples=tmp_mat,
                     log_posterior=calc_lpd,
                     data=object$data,
                     lb=lower, ub=upper),
                ell))
    }
    else {
      all_bridge_samp <- vector("list", n.chains)
      for (j in 1:n.chains) {
        # tmp <- coda::as.mcmc.list(select_chains(object, j))
        tmp_mat <- as.matrix(object_mcmc[[j]])
        if (!any(ell$silent)) {
          cat(paste0("\napplying \'bridge_sampler\' on chain ", j, "...\n\n"))
        }

        all_bridge_samp[[j]] <-
          do.call(bridgesampling::bridge_sampler,
                  c(list(samples=tmp_mat,
                         log_posterior=calc_lpd,
                         data=object$data,
                         lb=lower, ub=upper),
                    ell))
      }

      final_res <- all_bridge_samp[[1]]
      final_res$logml <- mean(vapply(all_bridge_samp,
                                     function(x) x$logml, 0))
      final_res$niter <- max(vapply(all_bridge_samp,
                                    function(x) x$niter, 0))
      final_res$method <- all_bridge_samp[[1]]$method
      final_res$q11 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q11))
      final_res$q12 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q12))
      final_res$q21 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q21))
      final_res$q22 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q22))

      final_res$all <- all_bridge_samp
      final_res
    }
  }


  else {

    calc_lpd <- function(par_vec, data) {
      allpar_mat <- matrix(par_vec,
                           nrow = length(object$par.name))
      # reparametrize if wnorm2
      if (object$model == "wnorm2") {
        allpar_mat <- apply(allpar_mat, 2,
                            function(x) c(x[1],
                                          unname(sigmas2kappas_wnorm2(x[2], x[3], x[4])),
                                          x[5:6]))
      }
      allpar_mat[1, ] <- allpar_mat[1, ]/sum(allpar_mat[1, ])
      rownames(allpar_mat) <- object$par.name
      inargs <- list_by_row(allpar_mat)
      inargs$x <- data
      if (grepl("wnorm", object$model))
        inargs$int.displ <- object$int.displ
      if (object$model == "vmcos")
        inargs$qrnd_grid <- object$qrnd_grid

      inargs$log <- TRUE
      llik <- suppressWarnings(
        sum(
          signif(do.call(paste0("d", object$model, "mix"), inargs)), 8
        )
      )


      par_mat <- allpar_mat[-1, ]
      pmix_vec <- allpar_mat[1, ]


      if (object$type == "bi") {
        lprior <- sum(-0.5*c((log(allpar_mat[1, ]))^2/norm.var[1],
                             (log(allpar_mat[2, ]))^2/norm.var[2],
                             ((allpar_mat[3, ]))^2/norm.var[3])) +
          sum(object$pmix.alpha*log(pmix_vec))
      } else {
        lprior <- sum(-0.5*log(allpar_mat[1, ])^2/norm.var) +
          sum(object$pmix.alpha*log(pmix_vec))
      }

      # if (object$type == "bi") {
      #   lprior <- sum((gam.loc - 1)*log(par_mat[1:2, ]) -
      #                   gam.rate*par_mat[1:2, ]) -
      #     0.5*sum(par_mat[3, ]^2)/object$norm.var +
      #     sum(object$pmix.alpha*log(pmix_vec))
      # } else {
      #   lprior <- sum((gam.loc - 1)*log(allpar_mat[1, ]) -
      #                   gam.rate*allpar_mat[1, ]) +
      #     sum(object$pmix.alpha*log(pmix_vec))
      # }

      unname(llik+lprior)
    }


    # lower and upper limits for reparameterized wnorm2

    if (object$model == "wnorm2") {
      lower <- rbind(0, object$modelpar.lower)
      upper <- rbind(Inf, object$modelpar.upper)
      upper[2:3, ] <- 1e5
      lower[4, ] <- -(1-1e-7)
      upper[4, ] <- (1-1e-7)
      lower <- c(lower)
      upper <- c(upper)
    }

    else {
      lower <- c(rbind(0, object$modelpar.lower))
      upper <- c(rbind(Inf, object$modelpar.upper))
    }

    names(lower) <- names(upper) <-
      colnames(object_mcmc_matrix)
    if (n.chains == 1 || !ave_over_chains) {
      tmp_mat <- do.call(rbind, lapply(1:n.chains,
                                       function(j)
                                         as.matrix(object_mcmc[[j]])))
      do.call(bridgesampling::bridge_sampler,
              c(list(samples=tmp_mat,
                     log_posterior=calc_lpd,
                     data=object$data,
                     lb=lower, ub=upper),
                ell))

    }
    else {
      all_bridge_samp <- vector("list", n.chains)
      for (j in 1:n.chains) {
        # tmp <- coda::as.mcmc.list(select_chains(object, j))
        tmp_mat <- as.matrix(object_mcmc[[j]])
        if (!any(ell$silent)) {
          cat(paste0("\napplying \'bridge_sampler\' on chain ", j, "...\n\n"))
        }

        all_bridge_samp[[j]] <-
          do.call(bridgesampling::bridge_sampler,
                  c(list(samples=tmp_mat,
                         log_posterior=calc_lpd,
                         data=object$data,
                         lb=lower, ub=upper),
                    ell))
      }

      final_res <- all_bridge_samp[[1]]
      final_res$logml <- mean(vapply(all_bridge_samp,
                                     function(x) x$logml, 0))
      final_res$niter <- max(vapply(all_bridge_samp,
                                    function(x) x$niter, 0))
      final_res$method <- all_bridge_samp[[1]]$method
      final_res$q11 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q11))
      final_res$q12 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q12))
      final_res$q21 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q21))
      final_res$q22 <- unlist(lapply(all_bridge_samp,
                                     function(x) x$q22))

      final_res$all <- all_bridge_samp
      final_res
    }
  }

}





#' Finding latent allocation (component indicators) from an angmcmc object
#' @inheritParams pointest
#' @param ... passed to \link{pointest} to estimate parameters.
#'
#' @details
#' In order to find the latent component indicators, estimates
#' of mixing proportions and model parameters are first computed via
#' pointest. Then, a data point is assigned label j, if the j-th
#' component gives highest density for that point.
#'
#' @return
#' Returns a vector of length n, where n is the length (if univariate) or
#' number of rows (if bivariate) of the data used in original fit.
#' i-th entry of the output vector provides component label for the i-th data point.
#'
#' @examples
#' # first fit a vmsin mixture model
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' # now find latent allocation
#' latent_allocation(fit.vmsin.20)
#' @export
latent_allocation <- function(object, ...)
{

  if (!is.angmcmc(object))
    stop("object must be an \'angmcmc\' object")

  ell <- list(...)
  if (any(!is.null(c(ell$comp.label, ell$par.name)))) {
    warning("\'par.name\' and \'comp.label\' are not used")
    ell$comp.label <- NULL
    ell$par.name <- NULL
  }


  if (object$ncomp > 1) {
    parest <- pointest(object, ...)

    pi.mix <- parest[1, ]
    par.mat <- parest[-1, ]

    if (object$model == "vmsin") {
      l_c_vmsin <- log_const_vmsin_all(par.mat)
      mem_p_model <- mem_p_sin(object$data, par.mat, pi.mix, l_c_vmsin)
    }
    else if (object$model == "vmcos") {
      l_c_vmcos <- log_const_vmcos_all(par.mat, object$qrnd_grid)
      mem_p_model <- mem_p_cos(object$data, par.mat, pi.mix, l_c_vmcos)
    }
    else if (object$model == "wnorm2") {
      l_c_wnorm2 <- log_const_wnorm2_all(par.mat)
      mem_p_model <- mem_p_wnorm2(object$data, par.mat, pi.mix, l_c_wnorm2, object$omega.2pi)
    }
    else if (object$model == "vm") {
      l_c_vm <- log_const_univm_all(par.mat)
      mem_p_model <- mem_p_univm(object$data, par.mat, pi.mix, l_c_vm)
    }
    else {
      l_c_wnorm <- log_const_uniwnorm_all(par.mat)
      mem_p_model <- mem_p_uniwnorm(object$data, par.mat, pi.mix, l_c_wnorm, object$omega.2pi)
    }

    labs <- apply(mem_p_model, 1, which.max_entry1)
  }

  else {
    warning("\'object\' has only one component. A vector of all 1\'s will is returned")
    labs <- rep(1, object$n.data)

  }

  labs
}

#' `BAMBI`: An R package for Bivariate Angular Mixture Models
#'
#' `BAMBI` is an R package that provides functions for fitting
#' (using Bayesian methods) and simulating mixtures of univariate
#' and bivariate angular distributions. Please see the reference for a
#' detailed description of the functionalities of `BAMBI`.
#'
#' @references
#'
#' Chakraborty, S., & Wong, S. W. (2021). BAMBI: An R package for
#' fitting bivariate angular mixture models. *Journal of Statistical Software*,
#' 99 (11), 1-69. \doi{10.18637/jss.v099.i11}
#'
#' @name BAMBI
#' @md
NULL
#> NULL


#' @useDynLib BAMBI, .registration = TRUE
#' @import stats
#' @importFrom stats4 mle
#' @import graphics
#' @importFrom grDevices colorRampPalette
#' @importFrom methods is
#' @importFrom utils tail txtProgressBar
#' @importFrom parallel detectCores
#' @importFrom Rcpp sourceCpp evalCpp

NULL





#' @export
print.angmcmc <- function(x, ...) {

  # browser()
  output <-  paste("Dataset consists of", x$n.data, "observations.")


  if(grepl(x$method, "hmc")) {
    output[2] <- paste(x$ncomp, "component", x$model, "mixture fitted via HMC for model parameters.",
                       ifelse(x$n.chains == 1, "Number of chain = ", "Number of chains = "),
                       paste0(x$n.chains, "."))

    if(x$epsilon.random) {
      output[3] <- paste("epsilon chosen randomly at each iteration with average epsilon =",
                         paste0(format(x$epsilon, scientific=TRUE, digits = 2), collapse = ", "))
    } else {
        output[3] <- paste("epsilon fixed at", x$epsilon )
    }
    if(x$L.random){
      output[4] <- paste("L chosen randomly at each iteration with average L =",
                         paste(round(x$L, 2), collapse = ", "),
                         "across the ", x$n.chains, ifelse(x$n.chains == 1, "chain.", "chains."))
    } else {
      output[4] <- paste("L fixed at", paste(x$L, collapse = ", "),
                         "across the", x$n.chains, ifelse(x$n.chains == 1, "chain.", "chains."))
    }
    output[5] <- paste("acceptance rate for model parameters = ",
                       round(100*mean(x$accpt.modelpar), 2), "%.")
  }


   else if(grepl(x$method, "rwmh")) {

    output[2] <- paste(x$ncomp, "component", x$model, "mixture fitted via RWMH for model parameters.",
                       ifelse(x$n.chains == 1, "Number of chain = ", "Number of chains = "),
                       paste0(x$n.chains, "."))

    output[3] <- paste("proposals are independent normal with variances",
                       paste(format(x$propscale.final, scientific=TRUE, digits = 2), sep = "", collapse = ", "),
                       "for", paste(x$par.name[-1], sep = "", collapse=", "))

    output[4] <- paste("acceptance rate for model parameters = ",
                       round(100*mean(x$accpt.modelpar), 2), "%.")
  }

  output[5] <- paste("Number of iterations =", x$n.iter)
  cat(output, sep = "\n")
}


.onUnload <- function (libpath) {
  library.dynam.unload("BAMBI", libpath)
}

#' @export
print.stepfit <- function(x, ...)
{
  if(x$check_min) {
    output <- paste("Optimum component size =", x$ncomp.best)
    output[2] <- paste("Extract the best fit using the function \'bestmodel()\'")
    cat("\n")
    cat(output, sep = "\n")
  } else {
    warning(paste(toupper(x$crit), "did not attend a first minimum. Probably more clusters are needed."))
    cat("\n")
    cat("Extract the best fit using the function \'bestmodel()\'")
  }
}


#' Angular MCMC (\code{angmcmc}) Object
#' @description Checking for and creating an angmcmc object
#' @param object any R object
#' @param ... arguments required to make an angmcmc object. See details
#' @return logical. Is the input an angmcmc object?
#' @details
#' \code{angmcmc} objects are classified lists that are created when any of the five mixture model fitting
#' functions, viz., \code{fit_vmmix}, \code{fit_wnormmix}, \code{fit_vmsinmix}, \code{fit_vmcosmix} and
#' \code{fit_wnorm2mix} is used. An \code{angmcmc} object contains a number of elements, including the dataset, the
#' model being fitted on the dataset and dimension of the model (univariate or bivariate), the tuning parameters
#' used, MCMC samples for the mixture model parameters, the (hidden) component or cluster indicators for  data
#' points in each iteration and the (iteration-wise) log likelihood and log posterior density values (both calculated
#' upto some normalizing constants). When printed, an angmcmc object returns a brief summary of the function
#' arguments used to produce the object and the average acceptance rate of the proposals (in HMC and RWMH) used
#' over iterations. An \code{angmcmc} object can be used as an argument for the diagnostic and post-processing
#' functions available in \code{BAMBI} for making further inferences.
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' is.angmcmc(fit.vmsin.20)
#' @export

is.angmcmc <- function(object) {
  inherits(object, "angmcmc")
}

#' @rdname is.angmcmc
#' @export
angmcmc <- function(...) {
 ell <- list(...)
 class(ell) <- "angmcmc"
 ell
}

# list of functions


#' The univariate von Mises distribution
#' @inheritParams rvmsin
#' @param x vector of angles (in radians) where the densities are to be evaluated.
#' @param mu vector of means.
#' @param kappa vector of concentration (inverse-variance) parameters; \code{kappa} > 0.
#' @details If \code{mu} and \code{kappa} are not specified they assume the default values of \code{0} and \code{1} respectively.
#' @details The univariate von Mises distribution has density
#' \deqn{f(x) = 1/(2\pi I_0 (\kappa)) \exp(\kappa \cos(x - mu))}
#' where \eqn{I_0 (\kappa)} denotes the modified Bessel function of the first kind with order 0 evaluated at the point \eqn{\kappa}.
#' @return \code{dvm} gives the density  and \code{rvm} generates random deviates.
#'
#' @examples
#'
#' kappa <- 1:3
#' mu <- 0:2
#' x <- 1:10
#' n <- 10
#'
#'
#' # when x and both parameters are scalars, dvm returns a single density
#' dvm(x[1], kappa[1], mu[1])
#'
#' # when x is a vector but both the parameters are scalars, dmv returns a vector of
#' # densities calculated at each entry of x with the same parameters
#' dvm(x, kappa[1], mu[1])
#'
#' # if x is scalar and at least one of the two paraemters is a vector, both parameters are
#' # recycled to the same length, and dvm returns a vector of with ith element being the
#' # density evaluated at x with parameter values kappa[i] and mu[i]
#' dvm(x[1], kappa, mu)
#'
#' # if x and at least one of the two paraemters is a vector, x and the two parameters are
#' # recycled to the same length, and dvm returns a vector of with ith element being the
#' # density at ith element of the (recycled) x with parameter values kappa[i] and mu[i]
#' dvm(x, kappa, mu)
#'
#' # when parameters are all scalars, number of observations generated by rvm is n
#' rvm(n, kappa[1], mu[1])
#'
#' # when at least one of the two parameters is a vector, both are recycled to the same length,
#' # n is ignored, and the number of observations generated by rvm is the same as the length of
#' # the recycled vectors
#' rvm(n, kappa, mu)
#'
#' @export

rvm <- function(n, kappa = 1, mu = 0)
{
  if(any(kappa < 0)) stop("kappa must be non_negative")
  if(any(mu < 0 | mu >= 2*pi)) mu <- prncp_reg(mu)

  if(max(length(kappa), length(mu)) > 1) {
    expanded <- expand_args(kappa, mu)
    kappa <- expanded[[1]]; mu <- expanded[[2]]
    m <- length(kappa)
    out <- rep(0, m)
    for(j in 1:m) {
      if (kappa[j] > 1e-7) {
        out[j] <- c(runivm_onepar(1, kappa[j], mu[j]))
      } else {
        out[j] <- runif(1, 0, 2*pi)
      }
    }

  } else {
    if (kappa> 1e-10) {
      out <- c(runivm_onepar(n, kappa, mu))
    } else {
      out <- runif(n, 0, 2*pi)
    }
  }

  out
}



#' @rdname rvm
#' @export

dvm <- function(x, kappa = 1, mu = 0, log = FALSE)
{
  if(any(kappa < 0)) stop("kappa must be non-negative")
  if(any(mu < 0 | mu >= 2*pi)) mu <- prncp_reg(mu)

  # if(max(length(kappa), length(mu)) > 1) {
  #   expanded <- expand_args(kappa, mu)
  #   kappa <- expanded[[1]]; mu <- expanded[[2]]
  #
  #   if(length(x) > 1) {
  #     x_set <- 1:length(x)
  #     par_set <- 1:length(kappa)
  #     expndn_set <- expand_args(x_set, par_set)
  #     x_set <- expndn_set[[1]]
  #     par_set <- expndn_set[[2]]
  #   } else{
  #     den <- as.vector(dunivm_onex_manypar(x, kappa, mu))
  #   }
  #
  # } else {
  #   if(length(x) > 1){
  #     den <- as.vector(dunivm_manyx_onepar(as.vector(x), kappa, mu))
  #   } else{
  #     den <- exp(ldunivmnum(as.vector(x), c(kappa, mu))) / const_univm(kappa)
  #   }
  # }

  if(max(length(kappa), length(mu)) > 1) {
    expanded <- expand_args(kappa, mu)
    kappa <- expanded[[1]]; mu <- expanded[[2]]
  }

  par.mat <- rbind(kappa, mu)
  n_par <- ncol(par.mat)
  n_x <- length(x)


  if (n_par == 1) {
    den <- c(dunivm_manyx_onepar(as.vector(x), kappa, mu))

  } else if (n_x == 1) {
    den <- c(dunivm_onex_manypar(x, kappa, mu))
  } else {
    x_set <- 1:n_x
    par_set <- 1:n_par
    expndn_set <- expand_args(x_set, par_set)
    x_set <- expndn_set[[1]]
    par_set <- expndn_set[[2]]
    den <- c(dunivm_manyx_manypar(x[x_set], kappa[par_set], mu[par_set]))
  }

  if (log) den <- log(den)

  den
}

#' The univariate von Mises mixtures
#' @inheritParams rvm
#' @param mu vector of component means.
#' @param kappa vector of component concentration (inverse-variance) parameters, \code{kappa > 0}.
#' @param pmix vector of mixing proportions.
#' @details \code{pmix}, \code{mu} and \code{kappa} must be of the same length, with \eqn{j}-th element corresponding to the \eqn{j}-th component of the mixture distribution.
#' @details The univariate von Mises mixture distribution with component size \code{K = \link{length}(pmix)} has density
#' \deqn{g(x) = p[1] * f(x; \kappa[1], \mu[1]) + ... + p[K] * f(x; \kappa[K], \mu[K])}
#' where \eqn{p[j], \kappa[j], \mu[j]} respectively denote the mixing proportion, concentration parameter and the mean parameter for the \eqn{j}-th component
#' and \eqn{f(. ; \kappa, \mu)} denotes the density function of the (univariate) von Mises distribution with mean parameter \eqn{\mu} and concentration parameter \eqn{\kappa}.
#' @return \code{dvmmix} computes the density  and \code{rvmmix} generates random deviates from the mixture density.
#'
#' @examples
#' kappa <- 1:3
#' mu <- 0:2
#' pmix <- c(0.3, 0.3, 0.4)
#' x <- 1:10
#' n <- 10
#'
#' # mixture densities calculated at each point in x
#' dvmmix(x, kappa, mu, pmix)
#'
#' # number of observations generated from the mixture distribution is n
#' rvmmix(n, kappa, mu, pmix)
#'
#' @export

rvmmix <- function(n, kappa, mu, pmix)
{

  allpar <- list(kappa=kappa, mu=mu, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 4) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$kappa < 0)) stop("kappa must be non-negative")
  if(any(allpar$mu < 0 | allpar$mu >= 2*pi)) allpar$mu <- prncp_reg(allpar$mu)

  out <- rep(0, n)
  ncomp <- allpar_len[1] # number of components
  comp_ind <- cID(tcrossprod(rep(1, n), allpar$pmix), ncomp, runif(n))
  # n samples from multinom(ncomp, pmix)
  for(j in seq_len(ncomp)) {
    obs_ind_j <- which(comp_ind == j)
    n_j <- length(obs_ind_j)
    if(n_j > 0) {
      out[obs_ind_j] <- rvm(n_j, kappa[j],  mu[j])
    }
  }

  out
}


#' @rdname rvmmix
#' @export
dvmmix <- function(x, kappa, mu, pmix, log=FALSE)
{
  allpar <- list(kappa=kappa, mu=mu, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 4) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$pmix < 0)) stop("pmix must be non-negative")
  if(sum(allpar$pmix) != 1) {
    allpar$pmix <- allpar$pmix/sum(allpar$pmix)
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$kappa < 0)) stop("kappa must be non-negative")
  if(any(allpar$mu < 0 | allpar$mu >= 2*pi)) allpar$mu <- prncp_reg(allpar$mu)


  ncomp <- length(kappa)

  allcompden <- vapply(1:ncomp,
                       function(j) dvm(x, kappa[j], mu[j], FALSE),
                       rep(0, length(x)))

  mixden <- c(allcompden %*% pmix)

  if (log) {
    log(mixden)
  } else {
    mixden
  }

}


#' Fitting univariate von Mises mixtures using MCMC
#' @inheritParams fit_vmsinmix
#'
#' @details
#' Wrapper for \link{fit_angmix} with \code{model = "vm"}.
#'
#'
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vm.20 <- fit_vmmix(wind$angle, ncomp = 3, n.iter =  20,
#'                        n.chains = 1)
#' fit.vm.20
#' @export


fit_vmmix <- function(...)
{
  fit_angmix(model="vm", ...)
}

bounded_hmc_biv <- function(lpr_grad,
                            init,
                            lower,
                            upper,
                            dep_cov = FALSE,
                            dep_cov_type = NULL,
                            zero_cov = FALSE,
                            nsteps = 1,
                            step)
{

  broken <- FALSE # will be true if breaks
  apr <- NULL #will  be replaced if doesn't break
  delta <- NULL #will be replaced if doesn't break

  lower1 <- lower
  upper1 <- upper

  ncomp <- 1
  npar <- 5
  kappa3_index <- 3
  # positions of kappa3 in vec(par_mat)

  lpr_grad.init <- lpr_grad(init)
  lpr.init <-   lpr_grad.init$lpr
  gr <- lpr_grad.init$grad

  init.p <- matrix(rnorm(npar), 5, ncomp)

  # Compute the kinetic energy at the start of the trajectory.
  kinetic.init <- sum(init.p^2) / 2

  # Compute the trajectory by the leapfrog method.
  q <- init
  p <- init.p

  if (zero_cov) {
    p[kappa3_index] <- 0
  }


  reflections <- 0

  # Make a half step for momentum at the beginning.

  p <- p + (step/2) * gr

  if (zero_cov) {
    p[kappa3_index] <- 0
  }

  for (i in 1:nsteps)
  {
    # Make a full step for the position.

    q <- q + step * p

    # check if broken
    if (any(is.nan(c(p, q)))) {
      broken <- TRUE
      #stop("Algorithm breaks. Try a smaller epsilon.")
      break
    }


    # Check for bound violations, and adjust position and momentum
    for(k in 1:npar) {
      # adjust the bound for the cov
      # param if dependent (for wnorm2)
      if(k == 3) {

        if (zero_cov) {
          next
        }
        else if (dep_cov) {

          # if (any(exp(q[1:2]) <= 0)) {
          #   broken <- TRUE
          #   break
          # }

          if (dep_cov_type %in% c("wnorm2_bound", "vmsin_unimodal")) {
            bd_k1k2 <- sqrt(exp(q[1])*exp(q[2]))
            lower[k] <- max(lower1[k], -bd_k1k2)
            upper[k] <- min(upper1[k], bd_k1k2)
          }
          else {
            # dep_cov_type == "vmcos_unimodal"
            lower[k] <- max(lower1[k], -exp(q[1])*exp(q[2])/(exp(q[1])+exp(q[2])))
          }

        }

        # if (any(is.nan(c(upper[k], lower[k])))) {
        #   broken <- TRUE
        #   break
        # }
      }

      reflections_curr <- 0L

      while(all(q[k] < lower[k] || q[k] > upper[k],
                reflections_curr <= 100)) {
        if (q[k]<lower[k]) {
          q[k] <- lower[k] + (lower[k] - q[k])
          p[k] <- -p[k]
          reflections_curr <- reflections_curr + 1L
        }

        else if (q[k]>upper[k]) {
          q[k] <- upper[k] - (q[k] - upper[k])
          p[k] <- -p[k]
          reflections_curr <- reflections_curr + 1L
        }

      }

      reflections <- reflections + reflections_curr

      if (reflections_curr >= 100) {
        broken <- TRUE
        break
      }

    }

    # check for broken, bound violation, or nan
    # if so then break from i loop
    if (any(is.nan(c(p, q)),
            broken)) {
      broken <- TRUE
      #stop("Algorithm breaks. Try a smaller epsilon.")
      break
    }


    # browser()

    # Evaluate the gradient at the new position, provided not broken

    lpr_grad_current <- lpr_grad(q)
    lr <- lpr_grad_current$lpr
    gr <- lpr_grad_current$grad


    # Make a full step for the momentum, except when we're coming to the end of
    # the trajectory.

    if (i != nsteps) {
      p <- p + step * gr
      if (zero_cov) {
        p[kappa3_index] <- 0
      }
    }
  }


  # Make a half step for momentum at the end.

  p <- p + (step/2) * gr
  if (zero_cov) {
    p[kappa3_index] <- 0
  }

  # Negate momentum at end of trajectory to make the proposal symmetric.

  p <- -p


  # Look at log probability and kinetic energy at the end of the trajectory.
  lpr.prop <- lr
  kinetic.prop <- sum(p^2) / 2

  # Accept or reject the state at the end of the trajectory.
  H.init <- -lpr.init + kinetic.init
  H.prop <- -lpr.prop + kinetic.prop
  delta <- H.prop - H.init
  apr <- min(1,exp(-delta))


  if (any(is.nan(c(p, q, apr)))) {
    broken <- TRUE
    apr <- NULL
    delta <- NULL
    #stop("Algorithm breaks. Try a smaller epsilon.")
  }


  if(broken) # reject
  {
    final.q <- init
    final.p <- init.p
    lpr.final <- lpr.init
    acc <- 0
  }
  else if (runif(1) > apr) # reject
  {
    final.q <- init
    final.p <- init.p
    lpr.final <- lpr.init
    acc <- 0
  }
  else  # accept
  {
    final.q <- q
    final.p <- p
    lpr.final <- lr
    acc <- 1
  }

  # Return new state, its log probability and gradient, plus additional
  # information, including the trajectory, if requested.

  out <- list (final=final.q, final.p=final.p, lpr=lpr.final, step=step,
               apr=apr, accpt=acc, delta=delta, reflections=reflections, broken = broken)

  out
}


bounded_hmc_uni <- function(lpr_grad,
                            init,
                            lower,
                            upper,
                            nsteps = 1,
                            step)
{

  broken <- FALSE # will be true if breaks
  apr <- NULL #will  be replaced if doesn't break
  delta <- NULL #will be replaced if doesn't break

  ncomp <- ncol(init)
  npar <- 2*ncomp

  lpr_grad.init <- lpr_grad(init)
  lpr.init <-   lpr_grad.init$lpr
  gr <- lpr_grad.init$grad

  init.p <- matrix(rnorm(npar), 2, ncomp)

  # Compute the kinetic energy at the start of the trajectory.
  kinetic.init <- sum(init.p^2) / 2

  # Compute the trajectory by the leapfrog method.
  q <- init
  p <- init.p


  reflections <- 0

  # Make a half step for momentum at the beginning.

  p <- p + (step/2) * gr

  # browser()
  # Alternate full steps for position and momentum.

  for (i in 1:nsteps)
  {
    # Make a full step for the position.

    q <- q + step * p

    # check if broken
    if (any(is.nan(c(p, q)))) {
      broken <- TRUE
      #stop("Algorithm breaks. Try a smaller epsilon.")
      break
    }


    # Check for bound violations, and adjust position and momentum
    for(k in 1:npar) {
      reflections_curr <- 0L

      while(all(q[k] < lower[k] || q[k] > upper[k],
                reflections_curr <= 100)) {
        if (q[k]<lower[k]) {
          q[k] <- lower[k] + (lower[k] - q[k])
          p[k] <- -p[k]
          reflections_curr <- reflections_curr + 1L
        }

        else if (q[k]>upper[k]) {
          q[k] <- upper[k] - (q[k] - upper[k])
          p[k] <- -p[k]
          reflections_curr <- reflections_curr + 1L
        }

      }

      reflections <- reflections + reflections_curr

      if (reflections_curr >= 100) {
        broken <- TRUE
        break
      }
    }

    # check if broken
    if (any(is.nan(c(p, q)), broken)) {
      broken <- TRUE
      #stop("Algorithm breaks. Try a smaller epsilon.")
      break
    }


    # Evaluate the gradient at the new position.

    lpr_grad_current <- lpr_grad(q)
    lr <- lpr_grad_current$lpr
    gr <- lpr_grad_current$grad


    # Make a full step for the momentum, except when we're coming to the end of
    # the trajectory.

    if (i != nsteps) {
      p <- p + step * gr
    }
  }


  # Make a half step for momentum at the end.

  p <- p + (step/2) * gr

  # Negate momentum at end of trajectory to make the proposal symmetric.

  p <- -p


  # Look at log probability and kinetic energy at the end of the trajectory.
  lpr.prop <- lr
  kinetic.prop <- sum(p^2) / 2

  # Accept or reject the state at the end of the trajectory.
  H.init <- -lpr.init + kinetic.init
  H.prop <- -lpr.prop + kinetic.prop
  delta <- H.prop - H.init
  apr <- min(1,exp(-delta))


  if (any(is.nan(c(p, q, apr)))) {
    broken <- TRUE
    apr <- NULL
    delta <- NULL
    #stop("Algorithm breaks. Try a smaller epsilon.")
  }


  if(broken) # reject
  {
    final.q <- init
    final.p <- init.p
    lpr.final <- lpr.init
    acc <- 0
  }
  else if (runif(1) > apr) # reject
  {
    final.q <- init
    final.p <- init.p
    lpr.final <- lpr.init
    acc <- 0
  }
  else  # accept
  {
    final.q <- q
    final.p <- p
    lpr.final <- lr
    acc <- 1
  }

  # Return new state, its log probability and gradient, plus additional
  # information, including the trajectory, if requested.

  out <- list (final=final.q, final.p=final.p, lpr=lpr.final, step=step,
               apr=apr, accpt=acc, delta=delta, reflections=reflections, broken = broken)

  out
}

# based on uniroot.all function from package rootSolve
# the package is now archived
uniroot.all <- function(
  f,
  interval,
  lower = min(interval),
  upper = max(interval),
  tol = .Machine$double.eps^0.2,
  maxiter = 1000,
  trace = 0,
  n = 100,
  ...
) {
  xseq <- seq(lower, upper, len = n + 1)
  f_xseq <- f(xseq, ...)
  out <- xseq[which(f_xseq == 0)]
  f_sign <- f_xseq[1:n] * f_xseq[2:(n + 1)]
  i_range <- which(f_sign < 0)
  for (i in i_range)  {
    out <- c(
      out,
      uniroot(f, lower = xseq[i],
              upper = xseq[i + 1],
              maxiter = maxiter,
              tol = tol,
              trace = trace,
              ...)$root
    )
  }
  out
}

# find_lscale <- function(x) {
#   y <- x
#   y[1:2] <- log(x[1:2])
#   y
# }

find_expscale <- function(x) {
  y <- x
  y[1:2] <- exp(x[1:2])
  y
}

find_expscale_uni <- function(x) {
  y <- x
  y[1] <- exp(x[1])
  y
}


find_lscale_mat <- function(x) {
  y <- x
  y[1:2, ] <- log(x[1:2, ])
  y
}

find_lscale_mat_uni <- function(x) {
  y <- x
  y[1, ] <- log(x[1, ])
  y
}





#' Fitting Bivariate and univariate angular mixture models
#'
#' @importFrom gtools rdirichlet
#' @importFrom future.apply future_lapply
#'
#' @param model angular model whose mixtures are to be fitted. Available choices are \code{"vmsin", "vmcos"} and \code{"wnorm2"} for
#' bivariate data, and \code{"vm"} and \code{"wnorm"} for univariate data.
#' @param data data matrix (if bivarate, in which case it must have two columns) or vector. If outside, the values
#' are transformed into the scale \eqn{[0, 2\pi)}. *Note:* BAMBI cannot handle missing data. Missing values must
#' either be removed or properly imputed.
#' @param ncomp number of components in the mixture model. Must be a positive integer. vector values are not allowed.
#' If \code{comp == 1}, a single component model is fitted.
#' @param start_par list with elements \code{pmix} (ignored if \code{comp == 1}), together with \code{kappa1, kappa2, mu1} and \code{mu2},
#' for bivariate models, and \code{kappa} and \code{mu} for univariate models,
#' all being vectors of length same as \code{ncomp}.
#' These provides the starting values for the Markov chain; with \eqn{j}-th component of each vector corresponding to the \eqn{j}-th
#' component of the mixture distribution. If missing, the data is first clustered into \code{ncomp} groups either via k-means (after
#' projecting onto a unit sphere), or randomly, depending on \code{rand_start},  and then moment estimators for components are used as
#' the starting points. Note that a very wrong starting point can potentially lead the chain to get stuck at a wrong solution for thousands
#' of iterations. As such, we recommend using the default option, which is k-means followed by moment estimation.
#' @param cov.restrict Should there be any restriction on the covariance parameter for a bivariate model. Available choices are
#' \code{"POSITIVE", "NEGATIVE", "ZERO"} and "NONE". Note that \code{"ZERO"} fits a mixture with product components. Defaults to
#' \code{"NONE"}.
#' @param unimodal.component logical. Should each component in the mixture model be unimodal? Only used if \code{model} is either \code{"vmsin"}
#' or \code{"vmcos"}. Defaults to FALSE.
#' @param n.chains number of chains to run. Must be a positive integer.
#' @param chains_parallel logical. Should the chains be run in parallel? Defaluts to TRUE, and ignored if \code{n.chains} = 1.
#' Note that parallelization is implemented via \link[future.apply]{future_lapply} from package \code{future.apply} which
#' uses futures for this purpose, and thus provides a convenient way of parallelization across various OSs and computing environments.
#' However, a proper \link[future]{plan} must be set for the parallization before running the chain. Otherwise the chains will run sequentially.
#' @param method MCMC strategy to be used for the model paramters:  \code{"hmc"} or \code{"rwmh"}.
#' @param perm_sampling logical. Should the permutation sampling algorithm of Fruhwirth-Schnatter (2001) be used?
#' If TRUE, at every iteration after burnin, once model parameters and mixing proportions are sampled,
#' a random permutation of 1, ..., ncomp is considered, and components are relabelled according
#' to this random permutation. This forced random label switchings may imporve the mixing rate of the chage. However, (automated) tuning
#' is very difficult with such a scheme, as there is no simple way of keeping track of the "original" component labels. This creates problem
#' with computing standard deviations of the generated model parameters, thus making the
#' scaling step used in tuning for \code{epsilon} or \code{paramscale} problematic as well. As such, \code{perm_sampling} is always turned
#' off during burn-in (even if \code{autotune = FALSE}), and turned on thereafter, if \code{TRUE}.
#' Defaults to and is set to \code{FALSE}.
#' @param int.displ absolute integer displacement for each coordinate for \code{wnorm} and \code{wnorm2} models (ignored otherwise). Default is 3.
#' Allowed minimum and maximum are 1 and 5 respectively.
#' @param epsilon,L  tuning parameters for HMC; ignored if \code{method = "rwmh"}. \code{epsilon} (step-size) is a single number,
#' or a vector of size \code{2*ncomp} for univariate models and \code{5*ncomp} for bivariate models. Note that the "mass matrix"
#' in HMC is assumed to be identity. As such, \code{epsilon}'s corresponding to different model parameters need to be in proper scale for
#' optimal acceptance rate. Can be autotuned during burnin. See \code{autotune}.
#' \code{L} (leapfrog steps) is a positive integer or a vector of positive integers of length \code{n.chains}.
#' If multiple chains are used, we suggest same \code{L} values acorss different chains to make the chains as homogenous as possible.
#'
#' @param epsilon.random logical. Should \code{epsilon*delta}, where \code{delta} is a random
#' number between \code{(1-epsilon.incr, 1+epsilon.incr)} be used instead of \code{epsilon} at each iteration?
#' Ignored if \code{method = "rwmh"}.
#' @param L.random logical. Should a random integer between \code{L.orig/exp(L.incr)} and \code{L.orig*exp(L.incr)}be used instead as \code{L}
#' at each iteration? Ignored if \code{method = "rwmh"}. Defaults to \code{TRUE}.
#' @param L.incr amount of randomness incorporated in L if \code{L.random = TRUE}.
#' @param epsilon.incr amount of randomness incorporated in \code{epsilon} if \code{epsilon.random = TRUE}.
#' @param propscale tuning parameters for RWMH; a vector of size 5 (for bivariate models) or 2 (for univariate models) representing
#' the variances for the proposal normal densities
#' for the model parameters. Ignored if \code{method = "hmc"}. Can be autotuned during burnin. See \code{autotune}.
#' @param n.iter number of iterations for the Markov Chain.
## @param gam.loc,gam.scale location and scale (hyper-) parameters for the gamma prior for \code{kappa1} and \code{kappa2}. See
## \link{dgamma}. Defaults are \code{gam.loc = 0, gam.scale = 1000} that makes the prior non-informative.
#' @param pmix.alpha concentration parameter(s) for the Dirichlet prior for \code{pmix}. Must either be a positive real number, or a vector
#' with positive entries and of length \code{ncomp}. The default is \eqn{(r+r(r+1)/2)/2+3}, where \eqn{r} is 1 or 2 according as whether
#' the model is univariate or bivariate. Note that it is recommended to use larger \code{alpha} values to ensure the a good posterior behavior,
#' especially when \link{fit_incremental_angmix} is used for model selection, which handles overfitting in "let two component-specific parameters be
#  identical", uses total number of components in the fitted model as the estimator for true component
#' size, and then penalizes for model complexity. See Fruhwirth-Schnatter (2011) for more details on this.
#' @param norm.var variance (hyper-) parameters in the normal prior for \code{log(kappa), log(kappa1), log(kappa2)} and \code{kappa3}. (Prior mean is zero).
#' Can be a vector. Default is 1000 that makes the prior non-informative.
#' @param burnin.prop proportion of iterations to used for burnin. Must be a be a number in [0, 1].
#' Default is 0.5.
#' @param thin thining size to be used. Must be a positive integer. If \code{thin = } n, then every nth iteration is reatained
#' in the final MCMC sample.
#' @param  autotune logical. Should the Markov chain auto-tune the parameter \code{epsilon} (in HMC) or
#' \code{propscale} (in RWMH) during burn-in?  Set to \code{TRUE} by default. An adaptive tuning strategy is implemented.
#' Here, at every 10th iteration during in burn-in, the acceptance ratio in the last \code{tune_ave_size}
#' iterations is calculated. Then the tuning parameter is decreased  (increased) by a factor of
#' \code{1-tune.incr} (\code{1+tune.incr}) if the calculated acceptance rate
#' falls below (above) \code{accpt.prob.lower} (\code{accpt.prob.upper}). In addditon, when \code{iter} is a multiple of
#' \code{tune_ave_size}, \code{epsilon} for each model parameter is rescaled via the standard deviation of
#' the corresponding parameter over the past \code{tune_ave_size} iterations.
#' @param tune.prop proportion of *\code{burnin}* used to tune the parameters (\code{epsilon} in HMC and
#' \code{propscale} in RWMH). Must be a number between 0 and 1; defaults to 1.  Ignored if \code{autotune == FALSE}.
#' @param show.progress logical. Should a progress bar be displayed?
#' @param accpt.prob.lower,accpt.prob.upper lower and upper limits of acceptance ratio to be maintained while tuning
#' during burn-in. Must be numbers between 0 and 1, which \code{accpt.prob.lower < accpt.prob.upper}. See \code{autotune}. Default to (0.6, 0,9) for HMC and  (0.3, 0.5) for RWMH.
#' Ignored if \code{autotune = FALSE}.
#' @param tune.incr how much should the tuning parameter be increased or decreased at each step while tuning during burn-in?
#' Must be a number between 0 and 1. See \code{autotune}. Defaults to 0.05. Ignored if \code{autotune = FALSE}.
#' @param rand_start logical. Should a random starting clustering be used? Must be either a scalar, or a vector of length \code{ncomp},
#' one for each chain. Ignored if \code{start_par} is supplied. See \code{start_par} for more details. Defaults to \code{FALSE}.
#' @param tune_ave_size number previous iterations used to compute the acceptance rate while tuning in burn-in. Must be a positive
#' integer. Defaults to 100.
#' @param qrnd,n_qrnd Used only if \code{method="vmcos"}. See \link{dvmcos} for details.
#' @param kappa_upper,kappa_lower upper and lower bounds for the concentration and (absolute) association parameters. Must be a positive integers. Defaults to 150 and 1e-4,
#' and parameter with value above or below these limits rarely make sense in practice.
#' Warning: values much larger or smaller than the default are not recommended as they can cause numerical instability.
#' @param return_llik_contri logical. Should the log likelihood contribution of each data point for each MCMC iteration in each chain be returned? This makes
#' computation of \link{waic.angmcmc} and \link{loo.angmcmc} much faster. *Warning*: Depending on the length of data and \code{n.iter}, this can be
#' very memory intensive. We suggest setting \code{return_llik_contri = TRUE} only if \link{waic.angmcmc} and \link{loo.angmcmc} are aimed for. Defaults to
#' \code{FALSE}.
#' @param return_tune_param logical. Should the values of the tuning parameters used at each iteration in each chain be returned? Defaults to \code{FALSE}.
#' @param ... Unused.
#'
#' @note
#' Sampling is done in log scale for the concentration parameters (kappa, kappa1 and kappa2).
#'
#'
#' Parallelization is done by default when more than one chain is used,
#' but the chains can be run sequentially as well by setting
#' \code{chains_parallel = FALSE}. To retain reproducibility while running
#' multiple chains in parallel, the same RNG state is passed at the
#' beginning of each chain. This is done by specifying \code{future.seed = TRUE}
#' in \code{future.apply::future_lapply} call. Then at the beginning of the i-th
#' chain, before drawing any parameters, i-many Uniform(0, 1) random numbers are
#' generated using \code{runif(i)} (and then thrown away). This ensures that the
#' RNG states across chains prior to random generation of the parameters are
#' different, and hence, no two chains can become identical, even if they have
#' the same starting and tuning parameters. This, however creates a difference
#' between a \code{fit_angmix} call with multiple chains which is run sequentially
#' by setting \code{chains_parallel = FALSE}, and another which is run sequentially
#' because of a sequential \code{plan()} (or no \code{plan()}), with
#' \code{chains_parallel = TRUE}. In the former, different RNG states are passed at
#' the initiation of each chain.
#'
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.vmsin.20 <- fit_angmix("vmsin", tim8,
#'   ncomp = 3, n.iter = 20,
#'   n.chains = 1
#' )
#' fit.vmsin.20
#'
#'
#' # Parallelization is implemented via future_lapply from the
#' # package future.apply. To parallelize, first provide a parallel
#' # plan(); otherwise the chains will run sequentially.
#' # Note that not all plan() might work on every OS, as they execute
#' # functions defined internally in fit_mixmodel. We suggest
#' # plan(multisession) which works on every OS.
#' \donttest{
#' library(future)
#' library(parallel)
#' # plan(multisession, gc = TRUE) # parallelize chains
#'
#' set.seed(1)
#' MC.fit <- fit_angmix("vmsin", tim8,
#'   ncomp = 3, n.iter = 5000,
#'   n.chains = 3
#' )
#'
#'
#' pointest(MC.fit)
#'
#' MC.fix <- fix_label(MC.fit)
#'
#' contour(MC.fit)
#' contour(MC.fix)
#' lpdtrace(MC.fit)
#' }
#'
#' @references
#' Fruhwirth-Schnatter, S. (2011). Label switching under model uncertainty. Mixtures: Estimation and Application, 213-239.
#'
#' Fruhwirth-Schnatter, S. (2001). Markov chain Monte Carlo estimation of classical and dynamic switching and mixture models. Journal of the American Statistical Association, 96(453), 194-209.
#'
#' @export

fit_angmix <- function(model = "vmsin",
                       data,
                       ncomp,
                       cov.restrict = "NONE",
                       unimodal.component = FALSE,
                       start_par = NULL,
                       rand_start = rep(FALSE, n.chains),
                       method = "hmc",
                       perm_sampling = FALSE,
                       n.chains = 3,
                       chains_parallel = TRUE,
                       return_llik_contri = FALSE,
                       int.displ = 3,
                       epsilon = 0.1,
                       L = 10,
                       epsilon.random = TRUE,
                       L.random = FALSE,
                       burnin.prop = 0.5,
                       tune.prop = 1,
                       thin = 1,
                       propscale = 0.05,
                       n.iter = 500,
                       # gam.loc = 0.001,
                       # gam.scale = 1000,
                       pmix.alpha = NULL,
                       norm.var = 1000,
                       autotune = TRUE,
                       show.progress = TRUE,
                       accpt.prob.upper,
                       accpt.prob.lower,
                       epsilon.incr = 0.05,
                       L.incr = 0.075,
                       tune.incr = 0.05,
                       tune_ave_size = 100,
                       kappa_upper = 150,
                       kappa_lower = 1e-4,
                       return_tune_param = FALSE,
                       qrnd = NULL,
                       n_qrnd = NULL, ...) {

  # if(is.null(dim(data)) | !(mode(data) %in% c("list", "numeric", "data.frame") && ncol(data) == 2)) stop("non-compatible data")


  dots <- list(...)

  backward_compatible <- FALSE
  backward_compatible <- dots$backward_compatible
  if (is.null(backward_compatible)) {
    backward_compatible <- FALSE
  }

  stopifnot(
    is.logical(backward_compatible)
  )

  # progress.backend <- "tcltk"
  # progressor <- function(...) NULL
  progress.backend <- "tcltk"


  signif_ <- if (backward_compatible) function(x, ...) x else function(x, ...) signif(x, 8)



  if (length(model) > 1) stop("\'model\' must be a scalar")

  if (model %in% c("vmsin", "vmcos", "wnorm2")) {
    type <- "bi"
  } else if (model %in% c("vm", "wnorm")) {
    type <- "uni"
  } else {
    stop("non-compatible model")
  }

  if (missing(ncomp)) {
    stop("\'ncomp\' is missing, with no default.")
  }

  if (length(ncomp) > 1) {
    stop("\'ncomp\' must be a scalar")
  }


  if (any(length(n.chains) > 1, length(n.chains) < 1, n.chains == 0)) {
    stop("Invalid n.chains")
  }

  zero_cov <- FALSE

  if (type == "bi") {
    if (!(is.matrix(data) | is.data.frame(data))) {
      stop("\'data\' must be a two column matrix for model = \'vmsin\', \'vmcos\' and \'wnorm2\'")
    }

    if (ncol(data) != 2) {
      stop("\'data\' must be a two column matrix for model = \'vmsin\', \'vmcos\' and \'wnorm2\'")
    }

    if (length(cov.restrict) > 1 |
        !cov.restrict %in% c("NONE", "POSITIVE", "NEGATIVE", "ZERO")) {
      stop("cov.restrict must be one of \"NONE\", \"POSITIVE\", \"NEGATIVE\" and  \"ZERO\"")
    }

    if (any(is.na(data))) {
      stop(paste(
        "\'data\' contains missing value(s). BAMBI cannot handle missing data.\n",
        "Either remove them or properly impute them."
      ))
    }
    data.rad <- rm_NA_rad(data)
    n.data <- nrow(data.rad)

    npar_1_comp <- 5
    par.names <- c("kappa1", "kappa2", "kappa3", "mu1", "mu2")

    par_lower <- replicate(ncomp, c(kappa_lower, kappa_lower, -kappa_upper, 0, 0))
    par_upper <- replicate(ncomp, c(
      kappa_upper, kappa_upper,
      kappa_upper, 2 * pi, 2 * pi
    ))

    if (cov.restrict == "POSITIVE") {
      par_lower[3, ] <- 0
    } else if (cov.restrict == "NEGATIVE") {
      par_upper[3, ] <- 0
    } else if (cov.restrict == "ZERO") {
      par_lower[3, ] <- par_upper[3, ] <- 0
      zero_cov <- TRUE
    }

    if (missing(pmix.alpha)) {
      pmix.alpha <- 5.5
    }
  } else {
    if ((is.matrix(data) | is.data.frame(data))) {
      data <- as.vector(as.matrix(data))
    }

    if (!is.numeric(data)) {
      stop("\'data\' must be a vector for \'model\' = \'vm\' and \'wnorm\'")
    }

    data <- as.numeric(data)
    data.rad <- rm_NA_rad(data)
    n.data <- length(data.rad)

    npar_1_comp <- 2
    par.names <- c("kappa", "mu")

    par_lower <- replicate(ncomp, c(kappa_lower, 0))
    par_upper <- replicate(ncomp, c(kappa_upper, 2 * pi))

    if (missing(pmix.alpha)) {
      pmix.alpha <- 4
    }
  }

  if (any(
    length(perm_sampling) > 1, length(perm_sampling) < 1,
    !is.logical(perm_sampling)
  )) {
    stop("Invalid perm_sampling")
  }


  if (any(
    length(chains_parallel) > 1, length(chains_parallel) < 1,
    !is.logical(chains_parallel)
  )) {
    stop("Invalid chains_parallel")
  }

  if (any(
    length(chains_parallel) > 1, length(chains_parallel) < 1,
    !is.logical(chains_parallel)
  )) {
    stop("Invalid chains_parallel")
  }

  if (any(
    length(autotune) > 1, length(autotune) < 1,
    !is.logical(autotune)
  )) {
    stop("Invalid autotune")
  }

  if (any(
    length(unimodal.component) > 1, length(unimodal.component) < 1,
    !is.logical(unimodal.component)
  )) {
    stop("Invalid unimodal.component")
  }

  if (any(
    length(return_llik_contri) > 1, length(return_llik_contri) < 1,
    !is.logical(return_llik_contri)
  )) {
    stop("Invalid return_llik_contri")
  }

  if (any(
    length(return_tune_param) > 1, length(return_tune_param) < 1,
    !is.logical(return_tune_param)
  )) {
    stop("Invalid return_tune_param")
  }





  if (length(pmix.alpha) == 1) {
    pmix.alpha <- rep(pmix.alpha, ncomp)
  } else if (length(pmix.alpha) != ncomp) {
    stop("length(pmix.alpha) and ncomp differ")
  }


  if (n.chains < 1) {
    stop("\'n.chains\' must be a positive integer")
  }

  if (n.chains == 1) {
    chains_parallel <- FALSE
  }




  if (length(burnin.prop) != 1 || burnin.prop < 0 || burnin.prop >= 1) {
    stop("\"burnin.prop\" must be a number in [0, 1)")
  }

  if (thin < 1) {
    stop("\"thin\" must be a positive integer")
  }

  if (tune.incr <= 0 | tune.incr >= 1) {
    stop("\'tune.incr\' must be between 0 and 1")
  }

  n.burnin <- ceiling(burnin.prop * n.iter)

  if (length(tune.prop) != 1 || tune.prop < 0 || tune.prop > 1) {
    stop("\"tune.prop\" must be in [0, 1]")
  }

  iter.tune <- ceiling(burnin.prop * tune.prop * n.iter)

  if (iter.tune > n.burnin) {
    iter.tune <- n.burnin
  }

  n.iter.final <- n.iter - n.burnin
  burnin_iter <- seq_len(n.burnin)
  thin <- round(thin)
  thin_filter <- c(TRUE, rep(FALSE, thin - 1))
  final_iter_set <- (seq_len(n.iter))[-burnin_iter][thin_filter]

  # gam.rate <- 1/gam.scale

  curr.model <- model

  if (length(rand_start) == 1) {
    rand_start <- rep(rand_start, n.chains)
  }


  if (method == "hmc") {
    if (any(L < 1)) {
      stop("\'L\' must be a positive integer")
    }

    L <- ceiling(L)

    if (length(L) == 1) {
      L <- rep(L, n.chains)
    } else if (length(L) != n.chains) {
      stop("\'L\' must be a vector of length \'n.chains\'")
    } else {
      rand_start <- rep(rand_start, n.chains)
    }

    if (missing(accpt.prob.upper)) {
      accpt.prob.upper <- 0.9
    }

    if (missing(accpt.prob.lower)) {
      accpt.prob.lower <- if (model %in% c("wnorm", "wnorm2")) 0.58 else 0.6
    }

    if (length(epsilon) == ncomp * npar_1_comp) {
      tune_param <- matrix(c(epsilon), npar_1_comp, ncomp)
    } else if (length(epsilon) == 1) {
      tune_param <- matrix(epsilon, npar_1_comp, ncomp)
    } else {
      stop("epsilon must either be a scalar or a vector of length 2*ncomp (univariate) or 5*ncomp (bivariate)")
    }
  }

  #
  #   if (method == "hmc" & sum(rand_start) == 0 & n.chains > 1 & chains_parallel) {
  #     if (length(L) == 1 | max(L) == min(L)) {
  #       warning(paste("same L accross multiple chains while running them in",
  #                     "parllel with rand_start = FALSE will just make",
  #                     "identical copies of the same chain. L is changed"))
  #       L <- seq(ceiling(L/2), ceiling(2*L), length.out = n.chains)
  #     }
  #   }


  if (grepl(method, "rwmh")) # using rwmh
  {
    if (missing(accpt.prob.upper)) {
      accpt.prob.upper <- 0.5
    }

    if (missing(accpt.prob.lower)) {
      accpt.prob.lower <- 0.3
    }

    if (length(propscale) == ncomp * npar_1_comp) {
      tune_param <- matrix(c(propscale), npar_1_comp, ncomp)
    } else if (length(propscale) == 1) {
      tune_param <- matrix(propscale, npar_1_comp, ncomp)
    } else {
      stop("propscale must either be a scalar or a vector of length 2*ncomp (univariate) or 5*ncomp (bivariate)")
    }
  }

  nterms <- tune_ave_size

  if (!model %in% c("wnorm", "wnorm2")) {
    int.displ <- omega.2pi <- NULL
  }


  if (model != "vmcos") {
    qrnd_grid <- NULL
    n_qrnd <- NULL
  }
  # Now rename the model specific compiled llik and grad functions


  if (model == "vmsin") {
    if (unimodal.component) {
      dep_cov <- TRUE
      dep_cov_type <- "vmsin_unimodal"
    } else {
      dep_cov <- FALSE
      dep_cov_type <- NULL
    }


    # # in log scale
    # lpd_grad_model_indep_1comp <-  function(data, par_vec_lscale,
    #                                         obs_group, n.clus) {
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   lpd_grad <- matrix(NA, 6, 1)
    #   if (n.clus > 0) {
    #     lpd_grad <- (grad_llik_vmsin_C(data[obs_group, , drop=FALSE],
    #                                    par_vec) +
    #                    c( # grad for lprior
    #                      (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #                      # lprior
    #                      sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #                            gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #                    )) * c(par_vec[1:2], rep(1, 4))
    #   } else {
    #     lpd_grad <-
    #       c( # grad for lprior
    #         (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #         # lprior
    #         sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #               gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #       ) *
    #       c(par_vec[1:2], rep(1, 4))
    #   }
    #
    #   list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    # }
    #
    #
    # lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group,
    #                                   n.clus) {
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   if (n.clus > 0) {
    #     # llik + prior
    #     res <-
    #       llik_vmsin_one_comp(data[obs_group, , drop=FALSE], par_vec,
    #                           log(const_vmsin(par_vec[1],
    #                                           par_vec[2], par_vec[3]))) +
    #       sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   } else{
    #     # only prior
    #     res <-
    #       sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   }
    #   res
    # }



    # in log scale
    lpd_grad_model_indep_1comp <- function(data, par_vec_lscale,
                                           obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      lpd_grad <- matrix(NA, 6, 1)
      if (n.clus > 0) {
        lpd_grad <- signif_(suppressWarnings(
          grad_llik_vmsin_C(
            data[obs_group, , drop = FALSE],
            par_vec
          )
        ) *
          c(par_vec[1:2], rep(1, 4)) +
          c( # grad for lprior
            -par_vec_lscale[1:3] / norm.var, 0, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
          ))
      } else {
        lpd_grad[] <-
          signif_(c( # grad for lprior
            -par_vec_lscale[1:3] / norm.var, 0, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
          ))
      }
      list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    }


    lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      if (n.clus > 0) {
        # llik + prior
        res <-
          signif_(llik_vmsin_one_comp(
            data[obs_group, , drop = FALSE], par_vec,
            log(const_vmsin(
              par_vec[1],
              par_vec[2], par_vec[3]
            ))
          ) +
            0.5 * sum(-par_vec_lscale[1:3]^2 / norm.var))
      } else {
        # only prior
        res <-
          signif_(0.5 * sum(-par_vec_lscale[1:3]^2 / norm.var))
      }
      res
    }




    llik_model_contri <- function(data, par_mat, pi_mix) {
      signif_(
        llik_vmsin_contri_C(
          data, par_mat, pi_mix,
          signif_(log_const_vmsin_all(par_mat))
        )
      )
    }

    mem_p_model <- function(data, par_mat, pi_mix) {
      signif_(
        mem_p_sin(
          data, par_mat, pi_mix,
          signif_(log_const_vmsin_all(par_mat)), 1
        )
      )
    }
  } else if (model == "vmcos") {
    ell <- list(qrnd_grid = qrnd, n_qrnd = n_qrnd)

    if (!is.null(ell$qrnd)) {
      qrnd_grid <- ell$qrnd
      dim_qrnd <- dim(qrnd_grid)
      if (!is.matrix(qrnd_grid) | is.null(dim_qrnd) |
          dim_qrnd[2] != 2) {
        stop("qrnd_grid must be a two column matrix")
      }
      n_qrnd <- dim_qrnd[1]
    } else if (!is.null(ell$n_qrnd)) {
      n_qrnd <- round(ell$n_qrnd)
      if (n_qrnd < 1) {
        stop("n_qrnd must be a positive integer")
      }
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    } else {
      n_qrnd <- 1e4
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    }

    if (unimodal.component) {
      dep_cov <- TRUE
      dep_cov_type <- "vmcos_unimodal"
    } else {
      dep_cov <- FALSE
      dep_cov_type <- NULL
    }



    # # in log scale
    # lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   lpd_grad <- matrix(NA, 6, 1)
    #   if (n.clus > 0) {
    #     lpd_grad[] <- (grad_llik_vmcos_C(data[obs_group, , drop=FALSE],
    #                                      par_vec[], qrnd_grid) +
    #                      c( # grad for lprior
    #                        (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #                        # lprior
    #                        sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #                              gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var)
    #     ) *
    #       c(par_vec[1:2], rep(1, 4))
    #   } else {
    #     lpd_grad[] <-
    #       c( # grad for lprior
    #         (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #         # lprior
    #         sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #               gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #       ) *
    #       c(par_vec[1:2], rep(1, 4))
    #   }
    #   list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    # }
    #
    # lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
    #
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   if (n.clus > 0) {
    #     # llik + prior
    #     res <-
    #       llik_vmcos_one_comp(data[obs_group, , drop=FALSE], par_vec[],
    #                           log(const_vmcos(par_vec[1],
    #                                           par_vec[2],
    #                                           par_vec[3],
    #                                           qrnd_grid))) +
    #       sum((gam.loc - 1)*log(par_vec[1:2])-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   } else{
    #     # only prior
    #     res <-
    #       sum((gam.loc - 1)*log(par_vec[1:2])-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   }
    #   res
    # }


    # in log scale
    lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      lpd_grad <- matrix(NA, 6, 1)
      if (n.clus > 0) {
        lpd_grad[] <- signif_(suppressWarnings(
          grad_llik_vmcos_C(
            data[obs_group, , drop = FALSE],
            par_vec[], qrnd_grid
          )
        ) *
          c(par_vec[1:2], rep(1, 4)) +
          c( # grad for lprior
            -par_vec_lscale[1:3] / norm.var, 0, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
          ))
      } else {
        lpd_grad[] <-
          signif_(c( # grad for lprior
            -par_vec_lscale[1:3] / norm.var, 0, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
          ))
      }
      list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    }

    lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
      if (n.clus > 0) {
        # llik + prior
        res <-
          signif_(llik_vmcos_one_comp(
            data[obs_group, , drop = FALSE], par_vec[],
            log(const_vmcos(
              par_vec[1],
              par_vec[2],
              par_vec[3],
              qrnd_grid
            ))
          ) -
            0.5 * sum(par_vec_lscale[1:3]^2 / norm.var))
      } else {
        # only prior
        res <-
          signif_(-0.5 * sum(par_vec_lscale[1:3]^2 / norm.var))
      }
      res
    }



    llik_model_contri <- function(data, par_mat, pi_mix) {
      signif_(
        llik_vmcos_contri_C(
          data, par_mat, pi_mix,
          signif_(log_const_vmcos_all(par_mat, qrnd_grid))
        )
      )
    }

    mem_p_model <- function(data, par_mat, pi_mix) {
      signif_(
        mem_p_cos(
          data, par_mat, pi_mix,
          signif_(log_const_vmcos_all(par_mat, qrnd_grid))
        )
      )
    }
  } else if (model == "wnorm2") {
    dep_cov <- TRUE
    dep_cov_type <- "wnorm2_bound"


    if (int.displ >= 5) {
      int.displ <- 5
    } else if (int.displ <= 1) int.displ <- 1

    int.displ <- floor(int.displ)
    omega.2pi.all <- expand.grid(-int.displ:int.displ, -int.displ:int.displ) * (2 * pi) # 2pi * integer displacements
    omega.2pi <- as.matrix(omega.2pi.all)

    # lpd_grad_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #   lpd_grad <- matrix(NA, 6, ncomp)
    #   for(j in 1:ncomp) {
    #     if (n.clus[j] > 0) {
    #       lpd_grad[, j] <- grad_llik_wnorm2_C(data[obs_group[[j]], , drop=FALSE],
    #                                           par_mat[, j], omega.2pi) +
    #         c( # grad for lprior
    #           (gam.loc - 1)/par_mat[1:2, j] - gam.rate,  -par_mat[3, j]/norm.var, 0, 0,
    #           # lprior
    #           sum((gam.loc - 1)*log(par_mat[1:2, j])-
    #                 gam.rate*par_mat[1:2, j]) - 0.5*par_mat[3, j]^2/norm.var
    #         )
    #     } else {
    #       lpd_grad[, j] <-
    #         c( # grad for lprior
    #           (gam.loc - 1)/par_mat[1:2, j] - gam.rate,  -par_mat[3, j]/norm.var, 0, 0,
    #           # lprior
    #           sum((gam.loc - 1)*log(par_mat[1:2, j])-
    #                 gam.rate*par_mat[1:2, j]) - 0.5*par_mat[3, j]^2/norm.var
    #         )
    #     }
    #   }
    #   list(lpr = sum(lpd_grad[6, ]), grad = lpd_grad[1:5, ])
    # }

    # lpd_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #   res <- 0
    #   for(j in 1:ncomp) {
    #     if (n.clus[j] > 0) {
    #       # llik + prior
    #       res <- res +
    #         llik_wnorm2_one_comp(data[obs_group[[j]], , drop=FALSE], par_mat[, j],
    #                              l_const_wnorm2(par_mat[, j]),
    #                              omega.2pi) +
    #         sum((gam.loc - 1)*log(par_mat[1:2, j])-
    #               gam.rate*par_mat[1:2, j]) - 0.5*par_mat[3, j]^2/norm.var
    #     } else{
    #       # only prior
    #       res <- res +
    #         sum((gam.loc - 1)*log(par_mat[1:2, j])-
    #               gam.rate*par_mat[1:2, j]) - 0.5*par_mat[3, j]^2/norm.var
    #     }
    #   }
    #   unname(res)
    # }


    # # in log scale
    # lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   lpd_grad <- matrix(NA, 6, 1)
    #   if (n.clus > 0) {
    #     lpd_grad[] <- (grad_llik_wnorm2_C(data[obs_group, , drop=FALSE],
    #                                       par_vec[], omega.2pi) +
    #                      c( # grad for lprior
    #                        (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #                        # lprior
    #                        sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #                              gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #                      )) *
    #       c(par_vec[1:2], rep(1, 4))
    #   } else {
    #     lpd_grad[] <-
    #       c( # grad for lprior
    #         (gam.loc - 1)/par_vec[1:2] - gam.rate,  -par_vec[3]/norm.var, 0, 0,
    #         # lprior
    #         sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #               gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #       ) *
    #       c(par_vec[1:2], rep(1, 4))
    #   }
    #   list(lpr = sum(lpd_grad[6]), grad = lpd_grad[1:5])
    # }
    #
    # lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
    #   par_vec <- c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5])
    #   if (n.clus > 0) {
    #     # llik + prior
    #     res <-
    #       llik_wnorm2_one_comp(data[obs_group, , drop=FALSE], par_vec[],
    #                            l_const_wnorm2(par_vec[]),
    #                            omega.2pi) +
    #       sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   } else{
    #     # only prior
    #     res <-
    #       sum((gam.loc - 1)*par_vec_lscale[1:2]-
    #             gam.rate*par_vec[1:2]) - 0.5*par_vec[3]^2/norm.var
    #   }
    #   unname(res)
    # }


    # in log scale
    lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- signif_(c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5]))
      lpd_grad <- matrix(NA, 6, 1)
      if (n.clus > 0) {
        lpd_grad[] <- signif_(
          grad_llik_wnorm2_C(
            data[obs_group, , drop = FALSE],
            par_vec[], omega.2pi
          ) *
            c(par_vec[1:2], rep(1, 4)) +
            c( # grad for lprior
              -par_vec_lscale[1:3] / norm.var, 0, 0,
              # lprior
              -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
            )
        )
      } else {
        lpd_grad[] <-
          signif_(c( # grad for lprior
            -par_vec_lscale[1:3] / norm.var, 0, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1:3]^2 / norm.var)
          ))
      }
      list(lpr = (lpd_grad[6]), grad = lpd_grad[1:5])
    }

    lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- signif_(c(exp(par_vec_lscale[1:2]), par_vec_lscale[3:5]))
      if (n.clus > 0) {
        # llik + prior
        res <-
          signif_(llik_wnorm2_one_comp(
            data[obs_group, , drop = FALSE], par_vec[],
            l_const_wnorm2(par_vec[]),
            omega.2pi
          ) -
            0.5 * sum(par_vec_lscale[1:3]^2 / norm.var))
      } else {
        # only prior
        res <-
          signif_(-0.5 * sum(par_vec_lscale[1:3]^2 / norm.var))
      }
      res
    }


    llik_model_contri <- function(data, par_mat, pi_mix) {
      signif_(
        llik_wnorm2_contri_C(
          data, par_mat, pi_mix,
          signif_(log_const_wnorm2_all(par_mat)),
          omega.2pi
        )
      )
    }

    mem_p_model <- function(data, par_mat, pi_mix) {
      signif_(
        mem_p_wnorm2(
          data, par_mat, pi_mix,
          signif_(log_const_wnorm2_all(par_mat)),
          omega.2pi
        )
      )
    }
  } else if (model == "vm") {
    dep_cov <- FALSE

    # lpd_grad_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #   lpd_grad <- matrix(NA, 3, ncomp)
    #   for(j in 1:ncomp) {
    #     if (n.clus[j] > 0) {
    #       lpd_grad[, j] <- grad_llik_univm_C(data[obs_group[[j]]],
    #                                          par_mat[, j]) +
    #         c( # grad for lprior
    #           (gam.loc - 1)/par_mat[1, j] - gam.rate,  0,
    #           # lprior
    #           (gam.loc - 1)*log(par_mat[1, j])-
    #             gam.rate*par_mat[1, j]
    #         )
    #
    #     } else {
    #       lpd_grad[, j] <-
    #         c( # grad for lprior
    #           (gam.loc - 1)/par_mat[1, j] - gam.rate,  0,
    #           # lprior
    #           (gam.loc - 1)*log(par_mat[1, j])-
    #             gam.rate*par_mat[1, j]
    #         )
    #     }
    #   }
    #   list(lpr = sum(lpd_grad[3, ]), grad = lpd_grad[1:2, ])
    # }
    #
    # lpd_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #   res <- 0
    #   for(j in 1:ncomp) {
    #     if (n.clus[j] > 0) {
    #       # llik + prior
    #       res <- res +
    #         llik_univm_one_comp(data[obs_group[[j]]], par_mat[, j],
    #                             log(const_univm(par_mat[1, j]))) +
    #         (gam.loc - 1)*log(par_mat[1, j]) - gam.rate*par_mat[1, j]
    #
    #     } else{
    #       # only prior
    #       res <- res +
    #         (gam.loc - 1)*log(par_mat[1, j]) - gam.rate*par_mat[1, j]
    #     }
    #   }
    #   unname(res)
    # }

    # lpd_grad_model_indep_1comp <- function(data, par_vec, obs_group, n.clus) {
    #   lpd_grad <- matrix(NA, 3, 1)
    #   if (n.clus > 0) {
    #     lpd_grad[] <- grad_llik_univm_C(data[obs_group],
    #                                     par_vec[]) +
    #       c( # grad for lprior
    #         (gam.loc - 1)/par_vec[1] - gam.rate,  0,
    #         # lprior
    #         (gam.loc - 1)*log(par_vec[1])-
    #           gam.rate*par_vec[1]
    #       )
    #
    #   } else {
    #     lpd_grad[] <-
    #       c( # grad for lprior
    #         (gam.loc - 1)/par_vec[1] - gam.rate,  0,
    #         # lprior
    #         (gam.loc - 1)*log(par_vec[1])-
    #           gam.rate*par_vec[1]
    #       )
    #   }
    #   list(lpr = (lpd_grad[3 ]), grad = lpd_grad[1:2 ])
    # }
    #

    # in log scale
    lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      lpd_grad <- matrix(NA, 3, 1)
      par_vec <- c(exp(par_vec_lscale[1]), par_vec_lscale[2])
      if (n.clus > 0) {
        lpd_grad[] <- signif_(suppressWarnings(
          grad_llik_univm_C(
            data[obs_group],
            par_vec[]
          )
        ) * c(par_vec[1], 1, 1) +

          c( # grad for lprior
            -par_vec_lscale[1] / norm.var, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1]^2 / norm.var)
          ))
      } else {
        lpd_grad[] <-
          signif_(c( # grad for lprior
            -par_vec_lscale[1] / norm.var, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1]^2 / norm.var)
          ))
      }
      list(lpr = (lpd_grad[3]), grad = lpd_grad[1:2])
    }

    lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1]), par_vec_lscale[2])
      if (n.clus > 0) {
        # llik + prior
        res <-
          signif_(
            llik_univm_one_comp(
              data[obs_group], par_vec[],
              log(const_univm(par_vec[1]))
            ) -
              0.5 * sum(par_vec_lscale[1]^2 / norm.var)
          )
      } else {
        # only prior
        res <-
          signif_(-0.5 * sum(par_vec_lscale[1]^2 / norm.var))
      }
      unname(res)
    }



    llik_model_contri <- function(data, par_mat, pi_mix) {
      signif_(
        llik_univm_contri_C(
          data, par_mat, pi_mix,
          signif_(log_const_univm_all(par_mat))
        )
      )
    }

    mem_p_model <- function(data, par_mat, pi_mix) {
      signif_(
        mem_p_univm(
          data, par_mat, pi_mix,
          signif_(log_const_univm_all(par_mat))
        )
      )
    }
  }

  # else if (model == "wnorm")
  else {
    dep_cov <- FALSE


    if (int.displ >= 5) {
      int.displ <- 5
    } else if (int.displ <= 1) int.displ <- 1

    int.displ <- floor(int.displ)
    omega.2pi <- (-int.displ):int.displ * (2 * pi) # 2pi * 1d integer displacements

    #
    #     lpd_grad_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #       lpd_grad <- matrix(NA, 3, ncomp)
    #       for(j in 1:ncomp) {
    #         if (n.clus[j] > 0) {
    #           lpd_grad[, j] <- grad_llik_uniwnorm_C(data[obs_group[[j]]],
    #                                                 par_mat[, j], omega.2pi) +
    #             c( # grad for lprior
    #               (gam.loc - 1)/par_mat[1, j] - gam.rate,  0,
    #               # lprior
    #               (gam.loc - 1)*log(par_mat[1, j])-
    #                 gam.rate*par_mat[1, j]
    #             )
    #
    #         } else {
    #           lpd_grad[, j] <-
    #             c( # grad for lprior
    #               (gam.loc - 1)/par_mat[1, j] - gam.rate,  0,
    #               # lprior
    #               (gam.loc - 1)*log(par_mat[1, j])-
    #                 gam.rate*par_mat[1, j]
    #             )
    #         }
    #       }
    #       list(lpr = sum(lpd_grad[3, ]), grad = lpd_grad[1:2, ])
    #     }
    #
    #     lpd_model_indep <- function(data, par_mat, obs_group, n.clus) {
    #       res <- 0
    #       for(j in 1:ncomp) {
    #         if (n.clus[j] > 0) {
    #           # llik + prior
    #           res <- res +
    #             llik_uniwnorm_one_comp(data[obs_group[[j]]], par_mat[, j],
    #                                    l_const_uniwnorm(par_mat[1, j]),
    #                                    omega.2pi) +
    #             (gam.loc - 1)*log(par_mat[1, j]) - gam.rate*par_mat[1, j]
    #
    #         } else{
    #           # only prior
    #           res <- res +
    #             (gam.loc - 1)*log(par_mat[1, j]) - gam.rate*par_mat[1, j]
    #         }
    #       }
    #       unname(res)
    #     }

    lpd_grad_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      lpd_grad <- matrix(NA, 3, 1)
      par_vec <- c(exp(par_vec_lscale[1]), par_vec_lscale[2])
      if (n.clus > 0) {
        lpd_grad[] <- signif_(grad_llik_uniwnorm_C(
          data[obs_group],
          par_vec[], omega.2pi
        ) * c(par_vec[1], 1, 1) +

          c( # grad for lprior
            -par_vec_lscale[1] / norm.var, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1]^2 / norm.var)
          ))
      } else {
        lpd_grad[] <-
          signif_(c( # grad for lprior
            -par_vec_lscale[1] / norm.var, 0,
            # lprior
            -0.5 * sum(par_vec_lscale[1]^2 / norm.var)
          ))
      }
      list(lpr = (lpd_grad[3]), grad = lpd_grad[1:2])
    }


    lpd_model_indep_1comp <- function(data, par_vec_lscale, obs_group, n.clus) {
      par_vec <- c(exp(par_vec_lscale[1]), par_vec_lscale[2])
      if (n.clus > 0) {
        # llik + prior
        res <-
          signif_(llik_uniwnorm_one_comp(
            data[obs_group], par_vec[],
            l_const_uniwnorm(par_vec[1]),
            omega.2pi
          ) -
            0.5 * sum(par_vec_lscale[1]^2 / norm.var))
      } else {
        # only prior
        res <-
          signif_(-0.5 * sum(par_vec_lscale[1]^2 / norm.var))
      }
      unname(res)
    }




    llik_model_contri <- function(data, par_mat, pi_mix) {
      signif_(
        llik_uniwnorm_contri_C(
          data, par_mat, pi_mix,
          signif_(log_const_uniwnorm_all(par_mat)),
          omega.2pi
        )
      )
    }

    mem_p_model <- function(data, par_mat, pi_mix) {
      signif_(
        mem_p_uniwnorm(
          data, par_mat, pi_mix,
          signif_(log_const_uniwnorm_all(par_mat)),
          omega.2pi
        )
      )
    }



    # grad_llik_uniwnorm_R <- function(data, par) {
    #   gr <- numDeriv::grad(function(par) llik_uniwnorm_one_comp(data, par,
    #                                                             l_const_uniwnorm(par[1]),
    #                                                             omega.2pi),
    #                        par)
    #   gr[3] <- llik_uniwnorm_one_comp(data, par,
    #                                   l_const_uniwnorm(par[1]), omega.2pi)
    #   gr
    # }

    # llik_uniwnorm_full(data.rad, par.mat, pi.mix, log_c = log_const_uniwnorm_all(par.mat), omega.2pi)
    #
    #
    # grad_llik_uniwnorm_C(data.rad, par.mat[, 1], omega.2pi)
    # grad_llik_uniwnorm_R(data.rad, par.mat[, 1])
  }




  if (!is.null(start_par)) {
    if (!is.list(start_par)) {
      stop("start_par must be a list")
    }
    if (!is.list(start_par[[1]])) {
      if (length(setdiff(c(par.names, "pmix"), names(start_par))) > 0) {
        stop("start_par does not have all parameters")
      }
      start_par <- lapply(1:n.chains, function(ii) start_par)
    } else {
      if (length(start_par) != n.chains) {
        stop("length(start_par) must be either 1 or equal to n.chains")
      }
      for (jj in 1:n.chains) {
        if (length(setdiff(c(par.names, "pmix"), names(start_par[[jj]]))) > 0) {
          stop(paste0("start_par[[", jj, "]] does not have all parameters"))
        }
      }
    }
  }

  if (any(is.null(start_par), !is.list(start_par[[1]])) &
      all(!rand_start)) {
    starting_1 <- process_startpar(
      start_par,
      data.rad,
      ncomp,
      model,
      FALSE
    )
    starting <- lapply(1:n.chains, function(j) starting_1)
  } else {
    starting <- lapply(
      1:n.chains,
      function(j) {
        process_startpar(
          start_par[[j]],
          data.rad,
          ncomp,
          model,
          rand_start[j]
        )
      }
    )
  }

  run_MC <- function(starting, L, chain_no) {

    # just to change the RNG state across chains, so that
    # no two c chains turn out to be identical
    change_rng_state <- runif(chain_no)

    if (method == "hmc") # using hmc
    {
      L_vec <- rep(L, n.iter)
      if (L.random) {
        L.orig <- L
      }
    }


    starting$par.mat[abs(starting$par.mat) >= kappa_upper / 2] <- kappa_upper / 2
    starting$par.mat[abs(starting$par.mat) <= 2 * kappa_lower] <- 2 * kappa_lower

    par.mat.all <- array(0, dim = c(npar_1_comp, ncomp, n.iter))
    par.mat_lscale.all <- par.mat.all
    pi.mix.all <- matrix(1, nrow = ncomp, ncol = n.iter)
    llik.all <- lprior.all <- lpd.all <- rep(-Inf, (n.iter))
    accpt.par.mat.all <- matrix(NA, ncomp, n.iter)
    modelpar.names <- par.names
    clus_ind.all <- matrix(1, nrow = n.data, ncol = n.iter)
    if (return_llik_contri) {
      llik_contri_all <- matrix(1, nrow = n.data, ncol = n.iter)
    } else {
      llik_contri_all <- NULL
    }
    mem_prob_all <- array(1, dim = c(n.data, ncomp, n.iter))



    epsilon_ave <- NULL
    L_ave <- NULL
    propscale_final <- NULL



    pi.mix <- starting$pi.mix
    par.mat <- as.matrix(starting$par.mat)

    if (zero_cov) {
      par.mat[3, ] <- 0
    } else if (cov.restrict == "POSITIVE") {
      par.mat[3, ] <- pmax(par.mat[3, ], 0)
    } else if (cov.restrict == "NEGATIVE") {
      par.mat[3, ] <- pmin(par.mat[3, ], 0)
    }


    if (type == "bi") {
      par.mat_lscale <- find_lscale_mat(par.mat)
      par_lower_lscale <- find_lscale_mat(par_lower)
      par_upper_lscale <- find_lscale_mat(par_upper)
    } else {
      par.mat_lscale <- find_lscale_mat_uni(par.mat)
      par_lower_lscale <- find_lscale_mat_uni(par_lower)
      par_upper_lscale <- find_lscale_mat_uni(par_upper)
    }

    # browser()

    if (ncomp == 1) {
      perm_sampling <- FALSE
      clus.ind <- clus.ind_curr <- rep(1, n.data)
      obs_group <- list(1:n.data)
      n.clus <- n.data
      post.wt <- matrix(1, n.data, 1)
      pi.mix <- 1
    }


    tcltk_fail <- FALSE

    if (show.progress & !exists("pb") & progress.backend == "tcltk") {
      pb <- tryCatch(tcltk::tkProgressBar(
        title = paste("Chain", chain_no),
        label = "Initializing...",
        min = 1, max = n.iter
      ),
      error = function(e) e
      )
      if (is(pb, "error") | is(pb, "warning")) {
        show.progress <- FALSE
        tcltk_fail <- TRUE
      }
    } else if (show.progress & !exists("pb") & progress.backend == "txt") {
      pb <- utils::txtProgressBar(title = paste("Chain", chain_no),
                                  label = "Initializing...",
                                  min = 1, max = n.iter, style = 3)
      tcltk_fail <- FALSE
    }

    if (tcltk_fail) {
      msg <- paste(
        "{tcltk} could not be loaded; 'show.progress' was set to FALSE."#,
        # "Consider setting progress.backend = \"txt\" for a text-based",
        # "progress bar. See ?fit_angmix for more details."
      )
      message(msg)
      show.progress <- FALSE
    }



    par.mat.all.order <- par.mat.all

    ntunes_up <- ntunes_down <- rep(0, ncomp)
    # tune_iter_no <- c()
    # ave_accpt_all <- c()

    tune_param_all <- matrix(NA, length(tune_param), n.iter)

    tune_status <- matrix(0, ncomp, n.iter)

    #  Run the Markov chain
    for (iter in seq_len(n.iter)) {
      # browser()

      if (method == "hmc") {
        if (epsilon.random) {
          tune_param_final <- tune_param * runif(1, 1 - epsilon.incr, 1 + epsilon.incr)
        } else {
          tune_param_final <- tune_param
        }
        if (L.random) {
          L_vec[iter] <- L <- sample(ceiling(L.orig / exp(L.incr)):ceiling(L.orig * exp(L.incr)), 1)
        }
      }

      #----------------------------------------------------------------------------------
      # generating mixture proportions if ncomp > 1
      #----------------------------------------------------------------------------------


      if (ncomp > 1) {
        post.wt <- mem_p_model(data.rad, par.mat, pi.mix)
        clus.ind <- cID(post.wt, ncomp, runif(n.data))
        # clus.ind <- apply(post.wt, 1, function(x) which.max(rmultinom(n = 1, size = 1, prob = x)))
        # n.clus <- tabulate(clus.ind_curr, nbins = ncomp)
        obs_group <- lapply(1:ncomp, function(j) which(clus.ind == j))
        n.clus <- listLen(obs_group)
        pi.mix <-
          as.numeric(rdirichlet(1, (pmix.alpha + n.clus))) # new mixture proportions
      }




      #----------------------------------------------------------------------------------
      # generating par.mat
      #----------------------------------------------------------------------------------

      if (type == "bi") {
        if (method == "hmc") {

          # hmc_curr <-
          #   bounded_hmc_biv(lpr_grad =
          #                     function(par_mat)
          #                       lpd_grad_model_indep(data.rad, par_mat,
          #                                            obs_group, n.clus),
          #                   init =  par.mat,
          #                   lower = par_lower,
          #                   upper = par_upper,
          #                   dep_cov = dep_cov,
          #                   dep_cov_type = dep_cov_type,
          #                   zero_cov = zero_cov,
          #                   nsteps = L,
          #                   step = tune_param_final)
          #
          # par.mat <- hmc_curr$final
          # accpt.par.mat.all[iter] <- hmc_curr$acc

          # par_mat = par.mat
          # lpd_grad_model_indep(data.rad, par_mat,
          #                      obs_group, n.clus)
          #
          #
          # lpd_grad_model_indep_1comp(data.rad, par_mat[, 1],
          #                      obs_group[[1]], n.clus[1])
          # lpd_grad_model_indep_1comp(data.rad, par_mat[, 2],
          #                            obs_group[[2]], n.clus[2])
          #
          #
          # lpd_model_indep(data.rad, par_mat,
          #                 obs_group, n.clus)
          #
          # lpd_model_indep_1comp(data.rad, par_mat[, 1],
          #                 obs_group[[1]], n.clus[1])
          #

          # browser()

          for (j in 1:ncomp) {
            # browser()
            hmc_curr <-
              bounded_hmc_biv(
                lpr_grad =
                  function(par_vec_lscale) {
                    lpd_grad_model_indep_1comp(
                      data.rad, par_vec_lscale,
                      obs_group[[j]], n.clus[j]
                    )
                  },
                init = par.mat_lscale[, j, drop = FALSE],
                lower = par_lower_lscale[, j, drop = FALSE],
                upper = par_upper_lscale[, j, drop = FALSE],
                dep_cov = dep_cov,
                dep_cov_type = dep_cov_type,
                zero_cov = zero_cov,
                nsteps = L,
                step = tune_param_final[, j, drop = FALSE]
              )

            par.mat[, j] <- signif_(find_expscale(hmc_curr$final))
            par.mat_lscale[, j] <- signif_(hmc_curr$final)
            accpt.par.mat.all[j, iter] <- hmc_curr$acc
          }
        } else {

          # rwmh_curr <- bounded_rwmh_biv(lpr =
          #                                 function(par_mat)
          #                                   lpd_model_indep(data.rad, par_mat,
          #                                                   obs_group, n.clus),
          #                               init =  par.mat,
          #                               lower = par_lower,
          #                               upper = par_upper,
          #                               dep_cov = dep_cov,
          #                               dep_cov_type = dep_cov_type,
          #                               zero_cov = zero_cov,
          #                               step = tune_param)
          #
          # par.mat <- rwmh_curr$final
          # accpt.par.mat.all[iter] <- rwmh_curr$accpt

          for (j in 1:ncomp) {
            rwmh_curr <- bounded_rwmh_biv(
              lpr =
                function(par_vec_lscale) {
                  lpd_model_indep_1comp(
                    data.rad, par_vec_lscale,
                    obs_group[[j]], n.clus[j]
                  )
                },
              init = par.mat_lscale[, j, drop = FALSE],
              lower = par_lower_lscale[, j, drop = FALSE],
              upper = par_upper_lscale[, j, drop = FALSE],
              dep_cov = dep_cov,
              dep_cov_type = dep_cov_type,
              zero_cov = zero_cov,
              step = tune_param[, j]
            )

            par.mat[, j] <- signif_(find_expscale(rwmh_curr$final))
            par.mat_lscale[, j] <- signif_(rwmh_curr$final)
            accpt.par.mat.all[j, iter] <- rwmh_curr$accpt
          }
        }


        lprior.all[iter] <-
          -0.5 * sum(par.mat_lscale[1:3, ]^2 / norm.var) +
          sum(pmix.alpha * log(pi.mix))
      }

      # if type == "uni"
      else {
        if (method == "hmc") {
          # hmc_curr <-
          #   bounded_hmc_uni(lpr_grad =
          #                     function(par_mat)
          #                       lpd_grad_model_indep(data.rad, par_mat,
          #                                            obs_group, n.clus),
          #                   init =  par.mat,
          #                   lower = par_lower,
          #                   upper = par_upper,
          #                   nsteps = L,
          #                   step = tune_param_final)
          #
          # par.mat <- hmc_curr$final
          # accpt.par.mat.all[iter] <- hmc_curr$acc

          for (j in 1:ncomp) {
            hmc_curr <-
              bounded_hmc_uni(
                lpr_grad =
                  function(par_vec_lscale) {
                    lpd_grad_model_indep_1comp(
                      data.rad, par_vec_lscale,
                      obs_group[[j]], n.clus[j]
                    )
                  },
                init = par.mat_lscale[, j, drop = FALSE],
                lower = par_lower_lscale[, j, drop = FALSE],
                upper = par_upper_lscale[, j, drop = FALSE],
                nsteps = L,
                step = tune_param_final[, j, drop = FALSE]
              )

            par.mat[, j] <- signif_(find_expscale_uni(hmc_curr$final))
            par.mat_lscale[, j] <- signif_(hmc_curr$final)
            accpt.par.mat.all[j, iter] <- hmc_curr$acc
          }
        } else {

          # rwmh_curr <- bounded_rwmh_uni(lpr =
          #                                 function(par_mat)
          #                                   lpd_model_indep(data.rad, par_mat,
          #                                                   obs_group, n.clus),
          #                               init =  par.mat,
          #                               lower = par_lower,
          #                               upper = par_upper,
          #                               step = tune_param)
          #
          # par.mat <- rwmh_curr$final
          # accpt.par.mat.all[iter] <- rwmh_curr$accpt

          for (j in 1:ncomp) {
            rwmh_curr <- bounded_rwmh_uni(
              lpr =
                function(par_vec_lscale) {
                  lpd_model_indep_1comp(
                    data.rad, par_vec_lscale,
                    obs_group[[j]], n.clus[j]
                  )
                },
              init = par.mat_lscale[, j, drop = FALSE],
              lower = par_lower_lscale[, j, drop = FALSE],
              upper = par_upper_lscale[, j, drop = FALSE],
              step = tune_param[, j]
            )

            par.mat[, j] <- signif_(find_expscale_uni(rwmh_curr$final))
            par.mat_lscale[, j] <- signif_(rwmh_curr$final)
            accpt.par.mat.all[j, iter] <- rwmh_curr$accpt
          }
        }

        lprior.all[iter] <-
          -0.5 * sum(par.mat_lscale[1, ]^2 / norm.var) +
          sum(pmix.alpha * log(pi.mix))
      }




      # do permutation sampling only after tuning
      if (perm_sampling & iter > n.burnin) {
        rand_perm <- sample(1:ncomp)
        clus.ind <- rand_perm[clus.ind] # random label switch
        post.wt <- post.wt[, rand_perm, drop = FALSE]
        par.mat <- par.mat[, rand_perm, drop = FALSE]
        par.mat_lscale <- par.mat_lscale[, rand_perm, drop = FALSE]
        pi.mix <- pi.mix[rand_perm, drop = FALSE]

        par.mat.all.order[, , iter] <- par.mat[, order(rand_perm), drop = FALSE]
        # needed for tuning

        # if (method == "hmc")
        tune_param <- tune_param[, rand_perm, drop = FALSE]
      }

      # browser()

      llik_contri <- llik_model_contri(data.rad, par.mat, pi.mix)

      if (return_llik_contri) {
        llik_contri_all[, iter] <- llik_contri
      }

      llik.all[iter] <- sum(llik_contri)
      lpd.all[iter] <- llik.all[iter] + lprior.all[iter]

      par.mat.all[, , iter] <- par.mat
      par.mat_lscale.all[, , iter] <- par.mat_lscale
      pi.mix.all[, iter] <- pi.mix
      mem_prob_all[, , iter] <- post.wt
      clus_ind.all[, iter] <- clus.ind




      tune_param_all[, iter] <- c(tune_param)

      # tuning tune_param during burnin
      if (autotune & iter >= nterms & iter %% 10 == 0 &
          iter <= iter.tune) {
        ave_accpt_all <- rep(0, ncomp)
        for (j in 1:ncomp) {
          ave_accpt_all[j] <-
            ave_accpt <- sum(accpt.par.mat.all[j, (iter - nterms + 1):iter]) / nterms

          if (ave_accpt > accpt.prob.upper) {
            tune_param[, j] <- tune_param[, j] * (1 + tune.incr)

            ntunes_up[j] <- ntunes_up[j] + 1
            tune_status[j, iter] <- 1
          } else if (ave_accpt < accpt.prob.lower) {
            tune_param[, j] <- tune_param[, j] * (1 - tune.incr)

            ntunes_down[j] <- ntunes_down[j] + 1
            tune_status[j, iter] <- -1
          }

          if (iter %% nterms == 0) {
            par.sd <- apply(par.mat_lscale.all[, j, (iter - nterms + 1):iter], 1, sd)
            mean_par.sd <- sum(par.sd) / (npar_1_comp)
            mean_tune_param_j <- sum(tune_param[, j]) / (npar_1_comp)
            if (mean_par.sd > 0) {
              tune_param[, j] <- 0.5 * par.sd / mean_par.sd * mean_tune_param_j + 0.5 * tune_param[, j]
            }

            #   else {
            #   par.sd <- apply(par.mat.all.order[, , (iter-nterms+1):iter], c(1:2), sd)
            #   par.sd <- par.sd[, rand_perm]
            #   mean_par.sd <- sum(par.sd)/(npar_1_comp*ncomp)
            #   mean_tune_param <- sum(tune_param)/(npar_1_comp*ncomp)
            #   if(mean_par.sd > 0)
            #     tune_param <- 0.5*par.sd/mean_par.sd*mean_tune_param + 0.5*tune_param
            #
            #   cat(ave_accpt, tune_param[1,],lpd.all[iter], "\n")
            # }
          }
        }

        # if(iter %% nterms == 0)
        #   cat(paste("(", paste0(ave_accpt_all, collapse = ","), ")"),
        #       tune_param[1,], lpd.all[iter], "\n")
      }

      # if (method == "hmc" & autotune & iter >= nterms
      #     & iter %% 5 == 0 & iter <= n.burnin) {
      #   acr <- cor(lpd.all[(iter-nterms+2):iter],
      #              lpd.all[(iter-nterms+1):(iter-1)])
      #   if (acr < acr.lower) {
      #     L <- ceiling(L*(1 - tune.incr))
      #   } else if (acr > acr.upper) {
      #     L <- ceiling(L*(1 + tune.incr))
      #   }
      # }



      if (show.progress) {
        message <- paste0(
          "Progress: ", round(iter / n.iter * 100), "% ",
          ifelse(iter <= n.burnin,
                 "(Burn-in)",
                 "(Sampling)"
          )
        )
        if (progress.backend == "tcltk") {
          tcltk::setTkProgressBar(pb, iter, label = message)
        } else if (progress.backend == "txt") {
          utils::setTxtProgressBar(pb, value = iter, label = message)
        }
      }
    }


    if (grepl(method, "hmc")) {
      epsilon_ave <- epsilon <- mean(tune_param_all)

      if (L.random) {
        L_ave <- sum(L_vec) / n.iter
      } else {
        L_ave <- L
      }
    }

    if (grepl(method, "rwmh")) {
      propscale_final <- propscale <- rowSums(tune_param_all) / n.iter
    }

    allpar_val <- array(1, dim = c(npar_1_comp + 1, ncomp, n.iter))
    allpar_val[1, , ] <- pi.mix.all
    allpar_val[-1, , ] <- par.mat.all
    rm(pi.mix.all, par.mat.all)

    allpar_name <- c("pmix", modelpar.names)
    dimnames(allpar_val)[[1]] <- c("pmix", modelpar.names)

    if (!return_tune_param) {
      rm(tune_param_all)
      tune_param_all <- NULL
    }


    result <- list(
      "par.value" = allpar_val, # [, , final_iter_set],
      "par.name" = allpar_name,
      "llik.contri" = llik_contri_all, # [, final_iter_set],
      "llik" = llik.all, # [final_iter_set],
      "lpd" = lpd.all, # [final_iter_set],
      "lprior" = lprior.all, # [final_iter_set],
      "accpt.modelpar" = accpt.par.mat.all, # [final_iter_set],
      "clus.ind" = clus_ind.all, # [, final_iter_set],
      "mem.prob" = mem_prob_all, # [, , final_iter_set],
      "epsilon" = epsilon_ave,
      "L" = L_ave,
      "propscale" = propscale_final,
      "tune_param" = tune_param_all,
      "par.upper" = par_upper,
      "par.lower" = par_lower,
      "tcltk_fail" = tcltk_fail
    )

    # if (show.progress & progress.backend == "tcltk")
    if (show.progress) close(pb)

    result
  }


  # seed <- floor(runif(1, 1,  100))

  # browser()
  # generate three chains in parallel, if possible
  if (chains_parallel) {
    res_list <- future_lapply(1:n.chains,
                              function(ii) {
                                run_MC(
                                  starting[[ii]],
                                  L[ii], ii
                                )
                              },
                              future.seed = TRUE
    )
  } else {
    res_list <- lapply(
      1:n.chains,
      function(ii) run_MC(starting[[ii]], L[ii], ii)
    )
  }



  # if (res_list[[1]]$tcltk_fail) {
  #   msg <- paste(
  #     "tcltk could not be loaded; 'show.progress' was set to FALSE.",
  #     "Consider setting progress.backend = \"txt\" for a text-based",
  #     "progress bar. See ?fit_angmix for more details."
  #   )
  #   warning()
  # }

  # combine the results from the lists
  allpar_val <- array(0, dim = c(npar_1_comp + 1, ncomp, n.iter, n.chains))
  llik_all <- lprior_all <- lpd_all <- matrix(0, n.iter, n.chains)
  accpt.modelpar_all <- array(0, dim = c(ncomp, n.iter, n.chains))
  clus.ind_all <- array(0, dim = c(n.data, n.iter, n.chains))
  if (return_llik_contri) {
    llik_contri_all <- array(0, dim = c(n.data, n.iter, n.chains))
  } else {
    llik_contri_all <- NULL
  }

  mem_prob_all <- array(1, dim = c(n.data, ncomp, n.iter, n.chains))

  if (return_tune_param) {
    tune_param_all <- array(NA, dim = c(length(tune_param), n.iter, n.chains))
  } else {
    tune_param_all <- NULL
  }

  for (j in 1:n.chains) {
    allpar_val[, , , j] <- res_list[[j]]$par.value
    llik_all[, j] <- res_list[[j]]$llik
    lprior_all[, j] <- res_list[[j]]$lprior
    lpd_all[, j] <- res_list[[j]]$lpd
    accpt.modelpar_all[, , j] <- res_list[[j]]$accpt.modelpar
    clus.ind_all[, , j] <- res_list[[j]]$clus.ind
    if (return_llik_contri) {
      llik_contri_all[, , j] <- res_list[[j]]$llik.contri
    }
    mem_prob_all[, , , j] <- res_list[[j]]$mem.prob
    if (return_tune_param) {
      tune_param_all[, , j] <- res_list[[j]]$tune_param
    }
  }

  if (method == "hmc") {
    epsilon_final <- do.call(cbind, lapply(res_list, function(x) x$epsilon))
    L_final <- unlist(lapply(res_list, function(x) x$L))
    propscale_final <- NULL
  } else {
    propscale_final <- do.call(cbind, lapply(res_list, function(x) x$propscale))
    epsilon_final <- NULL
    L_final <- NULL
  }

  out <- list(
    "par.value" = allpar_val,
    "clus.ind" = clus.ind_all,
    "par.name" = res_list[[1]]$par.name,
    "modelpar.lower" = res_list[[1]]$par.lower,
    "modelpar.upper" = res_list[[1]]$par.upper,
    "return_llik_contri" = return_llik_contri,
    "llik.contri" = llik_contri_all,
    "mem.prob" = mem_prob_all,
    "llik" = llik_all,
    "lpd" = lpd_all,
    "lprior" = lprior_all,
    "accpt.modelpar" = accpt.modelpar_all,
    "model" = curr.model,
    "method" = method,
    "perm_sampling" = perm_sampling,
    "epsilon.random" = epsilon.random,
    "epsilon" = epsilon_final,
    "L.random" = L.random,
    "L" = L_final,
    "iter.tune" = iter.tune,
    "propscale" = propscale_final,
    "tune_param" = tune_param_all,
    "return_tune_param" = return_tune_param,
    "type" = type,
    "data" = data.rad,
    "cov.restrict" = cov.restrict,
    "pmix.alpha" = pmix.alpha,
    "norm.var" = norm.var,
    "n.data" = n.data,
    "ncomp" = ncomp,
    "n.chains" = n.chains,
    "n.iter" = n.iter,
    "n.burnin" = n.burnin,
    "thin" = thin,
    "n.iter.final" = n.iter.final,
    "final_iter" = final_iter_set,
    "int.displ" = int.displ,
    "qrnd_grid" = qrnd_grid,
    "omega.2pi" = omega.2pi
  )

  class(out) <- "angmcmc"

  out
}

#' Backbone Dihedral Angles of Triose Phosphate Isomerase (8TIM)
#'
#' A dataset consisting of 490 pairs of backbone dihedral angles (in radian scale \eqn{[0, 2\pi)} )
#' \eqn{(\phi, \psi)} for the protein Triose Phosphate Isomerase (8TIM). The angles were obtained first by using
#' the DSSP software on the PDB file for 8TIM to get the backbone angles (in degrees),
#' and then by converting all angles into radians. Due to the presence of different secondary structures
#' (helices, sheets and loops) in the protein, the angular data show considerable variability, and is multimodal
#' with noticeably distinct clusters.
#'
#'
#' @format A data frame with 490 rows and 2 variables (backbone dihedral angles) phi and psi.
#' @source 8TIM PDB file: \url{http://www.rcsb.org/pdb/explore.do?structureId=8tim}.
#' @source DSSP software: \url{https://swift.cmbi.umcn.nl/gv/dssp/}.
#'
#' @usage
#' data(tim8)

"tim8"



#' Saturna Island wind directions
#'
#' @description
#' A dataset consisting of 239 observations on wind direction in radians (original measurements were
#' in 10s of degrees), measured at Saturna Island, British Columbia,
#' Canada during October 1-10, 2016 (obtained from Environment Canada website). There was a severe storm
#' during October 4-7, which caused significant fluctuations among the wind directions. As a result the
#' angular data show a clear multimodality.
#'
#' @format A data frame with 239 rows and 2 columns; the column "angle" provides the angular direction (in radian)
#' and the column day provides the days on which the data points were collected (ranges between 1-10, corresponding to
#' October 1-10, 2016).
#' @source Environment Canada: \url{https://climate.weather.gc.ca/climate_data/data_quality_e.html}.
#' @source CBC news on the storm: \url{https://www.cbc.ca/news/canada/british-columbia/storm-bc-1.3795204}.
#'
#' @usage
#' data(wind)

"wind"

bounded_rwmh_biv <- function(lpr,
                             init,
                             lower,
                             upper,
                             dep_cov = FALSE,
                             dep_cov_type = NULL,
                             zero_cov = FALSE,
                             step)
{

  lpr.init <- lpr(init)
  ncomp <- ncol(init)
  prop <- init

  if (zero_cov) prop[3, ] <- 0

  prop[1:2, ] <-  init[1:2, ] + rnorm(2*ncomp, 0, step[1:2])

  prop[4:5, ] <- prncp_reg(init[4:5, ] + rnorm(2*ncomp, 0, step[4:5]))

  if (!zero_cov) {
    prop[3, ] <-  init[3, ] + rnorm(ncomp, 0, step[3])

    if (dep_cov) {
      if (dep_cov_type %in% c("wnorm2_bound", "vmsin_unimodal")) {
        bd_k1k2 <- sqrt(exp(prop[1, ])*exp(prop[2, ]))
        lower[3, ] <- pmax(-bd_k1k2, lower[3, ])
        upper[3, ] <- pmin(bd_k1k2, upper[3, ])
      } else {
        # dep_cov_type == "vmcos_unimodal"
        lower[3, ] <- pmax(-exp(prop[1, ])*exp(prop[2, ])/
                             (exp(prop[1, ])+exp(prop[2, ])),
                           lower[3, ])
      }
    }
  }

  bd_err <- any(c(prop-lower, upper-prop) < 0)

  if (bd_err) {
    lpr.prop <- -Inf
  } else {
    lpr.prop <- lpr(prop)
  }

  aprob <- min(1, exp(lpr.prop-lpr.init))

  accpt <- (runif(1) < aprob)

  # if (is.na(propcheck)) browser()

  if (accpt) {
    final <- prop
    lpr.final <- lpr.prop
  } else {
    final <- init
    lpr.final <- lpr.init
  }


  out <- list (final=final, lpr=lpr.final, step=step,
               aprob=aprob, accpt=accpt*1)

  out
}


bounded_rwmh_uni <- function(lpr,
                             init,
                             lower,
                             upper,
                             step)
{

  lpr.init <- lpr(init)
  ncomp <- ncol(init)
  prop <- init

  prop[1, ] <-  pmax(init[1, ] + rnorm(ncomp, 0, step[1]), 1e-6)
  prop[2, ] <- prncp_reg(init[2, ] + rnorm(ncomp, 0, step[2]))

  bd_err <- any(c(prop-lower, upper-prop) < 0)

  if (bd_err) {
    lpr.prop <- -Inf
  } else {
    lpr.prop <- lpr(prop)
  }

  aprob <- min(1, exp(lpr.prop-lpr.init))

  accpt <- (runif(1) < aprob)

  # if (is.na(propcheck)) browser()

  if (accpt) {
    final <- prop
    lpr.final <- lpr.prop
  } else {
    final <- init
    lpr.final <- lpr.init
  }


  out <- list (final=final, lpr=lpr.final, step=step,
               aprob=aprob, accpt=accpt*1)

  out
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

rowVars <- function(mat_in) {
    .Call(`_BAMBI_rowVars`, mat_in)
}

par_mat_permute <- function(par_mat, perm_lab) {
    .Call(`_BAMBI_par_mat_permute`, par_mat, perm_lab)
}

cID <- function(probs, ncomp, Uv) {
    .Call(`_BAMBI_cID`, probs, ncomp, Uv)
}

change_labs <- function(orig, rand_perm) {
    .Call(`_BAMBI_change_labs`, orig, rand_perm)
}

calc_corr_tau_2 <- function(samp_mat) {
    .Call(`_BAMBI_calc_corr_tau_2`, samp_mat)
}

calc_corr_tau_1 <- function(samp_mat) {
    .Call(`_BAMBI_calc_corr_tau_1`, samp_mat)
}

calc_corr_fl <- function(samp_mat) {
    .Call(`_BAMBI_calc_corr_fl`, samp_mat)
}

const_univm <- function(k) {
    .Call(`_BAMBI_const_univm`, k)
}

log_const_univm_all <- function(par_mat) {
    .Call(`_BAMBI_log_const_univm_all`, par_mat)
}

ldunivmnum <- function(x, par) {
    .Call(`_BAMBI_ldunivmnum`, x, par)
}

dunivm_manyx_onepar <- function(x, k, mu) {
    .Call(`_BAMBI_dunivm_manyx_onepar`, x, k, mu)
}

dunivm_manyx_manypar <- function(x, k, mu) {
    .Call(`_BAMBI_dunivm_manyx_manypar`, x, k, mu)
}

dunivm_onex_manypar <- function(x, k, mu) {
    .Call(`_BAMBI_dunivm_onex_manypar`, x, k, mu)
}

runivm_onepar <- function(n, k, mu) {
    .Call(`_BAMBI_runivm_onepar`, n, k, mu)
}

runivm_manypar <- function(k, mu) {
    .Call(`_BAMBI_runivm_manypar`, k, mu)
}

univmmix <- function(x, par, pi, log_c_von) {
    .Call(`_BAMBI_univmmix`, x, par, pi, log_c_von)
}

univmmix_manyx <- function(x, par, pi, log_c) {
    .Call(`_BAMBI_univmmix_manyx`, x, par, pi, log_c)
}

mem_p_univm <- function(data, par, pi, log_c_von) {
    .Call(`_BAMBI_mem_p_univm`, data, par, pi, log_c_von)
}

llik_univm_one_comp <- function(data, par_vec, log_c) {
    .Call(`_BAMBI_llik_univm_one_comp`, data, par_vec, log_c)
}

grad_llik_univm_C <- function(data, par) {
    .Call(`_BAMBI_grad_llik_univm_C`, data, par)
}

llik_univm_contri_C <- function(data, par, pi, log_c) {
    .Call(`_BAMBI_llik_univm_contri_C`, data, par, pi, log_c)
}

l_const_uniwnorm <- function(k) {
    .Call(`_BAMBI_l_const_uniwnorm`, k)
}

const_uniwnorm <- function(k) {
    .Call(`_BAMBI_const_uniwnorm`, k)
}

log_const_uniwnorm_all <- function(par_mat) {
    .Call(`_BAMBI_log_const_uniwnorm_all`, par_mat)
}

lduniwnormnum <- function(x, par, omega_2pi_1d) {
    .Call(`_BAMBI_lduniwnormnum`, x, par, omega_2pi_1d)
}

duniwnorm_manyx_onepar <- function(x, k, mu, omega_2pi_1d) {
    .Call(`_BAMBI_duniwnorm_manyx_onepar`, x, k, mu, omega_2pi_1d)
}

duniwnorm_manyx_manypar <- function(x, k, mu, omega_2pi_1d) {
    .Call(`_BAMBI_duniwnorm_manyx_manypar`, x, k, mu, omega_2pi_1d)
}

duniwnorm_onex_manypar <- function(x, k, mu, omega_2pi_1d) {
    .Call(`_BAMBI_duniwnorm_onex_manypar`, x, k, mu, omega_2pi_1d)
}

mem_p_uniwnorm <- function(data, par, pi, log_c_von, omega_2pi_1d) {
    .Call(`_BAMBI_mem_p_uniwnorm`, data, par, pi, log_c_von, omega_2pi_1d)
}

llik_uniwnorm_contri_C <- function(data, par, pi, log_c, omega_2pi_1d) {
    .Call(`_BAMBI_llik_uniwnorm_contri_C`, data, par, pi, log_c, omega_2pi_1d)
}

llik_uniwnorm_one_comp <- function(data, par_vec, log_c, omega_2pi_1d) {
    .Call(`_BAMBI_llik_uniwnorm_one_comp`, data, par_vec, log_c, omega_2pi_1d)
}

grad_llik_uniwnorm_C <- function(data, par, omega_2pi_1d) {
    .Call(`_BAMBI_grad_llik_uniwnorm_C`, data, par, omega_2pi_1d)
}

BESSI0_C <- function(x) {
    .Call(`_BAMBI_BESSI0_C`, x)
}

const_vmcos_anltc <- function(k1, k2, k3) {
    .Call(`_BAMBI_const_vmcos_anltc`, k1, k2, k3)
}

const_vmcos_mc <- function(k1, k2, k3, uni_rand, return_log = FALSE) {
    .Call(`_BAMBI_const_vmcos_mc`, k1, k2, k3, uni_rand, return_log)
}

const_vmcos <- function(k1, k2, k3, uni_rand, return_log = FALSE) {
    .Call(`_BAMBI_const_vmcos`, k1, k2, k3, uni_rand, return_log)
}

d_const_vmcos_anltc <- function(k1, k2, k3) {
    .Call(`_BAMBI_d_const_vmcos_anltc`, k1, k2, k3)
}

d_const_vmcos_mc <- function(k1, k2, k3, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_d_const_vmcos_mc`, k1, k2, k3, uni_rand, ncores)
}

d_const_vmcos <- function(par, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_d_const_vmcos`, par, uni_rand, ncores)
}

log_const_vmcos_all <- function(par_mat, uni_rand) {
    .Call(`_BAMBI_log_const_vmcos_all`, par_mat, uni_rand)
}

ldcosnum <- function(x, y, par) {
    .Call(`_BAMBI_ldcosnum`, x, y, par)
}

dcos_onex_manypar <- function(x, k1, k2, k3, mu1, mu2, l_const_all) {
    .Call(`_BAMBI_dcos_onex_manypar`, x, k1, k2, k3, mu1, mu2, l_const_all)
}

dcos_manyx_onepar <- function(x, k1, k2, k3, mu1, mu2, l_const) {
    .Call(`_BAMBI_dcos_manyx_onepar`, x, k1, k2, k3, mu1, mu2, l_const)
}

dcos_manyx_manypar <- function(x, k1, k2, k3, mu1, mu2, l_const_all) {
    .Call(`_BAMBI_dcos_manyx_manypar`, x, k1, k2, k3, mu1, mu2, l_const_all)
}

rcos_unimodal <- function(n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmcos) {
    .Call(`_BAMBI_rcos_unimodal`, n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmcos)
}

rcos_bimodal <- function(n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmcos, mode_1, mode_2, vmpropn, unifpropn) {
    .Call(`_BAMBI_rcos_bimodal`, n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmcos, mode_1, mode_2, vmpropn, unifpropn)
}

rcos_onepar <- function(n, k1, k2, k3, mu1, mu2, I_upper_bd) {
    .Call(`_BAMBI_rcos_onepar`, n, k1, k2, k3, mu1, mu2, I_upper_bd)
}

mem_p_cos <- function(data, par, pi, log_c_von) {
    .Call(`_BAMBI_mem_p_cos`, data, par, pi, log_c_von)
}

llik_vmcos_contri_C <- function(data, par, pi, log_c) {
    .Call(`_BAMBI_llik_vmcos_contri_C`, data, par, pi, log_c)
}

llik_vmcos_one_comp <- function(data, par_vec, log_c) {
    .Call(`_BAMBI_llik_vmcos_one_comp`, data, par_vec, log_c)
}

grad_llik_vmcos_C <- function(data, par, uni_rand) {
    .Call(`_BAMBI_grad_llik_vmcos_C`, data, par, uni_rand)
}

vmcos_var_corr_anltc <- function(k1, k2, k3) {
    .Call(`_BAMBI_vmcos_var_corr_anltc`, k1, k2, k3)
}

vmcos_var_corr_mc <- function(k1, k2, k3, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_vmcos_var_corr_mc`, k1, k2, k3, uni_rand, ncores)
}

vmcos_var_cor_singlepar_cpp <- function(k1, k2, k3, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_vmcos_var_cor_singlepar_cpp`, k1, k2, k3, uni_rand, ncores)
}

ldcos_onex_manypar <- function(x, k1, k2, k3, mu1, mu2, l_const_all) {
    .Call(`_BAMBI_ldcos_onex_manypar`, x, k1, k2, k3, mu1, mu2, l_const_all)
}

ldcos_manyx_onepar <- function(x, k1, k2, k3, mu1, mu2, l_const) {
    .Call(`_BAMBI_ldcos_manyx_onepar`, x, k1, k2, k3, mu1, mu2, l_const)
}

ldcos_manyx_manypar <- function(x, k1, k2, k3, mu1, mu2, l_const_all) {
    .Call(`_BAMBI_ldcos_manyx_manypar`, x, k1, k2, k3, mu1, mu2, l_const_all)
}

const_vmsin <- function(k1, k2, lambda) {
    .Call(`_BAMBI_const_vmsin`, k1, k2, lambda)
}

d_const_vmsin <- function(par) {
    .Call(`_BAMBI_d_const_vmsin`, par)
}

log_const_vmsin_all <- function(par_mat) {
    .Call(`_BAMBI_log_const_vmsin_all`, par_mat)
}

ldsinnum <- function(x, y, par) {
    .Call(`_BAMBI_ldsinnum`, x, y, par)
}

dsin_onex_manypar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_dsin_onex_manypar`, x, k1, k2, k3, mu1, mu2)
}

dsin_manyx_onepar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_dsin_manyx_onepar`, x, k1, k2, k3, mu1, mu2)
}

dsin_manyx_manypar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_dsin_manyx_manypar`, x, k1, k2, k3, mu1, mu2)
}

rsin_unimodal <- function(n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmsin) {
    .Call(`_BAMBI_rsin_unimodal`, n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmsin)
}

rsin_bimodal <- function(n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmsin, mode_1, mode_2, vmpropn, unifpropn) {
    .Call(`_BAMBI_rsin_bimodal`, n, k1, k2, k3, mu1, mu2, kappa_opt, log_I0_kappa_opt, logK, log_const_vmsin, mode_1, mode_2, vmpropn, unifpropn)
}

rsin_onepar <- function(n, k1, k2, k3, mu1, mu2, I_upper_bd) {
    .Call(`_BAMBI_rsin_onepar`, n, k1, k2, k3, mu1, mu2, I_upper_bd)
}

mem_p_sin <- function(data, par, pi, log_c_von, ncores = 1L) {
    .Call(`_BAMBI_mem_p_sin`, data, par, pi, log_c_von, ncores)
}

llik_vmsin_contri_C <- function(data, par, pi, log_c) {
    .Call(`_BAMBI_llik_vmsin_contri_C`, data, par, pi, log_c)
}

llik_vmsin_one_comp <- function(data, par_vec, log_c) {
    .Call(`_BAMBI_llik_vmsin_one_comp`, data, par_vec, log_c)
}

grad_llik_vmsin_C <- function(data, par) {
    .Call(`_BAMBI_grad_llik_vmsin_C`, data, par)
}

vmsin_var_corr_anltc <- function(k1, k2, lambda) {
    .Call(`_BAMBI_vmsin_var_corr_anltc`, k1, k2, lambda)
}

vmsin_var_corr_mc <- function(k1, k2, k3, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_vmsin_var_corr_mc`, k1, k2, k3, uni_rand, ncores)
}

vmsin_var_cor_singlepar_cpp <- function(k1, k2, k3, uni_rand, ncores = 1L) {
    .Call(`_BAMBI_vmsin_var_cor_singlepar_cpp`, k1, k2, k3, uni_rand, ncores)
}

ldsin_onex_manypar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_ldsin_onex_manypar`, x, k1, k2, k3, mu1, mu2)
}

ldsin_manyx_onepar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_ldsin_manyx_onepar`, x, k1, k2, k3, mu1, mu2)
}

ldsin_manyx_manypar <- function(x, k1, k2, k3, mu1, mu2) {
    .Call(`_BAMBI_ldsin_manyx_manypar`, x, k1, k2, k3, mu1, mu2)
}

ldwnorm2_num <- function(x, par, omega_2pi) {
    .Call(`_BAMBI_ldwnorm2_num`, x, par, omega_2pi)
}

l_const_wnorm2 <- function(par) {
    .Call(`_BAMBI_l_const_wnorm2`, par)
}

const_wnorm2 <- function(par) {
    .Call(`_BAMBI_const_wnorm2`, par)
}

log_const_wnorm2_all <- function(par_mat) {
    .Call(`_BAMBI_log_const_wnorm2_all`, par_mat)
}

mem_p_wnorm2 <- function(data, par_mat, pi, log_c_wnorm, omega_2pi, ncores = 1L) {
    .Call(`_BAMBI_mem_p_wnorm2`, data, par_mat, pi, log_c_wnorm, omega_2pi, ncores)
}

llik_wnorm2_contri_C <- function(data, par, pi, log_c, omega_2pi) {
    .Call(`_BAMBI_llik_wnorm2_contri_C`, data, par, pi, log_c, omega_2pi)
}

llik_wnorm2_one_comp <- function(data, par_vec, log_c, omega_2pi) {
    .Call(`_BAMBI_llik_wnorm2_one_comp`, data, par_vec, log_c, omega_2pi)
}

grad_llik_wnorm2_C <- function(data, par, omega_2pi) {
    .Call(`_BAMBI_grad_llik_wnorm2_C`, data, par, omega_2pi)
}

grad_den_wnorm2_one_comp_i_unadj <- function(x, y, par, det_sig_inv, det_sig_inv_sqrt, omega_2pi) {
    .Call(`_BAMBI_grad_den_wnorm2_one_comp_i_unadj`, x, y, par, det_sig_inv, det_sig_inv_sqrt, omega_2pi)
}

dwnorm2_onex_manypar <- function(x, k1, k2, k3, mu1, mu2, omega_2pi) {
    .Call(`_BAMBI_dwnorm2_onex_manypar`, x, k1, k2, k3, mu1, mu2, omega_2pi)
}

dwnorm2_manyx_onepar <- function(x, k1, k2, k3, mu1, mu2, omega_2pi) {
    .Call(`_BAMBI_dwnorm2_manyx_onepar`, x, k1, k2, k3, mu1, mu2, omega_2pi)
}

dwnorm2_manyx_manypar <- function(x, k1, k2, k3, mu1, mu2, omega_2pi) {
    .Call(`_BAMBI_dwnorm2_manyx_manypar`, x, k1, k2, k3, mu1, mu2, omega_2pi)
}

#' Sample circular correlation coefficients
#' @param x two column matrix. NA values are not allowed.
#' @param type type of the circular correlation.
#' Must be one of "fl", "js", "tau1" and "tau2". See details.
#' @param alternative one of \code{"two.sided"}, \code{"less"} or \code{"greater"}
#' (defaults to \code{"two.sided"}).
#' Hypothesis test is performed only when \code{type} is either \code{"fl"} or \code{"js"},
#' in which case asymptotic standard error of the estimator is used to construct the test
#' statistic.
#' @param jackknife logical. Compute jackknifed estimate and standard error? Defaults to FALSE.
#' @param bootse logical. Compute bootstrap standard error? Defaults to FALSE.
#' @param n.boot number of bootstrapped samples to compute bootstrap standard error. Defaults to
#' 100. Ignored if \code{bootse} if FALSE.
#'
#' @details
#' \code{circ_cor} calculates the (sample) circular correlation between the columns of x.
#' Two parametric (the Jammalamadaka-Sarma (1988, equation 2.6) form \code{"js"}, and
#' the Fisher-Lee (1983, Section 3) form \code{"fl"})
#' and two non-parametric (two versions of Kendall's tau) correlation coefficients are considered.
#' The first version of Kendall's tau (\code{"tau1"}) is based on equation 2.1 in Fisher and Lee (1982),
#' whereas the second version (\code{"tau2"}) is computed using equations 6.7-6.8 in Zhan et al (2017).
#'
#' The cost-complexity for \code{"js"}, \code{"fl"}, \code{"tau2"} and \code{"tau1"} are \eqn{O(n), O(n^2), O(n^2)} and \eqn{O(n^3)}
#' respectively, where \eqn{n} denotes the number of rows in \code{x}. As such, for large \eqn{n} evaluation of
#' \code{"tau1"} will be slow.
#'
#'
#' @references
#'
#' Fisher, N. I. and Lee, A. J. (1982). Nonparametric measures of angular-angular association. Biometrika, 69(2), 315-321.
#'
#' Fisher, N. I. and Lee, A. J. (1983). A correlation coefficient for circular data. Biometrika, 70(2):327-332.
#'
#' Jammalamadaka, S. R. and Sarma, Y. (1988). A correlation coefficient for
#' angular variables. Statistical theory and data analysis II, pages 349-364.
#'
#' Zhan, X., Ma, T., Liu, S., & Shimizu, K. (2017). On circular correlation for data on the torus. Statistical Papers, 1-21.
#'
#'
#' @examples
#' # generate data from vmsin model
#' set.seed(1)
#' dat <- rvmsin(100, 2,3,-0.8,0,0)
#'
#' # now calculate circular correlation(s) between the 2 columns of dat
#' circ_cor(dat, type="js")
#' circ_cor(dat, type="fl")
#' circ_cor(dat, type="tau1")
#' circ_cor(dat, type="tau2")
#'
#'
#' @export


circ_cor <- function(x, type="js", alternative = "two.sided",
                     jackknife = FALSE, bootse = FALSE, n.boot = 100) {

  if (any(is.na(x)))
    stop("NA values in \'x\'")

  if (!is.matrix(x)) {
    stop("\'x\' must be a two-column matrix")
  }

  if (ncol(x)!=2) {
    stop("\'x\' must be a two-column matrix")
  }

  if (!type %in% c("fl", "js", "tau1", "tau2"))
    stop("\'type\' must be one of \'js\', \'fl\', \'tau1\' or \'tau2\'")

  x <- prncp_reg(x)
  n <- nrow(x)

  if (type == "fl") {
    calc_rho <- function(x) {

      rho <- calc_corr_fl(x)
      A_over_mu2 <- function(margin) {
        alpha <- sapply(1:2, function(p) sum(cos(p*margin))/n)
        beta <- sapply(1:2, function(p) sum(sin(p*margin))/n)
        A <- alpha[1]^2 + beta[1]^2 + alpha[2]*beta[1]^2 -
          alpha[1]^2*alpha[2] - 2*alpha[1]*beta[1]*beta[2]
        mu2 <- 0.5 * (1 - alpha[2]^2 - beta[2]^2)
        A/mu2
      }
      avar <- prod(apply(x, 2, A_over_mu2))

      se <- sqrt(avar)/sqrt(n)
      attr(rho, "se") <- se
      rho

    }
    # attr(rho_fl, "se") <- se
    # attr(rho_fl, "p.value") <- pval
    # rho
  } else if (type == "js") {
    calc_rho <- function(x) {
      sin_x_1_cent <- sin(x[, 1] - atan2(sum(sin(x[, 1])), sum(cos(x[, 1]))))
      sin_x_2_cent <- sin(x[, 2] - atan2(sum(sin(x[, 2])), sum(cos(x[, 2]))))
      num <- sum(sin_x_1_cent*sin_x_2_cent)
      den <- sqrt(sum(sin_x_1_cent^2)*sum(sin_x_2_cent^2))
      rho <- num/den

      # asymptotic variance
      # idx <- data.matrix(expand.grid(0:4, 0:4))
      idx <- rbind(
        c(2, 2),
        c(2, 0), c(0, 2),
        c(1, 3), c(3, 1),
        c(4, 0), c(0, 4)
      )
      rownames(idx) <- paste0(idx[, 1], idx[, 2])
      lambda <- apply(
        idx,
        1,
        function(ii) {
          sum(sin_x_1_cent^(ii[1]) * sin_x_2_cent^(ii[2]))/n
        }
      )
      avar <- unname(
        max(
          lambda["22"]/lambda["20"]*lambda["02"] -
            rho * (
              lambda["13"]/(lambda["20"]*sqrt(lambda["20"]*lambda["02"])) +
                lambda["31"]/(lambda["02"]*sqrt(lambda["20"]*lambda["02"]))
            ) +
            rho^2 / 4 * (
              1 +
                lambda["40"]/lambda["20"]^2 +
                lambda["04"]/lambda["02"]^2 +
                lambda["22"]/(lambda["20"]*lambda["02"])
            ),
          1e-10
        )
      )

      # browser()
      se <- sqrt(avar)/sqrt(n)
      # z <- rho_js/se
      attr(rho, "se") <- se
      # if (alternative == "two.sided") {
      #   pval <- 2 * pnorm(abs(z), lower.tail = FALSE)
      # } else if (alternative == "less") {
      #   pval <- pnorm(z, lower.tail = TRUE)
      # } else if (alternative == "greater") {
      #   pval <- pnorm(z, lower.tail = FALSE)
      # }
      #
      # attr(rho_js, "se") <- se
      # attr(rho_js, "p.value") <- pval
      # rho_js
      rho
    }
  } else if (type == "tau1") {
    calc_rho <- calc_corr_tau_1
  } else {
    calc_rho <- calc_corr_tau_2
  }

  # browser()

  rho <- calc_rho(x)
  rho_attr <- attributes(rho)

  if (jackknife) {
    vals <- vapply(
      1:n,
      function(ii){
        c(calc_rho(x[-ii, , drop = FALSE]))
      },
      0
    )
    vals_adj <- n*rho - (n-1)*vals
    rho_attr$jackknife.est <- sum(vals_adj)/n
    rho_attr$jackknife.se <- sqrt(var(vals_adj)/n)
  }

  if (bootse) {
    boot_vals <- vapply(
      1:n.boot,
      function(ii) {
        idx <- sample(1:n, replace = TRUE)
        c(calc_rho(x[idx, , drop = FALSE]))
      },
      0
    )
    rho_attr$bootse <- sd(boot_vals)
  }

  if (type %in% c("js", "fl")) {
    z <- c(rho)/rho_attr$se
    if (alternative == "two.sided") {
      rho_attr$pval <- 2 * pnorm(abs(z), lower.tail = FALSE)
    } else if (alternative == "less") {
      rho_attr$pval <- pnorm(z, lower.tail = TRUE)
    } else if (alternative == "greater") {
      rho_attr$pval <- pnorm(z, lower.tail = FALSE)
    }
  }

  attributes(rho) <- rho_attr
  rho

}


#' Analytic circular variances and correlations for bivariate angular models
#' @param model bivariate angular model. Must be one of \code{"vmsin"},
#' \code{"vmcos"}, or \code{"wnorm2"}.
#' @param kappa1,kappa2,kappa3 concentration and covariance parameters.
#' Recycled to the same size. kappa3^2 must be < kappa1*kappa2 in the wnorm2 model
#' (see \link{rwnorm2} for a detailed parameterization of \code{wnorm2}).
#' @param mu1,mu2 mean parameters. Ignored as they do not play any role in
#' the analytical formulas.
#' @param nsim Monte Carlo sample size. Ignored if all of \code{kappa1}, \code{kappa2}
#' and \code{abs(kappa3)} are < 150 or if model = \code{"wnorm2"}.
#' @inheritParams contour_model
#'
#' @return
#' Returns a list with elements \code{var1}, \code{var2} (circular variances for the
#' first and second coordinates), \code{rho_fl} and \code{rho_js} (circular correlations).
#' See details.
#'
#' @details
#' The function computes the analytic circular variances and correlations
#' (both Jammalamadaka-Sarma (JS) and Fisher-Lee (FL) forms) for von Mises sine,
#' von Mises cosine and bivariate wrapped normal distributions.
#'
#' For \code{wnorm2}, expressions for the circular variances,
#' JS and FL correlation coefficients can be found in Mardia and Jupp (2009),
#' Jammalamadaka and Sarma (1988) and Fisher and Lee (1983) respectively.
#' For \code{vmsin} and \code{vmcos} these expressions are provided in Chakraborty and Wong (2018).
#'
#' Because the analytic expressions in \code{vmsin} and \code{vmcos} models involve infinite sums
#' of product of Bessel functions,
#' if any of \code{kappa1}, \code{kappa2} and \code{abs(kappa3)} is larger
#' than or equal to 150, IID Monte Carlo with sample size \code{nsim} is used
#' to approximate \code{rho_js} for numerical stability.  From \code{rho_js},
#' \code{rho_fl} is computed using Corollary 2.2 in
#' Chakraborty and Wong (2018), which makes cost-complexity for
#' the \code{rho_fl} evaluation to be of order  O(\code{nsim}) for \code{vmsin}
#' and \code{vmcos} models. (In general,  \code{rho_fl} evaluation
#' is of order O(\code{nsim}^2)).
#'
#' In addition, for the \code{vmcos} model, when \code{-150 < kappa3 < -1}
#' or \code{50 < max(kappa1, kappa2, abs(kappa3)) <= 150}, the analytic formulas
#' in Chakraborty and Wong (2018) are used; however, the reciprocal of the normalizing
#' constant and its partial derivatives are all calculated numerically via (quasi) Monte carlo method for
#' numerical stability. These (quasi) random numbers can be provided through the
#' argument \code{qrnd}, which must be a two column matrix, with each element being
#' a  (quasi) random number between 0 and 1. Alternatively, if \code{n_qrnd} is
#' provided (and \code{qrnd} is missing), a two dimensional sobol sequence of size \code{n_qrnd} is
#' generated via the function \link[qrng]{sobol} from the R package \code{qrng}. If none of \code{qrnd}
#' or \code{n_qrnd} is available, a two dimensional sobol sequence of size 1e4 is used.
#'
#'
#' @examples
#' circ_varcor_model("vmsin", kappa1= 1, kappa2 = 2, kappa3 = 3)
#'
#' # Monte Carlo approximation
#' set.seed(1)
#' dat <- rvmsin(1000, 1, 2, 3)
#' # sample circular variance
#' circ_var <- function(x)
#'   1 - mean(cos(x - atan2(mean(sin(x)), mean(cos(x))) ))
#' circ_var(dat[, 1])
#' circ_var(dat[, 2])
#' circ_cor(dat, "fl")
#' circ_cor(dat, "js")
#'
#' @references
#' Fisher, N. I. and Lee, A. (1983). A correlation coefficient for circular data. Biometrika, 70(2):327-332.
#'
#' Jammalamadaka, S. R. and Sarma, Y. (1988). A correlation coefficient for
#' angular variables. Statistical theory and data analysis II, pages 349-364.
#'
#' Mardia, K. and Jupp, P. (2009). Directional Statistics. Wiley Series in Probability and Statistics. Wiley.
#'
#' Chakraborty, S. and Wong, S, W.K. (2018). On the circular correlation coefficients
#' for bivariate von Mises distributions on a torus. arXiv e-print.
#'
#' @export

circ_varcor_model <- function(model = "vmsin", kappa1 = 1, kappa2 = 1, kappa3 = 0,
                              mu1 = 0, mu2 = 0, nsim = 1e4, ...)
{

  if(any(c(kappa1, kappa2) < 0))
    stop("kappa1 and kappa2 must be non-negative")

  if (length(model) != 1 | !model %in% c("vmsin", "vmcos", "wnorm2"))
    stop("\'model\' must be one of \"vmsin\", \"vmcos\" or \"wnorm2\"")

  if (nsim <= 0)
    stop("\'nsim\' must be a positive integer")

  ell <- list(...)
  nsim <- round(nsim)

  if (model == "vmcos") {


    if (!is.null(ell$qrnd)) {
      qrnd_grid <- ell$qrnd
      dim_qrnd <- dim(qrnd_grid)
      if (!is.matrix(qrnd_grid) | is.null(dim_qrnd) |
          dim_qrnd[2] != 2)
        stop("\'qrnd\' must be a two column matrix")
      n_qrnd <- dim_qrnd[1]
    } else if (!is.null(ell$n_qrnd)){
      n_qrnd <- round(ell$n_qrnd)
      if (n_qrnd < 1)
        stop("n_qrnd must be a positive integer")
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    } else {
      n_qrnd <- 1e4
      qrnd_grid <- sobol(n_qrnd, 2, FALSE)
    }
  }

  if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    kappa1 <- expanded[[1]]
    kappa2 <- expanded[[2]]
    kappa3 <- expanded[[3]]
    if (model == "wnorm2" &
        any (kappa1*kappa2 - kappa3*kappa3 <= 1e-10))
      stop("abs(kappa3) must be less than sqrt(kappa1*kappa2) in wnorm2")
    lapply(1:length(kappa1),
           function(j) {
             inargs <- list(kappa1 = kappa1[j], kappa2 = kappa2[j],
                            kappa3 = kappa3[j])
             if (model == "vmcos") {
               inargs$qrnd_grid <- qrnd_grid
               # inargs$force_approx_const <- ell$force_approx_const
             }
             do.call(paste0(model, "_var_cor_singlepar"),
                     inargs)
           }
    )
  } else {
    if (model == "wnorm2" &
        (kappa1*kappa2 - kappa3*kappa3 <= 1e-10))
      stop("abs(kappa3) must be less than sqrt(kappa1*kappa2) in wnorm2")

    inargs <- list(kappa1 = kappa1, kappa2 = kappa2,
                   kappa3 = kappa3)
    if (model == "vmcos") inargs$qrnd_grid <- qrnd_grid
    do.call(paste0(model, "_var_cor_singlepar"),
            inargs)
  }

}

# list of functions


#' The univariate Wrapped Normal distribution
#' @inheritParams rvm
#' @param int.displ integer displacement. If \code{int.displ =} M, then the infinite sum in the
#' density is approximated by a sum over 2*M + 1 elements. (See Details.) The allowed values are 1, 2, 3, 4 and 5. Default is 3.
#' @details If \code{mu} and \code{kappa} are not specified they assume the default values of \code{0} and \code{1} respectively.
#' @details The univariate wrapped normal distribution has density
#' \deqn{f(x) = \sqrt(\kappa/(2\pi)) \sum \exp(-\kappa/2 (x - \mu(2\pi\omega))^2)}
#' where the sum extends over all integers \eqn{\omega},
#' and is approximated by a sum over \eqn{\omega} in \eqn{\{-M, -M+1, ..., M-1, M \}} if \code{int.displ = } \eqn{M}.
#' @return \code{dwnorm} gives the density  and \code{rwnorm} generates random deviates.
#'
#' @examples
#'
#' kappa <- 1:3
#' mu <- 0:2
#' x <- 1:10
#' n <- 10
#'
#'
#' # when x and both parameters are scalars, dwnorm returns a single density
#' dwnorm(x[1], kappa[1], mu[1])
#'
#' # when x is a vector but both the parameters are scalars, dmv returns a vector of
#' # densities calculated at each entry of x with the same parameters
#' dwnorm(x, kappa[1], mu[1])
#'
#' # if x is scalar and at least one of the two paraemters is a vector, both parameters are
#' # recycled to the same length, and dwnorm returns a vector of with ith element being the
#' # density evaluated at x with parameter values kappa[i] and mu[i]
#' dwnorm(x[1], kappa, mu)
#'
#' # if x and at least one of the two paraemters is a vector, x and the two parameters are
#' # recycled to the same length, and dwnorm returns a vector of with ith element being the
#' # density at ith element of the (recycled) x with parameter values kappa[i] and mu[i]
#' dwnorm(x, kappa, mu)
#'
#' # when parameters are all scalars, number of observations generated by rwnorm is n
#' rwnorm(n, kappa[1], mu[1])
#'
#' # when at least one of the two parameters is a vector, both are recycled to the same length,
#' # n is ignored, and the number of observations generated by rwnorm is the same as the length
#' # of the recycled vectors
#' rwnorm(n, kappa, mu)
#'
#' @export

rwnorm <- function(n=1, kappa = 1, mu = 0)
{
  if(any(kappa < 0)) stop("kappa must be non-negative")
  if(any(mu < 0 | mu >= 2*pi)) mu <- prncp_reg(mu)

  if (all(kappa > 1e-10)) {
    samp <- rnorm(n, mean = mu, sd = sqrt(1/kappa))
    prncp_reg(samp)
  } else {
    expndn_set <- expand_args(kappa, mu, 1:n)
    kappa <- expndn_set[[2]]
    mu <- expndn_set[[3]]
    # den <- rep(0, n_x)
    n_final <- length(kappa)
    which_unif <- which(kappa < 1e-10)
    n_unif <- length(which_unif)
    samp <- rep(0, n_final)
    samp[which_unif] <- runif(n_unif, 0, 2*pi)
    if (n_unif < n_final)
      samp[-which_unif] <- prncp_reg(rnorm(n_final - n_unif, mu[-which_unif],
                                           sqrt(1/kappa[-which_unif])))


    # for (i in 1:n_x) {
    #   if (kappa1[par_set[i]] < 1e-10 | kappa2[par_set[i]] < 1e-10) {
    #     den[i] <- c(dwnorm2_onex_manypar(x[x_set[i], ], kappa[par_set[i]],
    #                                      kappa2[par_set[i]], kappa3[par_set[i]],
    #                                      mu1[par_set[i]], mu2[par_set[i]], omega.2pi))
    #   } else {
    #     den[i] <- 1/(4*pi^2)
    #   }
    # }
    samp
  }

}

#' @rdname rwnorm
#' @export
dwnorm <- function(x, kappa = 1, mu = 0, int.displ,
                   log=FALSE)
{
  if(any(kappa < 0)) stop("kappa must be non-negative")
  if(any(mu < 0 | mu >= 2*pi)) mu <- prncp_reg(mu)

  if(missing(int.displ)) int.displ <- 3
  else if(int.displ >= 5) int.displ <- 5
  else if(int.displ <= 1) int.displ <- 1
  displ <- floor(int.displ)
  omega.2pi.1d <- (-displ):displ * (2*pi) # 2pi * 1d integer displacements


  if(max(length(kappa), length(mu)) > 1) {
    expanded <- expand_args(kappa, mu)
    kappa <- expanded[[1]]; mu <- expanded[[2]]
  }

  par.mat <- rbind(kappa, mu)
  n_par <- ncol(par.mat)
  n_x <- length(x)

  # browser()

  if (all(kappa > 1e-10)) {
    if (n_par == 1) {
      den <- c(duniwnorm_manyx_onepar(as.vector(x), kappa, mu, omega.2pi.1d))

    } else if (n_x == 1) {
      den <- c(duniwnorm_onex_manypar(x, kappa, mu, omega.2pi.1d))
    } else {
      x_set <- 1:n_x
      par_set <- 1:n_par
      expndn_set <- expand_args(x_set, par_set)
      x_set <- expndn_set[[1]]
      par_set <- expndn_set[[2]]
      den <- c(duniwnorm_manyx_manypar(x[x_set], kappa[par_set],
                                       mu[par_set], omega.2pi.1d))
    }
  } else {
    x_set <- 1:n_x
    par_set <- 1:n_par
    expndn_set <- expand_args(x_set, par_set)
    x_set <- expndn_set[[1]]
    par_set <- expndn_set[[2]]
    x_long <- x[x_set]
    kappa_long <- kappa[par_set]
    mu_long <- mu[par_set]
    which_unif <- which(kappa_long < 1e-10)
    n_x_long <- length(x_long)
    den <- rep(0, n_x_long)
    den[which_unif] <- 1/(2*pi)
    if (length(which_unif) < n_x_long)
      den[-which_unif] <- c(duniwnorm_manyx_manypar(x_long[-which_unif],
                                                    kappa_long[-which_unif],
                                                    mu_long[-which_unif],
                                                    omega.2pi.1d))
  }


  if (log) {
    den <- log(den)
  }
  den

}

#' The univariate Wrapped Normal mixtures
#' @inheritParams rvmmix
#' @inheritParams rwnorm
#' @details \code{pmix}, \code{mu} and \code{kappa} must be of the same length, with \eqn{j}-th element corresponding to the \eqn{j}-th component of the mixture distribution.
#' @details The univariate wrapped normal mixture distribution with component size \code{K = \link{length}(pmix)} has density
#' \deqn{g(x) = p[1] * f(x; \kappa[1], \mu[1]) + ... + p[K] * f(x; \kappa[K], \mu[K])}
#' where \eqn{p[j], \kappa[j], \mu[j]} respectively denote the mixing proportion, concentration parameter and the mean parameter for the \eqn{j}-th component
#' and \eqn{f(. ; \kappa, \mu)} denotes the density function of the (univariate) wrapped normal distribution with mean parameter \eqn{\mu} and concentration parameter \eqn{\kappa}.
#' @return \code{dwnormmix} computes the density  and \code{rwnormmix} generates random deviates from the mixture density.
#'
#' @examples
#' kappa <- 1:3
#' mu <- 0:2
#' pmix <- c(0.3, 0.3, 0.4)
#' x <- 1:10
#' n <- 10
#'
#' # mixture densities calculated at each point in x
#' dwnormmix(x, kappa, mu, pmix)
#'
#' # number of observations generated from the mixture distribution is n
#' rwnormmix(n, kappa, mu, pmix)
#'
#' @export

rwnormmix <- function(n=1, kappa, mu, pmix)
{
  allpar <- list(kappa=kappa, mu=mu, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$kappa <= 0)) stop("kappa must be positive in wnorm")
  if(any(allpar$mu < 0 | allpar$mu >= 2*pi)) allpar$mu <- prncp_reg(allpar$mu)

  clus_label <- cID(tcrossprod(rep(1, n), allpar$pmix), length(allpar$pmix), runif(n))
  samp <- rnorm(length(clus_label), mean = allpar$mu[clus_label], sd = 1/sqrt(allpar$kappa[clus_label]))

  prncp_reg(samp)
}

#' @rdname rwnormmix
#' @export
dwnormmix <- function(x, kappa, mu, pmix, int.displ=3, log=FALSE)
{
  allpar <- list(kappa=kappa, mu=mu, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$kappa <= 0)) stop("kappa must be positive")
  if(any(allpar$mu < 0 | allpar$mu >= 2*pi)) allpar$mu <- prncp_reg(allpar$mu)


  ncomp <- length(kappa)

  allcompden <- vapply(1:ncomp,
                       function(j) dwnorm(x, kappa[j], mu[j],
                                          int.displ, FALSE),
                       rep(0, length(x)))

  mixden <- c(allcompden %*% pmix)

  if (log) {
    log(mixden)
  } else {
    mixden
  }

}


#' Fitting univariate wrapped normal mixtures using MCMC
#' @inheritParams fit_vmsinmix
#'
#' @details
#' Wrapper for \link{fit_angmix} with \code{model = "wnorm"}.
#'
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.wnorm.20 <- fit_wnormmix(wind$angle, ncomp = 3, n.iter =  20,
#'                              n.chains = 1)
#' fit.wnorm.20
#'
#' @export

fit_wnormmix <- function(...)
{
  fit_angmix(model="wnorm", ...)
}

#' Stepwise fitting of angular mixture models with incremental component sizes and optimum model selection
#' @inheritParams pointest
#' @inheritParams fit_angmix
#' @param start_ncomp starting component size. A single component model is fitted if `start_ncomp` is equal to one.
#' @param max_ncomp maximum number of components allowed in the mixture model.
#' @param crit model selection criteria, one of `"LOOIC"`, `"WAIC"`, `"AIC"`, `"BIC"`, `"DIC"` or `"LOGML"`. Default is
#' `"LOOIC"`.
#' @param L HMC tuning parameter (trajectory length) passed to \link{fit_angmix}. Can be a numeric vetor (or scalar), in which case
#' the same `L` is passed to all \link{fit_angmix} calls, or can be a list of length `max_ncomp-start_ncomp+1`,
#' so that `L_list[[i]]` is passed as the argument `L` to \link{fit_angmix} call with `ncomp = max_ncomp+i-1`. See
#' \link{fit_angmix} for more details on `L` including its default values. Ignored if `method = "rwmh"`.
#' @param fn function to evaluate on MCMC samples to estimate parameters.
#' Defaults to `mean`, which computes the estimated posterior means. If `fn = max`,
#' then MAP estimate is calculated from the MCMC run. Used only if `crit = "DIC"`, and ignored otherwise.
#' @param prev_par logical. Should the MAP estimated parameters from the model with `ncomp = K` be used in the model
#' with `ncomp = K+1` as the starting parameters, with the component with largest mixing proportion appearing twice in the
#' bigger model?
#' @param form form of crit to be used. Available choices are 1 and 2. Used only if `crit` is `"DIC"` and ignored otherwise.
#' @param ... additional arguments passed to \link{fit_angmix}.
#' @param logml_maxiter maximum number of iterations (`maxiter`) passed to \link[bridgesampling]{bridge_sampler} for calculating
#' `LOGML`. Ignored if `crit` is not `LOGML`.
#' @param fix_label logical. Should the label switchings on the current fit (only the corresponding "best chain" if `use_best_chain = TRUE`)
#' be fixed before computing parameter estimates and model selection criterion? Defaults to `TRUE` if `perm_sampling` is true in
#' the \link{fit_angmix} call, or if `crit = "DIC"` and `form = 1`.
#' @param silent logical. Should the current status (such as what is the current component labels, which job is being done etc.)
#' be printed? Defaults to `TRUE`.
#' @param return_all logical. Should all angmcmc objects obtained during step-wise run be returned? *Warning*: depending on the
#' sizes of `n.iter`, `start_ncomp`, `max_ncomp` and `n.chains`, this can be very memory intesive. In such
#' cases, it is recommended that `return_all` be set to `FALSE`, and, if required, the intermediate fitted objects be
#' saved to file by setting `save_fits = TRUE`.
#' @param save_fits logical. Should the intermediate angmcmc objects obtained during step-wise run be saved
#'  to file using \link{save}? Defaults to TRUE. See `save_file` and `save_dir`.
#' @param save_file,save_dir `save_file` is a list of size `max_ncomp-start_ncomp+1`,
#' with k-th entry providing the `file`
#' argument used to \link{save} the intermediate angmcmc object with `ncomp = k` (titled `"fit_angmcmc"`).
#' If not provided, then k-th element
#' of `save_file[[k]]` is taken to be `paste(save_dir, "comp_k", sep="/")`. Both are ignored if
#' `save_fits = FALSE`.
#' @param use_best_chain logical. Should only the "best" chain obtained during each intermediate fit be used during
#' computation of model selection criterion? Here "best" means the chain
#' with largest (mean over iterations) log-posterior density. This can be helpful if one of the chains gets stuck at local optima. Defaults to TRUE.
#' @param return_llik_contri passed to \link{fit_angmix}. By default, set to `TRUE` if `crit` is either `"LOOIC"`
#' or `"WAIC"`, and to `FALSE` otherwise.
#' @param alpha significance level used in the test \eqn{H_{0K}}: expected log predictive density (elpd) for the fitted model with  K components >= elpd for the fitted model
#' with K + 1 components if `crit` is `"LOOIC"` or `"WAIC"`.
#' Must be a scalar between 0 and 1. Defaults to 0.05. See Details. Ignored for any other `crit`.
#' @param bonferroni_alpha logical. Should a Bonferroni correction be made on the test size `alpha` to adjust for
#' multiplicity due to (`max_ncomp` - `start_ncomp`) possible hypothesis tests? Defaults to TRUE.
#' Relevant only if `crit` is in  `c("LOOIC", "WAIC")`, and ignored otherwise. See Details.
#' @param bonferroni_adj_type character string. Denoting type of Bonferroni adjustment to make.
#' Possible choices are `"decreasing"` (default) and `"equal"`. Ignored if either `bonferroni_alpha`
#' is `FALSE`, or `crit` is outside `c("LOOIC", "WAIC")`. See Details.
#'
#' @details
#' The goal is to fit an angular mixture model with an optimally chosen component size K.
#' To obtain an optimum K, mixture models with incremental component sizes
#' between `start_ncomp` and `max_ncomp` are fitted incrementally using \link{fit_angmix},
#' starting from K = 1.
#' If the model selection criterion `crit` is `"LOOIC"` or `"WAIC"`, then a test of hypothesis
#' \eqn{H_{0K}}: expected log predictive density (elpd) for the fitted model with  K components >= elpd for the fitted model
#' with K + 1 components, is performed at every K >= 1. The test-statistic used for the test is an approximate z-score
#' based on the normalized estimated elpd difference between the two models obtained from \link[loo]{compare}, which provides
#' estimated elpd difference along with its standard error estimate. Because the computed standard error of elpd difference
#' can be overly optimistic when the elpd difference is small (in particular < 4),
#' a conservative worst-case estimate (equal to twice of the computed standard error)
#' is used in such cases. To account for multiplicity among the M =
#' (`max_ncomp` - `start_ncomp`) possible sequential tests performed,
#' by default a Bonferroni adjustment to the test level `alpha` is made.
#' Set `bonferroni_alpha = FALSE} to remove the adjustment. To encourage
#' parsimony in the final model, by default (`bonferroni_adj_type = "decreasing"`)
#' a decreasing sequence of adjusted alphas of the form `alpha * (0.5)^(1:M) / sum((0.5)^(1:M))`
#' is used. Set `bonferroni_adj_type = "equal"`
#' to use equal sequence of adjusted alphas (i.e., `alpha/M`) instead.
#'
#' The incremental fitting stops if  \eqn{H_{0K}} cannot be rejected
#' (at level `alpha`) for some K >= 1; this K is then regarded as the optimum number of components.
#' If `crit` is not `"LOOIC"` or `"WAIC"` then mixture model with the first minimum value of the model selection criterion `crit`
#' is taken as the best model.
#'
#' Note that in each intermediate fitted model, the total number of components (instead of the number of
#' "non-empty components") in the model is used to estimate of the true component
#' size, and then the fitted model is penalized for model complexity (via the model selection criterion used).
#' This approach of selecting an optimal K follows the perspective "let two component specific parameters
#' be identical" for overfitting mixtures, and as such the  Dirichlet prior hyper-parameters `pmix.alpha`
#' (passed to \link{fit_angmix}) should be large. See  Fruhwirth-Schnatter (2011) for more deltails.
#'
#' Note that the stability of \link[bridgesampling]{bridge_sampler} used in marginal likelihood estimation heavily depends on stationarity of the
#' chains. As such, while using this criterion, we recommending running the chain long enough, and setting `fix_label = TRUE`
#' for optimal performance.
#'
#' @references
#' Fruhwirth-Schnatter, S.: Label switching under model uncertainty. In: Mengerson, K., Robert, C., Titterington, D. (eds.) Mixtures:
#' Estimation and Application, pp. 213-239. Wiley, New York (2011).
#'
#'
#'
#'
#' @return Returns a named list (with class = `stepfit`) with the following seven elements:
#'
#' `fit.all` (if `return_all = TRUE`) - a list all angmcmc objects created at each component size;
#'
#' `fit.best` - angmcmc object corresponding to the optimum component size;
#'
#' `ncomp.best` - optimum component size (integer);
#'
#' `crit` - which model comparison criterion used (one of `"LOOIC", "WAIC", "AIC", "BIC", "DIC"` or `"LOGML"`);
#'
#' `crit.all` - all `crit` values calculated (for all component sizes);
#'
#' `crit.best` - `crit` value for the optimum component size; and
#'
#' `maxllik.all` - maximum (obtained from MCMC iterations) log likelihood for all fitted models
#'
#' `maxllik.best` - maximum log likelihodd for the optimal model; and
#'
#' `check_min` - logical; is the optimum component size less than `max_ncomp`?
#' @md
#' @examples
#' # illustration only - more iterations needed for convergence
#' set.seed(1)
#' fit.vmsin.step.15 <- fit_incremental_angmix("vmsin", tim8, "BIC", start_ncomp = 1,
#'                                           max_ncomp = 3, n.iter = 15,
#'                                           n.chains = 1, save_fits=FALSE)
#' (fit.vmsin.best.15 <- bestmodel(fit.vmsin.step.15))
#' lattice::densityplot(fit.vmsin.best.15)
#'
#'
#' @export


fit_incremental_angmix <- function(model, data,
                                   crit = "LOOIC",
                                   start_ncomp=1, max_ncomp=10,
                                   L = NULL,
                                   fn = mean,
                                   fix_label = NULL,
                                   form = 2,
                                   start_par = NULL,
                                   prev_par = TRUE,
                                   logml_maxiter = 1e4,
                                   return_all = FALSE,
                                   save_fits = FALSE,
                                   save_file = NULL,
                                   save_dir = "",
                                   silent = FALSE,
                                   return_llik_contri = (crit %in% c("LOOIC", "WAIC")),
                                   use_best_chain = TRUE,
                                   alpha = 0.05,
                                   bonferroni_alpha = TRUE,
                                   bonferroni_adj_type = "decreasing",
                                   ...)
{


  if (length(model) > 1)
    stop("\'model\' must be a scalar")
  if(missing(model))
    stop("argument \"model\" is missing, with no default")
  if(!crit %in% c("AIC", "BIC", "DIC", "WAIC", "LOOIC", "LOGML"))
    stop("non-compatible criterion")
  if (model %in% c("vmsin", "vmcos", "wnorm2")) {
    type <- "bi"
  } else if (model %in% c("vm", "wnorm")) {
    type <- "uni"
  } else  {
    stop("non-compatible model")
  }

  if (!missing(start_par)) {
    if (min(listLen(start_par)) != max(listLen(start_par)))
      stop("Lengths of elements in start_par differ")
    else if((listLen(start_par))[1] != start_ncomp)
      stop("Number of components in start_par must be equal to \'start_ncomp\'")
  }


  if(start_ncomp > max_ncomp)
    stop("\'start_ncomp\' cannot be smaller than \'max_ncomp\'")

  all_ncomp <- start_ncomp:max_ncomp
  n_ncomp <- length(all_ncomp)


  if (is.null(save_file)) {
    save_file <- lapply(all_ncomp, function(j) paste0(save_dir, "/comp_", j, ".Rdata"))
  }
  else if (!is.list(save_file) | length(save_file) != n_ncomp)
    stop("\'save_file\' must be a list of length max_ncomp-start_ncomp+1")


  crit_print <- crit

  if (crit == "LOGML")
    crit_print <- "(negative) LOGML"


  if (type == "bi") {
    if (!(is.matrix(data) | is.data.frame(data)))
      stop("\'data\' must be a two column matrix for model = \'vmsin\', \'vmcos\' and \'wnorm2\'")

    if (ncol(data) != 2)
      stop("\'data\' must be a two column matrix for model = \'vmsin\', \'vmcos\' and \'wnorm2\'")

    data.rad <- rm_NA_rad(data)
    n.data <- nrow(data.rad)

  }
  else  {
    if (!is.numeric(data))
      stop("\'data\' must be a vector for \'model\' = \'vm\' and \'wnorm\'")
  }


  fit_all <- NULL

  if (return_all)
    fit_all <- vector("list", n_ncomp)


  ell <- list(...)

  if (is.null(ell$perm_sampling))
    ell$perm_sampling <- formals(fit_angmix)$perm_sampling

  if (is.null(fix_label)) {
    if(any(form == 1 & crit == "DIC", ell$perm_sampling & prev_par,
           ell$perm_sampling & crit == "LOGML")) {
      fix_label <- TRUE
    } else {
      fix_label <- FALSE
    }

  }

  ntests_max <- n_ncomp - 1

  if (crit %in% c("LOOIC", "WAIC")) {

    if (any(length(alpha) != 1,
            !is.numeric(alpha),
            alpha < 0,
            alpha > 1)) {

      stop("\'alpha\' must be a scalar between 0 and 1")
    }



    stopifnot(
      is.logical(bonferroni_alpha),
      length(bonferroni_alpha) == 1,
      ! is.na(bonferroni_alpha),
      is.character(bonferroni_adj_type),
      length(bonferroni_adj_type) == 1,
      bonferroni_adj_type %in% c("equal", "decreasing"),
      ! is.na(bonferroni_adj_type)
    )

    alpha_vec <- rep(NA, ntests_max)
    if (!bonferroni_alpha) {
      alpha_vec <- rep(alpha, ntests_max)
    } else if (bonferroni_adj_type == "decreasing") {
      alpha_vec <- alpha * (0.5)^(1:ntests_max) / sum((0.5)^(1:ntests_max))
    } else if (bonferroni_adj_type == "equal") {
      alpha_vec <- alpha / ntests_max
    }

  }


  # q_norm <- qnorm(alpha, lower.tail = FALSE)


  # all_fit <- vector("list", n_ncomp)
  all_input <- list("data" = data, "model" = model,
                    return_llik_contri = return_llik_contri,
                    ...)

  formal_fit_angmix <- formals(fit_angmix)

  if (is.null(all_input$n.chains))
    all_input$n.chains <- formal_fit_angmix$n.chains

  n.chains <- all_input$n.chains


  # when missing L, get default value of L and make L_list
  if (is.null(L)) {
    L_list <- lapply(1:(max_ncomp-start_ncomp+1),
                     function(j) {
                       ncomp <- start_ncomp-j+1
                       eval(formal_fit_angmix$L)
                     })
  }
  # when L is not null, check if it's in correct format first
  else if (all(!is.list(L) & !is.numeric(L),
               is.list(L) & length(L) != (max_ncomp - start_ncomp + 1),
               !is.numeric(L)))
    stop("L must either be a list of length max_ncomp-start_ncomp+1 or a vector")

  else if (!is.list(L) & is.numeric(L)) {
    L_list <- lapply(1:(max_ncomp - start_ncomp + 1),
                     function(j) L)
    # all elements of L_list are the same
  }

  else if (is.list(L)) {
    L_list <- L
    # L_list is just L, when given properly
  }


  all_par_est <- vector("list", length = max_ncomp-start_ncomp+1)


  # all_crit <- rep(0, n_ncomp)
  all_crit <- vector("list", length = n_ncomp)
  all_maxllik <- rep(0, n_ncomp)

  if(!form %in% 1:2) form <- 1

  check_min <- FALSE

  curr_seed <- NULL
  if (exists(".Random.seed", .GlobalEnv)) {
    curr_seed <- .GlobalEnv$.Random.seed
  }



  for(j in seq_len(length(all_ncomp))) {
    all_input$ncomp <- all_ncomp[j]
    all_input$L <- L_list[[j]]

    # copy the previous fit as fit_prev if j > 1
    if (j > 1) {
      fit_prev <- fit_angmcmc
      rm(fit_angmcmc)
      gc()
    }

    # starting parameters for j > 1, ncomp >= 3
    if (j > 1 & prev_par & all_ncomp[j] > 2) {
      all_par <- all_par_est[[j-1]]
      # find the component with largest mix_prop
      copy_comp_id <- which.max(all_par[1, ])[1]
      new_comp <- all_par[, copy_comp_id]
      new_comp_id <- all_input$ncomp
      all_par <- cbind(all_par, new_comp)
      # distribute the weights between the "new" and "old" components
      all_par[1, c(copy_comp_id, new_comp_id)] <-
        all_par[1, copy_comp_id]/2

      colnames(all_par) <- 1:new_comp_id
      start_par <- list_by_row(all_par)
      all_input$start_par <- start_par
    }

    else if (j == 1 & !missing(start_par)) {
      all_input$start_par <- start_par
    }
    else {
      all_input$start_par <- NULL
    }







    if (!silent) {
      cat("**************\n")
      cat(paste("Fitting", model,
                "mixture model with ncomp = ",
                all_ncomp[j], "...\n") )

    }

    if (!is.null(curr_seed)) {
      msg <- paste(
        "\n***Restoring RNG state to specified seed***\n"
      )
      # cat(msg)
      .GlobalEnv$.Random.seed <- curr_seed
    }

    fit_angmcmc <- do.call(fit_angmix, all_input)


    all_maxllik[j] <- maxllik_curr <- max(fit_angmcmc$llik[fit_angmcmc$final_iter, ])

    if (!silent)
      cat(paste("\nMaximum log likelihood (from MCMC iterations) =",
                round(maxllik_curr, 3), "\n"))


    if (use_best_chain) {
      best.chain.id <- which.max(
        vapply(1:fit_angmcmc$n.chains,
               function (j) mean(fit_angmcmc$lpd[fit_angmcmc$final_iter, j]),
               0))
      fit_angmcmc_adj <- select_chains(fit_angmcmc, best.chain.id)
    } else {
      fit_angmcmc_adj <- fit_angmcmc
    }


    if (fix_label & all_ncomp[j] > 1) {
      if(!silent)
        cat("\nCalling fix_label with default settings ...\n")

      fit_angmcmc_adj <- fix_label(fit_angmcmc_adj)

      # replace fit_angmcmc by fit_angmcmc_adj if fix_label and !use_best_chain
      if (!use_best_chain) {
        fit_angmcmc <- fit_angmcmc_adj
      }

    }
    else if (!silent) {
      cat("\nSkipping fix_label call ...\n")
    }



    all_par_est[[j]] <- pointest(fit_angmcmc_adj, fn = "MODE")



    if(save_fits) {
      if (!silent)
        cat(paste0("\nSaving the output (titled \'fit_angmcmc\') with filename = \'",
                   save_file[[j]],  "\' ...\n"))
      save(fit_angmcmc, file=save_file[[j]])
    }


    if (return_all)
      fit_all[[j]] <- fit_angmcmc



    if(!silent)
      cat("\nComputing model selection criterion ...\n")

    if (crit == "WAIC") {
      curr_crit <- suppressWarnings(loo::waic(fit_angmcmc_adj))
      all_crit[[j]] <- curr_crit
    }

    else if (crit == "LOOIC") {
      curr_crit <- suppressWarnings(loo::loo(fit_angmcmc_adj))
      all_crit[[j]] <- curr_crit
    }

    else if (crit == "LOGML") {
      curr_crit <- tryCatch(bridgesampling::bridge_sampler(fit_angmcmc_adj, silent = TRUE, maxiter = logml_maxiter),
                            error = function(e) "error")

      if (unlist(curr_crit)[1] == "error")
        stop(paste0("log posterior too unstable with ncomp = ",
                    all_ncomp[j], " to calculate log ML. Try a different criterion."))
      all_crit[[j]] <- -curr_crit$logml
    }

    else if (crit == "DIC") {
      curr_crit <- DIC(fit_angmcmc_adj, form=form)
      all_crit[[j]] <- curr_crit["DIC"]
    }

    else if (crit == "AIC") {
      all_crit[[j]] <- AIC(fit_angmcmc_adj)
    }

    else {
      all_crit[[j]] <- BIC(fit_angmcmc_adj)
    }

    # if(j > start_ncomp) cat("\n")

    if(!silent) {
      crit_val_print <- ""
      if (!crit %in% c("LOOIC", "WAIC"))
        crit_val_print <- round(all_crit[[j]], 3)

      cat(paste("\t", "ncomp = ", all_ncomp[j], ",\t",
                crit_print, ":", crit_val_print))

      if (crit %in% c("LOOIC", "WAIC")) {
        cat("\n")
        cat(suppressWarnings(utils::capture.output(all_crit[[j]])), sep = "\n")
      }
      cat("\n")
      cat("**************\n\n")
    }

    if (j > 1 ) {
      if (crit %in% c("LOOIC", "WAIC")) {
        # browser()
        crit_list <- list(
          comp_j_minus_1 = all_crit[[j-1]],
          comp_j = all_crit[[j]]
        )

        compare_crit_obj <- loo::loo_compare(
          x = crit_list
        )

        if ("model" %in% colnames(compare_crit_obj)) {
          idx_curr <- which(compare_crit_obj$model == "comp_j")
          idx_prev <- which(compare_crit_obj$model == "comp_j_minus_1")
        } else {
          idx_curr <- which(rownames(compare_crit_obj) == "comp_j")
          idx_prev <- which(rownames(compare_crit_obj) == "comp_j_minus_1")
        }

        E_diff <- compare_crit_obj[idx_curr, "elpd_diff"] -
          compare_crit_obj[idx_prev, "elpd_diff"]
        E_diff_se <- sum(compare_crit_obj[c(idx_curr, idx_prev), "se_diff"])

        if (abs(E_diff) < 4) {
          E_diff_se <- 2 * E_diff_se
        }

        # test for signif improvement in elpd
        # H0: curr elpd - prev elpd <= 0 vs Ha: >
        zscore <- E_diff/E_diff_se
        pval_curr <- pnorm(zscore, lower.tail = FALSE)


        # zscore <- compare_crit[1]/compare_crit[2]
        if (pval_curr >= alpha_vec[j-1]) {
          # fail to reject null at alpha --
          # so no signific improvement in curr elpd compared to prev
          check_min <- TRUE
          j.best <- j-1
          pval_txt <- format(pval_curr, digits = 3, scientific = TRUE)
          alpha_txt <- format(alpha_vec[j-1], digits = 3, scientific = TRUE)

          msg <- paste0(
            "\nImprovement in predicitive accuracy not significant",
            " (p=", pval_txt,
            ">=level=", alpha_txt,
            "). Stopping...\n\n"
          )
          cat(msg)
          fit_best <- fit_prev #previous fit is best
          break
        }
      } else if (all_crit[[j]] > all_crit[[j-1]]) {
        check_min <- TRUE
        j.best <- j-1
        cat("\nFirst minimum attained. Stopping...\n")
        fit_best <- fit_prev #previous fit is best
        break
      }
    }


    if (all_ncomp[j] == max_ncomp) {
      cat("\n\'max_ncomp\' reached. Stopping...\n")
      j.best <- j
      fit_best <- fit_angmcmc
    }
  }

  result <- list("fit.all" = fit_all[1:j], "fit.best" = fit_best,
                 "ncomp.best" = all_ncomp[j.best], "crit" = crit,
                 "crit.all" = all_crit[1:j],
                 "crit.best" = all_crit[[j.best]],
                 "maxllik.all" = all_maxllik[1:j],
                 "maxllik.best" = all_maxllik[j.best],
                 "check_min" = check_min)
  class(result) <- "stepfit"

  result

}


#' Convenience function for extracting angmcmc object, and the value of the model
#' selection criterion corresponding to the best fitted model in stepwise fits
#'
#' @param step_object stepwise fitted object obtained from \link{fit_incremental_angmix}.
#'
#' @return `bestmodel` returns an `angmcmc` object, and
#' `bestcriterion` returns the  corresponding value of model selection criterion  for the best fitted model in `step_object`.
#'
#' @details
#' These are convenience functions; the best fitted model and the corresponding value of model selection criterion
#' can also be directly obtained by
#' extracting the elements `"fit.best"` and `"crit.best"` from `step_object` respectively.
#' Note that `bestcriterion} returns:
#' (a) a scalar number (class = `numeric`) if `crit`
#' used in original `fit_incremental_angmix` call is `'AIC'`, `'BIC'` or `'DIC'`,
#' (b) an element of class `bridge` from package `bridgesampling` if `crit` is
#' `LOGML`, (c) an element of class `c("waic", "loo")` if `crit = 'WAIC'`, and (d) an element of
#' class `c("psis_loo", "loo")` if `crit = "LOOIC"`. See documentations of these model
#' selection criteria for more details.
#'
#' @md
#' @examples
#' # illustration only - more iterations needed for convergence
#' set.seed(1)
#' fit.vmsin.step.15 <- fit_incremental_angmix("vmsin", tim8, start_ncomp = 1,
#'                                             max_ncomp = 3, n.iter = 15,
#'                                             n.chains = 1,
#'                                             crit = "WAIC")
#' fit.vmsin.best.15 <- bestmodel(fit.vmsin.step.15)
#' fit.vmsin.best.15
#'
#' crit.best <- bestcriterion(fit.vmsin.step.15)
#' crit.best
#'
#' @export

bestmodel <- function(step_object) {
  if (!is(step_object, "stepfit")) stop("\'step_object\' is not a stepwise fitted object")
  step_object$fit.best
}


#' @rdname bestmodel
#' @export
bestcriterion <- function(step_object) {
  if (!is(step_object, "stepfit")) stop("\'step_object\' is not a stepwise fitted object")
  step_object$crit.best
}

#' The bivariate Wrapped Normal distribution
#' @inheritParams rvmsin
#' @inheritParams rwnorm
#' @param int.displ integer displacement. If \code{int.displ =} M, then each infinite sum in the
#' density is approximated by a finite sum over 2*M + 1 elements. (See Details.) The allowed values are 1, 2, 3, 4 and 5. Default is 3.
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0},
#' and \code{kappa3^2 < kappa1*kappa2}.
#' @param ... additional arguments passed to \link[mvtnorm]{rmvnorm} from package \code{mvtnorm}
#' @importFrom mvtnorm rmvnorm
#'
#' @details
#' The bivariate wrapped normal density at the point \eqn{x = (x_1, x_2)} is given by,
#' \deqn{f(x) = \sqrt((\kappa_1 \kappa_2 - (\kappa_3)^2)) / (2\pi) \sum \exp(-1/2 * (\kappa_1 (T_1)^2 + \kappa_2 (T_2)^2 + 2 \kappa_3 (T_1) (T_2)) )}
#' where
#' \deqn{T_1 = T_1(x, \mu, \omega) = (x_1 - \mu_1(2\pi\omega_1))}
#' \deqn{T_2 = T_2(x, \mu, \omega) = (x_2 - \mu_1(2\pi\omega_2))}
#' the sum extends over all pairs of integers \eqn{\omega = (\omega_1, \omega_2)},
#' and is approximated by a sum over \eqn{(\omega_1, \omega_2)} in \eqn{\{-M, -M+1, ..., M-1, M \}^2} if \code{int.displ = } \eqn{M}.
#'
#' Note that above density is essentially the "wrapped" version of a bivariate normal density with mean
#' \deqn{\mu = (\mu_1, \mu_2)}
#' and dispersion matrix  \eqn{\Sigma = \Delta^{-1}}, with \eqn{\Delta} being a \eqn{2 \times 2}
#' symmetric, positive definite matrix
#' with diagonal entries \eqn{\kappa_1} and \eqn{\kappa_2} and the off-diagonal entries
#' \eqn{\kappa_3}.
#'
#'
#' @return \code{dwnorm2} gives the density  and \code{rwnorm2} generates random deviates.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # when x is a bivariate vector and parameters are all scalars,
#' # dwnorm2 returns single density
#' dwnorm2(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when x is a two column matrix and parameters are all scalars,
#' # dmvsin returns a vector of densities calculated at the rows of
#' # x with the same parameters
#' dwnorm2(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # if x is a bivariate vector and at least one of the parameters is
#' # a vector, all parameters are recycled to the same length, and
#' # dwnorm2 returns a vector of with ith element being the density
#' # evaluated at x with parameter values kappa1[i], kappa2[i],
#' # kappa3[i], mu1[i] and mu2[i]
#' dwnorm2(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # if x is a two column matrix and at least one of the parameters is
#' # a vector, rows of x and the parameters are recycled to the same
#' # length, and dwnorm2 returns a vector of with ith element being the
#' # density evaluated at ith row of x with parameter values kappa1[i],
#' # kappa2[i], # kappa3[i], mu1[i] and mu2[i]
#' dwnorm2(x, kappa1, kappa2, kappa3, mu1, mu2)
#'
#' # when parameters are all scalars, number of observations generated
#' # by rwnorm2 is n
#' rwnorm2(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])
#'
#' # when at least one of the parameters is a vector, all parameters are
#' # recycled to the same length, n is ignored, and the number of
#' # observations generated by rwnorm2 is the same as the length of the
#' # recycled vectors
#' rwnorm2(n, kappa1, kappa2, kappa3, mu1, mu2)
#'
#' @export

rwnorm2 <- function(n, kappa1=1, kappa2=1,
                    kappa3=0, mu1=0, mu2=0, ...)
{
  if(any(c(kappa1, kappa2) < 0))
    stop("kappa1 and kappa2 must be non-negative")
  if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)

  if(max(length(kappa1), length(kappa2), length(kappa3),
         length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    kappa1 <- expanded[[1]]; kappa2 <- expanded[[2]]; kappa3 <- expanded[[3]]
    mu1 <- expanded[[4]]; mu2 <- expanded[[5]]
    if(any(kappa1*kappa2 - kappa3*kappa3 < 0))
      stop("abs(kappa3) must be less than or equal to sqrt(kappa1*kappa2) in wnorm2")
    m <- length(kappa1)

    samp <- matrix(0, m, 2)

    for (j in 1:m) {
      if (kappa1[j] < 1e-10 & kappa2[j] < 1e-10) {
        samp[j, ] <- runif(2, 0, 2*pi)
      }

      else {
        mu_curr <-  c(mu1[j], mu2[j])
        sigma_curr <-  matrix(c(kappa2[j], -kappa3[j], -kappa3[j],
                                kappa1[j])/(kappa1[j]*kappa2[j] - kappa3[j]^2), 2)
        # samp <- t(sapply(1:m, function(j) rnorm2(1, mu_list[[j]], sigma_list[[j]])))

        samp[j, ] <-
          rmvnorm(1, mean=mu_curr, sigma=sigma_curr, ...)
      }
    }
  }

  else {
    if(kappa1*kappa2 - kappa3*kappa3 < 0)
      stop("abs(kappa3) must be less than or equal to sqrt(kappa1*kappa2) in wnorm2")

    if (kappa1*kappa2 < 1e-10) {
      samp <- matrix(runif(2*n, 0, 2*pi), ncol=2)
    }

    else {
      mu <- c(mu1, mu2)
      sigma <- matrix(c(kappa2, -kappa3, -kappa3, kappa1)/(kappa1*kappa2 - kappa3*kappa3), 2)
      samp <- rmvnorm(n, mean=mu, sigma=sigma, ...)
    }
  }
  prncp_reg(samp)
}

#' @rdname rwnorm2
#' @export

dwnorm2 <- function(x, kappa1=1, kappa2=1, kappa3=0, mu1=0,
                    mu2=0, int.displ, log=FALSE)
{
  if(missing(int.displ)) int.displ <- 3
  else if(int.displ >= 5) int.displ <- 5
  else if(int.displ <= 1) int.displ <- 1
  displ <- floor(int.displ)
  omega.2pi.all <- expand.grid(-displ:displ,-displ:displ) * (2*pi) # 2pi * integer displacements
  omega.2pi <- as.matrix(omega.2pi.all)

  if(any(c(kappa1, kappa2) < 0))
    stop("kappa1 and kappa2 must be non-negative")
  if(any(mu1 < 0 | mu1 >= 2*pi)) mu1 <- prncp_reg(mu1)
  if(any(mu2 < 0 | mu2 >= 2*pi)) mu2 <- prncp_reg(mu2)
  if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
     || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")


  if(max(length(kappa1), length(kappa2), length(kappa3), length(mu1), length(mu2)) > 1) {
    expanded <- expand_args(kappa1, kappa2, kappa3, mu1, mu2)
    kappa1 <- expanded[[1]]
    kappa2 <- expanded[[2]]
    kappa3 <- expanded[[3]]
    mu1 <- expanded[[4]]
    mu2 <- expanded[[5]]
  }

  par.mat <- rbind(kappa1, kappa2, kappa3, mu1, mu2)
  n_par <- ncol(par.mat)
  if (length(x) == 2) x <- matrix(x, nrow = 1)
  n_x <- nrow(x)

  if (all (kappa1 > 1e-10 | kappa2 > 1e-10)) {
    # regular wnorm2 density
    if (n_par == 1) {
      den <- c(dwnorm2_manyx_onepar(x, kappa1, kappa2, kappa3,
                                    mu1, mu2, omega.2pi))
    } else if (n_x == 1) {
      den <- c(dwnorm2_onex_manypar(c(x), kappa1, kappa2, kappa3,
                                    mu1, mu2, omega.2pi))
    } else {
      x_set <- 1:n_x
      par_set <- 1:n_par
      expndn_set <- expand_args(x_set, par_set)
      x_set <- expndn_set[[1]]
      par_set <- expndn_set[[2]]
      den <- c(dwnorm2_manyx_manypar(x[x_set, ], kappa1[par_set],
                                     kappa2[par_set], kappa3[par_set],
                                     mu1[par_set], mu2[par_set], omega.2pi))
    }
  }
  else {
    # some can be uniform
    x_set <- 1:n_x
    par_set <- 1:n_par
    expndn_set <- expand_args(x_set, par_set)
    x_set <- expndn_set[[1]]
    par_set <- expndn_set[[2]]
    x_long <- x[x_set, , drop=FALSE]
    kappa1_long <- kappa1[par_set]
    kappa2_long <- kappa2[par_set]
    kappa3_long <- kappa3[par_set]
    mu1_long <- mu1[par_set]
    mu2_long <- mu2[par_set]
    n_x_final <- nrow(x_long)
    den <- rep(0, n_x_final)
    which_unif <- which(kappa1_long < 1e-10 & kappa2_long < 1e-10)
    n_unif <- length(which_unif)


    # browser()
    den[which_unif] <- 1/(4*pi^2)
    if (n_unif < n_x_final)
      den[-which_unif] <- c(dwnorm2_manyx_manypar(x_long[-which_unif, , drop=FALSE], kappa1_long[-which_unif],
                                                 kappa2_long[-which_unif], kappa3_long[-which_unif],
                                                 mu1_long[-which_unif], mu2_long[-which_unif],
                                                 omega.2pi))


  }

  if (log)
    den <- log(den)

  den
}

#' The bivariate Wrapped Normal mixtures
#' @inheritParams rvmsinmix
#' @inheritParams rwnorm2
#' @param mu1,mu2 vectors of mean parameters.
#' @param kappa1,kappa2,kappa3 vectors of concentration parameters; \code{kappa1, kappa2 > 0, kappa3^2 < kappa1*kappa2} for each component.
#'
#' @details All the argument vectors \code{pmix, kappa1, kappa2, kappa3, mu1} and \code{mu2} must be of the same length,
#' with \eqn{j}-th element corresponding to the \eqn{j}-th component of the mixture distribution.
#' @details The bivariate wrapped normal mixture distribution with component size \code{K = \link{length}(pmix)} has density
#' \deqn{g(x) = \sum p[j] * f(x; \kappa_1[j], \kappa_2[j], \kappa_3[j], \mu_1[j], \mu_2[j])}
#' where the sum extends over \eqn{j}; \eqn{p[j]; \kappa_1[j], \kappa_2[j], \kappa_3[j]}; and \eqn{\mu_1[j], \mu_2[j]} respectively denote the mixing proportion,
#' the three concentration parameters and the two mean parameter for the \eqn{j}-th component, \eqn{j = 1, ..., K},
#' and \eqn{f(. ; \kappa_1, \kappa_2, \kappa_3, \mu_1, \mu_2)} denotes the density function of the wrapped normal distribution
#' with concentration parameters \eqn{\kappa_1, \kappa_2, \kappa_3} and  mean parameters \eqn{\mu_1, \mu_2}.
#' @return \code{dwnorm2mix} computes the density  and \code{rwnorm2mix} generates random deviates from the mixture density.
#'
#' @examples
#' kappa1 <- c(1, 2, 3)
#' kappa2 <- c(1, 6, 5)
#' kappa3 <- c(0, 1, 2)
#' mu1 <- c(1, 2, 5)
#' mu2 <- c(0, 1, 3)
#' pmix <- c(0.3, 0.4, 0.3)
#' x <- diag(2, 2)
#' n <- 10
#'
#' # mixture densities calculated at the rows of x
#' dwnorm2mix(x, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' # number of observations generated from the mixture distribution is n
#' rwnorm2mix(n, kappa1, kappa2, kappa3, mu1, mu2, pmix)
#'
#' @export

rwnorm2mix <- function(n, kappa1, kappa2, kappa3,
                       mu1, mu2, pmix, ...)
{
  allpar <- list(kappa1=kappa1, kappa2=kappa2, kappa3=kappa3,
                 mu1=mu1, mu2=mu2, pmix=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len))
    stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(c(allpar$kappa1, allpar$kappa2) <= 0))
    stop("kappa1 and kappa2 must be positive in wnorm2")
  if(any(allpar$kappa1*allpar$kappa2 - allpar$kappa3^2 <= 1e-10))
    stop("abs(kappa3) must be less than sqrt(kappa1*kappa2) in wnorm2")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)

  out <- matrix(0, n, 2)
  ncomp <- allpar_len[1] # number of components
  comp_ind <- cID(tcrossprod(rep(1, n), allpar$pmix), ncomp, runif(n))
  # n samples from multinom(ncomp, pmix)
  for(j in seq_len(ncomp)) {
    obs_ind_j <- which(comp_ind == j)
    n_j <- length(obs_ind_j)
    if(n_j > 0) {
      out[obs_ind_j, ] <- rwnorm2(n_j, kappa1[j], kappa2[j],
                                  kappa3[j], mu1[j], mu2[j], ...)
    }
  }

  out
}


#' @rdname rwnorm2mix
#' @export
dwnorm2mix <- function(x, kappa1, kappa2, kappa3,
                       mu1, mu2, pmix, int.displ, log=FALSE)
{
  if(missing(int.displ)) int.displ <- 3
  else if(int.displ >= 5) int.displ <- 5
  else if(int.displ <= 1) int.displ <- 1
  displ <- floor(int.displ)
  omega.2pi.all <- expand.grid(-displ:displ,-displ:displ) * (2*pi) # 2pi * integer displacements
  omega.2pi <- as.matrix(omega.2pi.all)

  allpar <- list("kappa1"=kappa1, "kappa2"=kappa2, "kappa3"=kappa3,
                 "mu1"=mu1, "mu2"=mu2, "pmix"=pmix)

  allpar_len <- listLen(allpar)
  if(min(allpar_len) != max(allpar_len)) stop("component size mismatch: number of components of the input parameter vectors differ")

  if(any(allpar$pmix < 0)) stop("\'pmix\' must be non-negative")
  sum_pmix <- sum(allpar$pmix)
  if(signif(sum_pmix, 5) != 1) {
    if(sum_pmix <= 0) stop("\'pmix\' must have at least one positive element")
    allpar$pmix <- allpar$pmix/sum_pmix
    warning("\'pmix\' is rescaled to add up to 1")
  }

  if(any(allpar$kappa1*allpar$kappa2 - allpar$kappa3*allpar$kappa3 <= 1e-10))
    stop("abs(kappa3) must be less than sqrt(kappa1*kappa2) in wnorm2")

  if(any(c(allpar$kappa1, allpar$kappa2) <= 0))
    stop("kappa1 and kappa2 must be positive in wnorm2")
  if(any(allpar$mu1 < 0 | allpar$mu1 >= 2*pi)) allpar$mu1 <- prncp_reg(allpar$mu1)
  if(any(allpar$mu2 < 0 | allpar$mu2 >= 2*pi)) allpar$mu2 <- prncp_reg(allpar$mu2)
  if((length(dim(x)) < 2 && length(x) != 2) || (length(dim(x)) == 2 && tail(dim(x), 1) != 2)
     || (length(dim(x)) > 2)) stop("x must either be a bivariate vector or a two-column matrix")


  ncomp <- length(kappa1)

  if (length(x) == 2) x <- matrix(x, nrow=1)

  allcompden <- vapply(1:ncomp,
                       function(j) dwnorm2(x, kappa1[j], kappa2[j],
                                           kappa3[j], mu1[j], mu2[j],
                                           int.displ, FALSE),
                       rep(0, nrow(x)))

  mixden <- c(allcompden %*% pmix)

  if (log) {
    log(mixden)
  } else {
    mixden
  }
}


#' Fitting bivariate wrapped normal model mixtures using MCMC
#' @inheritParams fit_vmsinmix
#'
#' @details
#' Wrapper for \link{fit_angmix} with \code{model = "wnorm2"}.
#'
#'
#' @examples
#' # illustration only - more iterations needed for convergence
#' fit.wnorm2.10 <- fit_wnorm2mix(tim8, ncomp = 3, n.iter =  10,
#'                              n.chains = 1)
#' fit.wnorm2.10
#'
#' @export

fit_wnorm2mix <- function(...)
{
  fit_angmix(model="wnorm2", ...)
}


wnorm2_var_cor_singlepar <- function(kappa1, kappa2, kappa3) {
  den <- kappa1*kappa2 - kappa3^2
  sig1_sq <- kappa2/den
  sig2_sq <- kappa1/den
  sig12 <- -kappa3/den

  rho_fl <- sinh(2*sig12) /
    sqrt(sinh(2*sig1_sq)* sinh(2*sig2_sq))

  rho_js <- sinh(sig12) /
    sqrt(sinh(sig1_sq)* sinh(sig2_sq))

  list(var1 = 1-exp(-0.5*sig1_sq),
       var2 = 1-exp(-0.5*sig2_sq),
       rho_fl = rho_fl,
       rho_js = rho_js
  )
}

expand_args <- function(...){
  ell <- list(...)
  max_length <- max(vapply(ell, length, 0))
  lapply(ell, rep, length.out = max_length)
}

ldgamanum <- function(x, loc, scale) {
  (loc-1)*log(x) - x/scale
}


# mean <- function(x) {
#   sum(x)/length(x)
# }


har_mean <- function(x) {
  if(sum(x == 0) > 0) stop("zero value in harmonic mean")
  1/mean(1/x)
}


sum_sq <- function(x) sum(x^2)


prncp_reg <- function(x) x %% (2*pi)
#makes and angle in [0, 2*pi]


prncp_reg.minuspi.pi <- function(x) {
  y <- (x + pi) %% (2*pi)
  y_neg <- which(y < 0)
  y[y_neg] <- 2*pi + y[y_neg]
  y - pi
}  #makes a single angle in [-pi, pi]


atan3 <- function(x) prncp_reg(atan(x[2]/x[1]))


sph2cart <- function(x)
  c(cos(x[1])*sin(x[2]), sin(x[1])*sin(x[2]), cos(x[2]))
#calculates unit vectors from pair of angles


listLen <- function(l)
  vapply(1:length(l), function(i) length(l[[i]]), 0)


rm_NA_rad <- function(data, rad = TRUE) {
  if(length(dim(data)) == 2) {
    phi.no.na <- data[,1]
    psi.no.na <- data[,2]
    na.phi.id <- NULL
    na.psi.id <- NULL
    is.na.phi <- is.na(data[,1])
    is.na.psi <- is.na(data[,2])
    if(sum(is.na.phi) > 0)
      na.phi.id <- which(is.na.phi)
    if(sum(is.na.psi) > 0)
      na.psi.id <- which(is.na.psi)
    na.id <- union(na.phi.id, na.psi.id)
    if(length(na.id) > 0){
      phi.no.na <- data[,1][-na.id]
      psi.no.na <- data[,2][-na.id]
    }
    if(rad)    res <- prncp_reg(cbind(phi.no.na, psi.no.na))
    else res <- prncp_reg(cbind(phi.no.na, psi.no.na) * pi / 180)
    colnames(res) <- colnames(data)
  } else {
    data.no.na <- data
    na.id <- NULL
    is.na.data <- is.na(data)
    if(sum(is.na.data) > 0){
      na.id <- which(is.na.data)
      data.no.na <- data[-na.id]
    }
    if(rad) res <- prncp_reg(data.no.na)
    else res <- prncp_reg(data.no.na * pi / 180)
  }
  res
} #removes NA and converts into radians


# rdirichlet <- function (n, alpha)  # random generation from dirichlet
# {
#   len <- length(alpha)
#   x <- matrix(rgamma(len * n, alpha), ncol = len, byrow = TRUE)
#   tot <- x %*% rep(1, len)
#   x/as.vector(tot)
# }


# rnorm2 <- function (n = 1, mu, Sigma)  # random generation from biv normal
# {
#   p <- 2L
#   eS <- eigen(Sigma, symmetric = TRUE)
#   ev <- eS$values
#   X <- matrix(rnorm(p * n), n)
#   X <- drop(mu) + eS$vectors %*% diag(sqrt(pmax(ev, 0)), p) %*% t(X)
#   if (n == 1) drop(X)
#   else t(X)
# }


list_by_row <- function(mat, row_index)  # create a list with elements being rows of a matrix
{
  mat.list <- lapply(1:nrow(mat), function(j) mat[j, ])
  names(mat.list) <- rownames(mat)
  mat.list
}

addtolist <- function(list_in, ...)  # add element to a list
{
  ell <- list(...)
  c(list_in, ell)
}


press_enter <- function()  # waits for the user to press [enter]
{
  cat("Press [enter] to continue")
  line <- readline()
}


kappas2sigmas_wnorm2 <- function(kappa1, kappa2, kappa3) {
  den <- kappa1*kappa2 - kappa3^2
  sigma1 <- kappa2/den
  sigma2 <- kappa1/den
  rho <- -kappa3/sqrt(kappa1*kappa2)
  c(sigma11 = sigma1, sigma22 = sigma2, rho = rho)
}

sigmas2kappas_wnorm2 <- function(sigma11, sigma22, rho) {
  den <- sigma11*sigma22*(1-rho^2)
  kappa1 <- sigma22/den
  kappa2 <- sigma11/den
  kappa3 <- -rho/((1-rho^2)*sqrt(sigma11*sigma22))
  c(kappa1 = kappa1, kappa2 = kappa2, kappa3 = kappa3)
}

which.max_entry1 <- function(x) {
  which.max(x)[1]
}


signif_or_round <- function(x, ...) {
  for(j in length(x)) {
    if (abs(x[j]) > 1) return(round(x[j], ...))
    else return(signif(x[j], ...))
  }
}


# print est (ci_lower, ci_upper) for each element
est_ci <- function(est, lower, upper, digits = 2)
{
  out_mat <- est
  for(j in 1:length(est)) {
    out_mat[j] <- paste0(format(signif_or_round(est[j], digits), nsmall = digits),
                         " (",
                         format(signif_or_round(lower[j], digits), nsmall = digits),
                         ", ",
                         format(signif_or_round(upper[j], digits), nsmall = digits),
                         ")")
  }

  as.data.frame(out_mat)
}

# Bambi Community Code of Conduct

Be kind to others. Do not insult or put down others.
Behave professionally. Remember that harassment and sexist, racist,
or exclusionary jokes are not appropriate.

All communication should be appropriate for a professional audience
including people of many different backgrounds. Sexual language and
imagery are not appropriate.

Bambi is dedicated to providing a harassment-free community for everyone,
regardless of gender, sexual orientation, gender identity, and
expression, disability, physical appearance, body size, race,
or religion. We do not tolerate harassment of community members
in any form.

Thank you for helping make this a welcoming, friendly community for all.

# Guidelines for Contributing

As a scientific community-driven software project, Bambi welcomes contributions from interested individuals or groups. These guidelines are provided to give potential contributors information to make their contribution compliant with the conventions of the Bambi project, and maximize the probability of such contributions to be merged as quickly and efficiently as possible.

There are 4 main ways of contributing to the Bambi project (in descending order of difficulty or scope):

- Adding new or improved functionality to the existing codebase
- Fixing outstanding issues (bugs) with the existing codebase. They range from low-level software bugs to higher-level design problems
- Contributing or improving the documentation (`docs`) or examples (`bambi/examples`)
- Submitting issues related to bugs or desired enhancements

## Opening issues

We appreciate being notified of problems with the existing Bambi code. We prefer that issues be filed the on [Github Issue Tracker](https://github.com/bambinos/bambi/issues), rather than on social media or by direct email to the developers.

Please verify that your issue is not being currently addressed by other issues or pull requests by using the GitHub search tool to look for key words in the project issue tracker.

## Contributing code via pull requests

While issue reporting is valuable, we strongly encourage users who are inclined to do so to submit patches for new or existing issues via pull
requests. This is particularly the case for simple fixes, such as typos or tweaks to documentation, which do not require a heavy investment
of time and attention.

Contributors are also encouraged to contribute new code to enhance Bambi's functionality, also via pull requests.

The preferred workflow for contributing to Bambi is to fork
the [GitHub repository](https://github.com/bambinos/bambi/), clone it to your local machine, and develop on a feature branch.

For more instructions see the
[Pull request checklist](#pull-request-checklist)

### Code Formatting

For code generally follow the
[TensorFlow's style guide](https://www.tensorflow.org/versions/master/how_tos/style_guide.html)
or the [Google style guide](https://github.com/google/styleguide/blob/gh-pages/pyguide.md)
Both more or less follows PEP 8.

### Docstring formatting

Docstrings should follow the
[numpy docstring guide](https://numpydoc.readthedocs.io/en/latest/format.html)
Please reasonably document any additions or changes to the codebase, when in doubt, add a docstring.

## Steps

1. Fork the [project repository](https://github.com/bambinos/bambi/) by clicking on the 'Fork' button near the top right of the main repository page. This creates a copy of the code under your GitHub user account.

2. Clone your fork of the Bambi repo from your GitHub account to your local disk, and add the base repository as a remote:

   ```bash
   git clone git@github.com:<your GitHub handle>/bambi.git
   cd bambi
   git remote add upstream git@github.com:bambinos/bambi.git
   ```

3. Create a `feature` branch to hold your development changes:

   ```bash
   git checkout -b my-feature
   ```

   Always use a `feature` branch. It's good practice to never routinely work on the `master` branch of any repository.

4. Project requirements are in `requirements.txt`, and libraries used for development are in `requirements-dev.txt`. To set up a development environment, you may run:

   ```bash
   pip install -r requirements.txt
   pip install -r requirements-dev.txt
   ```

5. Develop the feature on your feature branch. Add changed files using `git add` and then `git commit` files:

   ```bash
   git add modified_files
   git commit
   ```

   to record your changes locally.
   After committing, it is a good idea to sync with the base repository in case there have been any changes:

   ```bash
   git fetch upstream
   git rebase upstream/master
   ```

   Then push the changes to your GitHub account with:

   ```bash
   git push -u origin my-feature
   ```

6. Go to the GitHub web page of your fork of the Bambi repo. Click the 'Pull request' button to send your changes to the project's maintainers for review. This will send an email to the committers.

## Pull request checklist

We recommend that your contribution complies with the following guidelines before you submit a pull request:

- If your pull request addresses an issue, please use the pull request title to describe the issue and mention the issue number in the pull request description. This will make sure a link back to the original issue is created.

- All public methods must have informative docstrings with sample usage when appropriate.

- To indicate a work in progress please mark the PR as `draft`. Drafts may be useful to (1) indicate you are working on something to avoid duplicated work, (2) request broad review of functionality or API, or (3) seek collaborators.

- All other tests pass when everything is rebuilt from scratch.

- When adding additional functionality, provide at least one example script or Jupyter Notebook in the `bambi/examples/` folder. Have a look at other examples for reference. Examples should demonstrate why the new functionality is useful in practice and, if possible, compare it to other methods available in Bambi.

- Added tests follow the [pytest fixture pattern](https://docs.pytest.org/en/latest/fixture.html#fixture)

- Documentation and high-coverage tests are necessary for enhancements to be accepted.

- Documentation follows Numpy style guide

- Run any of the pre-existing examples in `examples` that contain analyses that would be affected by your changes to ensure that nothing breaks. This is a useful opportunity to not only check your work for bugs that might not be revealed by unit test, but also to show how your contribution improves Bambi for end users.

- Code coverage **cannot** decrease. Coverage can be checked with **pytest-cov** package:

  ```bash
  pip install pytest pytest-cov coverage
  pytest --cov=bambi --cov-report=html bambi/tests/
  ```

- Your code passes pylint

  ```bash
  pip install pylint
  pylint bambi/
  ```

**This guide was derived from the [arviz guide to contributing](https://github.com/arviz-devs/arviz/blob/master/CONTRIBUTING.md)**

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{as.mcmc.list.angmcmc}
\alias{as.mcmc.list.angmcmc}
\title{Create an mcmc.list object from an angmcmc object}
\usage{
\method{as.mcmc.list}{angmcmc}(x, ...)
}
\arguments{
\item{x}{angmcmc object}

\item{...}{unused}
}
\description{
Create an mcmc.list object from an angmcmc object
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)

# now convert it to mcmc.list
library(coda)
fit.vmsin.20.mcmc <- as.mcmc.list(fit.vmsin.20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{DIC}
\alias{DIC}
\title{Deviance Information Criterion (DIC) for angmcmc objects}
\usage{
DIC(object, form = 2, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{form}{form of DIC to use. Available choices are 1 and 2 (default). See details.}

\item{...}{additional model specific arguments to be passed to \code{DIC}. For example, \code{int.displ}
specifies integer dispacement in wnorm and wnorm2 models. See \link{fit_wnormmix} and
\link{fit_wnorm2mix} for more details.}
}
\value{
Computes the DIC for a given angmcmc object
}
\description{
Deviance Information Criterion (DIC) for angmcmc objects
}
\details{
Given a deviance function \eqn{D(\theta) = -2 log(p(y|\theta))}, and an estimate
\eqn{\theta* = (\sum \theta_i) / N} of the posterior mean
\eqn{E(\theta|y)}, where \eqn{y} denote the data, \eqn{\theta} are the unknown
parameters of the model, \eqn{\theta_1, ..., \theta_N} are MCMC samples from the posterior
distribution of \eqn{\theta} given \eqn{y} and \eqn{p(y|\theta)} is the likelihood function,
the (form 1 of) Deviance Infomation Criterion (DIC) is defined as
\deqn{DIC = 2 ( (\sum_{s=1}^N D(\theta_s)) / N - D(\theta*) )}
The second form for DIC is given by
\deqn{DIC = D(\theta*) - 4 \hat{var} \log p(y|\theta_s)}
where for \eqn{i = 1, ..., n}, \eqn{\hat{var} \log p(y|\theta)} denotes the estimated variance
of the log likelihood based on the realizations \eqn{\theta_1, ..., \theta_N}.

Like AIC and BIC, DIC is an asymptotic approximation for large samples, and
is only valid when the posterior distribution is approximately normal.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
DIC(fit.vmsin.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{logLik.angmcmc}
\alias{logLik.angmcmc}
\title{Extract Log-Likelihood from angmcmc objects}
\usage{
\method{logLik}{angmcmc}(object, method = 1, fn, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{method}{interger specifying method of estimating the log likelihood. Must be 1 or 2. Defaults to 1. See details.}

\item{fn}{function to evaluate on the iteration-wise log-likelihood values obtained during MCMC run if \code{method = 1}; or,
if \code{method = 2}, function to evaluate on the MCMC samples for parameter estimation (passed to \link{pointest}).
Defaults to \code{max} if \code{method = 1} and \code{mean} if \code{method = 2}.}

\item{...}{additional arguments to be passed to the function.}
}
\value{
Returns an object of class \link{logLik}. This is a number (the estimated log likelihood) with attributes "df"
(degrees of freedom) and "nobs" (number of observations).
}
\description{
Extract Log-Likelihood from angmcmc objects
}
\details{
There are two ways to estimate the log likelihood from the model. If \code{method = 1},
then log likelihood is estimated by applying \code{fn} (defaults to max, if method = 1)
direclty on the log likelihood values from observed during the MCMC run.
On the other hand, if \code{method == 2}, then  parameter estimates
are first computed using \code{pointest} with \code{fn}
(defaults to "MODE", if \code{method == 2}) applied on the MCMC samples,
and then then log likelihood is evaluated at the parameter estimates.


The degrees of the likelihood function is the total number of free parameters estimated in the mixture models,
which is equal to \eqn{6K - 1} for bivariate models (vmsin, vmcos and wnorm2), or \eqn{3K - 1} for univariate
models (vm and wnorm), where \eqn{K} denotes the number of components in the mixture model.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
logLik(fit.vmsin.20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_mixmodel.R
\name{fit_angmix}
\alias{fit_angmix}
\title{Fitting Bivariate and univariate angular mixture models}
\usage{
fit_angmix(
  model = "vmsin",
  data,
  ncomp,
  cov.restrict = "NONE",
  unimodal.component = FALSE,
  start_par = NULL,
  rand_start = rep(FALSE, n.chains),
  method = "hmc",
  perm_sampling = FALSE,
  n.chains = 3,
  chains_parallel = TRUE,
  return_llik_contri = FALSE,
  int.displ = 3,
  epsilon = 0.1,
  L = 10,
  epsilon.random = TRUE,
  L.random = FALSE,
  burnin.prop = 0.5,
  tune.prop = 1,
  thin = 1,
  propscale = 0.05,
  n.iter = 500,
  pmix.alpha = NULL,
  norm.var = 1000,
  autotune = TRUE,
  show.progress = TRUE,
  accpt.prob.upper,
  accpt.prob.lower,
  epsilon.incr = 0.05,
  L.incr = 0.075,
  tune.incr = 0.05,
  tune_ave_size = 100,
  kappa_upper = 150,
  kappa_lower = 1e-04,
  return_tune_param = FALSE,
  qrnd = NULL,
  n_qrnd = NULL,
  ...
)
}
\arguments{
\item{model}{angular model whose mixtures are to be fitted. Available choices are \code{"vmsin", "vmcos"} and \code{"wnorm2"} for
bivariate data, and \code{"vm"} and \code{"wnorm"} for univariate data.}

\item{data}{data matrix (if bivarate, in which case it must have two columns) or vector. If outside, the values
are transformed into the scale \eqn{[0, 2\pi)}. *Note:* BAMBI cannot handle missing data. Missing values must
either be removed or properly imputed.}

\item{ncomp}{number of components in the mixture model. Must be a positive integer. vector values are not allowed.
If \code{comp == 1}, a single component model is fitted.}

\item{cov.restrict}{Should there be any restriction on the covariance parameter for a bivariate model. Available choices are
\code{"POSITIVE", "NEGATIVE", "ZERO"} and "NONE". Note that \code{"ZERO"} fits a mixture with product components. Defaults to
\code{"NONE"}.}

\item{unimodal.component}{logical. Should each component in the mixture model be unimodal? Only used if \code{model} is either \code{"vmsin"}
or \code{"vmcos"}. Defaults to FALSE.}

\item{start_par}{list with elements \code{pmix} (ignored if \code{comp == 1}), together with \code{kappa1, kappa2, mu1} and \code{mu2},
for bivariate models, and \code{kappa} and \code{mu} for univariate models,
all being vectors of length same as \code{ncomp}.
These provides the starting values for the Markov chain; with \eqn{j}-th component of each vector corresponding to the \eqn{j}-th
component of the mixture distribution. If missing, the data is first clustered into \code{ncomp} groups either via k-means (after
projecting onto a unit sphere), or randomly, depending on \code{rand_start},  and then moment estimators for components are used as
the starting points. Note that a very wrong starting point can potentially lead the chain to get stuck at a wrong solution for thousands
of iterations. As such, we recommend using the default option, which is k-means followed by moment estimation.}

\item{rand_start}{logical. Should a random starting clustering be used? Must be either a scalar, or a vector of length \code{ncomp},
one for each chain. Ignored if \code{start_par} is supplied. See \code{start_par} for more details. Defaults to \code{FALSE}.}

\item{method}{MCMC strategy to be used for the model paramters:  \code{"hmc"} or \code{"rwmh"}.}

\item{perm_sampling}{logical. Should the permutation sampling algorithm of Fruhwirth-Schnatter (2001) be used?
If TRUE, at every iteration after burnin, once model parameters and mixing proportions are sampled,
a random permutation of 1, ..., ncomp is considered, and components are relabelled according
to this random permutation. This forced random label switchings may imporve the mixing rate of the chage. However, (automated) tuning
is very difficult with such a scheme, as there is no simple way of keeping track of the "original" component labels. This creates problem
with computing standard deviations of the generated model parameters, thus making the
scaling step used in tuning for \code{epsilon} or \code{paramscale} problematic as well. As such, \code{perm_sampling} is always turned
off during burn-in (even if \code{autotune = FALSE}), and turned on thereafter, if \code{TRUE}.
Defaults to and is set to \code{FALSE}.}

\item{n.chains}{number of chains to run. Must be a positive integer.}

\item{chains_parallel}{logical. Should the chains be run in parallel? Defaluts to TRUE, and ignored if \code{n.chains} = 1.
Note that parallelization is implemented via \link[future.apply]{future_lapply} from package \code{future.apply} which
uses futures for this purpose, and thus provides a convenient way of parallelization across various OSs and computing environments.
However, a proper \link[future]{plan} must be set for the parallization before running the chain. Otherwise the chains will run sequentially.}

\item{return_llik_contri}{logical. Should the log likelihood contribution of each data point for each MCMC iteration in each chain be returned? This makes
computation of \link{waic.angmcmc} and \link{loo.angmcmc} much faster. *Warning*: Depending on the length of data and \code{n.iter}, this can be
very memory intensive. We suggest setting \code{return_llik_contri = TRUE} only if \link{waic.angmcmc} and \link{loo.angmcmc} are aimed for. Defaults to
\code{FALSE}.}

\item{int.displ}{absolute integer displacement for each coordinate for \code{wnorm} and \code{wnorm2} models (ignored otherwise). Default is 3.
Allowed minimum and maximum are 1 and 5 respectively.}

\item{epsilon, L}{tuning parameters for HMC; ignored if \code{method = "rwmh"}. \code{epsilon} (step-size) is a single number,
or a vector of size \code{2*ncomp} for univariate models and \code{5*ncomp} for bivariate models. Note that the "mass matrix"
in HMC is assumed to be identity. As such, \code{epsilon}'s corresponding to different model parameters need to be in proper scale for
optimal acceptance rate. Can be autotuned during burnin. See \code{autotune}.
\code{L} (leapfrog steps) is a positive integer or a vector of positive integers of length \code{n.chains}.
If multiple chains are used, we suggest same \code{L} values acorss different chains to make the chains as homogenous as possible.}

\item{epsilon.random}{logical. Should \code{epsilon*delta}, where \code{delta} is a random
number between \code{(1-epsilon.incr, 1+epsilon.incr)} be used instead of \code{epsilon} at each iteration?
Ignored if \code{method = "rwmh"}.}

\item{L.random}{logical. Should a random integer between \code{L.orig/exp(L.incr)} and \code{L.orig*exp(L.incr)}be used instead as \code{L}
at each iteration? Ignored if \code{method = "rwmh"}. Defaults to \code{TRUE}.}

\item{burnin.prop}{proportion of iterations to used for burnin. Must be a be a number in [0, 1].
Default is 0.5.}

\item{tune.prop}{proportion of *\code{burnin}* used to tune the parameters (\code{epsilon} in HMC and
\code{propscale} in RWMH). Must be a number between 0 and 1; defaults to 1.  Ignored if \code{autotune == FALSE}.}

\item{thin}{thining size to be used. Must be a positive integer. If \code{thin = } n, then every nth iteration is reatained
in the final MCMC sample.}

\item{propscale}{tuning parameters for RWMH; a vector of size 5 (for bivariate models) or 2 (for univariate models) representing
the variances for the proposal normal densities
for the model parameters. Ignored if \code{method = "hmc"}. Can be autotuned during burnin. See \code{autotune}.}

\item{n.iter}{number of iterations for the Markov Chain.}

\item{pmix.alpha}{concentration parameter(s) for the Dirichlet prior for \code{pmix}. Must either be a positive real number, or a vector
with positive entries and of length \code{ncomp}. The default is \eqn{(r+r(r+1)/2)/2+3}, where \eqn{r} is 1 or 2 according as whether
the model is univariate or bivariate. Note that it is recommended to use larger \code{alpha} values to ensure the a good posterior behavior,
especially when \link{fit_incremental_angmix} is used for model selection, which handles overfitting in "let two component-specific parameters be
size, and then penalizes for model complexity. See Fruhwirth-Schnatter (2011) for more details on this.}

\item{norm.var}{variance (hyper-) parameters in the normal prior for \code{log(kappa), log(kappa1), log(kappa2)} and \code{kappa3}. (Prior mean is zero).
Can be a vector. Default is 1000 that makes the prior non-informative.}

\item{autotune}{logical. Should the Markov chain auto-tune the parameter \code{epsilon} (in HMC) or
\code{propscale} (in RWMH) during burn-in?  Set to \code{TRUE} by default. An adaptive tuning strategy is implemented.
Here, at every 10th iteration during in burn-in, the acceptance ratio in the last \code{tune_ave_size}
iterations is calculated. Then the tuning parameter is decreased  (increased) by a factor of
\code{1-tune.incr} (\code{1+tune.incr}) if the calculated acceptance rate
falls below (above) \code{accpt.prob.lower} (\code{accpt.prob.upper}). In addditon, when \code{iter} is a multiple of
\code{tune_ave_size}, \code{epsilon} for each model parameter is rescaled via the standard deviation of
the corresponding parameter over the past \code{tune_ave_size} iterations.}

\item{show.progress}{logical. Should a progress bar be displayed?}

\item{accpt.prob.lower, accpt.prob.upper}{lower and upper limits of acceptance ratio to be maintained while tuning
during burn-in. Must be numbers between 0 and 1, which \code{accpt.prob.lower < accpt.prob.upper}. See \code{autotune}. Default to (0.6, 0,9) for HMC and  (0.3, 0.5) for RWMH.
Ignored if \code{autotune = FALSE}.}

\item{epsilon.incr}{amount of randomness incorporated in \code{epsilon} if \code{epsilon.random = TRUE}.}

\item{L.incr}{amount of randomness incorporated in L if \code{L.random = TRUE}.}

\item{tune.incr}{how much should the tuning parameter be increased or decreased at each step while tuning during burn-in?
Must be a number between 0 and 1. See \code{autotune}. Defaults to 0.05. Ignored if \code{autotune = FALSE}.}

\item{tune_ave_size}{number previous iterations used to compute the acceptance rate while tuning in burn-in. Must be a positive
integer. Defaults to 100.}

\item{kappa_upper, kappa_lower}{upper and lower bounds for the concentration and (absolute) association parameters. Must be a positive integers. Defaults to 150 and 1e-4,
and parameter with value above or below these limits rarely make sense in practice.
Warning: values much larger or smaller than the default are not recommended as they can cause numerical instability.}

\item{return_tune_param}{logical. Should the values of the tuning parameters used at each iteration in each chain be returned? Defaults to \code{FALSE}.}

\item{qrnd, n_qrnd}{Used only if \code{method="vmcos"}. See \link{dvmcos} for details.}

\item{...}{Unused.}
}
\description{
Fitting Bivariate and univariate angular mixture models
}
\note{
Sampling is done in log scale for the concentration parameters (kappa, kappa1 and kappa2).


Parallelization is done by default when more than one chain is used,
but the chains can be run sequentially as well by setting
\code{chains_parallel = FALSE}. To retain reproducibility while running
multiple chains in parallel, the same RNG state is passed at the
beginning of each chain. This is done by specifying \code{future.seed = TRUE}
in \code{future.apply::future_lapply} call. Then at the beginning of the i-th
chain, before drawing any parameters, i-many Uniform(0, 1) random numbers are
generated using \code{runif(i)} (and then thrown away). This ensures that the
RNG states across chains prior to random generation of the parameters are
different, and hence, no two chains can become identical, even if they have
the same starting and tuning parameters. This, however creates a difference
between a \code{fit_angmix} call with multiple chains which is run sequentially
by setting \code{chains_parallel = FALSE}, and another which is run sequentially
because of a sequential \code{plan()} (or no \code{plan()}), with
\code{chains_parallel = TRUE}. In the former, different RNG states are passed at
the initiation of each chain.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_angmix("vmsin", tim8,
  ncomp = 3, n.iter = 20,
  n.chains = 1
)
fit.vmsin.20


# Parallelization is implemented via future_lapply from the
# package future.apply. To parallelize, first provide a parallel
# plan(); otherwise the chains will run sequentially.
# Note that not all plan() might work on every OS, as they execute
# functions defined internally in fit_mixmodel. We suggest
# plan(multisession) which works on every OS.
\donttest{
library(future)
library(parallel)
# plan(multisession, gc = TRUE) # parallelize chains

set.seed(1)
MC.fit <- fit_angmix("vmsin", tim8,
  ncomp = 3, n.iter = 5000,
  n.chains = 3
)


pointest(MC.fit)

MC.fix <- fix_label(MC.fit)

contour(MC.fit)
contour(MC.fix)
lpdtrace(MC.fit)
}

}
\references{
Fruhwirth-Schnatter, S. (2011). Label switching under model uncertainty. Mixtures: Estimation and Application, 213-239.

Fruhwirth-Schnatter, S. (2001). Markov chain Monte Carlo estimation of classical and dynamic switching and mixture models. Journal of the American Statistical Association, 96(453), 194-209.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrap_angles.R
\name{zero_to_2pi}
\alias{zero_to_2pi}
\alias{minuspi_to_pi}
\title{Wrap angles into \code{[-pi, pi]} or \code{[0, 2*pi]}}
\usage{
zero_to_2pi(x)

minuspi_to_pi(x)
}
\arguments{
\item{x}{numeric vector or matrix or data.frame.}
}
\description{
Wrap angles into \code{[-pi, pi]} or \code{[0, 2*pi]}
}
\details{
\code{minuspi_to_pi} wraps \code{x} into \code{[-pi, pi]},
while \code{zero_to_pi} wraps \code{x} into \code{[0, 2*pi]}.
}
\examples{
dat <- matrix(runif(100, -pi, pi), ncol=2)
dat1 <- zero_to_2pi(dat)
dat2 <- minuspi_to_pi(dat1)
all.equal(dat, dat2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{select_chains}
\alias{select_chains}
\title{Select chains from angmcmc objects}
\usage{
select_chains(object, chain.no, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{chain.no}{labels of chains to be retained in the final sample. If missing,
all chains are used.}

\item{...}{unused}
}
\value{
Returns another angmcmc object with only selected chains passed through \code{chain.no}
}
\description{
Select chains from angmcmc objects
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             L = c(10, 12), chains_parallel = FALSE,
                             n.chains = 2)
fit.vmsin.20
fit.vmsin.20.1 <- select_chains(fit.vmsin.20, 1)
fit.vmsin.20.1

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{d_fitted}
\alias{d_fitted}
\alias{r_fitted}
\title{Density and random deviates from an angmcmc object}
\usage{
d_fitted(x, object, type = "point-est", fn = mean, log = FALSE, chain.no, ...)

r_fitted(n = 1, object, type = "point-est", fn = mean, chain.no, ...)
}
\arguments{
\item{x}{vector, if univariate or a two column matrix, if bivariate, with each row a 2-D vector, (can
also be a data frame of similar dimensions) of points where the
densities are to be computed.}

\item{object}{angular MCMC object. The dimension of the model must match with \code{x}.}

\item{type}{Method of estimating density/generating random deviates. Possible choices are
\code{"post-pred"} and \code{"point-est"}. See details. Defaults to \code{"point-est"}.}

\item{fn}{function, or a single character string specifying its name, to evaluate on MCMC samples to estimate
parameters.  Defaults to \code{mean}, which computes the estimated posterior mean.
Note that if \code{fn = "MODE"} (warning: not \code{"mode"}) or \code{fn = "MAP"}, then the maximum aposteriori estimate (MAP) is
calculated.}

\item{log}{logical. Should the log density be returned instead?}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{...}{additional arguments to be passed to the function.}

\item{n}{number of observations to be generated.}
}
\value{
\code{d_fitted} gives a vector the densities computed at the given points  and \code{r_fitted}
creates a vector (if univariate) or a matrix (if bivariate) with each row being a 2-D point, of random deviates.
}
\description{
Density and random deviates from an angmcmc object
}
\details{
If \code{type = 'point-est'}, density is evaluated/random samples are generated at a point estimate of
the parameter values.  To estimate the mixture density, first the parameter vector \eqn{\eta} is estimated
by applying \code{fn} on the MCMC samples (using the function \link{pointest}), yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}.
Then the mixture density \eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by
\eqn{f(x|\hat{\eta})}. The random deviates are generated from the estimated mixture density \eqn{f(x|\hat{\eta})}.

If \code{type == 'post-pred'}, posterior predictive samples and densities are returned. That
is, the average density \eqn{S^{-1} \sum_{s = 1}^S f(x | \eta_s)} is returned in \code{d_fitted},
where \eqn{\eta_1, \dots, \eta_S} is the set posterior MCMC samples obtained from \code{object}. In
\code{r_fitted}, first a random sub-sample \eqn{\eta_{(1)}, \dots, \eta_{(n)}} of size \code{n} from the
set of posterior samples \eqn{\eta_1, \dots, \eta_S} is drawn (with replacement if \code{n} > S). Then
the i-th posterior predictive data point is generated from the mixture density
\eqn{f(x|\eta_{(i)})} for i = 1,..., n.
}
\examples{
set.seed(1)
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
d_fitted(c(0,0), fit.vmsin.20, type = "post-pred")
d_fitted(c(0,0), fit.vmsin.20, type = "point-est")

r_fitted(10, fit.vmsin.20, type = "post-pred")
r_fitted(10, fit.vmsin.20, type = "point-est")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vm2_mle.R
\name{vm2_mle}
\alias{vm2_mle}
\title{Maximum likelihood estimation of bivariate von Mises parameters}
\usage{
vm2_mle(data, model = c("vmsin", "vmcos"), ...)
}
\arguments{
\item{data}{data matrix (if bivarate, in which case it must have two columns) or vector. If outside, the values
are transformed into the scale \eqn{[0, 2\pi)}. *Note:* BAMBI cannot handle missing data. Missing values must
either be removed or properly imputed.}

\item{model}{Bivariate von Mises model. One of "vmsin", "vmcos" or "indep".}

\item{...}{Additional arguments. See details.}
}
\value{
An object of class \link{mle-class}.
}
\description{
Maximum likelihood estimation of bivariate von Mises parameters
}
\details{
The parameters \code{kappa1} and \code{kappa2} are optimized
in log scales. The method of optimization used (passed to \link{optim})
can be specified through \code{method} in \code{...}
(defaults to \code{"L-BFGS-B"}). Note, however, that
lower (0)  and upper (2*pi) bounds for \code{mu1} and \code{mu2}
are specified; so not all methods implemented in \link{optim} will work.
}
\examples{
pars <- list(kappa1 = 3, kappa2 = 2, kappa3 = 1.5, mu1 = 0.5, mu2 = 1.5)
nsamp <- 2000
model <- "vmsin"
set.seed(100)
dat_gen <- do.call(paste0("r", model), c(list(n = nsamp), pars))

est <- vm2_mle(dat_gen, model = model)
library(stats4)
coef(est)
vcov(est)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_diagnostics.R
\name{plot.angmcmc}
\alias{plot.angmcmc}
\title{Summary plots for angmcmc objects}
\usage{
\method{plot}{angmcmc}(
  x,
  par.name,
  comp.label,
  chain.no,
  do.paramtrace = TRUE,
  do.lpdtrace = TRUE,
  use.llik = FALSE,
  ...
)
}
\arguments{
\item{x}{angmcmc object}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{do.paramtrace}{logical. Should the trace(s) for the
parameter(s) be plotted?}

\item{do.lpdtrace}{logical. Should the log posterior trace
be plotted?}

\item{use.llik}{logical. Should the log likelihood be plotted
instead? Ignored if \code{do.lpdtrace == FALSE}.}

\item{...}{unused}
}
\description{
Summary plots for angmcmc objects
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
plot(fit.vmsin.20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BAMBI-package.R
\name{BAMBI}
\alias{BAMBI}
\title{\code{BAMBI}: An R package for Bivariate Angular Mixture Models}
\description{
\code{BAMBI} is an R package that provides functions for fitting
(using Bayesian methods) and simulating mixtures of univariate
and bivariate angular distributions. Please see the reference for a
detailed description of the functionalities of \code{BAMBI}.
}
\references{
Chakraborty, S., & Wong, S. W. (2021). BAMBI: An R package for
fitting bivariate angular mixture models. \emph{Journal of Statistical Software},
99 (11), 1-69. \doi{10.18637/jss.v099.i11}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tim8}
\alias{tim8}
\title{Backbone Dihedral Angles of Triose Phosphate Isomerase (8TIM)}
\format{
A data frame with 490 rows and 2 variables (backbone dihedral angles) phi and psi.
}
\source{
8TIM PDB file: \url{http://www.rcsb.org/pdb/explore.do?structureId=8tim}.

DSSP software: \url{https://swift.cmbi.umcn.nl/gv/dssp/}.
}
\usage{
data(tim8)
}
\description{
A dataset consisting of 490 pairs of backbone dihedral angles (in radian scale \eqn{[0, 2\pi)} )
\eqn{(\phi, \psi)} for the protein Triose Phosphate Isomerase (8TIM). The angles were obtained first by using
the DSSP software on the PDB file for 8TIM to get the backbone angles (in degrees),
and then by converting all angles into radians. Due to the presence of different secondary structures
(helices, sheets and loops) in the protein, the angular data show considerable variability, and is multimodal
with noticeably distinct clusters.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_vmsin_fns.R
\name{rvmsin}
\alias{rvmsin}
\alias{dvmsin}
\title{The bivariate von Mises sine model}
\usage{
rvmsin(
  n,
  kappa1 = 1,
  kappa2 = 1,
  kappa3 = 0,
  mu1 = 0,
  mu2 = 0,
  method = "naive"
)

dvmsin(x, kappa1 = 1, kappa2 = 1, kappa3 = 0, mu1 = 0, mu2 = 0, log = FALSE)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa1, kappa2, kappa3}{vectors of concentration parameters; \code{kappa1, kappa2 > 0}.}

\item{mu1, mu2}{vectors of mean parameters.}

\item{method}{Rejection sampling method to be used. Available choices are \code{"naive"} (default) or \code{"vmprop"}. See details.}

\item{x}{bivariate vector  or a two-column matrix with each row being a bivariate vector of angles
(in radians)  where the densities are to be evaluated.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dvmsin} gives the density  and \code{rvmsin} generates random deviates.
}
\description{
The bivariate von Mises sine model
}
\details{
The bivariate von Mises sine model density at the point \eqn{x = (x_1, x_2)} is given by
\deqn{f(x) = C_s (\kappa_1, \kappa_2, \kappa_3) \exp(\kappa_1 \cos(T_1) + \kappa_2 \cos(T_2) + \kappa_3 \sin(T_1) \sin(T_2))}
where
\deqn{T_1 = x_1 - \mu_1;  T_2 = x_2 - \mu_2}
and \eqn{C_s (\kappa_1, \kappa_2, \kappa_3)} denotes the normalizing constant for the sine model.

Two different rejection sampling methods are implemented for random generation. If \code{method = "vmprop"}, then first the y-marginal
is drawn from the associated marginal density, and then x is generated from the conditional distributio of x given y. The marginal generation of
y is implemented in a rejection sampling scheme with proposal being either von Mises (if the target marginal density is unimodal), or a mixture of
von Mises (if bimodal), with optimally chosen concentration. This the method suggested in Mardia et al. (2007). On the other hand, when
\code{method = "naive"} (default) a (naive) bivariate rejection sampling scheme with (bivariate) uniform propsoal is used.

Note that although method = \code{"vmprop"} may provide better efficiency when the density is highly concentrated, it does have
an (often substantial) overhead due to the optimziation step required to find a reasonable proposal concentration parameter.
This can compensate the efficiency gains of this method, especially when \code{n} is not large.
}
\examples{
kappa1 <- c(1, 2, 3)
kappa2 <- c(1, 6, 5)
kappa3 <- c(0, 1, 2)
mu1 <- c(1, 2, 5)
mu2 <- c(0, 1, 3)
x <- diag(2, 2)
n <- 10

# when x is a bivariate vector and parameters are all scalars,
# dvmsin returns single density
dvmsin(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when x is a two column matrix and parameters are all scalars,
# dmvsin returns a vector of densities calculated at the rows of
# x with the same parameters
dvmsin(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# if x is a bivariate vector and at least one of the parameters is
# a vector, all parameters are recycled to the same length, and
# dvmsin returns a vector of with ith element being the density
# evaluated at x with parameter values kappa1[i], kappa2[i],
# kappa3[i], mu1[i] and mu2[i]
dvmsin(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)

# if x is a two column matrix and at least one of the parameters is
# a vector, rows of x and the parameters are recycled to the same
# length, and dvmsin returns a vector of with ith element being the
# density evaluated at ith row of x with parameter values kappa1[i],
# kappa2[i], # kappa3[i], mu1[i] and mu2[i]
dvmsin(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)

# when parameters are all scalars, number of observations generated
# by rvmsin is n
rvmsin(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when at least one of the parameters is a vector, all parameters are
# recycled to the same length, n is ignored, and the number of
# observations generated by rvmsin is the same as the length of the
# recycled vectors
rvmsin(n, kappa1, kappa2, kappa3, mu1, mu2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{quantile.angmcmc}
\alias{quantile.angmcmc}
\title{Quantile estimates for parameters from an angmcmc object}
\usage{
\method{quantile}{angmcmc}(x, par.name, comp.label, chain.no, probs = seq(0, 1, 0.25), ...)
}
\arguments{
\item{x}{angmcmc object}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{probs}{numeric vector of probabilities with values in
    \eqn{[0,1]}.  (Values up to \samp{2e-14} outside that
    range are accepted and moved to the nearby endpoint.)}

\item{...}{further arguments to pass to \code{quantile}.  In particular, \code{probs = seq(0, 1, 0.25)}
is the default vector of quantiles computed for each parameter.}
}
\value{
Returns a three dimensional array of quantiles, or a matrix (vector) of quantiles
 if one (or two) among \code{par.name},  \code{comp.label}, \code{probs} has length 1.
}
\description{
Quantile estimates for parameters from an angmcmc object
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# 0.025th quantiles
(quant_025 <- quantile(fit.vmsin.20, prob = 0.025))
# 0.975th quantiles
(quant_975 <- quantile(fit.vmsin.20, prob = 0.975))
# default quantiles
(quant_def <- quantile(fit.vmsin.20))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{add_burnin_thin}
\alias{add_burnin_thin}
\title{Add (extra) burnin and thin to angmcmc object after original run}
\usage{
add_burnin_thin(object, burnin.prop = 0, thin = 1)
}
\arguments{
\item{object}{angmcmc object}

\item{burnin.prop}{proportion of iterations to used for burnin. Must be a be a number in [0, 1].
Default is 0.5.}

\item{thin}{thining size to be used. Must be a positive integer. If \code{thin = } n, then every nth iteration is reatained
in the final MCMC sample.}
}
\description{
Add (extra) burnin and thin to angmcmc object after original run
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
lpdtrace(fit.vmsin.20)
# Now add extra burn-in
fit.vmsin.20.burn <- add_burnin_thin(fit.vmsin.20, 0.3)
lpdtrace(fit.vmsin.20.burn)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_vm_fns.R
\name{rvm}
\alias{rvm}
\alias{dvm}
\title{The univariate von Mises distribution}
\usage{
rvm(n, kappa = 1, mu = 0)

dvm(x, kappa = 1, mu = 0, log = FALSE)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa}{vector of concentration (inverse-variance) parameters; \code{kappa} > 0.}

\item{mu}{vector of means.}

\item{x}{vector of angles (in radians) where the densities are to be evaluated.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dvm} gives the density  and \code{rvm} generates random deviates.
}
\description{
The univariate von Mises distribution
}
\details{
If \code{mu} and \code{kappa} are not specified they assume the default values of \code{0} and \code{1} respectively.

The univariate von Mises distribution has density
\deqn{f(x) = 1/(2\pi I_0 (\kappa)) \exp(\kappa \cos(x - mu))}
where \eqn{I_0 (\kappa)} denotes the modified Bessel function of the first kind with order 0 evaluated at the point \eqn{\kappa}.
}
\examples{

kappa <- 1:3
mu <- 0:2
x <- 1:10
n <- 10


# when x and both parameters are scalars, dvm returns a single density
dvm(x[1], kappa[1], mu[1])

# when x is a vector but both the parameters are scalars, dmv returns a vector of
# densities calculated at each entry of x with the same parameters
dvm(x, kappa[1], mu[1])

# if x is scalar and at least one of the two paraemters is a vector, both parameters are
# recycled to the same length, and dvm returns a vector of with ith element being the
# density evaluated at x with parameter values kappa[i] and mu[i]
dvm(x[1], kappa, mu)

# if x and at least one of the two paraemters is a vector, x and the two parameters are
# recycled to the same length, and dvm returns a vector of with ith element being the
# density at ith element of the (recycled) x with parameter values kappa[i] and mu[i]
dvm(x, kappa, mu)

# when parameters are all scalars, number of observations generated by rvm is n
rvm(n, kappa[1], mu[1])

# when at least one of the two parameters is a vector, both are recycled to the same length,
# n is ignored, and the number of observations generated by rvm is the same as the length of
# the recycled vectors
rvm(n, kappa, mu)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BAMBI-package.R
\name{is.angmcmc}
\alias{is.angmcmc}
\alias{angmcmc}
\title{Angular MCMC (\code{angmcmc}) Object}
\usage{
is.angmcmc(object)

angmcmc(...)
}
\arguments{
\item{object}{any R object}

\item{...}{arguments required to make an angmcmc object. See details}
}
\value{
logical. Is the input an angmcmc object?
}
\description{
Checking for and creating an angmcmc object
}
\details{
\code{angmcmc} objects are classified lists that are created when any of the five mixture model fitting
functions, viz., \code{fit_vmmix}, \code{fit_wnormmix}, \code{fit_vmsinmix}, \code{fit_vmcosmix} and
\code{fit_wnorm2mix} is used. An \code{angmcmc} object contains a number of elements, including the dataset, the
model being fitted on the dataset and dimension of the model (univariate or bivariate), the tuning parameters
used, MCMC samples for the mixture model parameters, the (hidden) component or cluster indicators for  data
points in each iteration and the (iteration-wise) log likelihood and log posterior density values (both calculated
upto some normalizing constants). When printed, an angmcmc object returns a brief summary of the function
arguments used to produce the object and the average acceptance rate of the proposals (in HMC and RWMH) used
over iterations. An \code{angmcmc} object can be used as an argument for the diagnostic and post-processing
functions available in \code{BAMBI} for making further inferences.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
is.angmcmc(fit.vmsin.20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_wnorm_fns.R
\name{rwnormmix}
\alias{rwnormmix}
\alias{dwnormmix}
\title{The univariate Wrapped Normal mixtures}
\usage{
rwnormmix(n = 1, kappa, mu, pmix)

dwnormmix(x, kappa, mu, pmix, int.displ = 3, log = FALSE)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa}{vector of component concentration (inverse-variance) parameters, \code{kappa > 0}.}

\item{mu}{vector of component means.}

\item{pmix}{vector of mixing proportions.}

\item{x}{vector of angles (in radians) where the densities are to be evaluated.}

\item{int.displ}{integer displacement. If \code{int.displ =} M, then the infinite sum in the
density is approximated by a sum over 2*M + 1 elements. (See Details.) The allowed values are 1, 2, 3, 4 and 5. Default is 3.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dwnormmix} computes the density  and \code{rwnormmix} generates random deviates from the mixture density.
}
\description{
The univariate Wrapped Normal mixtures
}
\details{
\code{pmix}, \code{mu} and \code{kappa} must be of the same length, with \eqn{j}-th element corresponding to the \eqn{j}-th component of the mixture distribution.

The univariate wrapped normal mixture distribution with component size \code{K = \link{length}(pmix)} has density
\deqn{g(x) = p[1] * f(x; \kappa[1], \mu[1]) + ... + p[K] * f(x; \kappa[K], \mu[K])}
where \eqn{p[j], \kappa[j], \mu[j]} respectively denote the mixing proportion, concentration parameter and the mean parameter for the \eqn{j}-th component
and \eqn{f(. ; \kappa, \mu)} denotes the density function of the (univariate) wrapped normal distribution with mean parameter \eqn{\mu} and concentration parameter \eqn{\kappa}.
}
\examples{
kappa <- 1:3
mu <- 0:2
pmix <- c(0.3, 0.3, 0.4)
x <- 1:10
n <- 10

# mixture densities calculated at each point in x
dwnormmix(x, kappa, mu, pmix)

# number of observations generated from the mixture distribution is n
rwnormmix(n, kappa, mu, pmix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/circ_cor.R
\name{circ_varcor_model}
\alias{circ_varcor_model}
\title{Analytic circular variances and correlations for bivariate angular models}
\usage{
circ_varcor_model(
  model = "vmsin",
  kappa1 = 1,
  kappa2 = 1,
  kappa3 = 0,
  mu1 = 0,
  mu2 = 0,
  nsim = 10000,
  ...
)
}
\arguments{
\item{model}{bivariate angular model. Must be one of \code{"vmsin"},
\code{"vmcos"}, or \code{"wnorm2"}.}

\item{kappa1, kappa2, kappa3}{concentration and covariance parameters.
Recycled to the same size. kappa3^2 must be < kappa1*kappa2 in the wnorm2 model
(see \link{rwnorm2} for a detailed parameterization of \code{wnorm2}).}

\item{mu1, mu2}{mean parameters. Ignored as they do not play any role in
the analytical formulas.}

\item{nsim}{Monte Carlo sample size. Ignored if all of \code{kappa1}, \code{kappa2}
and \code{abs(kappa3)} are < 150 or if model = \code{"wnorm2"}.}

\item{...}{additional model specific argment}
}
\value{
Returns a list with elements \code{var1}, \code{var2} (circular variances for the
first and second coordinates), \code{rho_fl} and \code{rho_js} (circular correlations).
See details.
}
\description{
Analytic circular variances and correlations for bivariate angular models
}
\details{
The function computes the analytic circular variances and correlations
(both Jammalamadaka-Sarma (JS) and Fisher-Lee (FL) forms) for von Mises sine,
von Mises cosine and bivariate wrapped normal distributions.

For \code{wnorm2}, expressions for the circular variances,
JS and FL correlation coefficients can be found in Mardia and Jupp (2009),
Jammalamadaka and Sarma (1988) and Fisher and Lee (1983) respectively.
For \code{vmsin} and \code{vmcos} these expressions are provided in Chakraborty and Wong (2018).

Because the analytic expressions in \code{vmsin} and \code{vmcos} models involve infinite sums
of product of Bessel functions,
if any of \code{kappa1}, \code{kappa2} and \code{abs(kappa3)} is larger
than or equal to 150, IID Monte Carlo with sample size \code{nsim} is used
to approximate \code{rho_js} for numerical stability.  From \code{rho_js},
\code{rho_fl} is computed using Corollary 2.2 in
Chakraborty and Wong (2018), which makes cost-complexity for
the \code{rho_fl} evaluation to be of order  O(\code{nsim}) for \code{vmsin}
and \code{vmcos} models. (In general,  \code{rho_fl} evaluation
is of order O(\code{nsim}^2)).

In addition, for the \code{vmcos} model, when \code{-150 < kappa3 < -1}
or \code{50 < max(kappa1, kappa2, abs(kappa3)) <= 150}, the analytic formulas
in Chakraborty and Wong (2018) are used; however, the reciprocal of the normalizing
constant and its partial derivatives are all calculated numerically via (quasi) Monte carlo method for
numerical stability. These (quasi) random numbers can be provided through the
argument \code{qrnd}, which must be a two column matrix, with each element being
a  (quasi) random number between 0 and 1. Alternatively, if \code{n_qrnd} is
provided (and \code{qrnd} is missing), a two dimensional sobol sequence of size \code{n_qrnd} is
generated via the function \link[qrng]{sobol} from the R package \code{qrng}. If none of \code{qrnd}
or \code{n_qrnd} is available, a two dimensional sobol sequence of size 1e4 is used.
}
\examples{
circ_varcor_model("vmsin", kappa1= 1, kappa2 = 2, kappa3 = 3)

# Monte Carlo approximation
set.seed(1)
dat <- rvmsin(1000, 1, 2, 3)
# sample circular variance
circ_var <- function(x)
  1 - mean(cos(x - atan2(mean(sin(x)), mean(cos(x))) ))
circ_var(dat[, 1])
circ_var(dat[, 2])
circ_cor(dat, "fl")
circ_cor(dat, "js")

}
\references{
Fisher, N. I. and Lee, A. (1983). A correlation coefficient for circular data. Biometrika, 70(2):327-332.

Jammalamadaka, S. R. and Sarma, Y. (1988). A correlation coefficient for
angular variables. Statistical theory and data analysis II, pages 349-364.

Mardia, K. and Jupp, P. (2009). Directional Statistics. Wiley Series in Probability and Statistics. Wiley.

Chakraborty, S. and Wong, S, W.K. (2018). On the circular correlation coefficients
for bivariate von Mises distributions on a torus. arXiv e-print.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{extractsamples}
\alias{extractsamples}
\title{Extract MCMC samples for parameters from an angmcmc object}
\usage{
extractsamples(object, par.name, comp.label, chain.no, drop = TRUE, ...)
}
\arguments{
\item{object}{angular MCMC object}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{drop}{logical. Should the dimension of the output be dropped, if \code{par.name},
\code{comp.label} or \code{chain.no} has a single level?}

\item{...}{additional arguments to be passed to the function.}
}
\value{
Returns  a four dimensional array with

dimension 1 - model parameters and mixing proportions
dimention 2 - components
dimension 3 - MCMC iterations
dimension 4 - chain number
}
\description{
Extract MCMC samples for parameters from an angmcmc object
}
\details{
The default for both \code{par.name} and \code{comp.label} are the all possible choices
available in \code{object}.
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# extract Markov chain realizations for kappa1 from component 1
extr_kappa1_1 <- extractsamples(fit.vmsin.20, "kappa1", 1)
# for kappa1 from component from all components
extr_kappa1 <- extractsamples(fit.vmsin.20, "kappa1")
# for all parameters in component 1
extr_1 <- extractsamples(fit.vmsin.20, comp.label = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_wnorm2_fns.R
\name{fit_wnorm2mix}
\alias{fit_wnorm2mix}
\title{Fitting bivariate wrapped normal model mixtures using MCMC}
\usage{
fit_wnorm2mix(...)
}
\arguments{
\item{...}{arguments (other than \code{model}) passed to \link{fit_angmix}}
}
\description{
Fitting bivariate wrapped normal model mixtures using MCMC
}
\details{
Wrapper for \link{fit_angmix} with \code{model = "wnorm2"}.
}
\examples{
# illustration only - more iterations needed for convergence
fit.wnorm2.10 <- fit_wnorm2mix(tim8, ncomp = 3, n.iter =  10,
                             n.chains = 1)
fit.wnorm2.10

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_vmsin_fns.R
\name{rvmsinmix}
\alias{rvmsinmix}
\alias{dvmsinmix}
\title{The bivariate von Mises sine model mixtures}
\usage{
rvmsinmix(n, kappa1, kappa2, kappa3, mu1, mu2, pmix, method = "naive")

dvmsinmix(x, kappa1, kappa2, kappa3, mu1, mu2, pmix, log = FALSE)
}
\arguments{
\item{n}{number of observations.}

\item{kappa1, kappa2, kappa3}{vectors of concentration parameters; \code{kappa1, kappa2 > 0} for each component.}

\item{mu1, mu2}{vectors of mean parameters.}

\item{pmix}{vector of mixture proportions.}

\item{method}{Rejection sampling method to be used. Available choices are \code{"naive"} (default) or \code{"vmprop"}. See details.}

\item{x}{matrix of angles (in radians) where the density is to be evaluated, with each row being a
single bivariate vector of angles.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dvmsinmix} computes the density (vector if x is a two column matrix with more than one row)
and \code{rvmsinmix} generates random deviates from the mixture density.
}
\description{
The bivariate von Mises sine model mixtures
}
\details{
All the argument vectors \code{pmix, kappa1, kappa2, kappa3, mu1} and \code{mu2} must be of
the same length ( = component size of the mixture model), with \eqn{j}-th element corresponding to the
\eqn{j}-th component of the mixture distribution.

The bivariate von Mises sine model mixture distribution with component size \code{K = \link{length}(p.mix)} has density
\deqn{g(x) = \sum p[j] * f(x; \kappa_1[j], \kappa_2[j], \kappa_3[j], \mu_1[j], \mu_2[j])}
where the sum extends over \eqn{j}; \eqn{p[j]; \kappa_1[j], \kappa_2[j], \kappa_3[j]}; and \eqn{\mu_1[j], \mu_2[j]} respectively denote the mixing proportion,
the three concentration parameters and the two mean parameter for the \eqn{j}-th component, \eqn{j = 1, ..., K},
and \eqn{f(. ; \kappa_1, \kappa_2, \kappa_3, \mu_1, \mu_2)} denotes the density function of the von Mises sine model
with concentration parameters \eqn{\kappa_1, \kappa_2, \kappa_3} and  mean parameters \eqn{\mu_1, \mu_2}.
}
\examples{
kappa1 <- c(1, 2, 3)
kappa2 <- c(1, 6, 5)
kappa3 <- c(0, 1, 2)
mu1 <- c(1, 2, 5)
mu2 <- c(0, 1, 3)
pmix <- c(0.3, 0.4, 0.3)
x <- diag(2, 2)
n <- 10

# mixture densities calculated at the rows of x
dvmsinmix(x, kappa1, kappa2, kappa3, mu1, mu2, pmix)

# number of observations generated from the mixture distribution is n
rvmsinmix(n, kappa1, kappa2, kappa3, mu1, mu2, pmix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_vm_fns.R
\name{rvmmix}
\alias{rvmmix}
\alias{dvmmix}
\title{The univariate von Mises mixtures}
\usage{
rvmmix(n, kappa, mu, pmix)

dvmmix(x, kappa, mu, pmix, log = FALSE)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa}{vector of component concentration (inverse-variance) parameters, \code{kappa > 0}.}

\item{mu}{vector of component means.}

\item{pmix}{vector of mixing proportions.}

\item{x}{vector of angles (in radians) where the densities are to be evaluated.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dvmmix} computes the density  and \code{rvmmix} generates random deviates from the mixture density.
}
\description{
The univariate von Mises mixtures
}
\details{
\code{pmix}, \code{mu} and \code{kappa} must be of the same length, with \eqn{j}-th element corresponding to the \eqn{j}-th component of the mixture distribution.

The univariate von Mises mixture distribution with component size \code{K = \link{length}(pmix)} has density
\deqn{g(x) = p[1] * f(x; \kappa[1], \mu[1]) + ... + p[K] * f(x; \kappa[K], \mu[K])}
where \eqn{p[j], \kappa[j], \mu[j]} respectively denote the mixing proportion, concentration parameter and the mean parameter for the \eqn{j}-th component
and \eqn{f(. ; \kappa, \mu)} denotes the density function of the (univariate) von Mises distribution with mean parameter \eqn{\mu} and concentration parameter \eqn{\kappa}.
}
\examples{
kappa <- 1:3
mu <- 0:2
pmix <- c(0.3, 0.3, 0.4)
x <- 1:10
n <- 10

# mixture densities calculated at each point in x
dvmmix(x, kappa, mu, pmix)

# number of observations generated from the mixture distribution is n
rvmmix(n, kappa, mu, pmix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_model_select.R
\name{fit_incremental_angmix}
\alias{fit_incremental_angmix}
\title{Stepwise fitting of angular mixture models with incremental component sizes and optimum model selection}
\usage{
fit_incremental_angmix(
  model,
  data,
  crit = "LOOIC",
  start_ncomp = 1,
  max_ncomp = 10,
  L = NULL,
  fn = mean,
  fix_label = NULL,
  form = 2,
  start_par = NULL,
  prev_par = TRUE,
  logml_maxiter = 10000,
  return_all = FALSE,
  save_fits = FALSE,
  save_file = NULL,
  save_dir = "",
  silent = FALSE,
  return_llik_contri = (crit \%in\% c("LOOIC", "WAIC")),
  use_best_chain = TRUE,
  alpha = 0.05,
  bonferroni_alpha = TRUE,
  bonferroni_adj_type = "decreasing",
  ...
)
}
\arguments{
\item{model}{angular model whose mixtures are to be fitted. Available choices are \code{"vmsin", "vmcos"} and \code{"wnorm2"} for
bivariate data, and \code{"vm"} and \code{"wnorm"} for univariate data.}

\item{data}{data matrix (if bivarate, in which case it must have two columns) or vector. If outside, the values
are transformed into the scale \eqn{[0, 2\pi)}. *Note:* BAMBI cannot handle missing data. Missing values must
either be removed or properly imputed.}

\item{crit}{model selection criteria, one of \code{"LOOIC"}, \code{"WAIC"}, \code{"AIC"}, \code{"BIC"}, \code{"DIC"} or \code{"LOGML"}. Default is
\code{"LOOIC"}.}

\item{start_ncomp}{starting component size. A single component model is fitted if \code{start_ncomp} is equal to one.}

\item{max_ncomp}{maximum number of components allowed in the mixture model.}

\item{L}{HMC tuning parameter (trajectory length) passed to \link{fit_angmix}. Can be a numeric vetor (or scalar), in which case
the same \code{L} is passed to all \link{fit_angmix} calls, or can be a list of length \code{max_ncomp-start_ncomp+1},
so that \code{L_list[[i]]} is passed as the argument \code{L} to \link{fit_angmix} call with \code{ncomp = max_ncomp+i-1}. See
\link{fit_angmix} for more details on \code{L} including its default values. Ignored if \code{method = "rwmh"}.}

\item{fn}{function to evaluate on MCMC samples to estimate parameters.
Defaults to \code{mean}, which computes the estimated posterior means. If \code{fn = max},
then MAP estimate is calculated from the MCMC run. Used only if \code{crit = "DIC"}, and ignored otherwise.}

\item{fix_label}{logical. Should the label switchings on the current fit (only the corresponding "best chain" if \code{use_best_chain = TRUE})
be fixed before computing parameter estimates and model selection criterion? Defaults to \code{TRUE} if \code{perm_sampling} is true in
the \link{fit_angmix} call, or if \code{crit = "DIC"} and \code{form = 1}.}

\item{form}{form of crit to be used. Available choices are 1 and 2. Used only if \code{crit} is \code{"DIC"} and ignored otherwise.}

\item{start_par}{list with elements \code{pmix} (ignored if \code{comp == 1}), together with \code{kappa1, kappa2, mu1} and \code{mu2},
for bivariate models, and \code{kappa} and \code{mu} for univariate models,
all being vectors of length same as \code{ncomp}.
These provides the starting values for the Markov chain; with \eqn{j}-th component of each vector corresponding to the \eqn{j}-th
component of the mixture distribution. If missing, the data is first clustered into \code{ncomp} groups either via k-means (after
projecting onto a unit sphere), or randomly, depending on \code{rand_start},  and then moment estimators for components are used as
the starting points. Note that a very wrong starting point can potentially lead the chain to get stuck at a wrong solution for thousands
of iterations. As such, we recommend using the default option, which is k-means followed by moment estimation.}

\item{prev_par}{logical. Should the MAP estimated parameters from the model with \code{ncomp = K} be used in the model
with \code{ncomp = K+1} as the starting parameters, with the component with largest mixing proportion appearing twice in the
bigger model?}

\item{logml_maxiter}{maximum number of iterations (\code{maxiter}) passed to \link[bridgesampling]{bridge_sampler} for calculating
\code{LOGML}. Ignored if \code{crit} is not \code{LOGML}.}

\item{return_all}{logical. Should all angmcmc objects obtained during step-wise run be returned? \emph{Warning}: depending on the
sizes of \code{n.iter}, \code{start_ncomp}, \code{max_ncomp} and \code{n.chains}, this can be very memory intesive. In such
cases, it is recommended that \code{return_all} be set to \code{FALSE}, and, if required, the intermediate fitted objects be
saved to file by setting \code{save_fits = TRUE}.}

\item{save_fits}{logical. Should the intermediate angmcmc objects obtained during step-wise run be saved
to file using \link{save}? Defaults to TRUE. See \code{save_file} and \code{save_dir}.}

\item{save_file, save_dir}{\code{save_file} is a list of size \code{max_ncomp-start_ncomp+1},
with k-th entry providing the \code{file}
argument used to \link{save} the intermediate angmcmc object with \code{ncomp = k} (titled \code{"fit_angmcmc"}).
If not provided, then k-th element
of \code{save_file[[k]]} is taken to be \code{paste(save_dir, "comp_k", sep="/")}. Both are ignored if
\code{save_fits = FALSE}.}

\item{silent}{logical. Should the current status (such as what is the current component labels, which job is being done etc.)
be printed? Defaults to \code{TRUE}.}

\item{return_llik_contri}{passed to \link{fit_angmix}. By default, set to \code{TRUE} if \code{crit} is either \code{"LOOIC"}
or \code{"WAIC"}, and to \code{FALSE} otherwise.}

\item{use_best_chain}{logical. Should only the "best" chain obtained during each intermediate fit be used during
computation of model selection criterion? Here "best" means the chain
with largest (mean over iterations) log-posterior density. This can be helpful if one of the chains gets stuck at local optima. Defaults to TRUE.}

\item{alpha}{significance level used in the test \eqn{H_{0K}}: expected log predictive density (elpd) for the fitted model with  K components >= elpd for the fitted model
with K + 1 components if \code{crit} is \code{"LOOIC"} or \code{"WAIC"}.
Must be a scalar between 0 and 1. Defaults to 0.05. See Details. Ignored for any other \code{crit}.}

\item{bonferroni_alpha}{logical. Should a Bonferroni correction be made on the test size \code{alpha} to adjust for
multiplicity due to (\code{max_ncomp} - \code{start_ncomp}) possible hypothesis tests? Defaults to TRUE.
Relevant only if \code{crit} is in  \code{c("LOOIC", "WAIC")}, and ignored otherwise. See Details.}

\item{bonferroni_adj_type}{character string. Denoting type of Bonferroni adjustment to make.
Possible choices are \code{"decreasing"} (default) and \code{"equal"}. Ignored if either \code{bonferroni_alpha}
is \code{FALSE}, or \code{crit} is outside \code{c("LOOIC", "WAIC")}. See Details.}

\item{...}{additional arguments passed to \link{fit_angmix}.}
}
\value{
Returns a named list (with class = \code{stepfit}) with the following seven elements:

\code{fit.all} (if \code{return_all = TRUE}) - a list all angmcmc objects created at each component size;

\code{fit.best} - angmcmc object corresponding to the optimum component size;

\code{ncomp.best} - optimum component size (integer);

\code{crit} - which model comparison criterion used (one of \verb{"LOOIC", "WAIC", "AIC", "BIC", "DIC"} or \code{"LOGML"});

\code{crit.all} - all \code{crit} values calculated (for all component sizes);

\code{crit.best} - \code{crit} value for the optimum component size; and

\code{maxllik.all} - maximum (obtained from MCMC iterations) log likelihood for all fitted models

\code{maxllik.best} - maximum log likelihodd for the optimal model; and

\code{check_min} - logical; is the optimum component size less than \code{max_ncomp}?
}
\description{
Stepwise fitting of angular mixture models with incremental component sizes and optimum model selection
}
\details{
The goal is to fit an angular mixture model with an optimally chosen component size K.
To obtain an optimum K, mixture models with incremental component sizes
between \code{start_ncomp} and \code{max_ncomp} are fitted incrementally using \link{fit_angmix},
starting from K = 1.
If the model selection criterion \code{crit} is \code{"LOOIC"} or \code{"WAIC"}, then a test of hypothesis
\eqn{H_{0K}}: expected log predictive density (elpd) for the fitted model with  K components >= elpd for the fitted model
with K + 1 components, is performed at every K >= 1. The test-statistic used for the test is an approximate z-score
based on the normalized estimated elpd difference between the two models obtained from \link[loo]{compare}, which provides
estimated elpd difference along with its standard error estimate. Because the computed standard error of elpd difference
can be overly optimistic when the elpd difference is small (in particular < 4),
a conservative worst-case estimate (equal to twice of the computed standard error)
is used in such cases. To account for multiplicity among the M =
(\code{max_ncomp} - \code{start_ncomp}) possible sequential tests performed,
by default a Bonferroni adjustment to the test level \code{alpha} is made.
Set \verb{bonferroni_alpha = FALSE\} to remove the adjustment. To encourage parsimony in the final model, by default (}bonferroni_adj_type = "decreasing"\verb{) a decreasing sequence of adjusted alphas of the form }alpha * (0.5)^(1:M) / sum((0.5)^(1:M))\verb{is used. Set}bonferroni_adj_type = "equal"\verb{to use equal sequence of adjusted alphas (i.e.,}alpha/M`) instead.

The incremental fitting stops if  \eqn{H_{0K}} cannot be rejected
(at level \code{alpha}) for some K >= 1; this K is then regarded as the optimum number of components.
If \code{crit} is not \code{"LOOIC"} or \code{"WAIC"} then mixture model with the first minimum value of the model selection criterion \code{crit}
is taken as the best model.

Note that in each intermediate fitted model, the total number of components (instead of the number of
"non-empty components") in the model is used to estimate of the true component
size, and then the fitted model is penalized for model complexity (via the model selection criterion used).
This approach of selecting an optimal K follows the perspective "let two component specific parameters
be identical" for overfitting mixtures, and as such the  Dirichlet prior hyper-parameters \code{pmix.alpha}
(passed to \link{fit_angmix}) should be large. See  Fruhwirth-Schnatter (2011) for more deltails.

Note that the stability of \link[bridgesampling]{bridge_sampler} used in marginal likelihood estimation heavily depends on stationarity of the
chains. As such, while using this criterion, we recommending running the chain long enough, and setting \code{fix_label = TRUE}
for optimal performance.
}
\examples{
# illustration only - more iterations needed for convergence
set.seed(1)
fit.vmsin.step.15 <- fit_incremental_angmix("vmsin", tim8, "BIC", start_ncomp = 1,
                                          max_ncomp = 3, n.iter = 15,
                                          n.chains = 1, save_fits=FALSE)
(fit.vmsin.best.15 <- bestmodel(fit.vmsin.step.15))
lattice::densityplot(fit.vmsin.best.15)


}
\references{
Fruhwirth-Schnatter, S.: Label switching under model uncertainty. In: Mengerson, K., Robert, C., Titterington, D. (eds.) Mixtures:
Estimation and Application, pp. 213-239. Wiley, New York (2011).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_model_select.R
\name{bestmodel}
\alias{bestmodel}
\alias{bestcriterion}
\title{Convenience function for extracting angmcmc object, and the value of the model
selection criterion corresponding to the best fitted model in stepwise fits}
\usage{
bestmodel(step_object)

bestcriterion(step_object)
}
\arguments{
\item{step_object}{stepwise fitted object obtained from \link{fit_incremental_angmix}.}
}
\value{
\code{bestmodel} returns an \code{angmcmc} object, and
\code{bestcriterion} returns the  corresponding value of model selection criterion  for the best fitted model in \code{step_object}.
}
\description{
Convenience function for extracting angmcmc object, and the value of the model
selection criterion corresponding to the best fitted model in stepwise fits
}
\details{
These are convenience functions; the best fitted model and the corresponding value of model selection criterion
can also be directly obtained by
extracting the elements \code{"fit.best"} and \code{"crit.best"} from \code{step_object} respectively.
Note that \verb{bestcriterion\} returns: (a) a scalar number (class = }numeric\verb{) if }crit\verb{used in original}fit_incremental_angmix\verb{call is}'AIC'\verb{, }'BIC'\code{or}'DIC'\verb{, (b) an element of class }bridge\verb{from package}bridgesampling\code{if}crit\code{is}LOGML\verb{, (c) an element of class }c("waic", "loo")\code{if}crit = 'WAIC'\verb{, and (d) an element of class }c("psis_loo", "loo")\code{if}crit = "LOOIC"`. See documentations of these model
selection criteria for more details.
}
\examples{
# illustration only - more iterations needed for convergence
set.seed(1)
fit.vmsin.step.15 <- fit_incremental_angmix("vmsin", tim8, start_ncomp = 1,
                                            max_ncomp = 3, n.iter = 15,
                                            n.chains = 1,
                                            crit = "WAIC")
fit.vmsin.best.15 <- bestmodel(fit.vmsin.step.15)
fit.vmsin.best.15

crit.best <- bestcriterion(fit.vmsin.step.15)
crit.best

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{fix_label}
\alias{fix_label}
\title{Fix label switching in angmcmc objects}
\usage{
fix_label(object, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{...}{arguments other than \code{z, K, complete, mcmc, p}
and \code{data} passed to \link[label.switching]{label.switching}. See details.}
}
\value{
Returns a single \code{angmcmc} object or a list of \code{angmcmc} objects (according as whether
the argument \code{method} is a scalar or vector) with label switchings corrected (after burn-in and thin)
according to the resulting permutation from \link[label.switching]{label.switching}.
}
\description{
Fix label switching in angmcmc objects
}
\details{
\code{fix_label} is a wrapper for  \link[label.switching]{label.switching} from
package \code{label.switching} for \code{angmcmc} objects. The arguments
\code{z, K, complete, mcmc, p} and \code{data} are appropriately filled in
from \code{object}. The \code{label.switching} argument \code{method} can
be a scalar or vector; for this wrapper it defaults to \code{"STEPHENS"} if the \code{angmcmc} was
created with permutation sampling (by setting perm_sampling = TRUE in
\link{fit_angmix}), and to \code{"DATA-BASED"} otherwise.
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# now apply fix_label
fit.vmsin.20.fix <- fix_label(fit.vmsin.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_diagnostics.R
\name{lpdtrace}
\alias{lpdtrace}
\title{Trace and autocorrelation plots of log posterior density or log likelihood from an angmcmc object}
\usage{
lpdtrace(
  object,
  chain.no,
  use.llik = FALSE,
  plot.autocor = FALSE,
  lag.max = NULL,
  ...
)
}
\arguments{
\item{object}{angular MCMC object.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{use.llik}{logical. Should log likelihood be plotted instead of log posterior? Set
to \code{FALSE} by default.}

\item{plot.autocor}{logical. Should the autocorrelations be plotted as well?}

\item{lag.max}{maximum lag for autocorrelation.  Passed to \link{acf}. Ignored if
\code{plot.autocor = FALSE}.}

\item{...}{unused}
}
\description{
Trace and autocorrelation plots of log posterior density or log likelihood from an angmcmc object
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# log posterior density trace
lpdtrace(fit.vmsin.20)
# log likelihood trace
lpdtrace(fit.vmsin.20, use.llik = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/circ_cor.R
\name{circ_cor}
\alias{circ_cor}
\title{Sample circular correlation coefficients}
\usage{
circ_cor(
  x,
  type = "js",
  alternative = "two.sided",
  jackknife = FALSE,
  bootse = FALSE,
  n.boot = 100
)
}
\arguments{
\item{x}{two column matrix. NA values are not allowed.}

\item{type}{type of the circular correlation.
Must be one of "fl", "js", "tau1" and "tau2". See details.}

\item{alternative}{one of \code{"two.sided"}, \code{"less"} or \code{"greater"}
(defaults to \code{"two.sided"}).
Hypothesis test is performed only when \code{type} is either \code{"fl"} or \code{"js"},
in which case asymptotic standard error of the estimator is used to construct the test
statistic.}

\item{jackknife}{logical. Compute jackknifed estimate and standard error? Defaults to FALSE.}

\item{bootse}{logical. Compute bootstrap standard error? Defaults to FALSE.}

\item{n.boot}{number of bootstrapped samples to compute bootstrap standard error. Defaults to
100. Ignored if \code{bootse} if FALSE.}
}
\description{
Sample circular correlation coefficients
}
\details{
\code{circ_cor} calculates the (sample) circular correlation between the columns of x.
Two parametric (the Jammalamadaka-Sarma (1988, equation 2.6) form \code{"js"}, and
the Fisher-Lee (1983, Section 3) form \code{"fl"})
and two non-parametric (two versions of Kendall's tau) correlation coefficients are considered.
The first version of Kendall's tau (\code{"tau1"}) is based on equation 2.1 in Fisher and Lee (1982),
whereas the second version (\code{"tau2"}) is computed using equations 6.7-6.8 in Zhan et al (2017).

The cost-complexity for \code{"js"}, \code{"fl"}, \code{"tau2"} and \code{"tau1"} are \eqn{O(n), O(n^2), O(n^2)} and \eqn{O(n^3)}
respectively, where \eqn{n} denotes the number of rows in \code{x}. As such, for large \eqn{n} evaluation of
\code{"tau1"} will be slow.
}
\examples{
# generate data from vmsin model
set.seed(1)
dat <- rvmsin(100, 2,3,-0.8,0,0)

# now calculate circular correlation(s) between the 2 columns of dat
circ_cor(dat, type="js")
circ_cor(dat, type="fl")
circ_cor(dat, type="tau1")
circ_cor(dat, type="tau2")


}
\references{
Fisher, N. I. and Lee, A. J. (1982). Nonparametric measures of angular-angular association. Biometrika, 69(2), 315-321.

Fisher, N. I. and Lee, A. J. (1983). A correlation coefficient for circular data. Biometrika, 70(2):327-332.

Jammalamadaka, S. R. and Sarma, Y. (1988). A correlation coefficient for
angular variables. Statistical theory and data analysis II, pages 349-364.

Zhan, X., Ma, T., Liu, S., & Shimizu, K. (2017). On circular correlation for data on the torus. Statistical Papers, 1-21.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_model_plots.R
\name{contour_model}
\alias{contour_model}
\title{Contourplot for bivariate angular mixture model densities}
\usage{
contour_model(
  model = "vmsin",
  kappa1,
  kappa2,
  kappa3,
  mu1,
  mu2,
  pmix = rep(1/length(kappa1), length(kappa1)),
  xpoints = seq(0, 2 * pi, length.out = 100),
  ypoints = seq(0, 2 * pi, length.out = 100),
  levels,
  nlevels = 20,
  xlab = "x",
  ylab = "y",
  col = "black",
  lty = 1,
  main,
  ...
)
}
\arguments{
\item{model}{bivariate angular model whose mixture is of interest. Must be one of
"vmsin", "vmcos" and "wnorm2".}

\item{kappa1, kappa2, kappa3, mu1, mu2, pmix}{model parameters and mixing
proportions. See the respective mixture model densities (\link{dvmsinmix}, \link{dvmcosmix},
\link{dwnorm2mix}) for more details.}

\item{xpoints}{Points on the first (x-) coordinate where the density is to be evaluated.
Default to seq(0, 2*pi, length.out=100).}

\item{ypoints}{Points on the first (x-) coordinate where the density is to be evaluated.
Default to seq(0, 2*pi, length.out=100).}

\item{levels}{numeric vector of levels at which to draw contour lines;
passed to the \link{contour} function in graphics.}

\item{nlevels}{number of contour levels desired \strong{if} levels is not supplied;
passed to the \link{contour} function in graphics.}

\item{xlab, ylab, col, lty, main}{graphical parameters passed to \link{contour}.}

\item{...}{additional model specific argment}
}
\description{
Contourplot for bivariate angular mixture model densities
}
\examples{
contour_model('vmsin', 1, 1, 1.5, pi, pi)
contour_model('vmcos', 1, 1, 1.5, pi, pi)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{pointest}
\alias{pointest}
\title{Point estimates for parameters from an angmcmc object}
\usage{
pointest(object, fn = mean, par.name, comp.label, chain.no, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{fn}{function, or a single character string specifying its name, to evaluate on MCMC samples to estimate
parameters.  Defaults to \code{mean}, which computes the estimated posterior mean.
Note that if \code{fn = "MODE"} (warning: not \code{"mode"}) or \code{fn = "MAP"}, then the maximum aposteriori estimate (MAP) is
calculated.}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{...}{additional arguments to be passed to the function.}
}
\value{
Returns a matrix of point estimates, or vector of point estimates if \code{length(par.name)==1} or \code{length(comp.label)==1}.
}
\description{
Point estimates for parameters from an angmcmc object
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# estimate parameters by sample mean
(est_mean <- pointest(fit.vmsin.20))
# estimate parameters by sample median
(est_median <- pointest(fit.vmsin.20, fn = median))
# estimate parameters by MAP
(est_median <- pointest(fit.vmsin.20, fn = "MODE"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_wnorm2_fns.R
\name{rwnorm2}
\alias{rwnorm2}
\alias{dwnorm2}
\title{The bivariate Wrapped Normal distribution}
\usage{
rwnorm2(n, kappa1 = 1, kappa2 = 1, kappa3 = 0, mu1 = 0, mu2 = 0, ...)

dwnorm2(
  x,
  kappa1 = 1,
  kappa2 = 1,
  kappa3 = 0,
  mu1 = 0,
  mu2 = 0,
  int.displ,
  log = FALSE
)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa1, kappa2, kappa3}{vectors of concentration parameters; \code{kappa1, kappa2 > 0},
and \code{kappa3^2 < kappa1*kappa2}.}

\item{mu1, mu2}{vectors of mean parameters.}

\item{...}{additional arguments passed to \link[mvtnorm]{rmvnorm} from package \code{mvtnorm}}

\item{x}{bivariate vector  or a two-column matrix with each row being a bivariate vector of angles
(in radians)  where the densities are to be evaluated.}

\item{int.displ}{integer displacement. If \code{int.displ =} M, then each infinite sum in the
density is approximated by a finite sum over 2*M + 1 elements. (See Details.) The allowed values are 1, 2, 3, 4 and 5. Default is 3.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dwnorm2} gives the density  and \code{rwnorm2} generates random deviates.
}
\description{
The bivariate Wrapped Normal distribution
}
\details{
The bivariate wrapped normal density at the point \eqn{x = (x_1, x_2)} is given by,
\deqn{f(x) = \sqrt((\kappa_1 \kappa_2 - (\kappa_3)^2)) / (2\pi) \sum \exp(-1/2 * (\kappa_1 (T_1)^2 + \kappa_2 (T_2)^2 + 2 \kappa_3 (T_1) (T_2)) )}
where
\deqn{T_1 = T_1(x, \mu, \omega) = (x_1 - \mu_1(2\pi\omega_1))}
\deqn{T_2 = T_2(x, \mu, \omega) = (x_2 - \mu_1(2\pi\omega_2))}
the sum extends over all pairs of integers \eqn{\omega = (\omega_1, \omega_2)},
and is approximated by a sum over \eqn{(\omega_1, \omega_2)} in \eqn{\{-M, -M+1, ..., M-1, M \}^2} if \code{int.displ = } \eqn{M}.

Note that above density is essentially the "wrapped" version of a bivariate normal density with mean
\deqn{\mu = (\mu_1, \mu_2)}
and dispersion matrix  \eqn{\Sigma = \Delta^{-1}}, with \eqn{\Delta} being a \eqn{2 \times 2}
symmetric, positive definite matrix
with diagonal entries \eqn{\kappa_1} and \eqn{\kappa_2} and the off-diagonal entries
\eqn{\kappa_3}.
}
\examples{
kappa1 <- c(1, 2, 3)
kappa2 <- c(1, 6, 5)
kappa3 <- c(0, 1, 2)
mu1 <- c(1, 2, 5)
mu2 <- c(0, 1, 3)
x <- diag(2, 2)
n <- 10

# when x is a bivariate vector and parameters are all scalars,
# dwnorm2 returns single density
dwnorm2(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when x is a two column matrix and parameters are all scalars,
# dmvsin returns a vector of densities calculated at the rows of
# x with the same parameters
dwnorm2(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# if x is a bivariate vector and at least one of the parameters is
# a vector, all parameters are recycled to the same length, and
# dwnorm2 returns a vector of with ith element being the density
# evaluated at x with parameter values kappa1[i], kappa2[i],
# kappa3[i], mu1[i] and mu2[i]
dwnorm2(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)

# if x is a two column matrix and at least one of the parameters is
# a vector, rows of x and the parameters are recycled to the same
# length, and dwnorm2 returns a vector of with ith element being the
# density evaluated at ith row of x with parameter values kappa1[i],
# kappa2[i], # kappa3[i], mu1[i] and mu2[i]
dwnorm2(x, kappa1, kappa2, kappa3, mu1, mu2)

# when parameters are all scalars, number of observations generated
# by rwnorm2 is n
rwnorm2(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when at least one of the parameters is a vector, all parameters are
# recycled to the same length, n is ignored, and the number of
# observations generated by rwnorm2 is the same as the length of the
# recycled vectors
rwnorm2(n, kappa1, kappa2, kappa3, mu1, mu2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_diagnostics.R
\name{contour.angmcmc}
\alias{contour.angmcmc}
\title{Contour plot for angmcmc objects with bivariate data}
\usage{
\method{contour}{angmcmc}(
  x,
  fn = "MAP",
  type = "point-est",
  show.data = TRUE,
  xpoints = seq(0, 2 * pi, length.out = 100),
  ypoints = seq(0, 2 * pi, length.out = 100),
  levels,
  nlevels = 20,
  cex = 1,
  col = "red",
  alpha = 0.4,
  pch = 19,
  ...
)
}
\arguments{
\item{x}{angular MCMC object (with bivariate data).}

\item{fn}{function, or a single character string specifying its name, to evaluate on MCMC samples to estimate
parameters.  Defaults to \code{mean}, which computes the estimated posterior mean.
Note that if \code{fn = "MODE"} (warning: not \code{"mode"}) or \code{fn = "MAP"}, then the maximum aposteriori estimate (MAP) is
calculated.}

\item{type}{Passed to \link{d_fitted}. Possible choices are "point-est" and "post-pred".}

\item{show.data}{logical. Should the data points be added to the contour plot? Ignored if \code{object} is NOT supplied.}

\item{xpoints}{Points on the first (x-) coordinate where the density is to be evaluated.
Default to seq(0, 2*pi, length.out=100).}

\item{ypoints}{Points on the first (x-) coordinate where the density is to be evaluated.
Default to seq(0, 2*pi, length.out=100).}

\item{levels}{numeric vector of levels at which to draw contour lines;
passed to the \link{contour} function in graphics.}

\item{nlevels}{number of contour levels desired \strong{if} levels is not supplied;
passed to the \link{contour} function in graphics.}

\item{cex, col, pch}{graphical parameters passed to \link{points} from graphics for plotting the data points.
Ignored if \code{show.data == FALSE}.}

\item{alpha}{color transparency for the data points, implemented via \link[scales]{alpha} from package \code{scales}.
Ignored if \code{show.data == FALSE}.}

\item{...}{additional arguments to be passed to the function \link{contour}.}
}
\description{
Contour plot for angmcmc objects with bivariate data
}
\details{
\code{contour.angmcmc} is an S3 function for \code{angmcmc} objects that calls \link{contour} from graphics.

To estimate the mixture density required to construct the contour plot, first the parameter vector \eqn{\eta} is estimated
by applying \code{fn} on the MCMC samples, yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}. Then the mixture density
\eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by \eqn{f(x|\hat{\eta})}.
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# now create a contour plot
contour(fit.vmsin.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_diagnostics.R
\name{densityplot.angmcmc}
\alias{densityplot.angmcmc}
\title{Density plots for angmcmc objects}
\usage{
\method{densityplot}{angmcmc}(
  x,
  data = NULL,
  fn = mean,
  type = "point-est",
  log.density = FALSE,
  xpoints = seq(0, 2 * pi, length.out = 35),
  ypoints = seq(0, 2 * pi, length.out = 35),
  plot = TRUE,
  show.hist = ifelse(log.density, FALSE, TRUE),
  xlab,
  ylab,
  zlab = ifelse(log.density, "Log Density", "Density"),
  main,
  ...
)
}
\arguments{
\item{x}{angmcmc object.}

\item{data}{unused. The parameter is already filled with results from fitted angular model. It is kept
to ensure compatibility with the lattice S3 generic \code{densityplot}.}

\item{fn}{function, or a single character string specifying its name, to evaluate on MCMC samples to estimate
parameters.  Defaults to \code{mean}, which computes the estimated posterior mean.
Note that if \code{fn = "MODE"} (warning: not \code{"mode"}) or \code{fn = "MAP"}, then the maximum aposteriori estimate (MAP) is
calculated.}

\item{type}{Passed to \link{d_fitted}. Possible choices are "point-est" and "post-pred".}

\item{log.density}{logical. Should log density be used for the plot?}

\item{xpoints, ypoints}{Points on the  x and y coordinates (if bivariate) or only x coordinate
(if univariate) where the density is to be evaluated. Each defaults to seq(0, 2*pi, length.out=100).}

\item{plot}{logical. Should the density surface (if the fitted data is bivariate) or the density
curve (if univariate) be plotted?}

\item{show.hist}{logical. Should a histogram for the data
points be added to the plot, if the fitted data is univariate? Ignored if data is
bivariate.}

\item{xlab, ylab, zlab, main}{graphical parameters passed to \code{lattice::wireframe} (if
bivariate) or \link{plot} (if univariate). If the data is univariate, \code{zlab} and \code{ylab} can be
used interchangeably (both correspond to the density).}

\item{...}{additional arguments passed to \code{lattice::wireframe} if
fitted data is bivariate, or to \link{hist} (if (\code{show.hist == TRUE})), if the fitted data is univariate}
}
\description{
Plot fitted angular mixture model density surfaces or curves.
}
\details{
When \code{plot==TRUE}, \code{densityplot.angmcmc} calls \code{lattice::wireframe} or
\link{plot} from graphics to draw the surface or curve.

To estimate the mixture density, first the parameter vector \eqn{\eta} is estimated
by applying \code{fn} on the MCMC samples, yielding the (consistent) Bayes estimate \eqn{\hat{\eta}}. Then the mixture density
\eqn{f(x|\eta)} at any point \eqn{x} is (consistently) estimated by \eqn{f(x|\hat{\eta})}.


Note that \code{densityplot.angmcmc} \strong{does not} plot the kernel densitie estimates
of the MCMC parameters. (These plots can be obtained by first converting an \code{angmcmc}
object to an \code{mcmc} object via \link[coda]{as.mcmc.list}, and then
by using \code{densplot} from package coda on the resulting \code{mcmc.list} object. Instead,
\code{densityplot.angmcmc} returns the surface (if 2-D) or the curve (if 1-D)
of the fitted model density evaluated at the estimated parameter vector (obtain through \link{pointest}).
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# now create density surface with the default first 1/3 as burn-in and thin = 1
library(lattice)
densityplot(fit.vmsin.20)
# the viewing angles can be changed through the argument 'screen'
# (passed to lattice::wireframe)
densityplot(fit.vmsin.20, screen = list(z=-30, x=-60))
densityplot(fit.vmsin.20, screen = list(z=30, x=-60))
# the colors can be changed through 'col.regions'
cols <- grDevices::colorRampPalette(c("blue", "green",
                                      "yellow", "orange", "red"))(100)
densityplot(fit.vmsin.20, col.regions = cols)

# Now fit a vm mixture model
# illustration only - more iterations needed for convergence
fit.vm.20 <- fit_vmmix(wind$angle, ncomp = 3, n.iter =  20,
                             n.chains = 1)
densityplot(fit.vm.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_vmcos_fns.R
\name{rvmcos}
\alias{rvmcos}
\alias{dvmcos}
\title{The bivariate von Mises cosine model}
\usage{
rvmcos(
  n,
  kappa1 = 1,
  kappa2 = 1,
  kappa3 = 0,
  mu1 = 0,
  mu2 = 0,
  method = "naive"
)

dvmcos(
  x,
  kappa1 = 1,
  kappa2 = 1,
  kappa3 = 0,
  mu1 = 0,
  mu2 = 0,
  log = FALSE,
  ...
)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa1, kappa2, kappa3}{vectors of concentration parameters; \code{kappa1, kappa2 > 0}.}

\item{mu1, mu2}{vectors of mean parameters.}

\item{method}{Rejection sampling method to be used. Available choices are \code{"naive"} (default) or \code{"vmprop"}. See details.}

\item{x}{bivariate vector  or a two-column matrix with each row being a bivariate vector of angles
(in radians)  where the densities are to be evaluated.}

\item{log}{logical. Should the log density be returned instead?}

\item{...}{additional arguments to be passed to dvmcos. See details.}
}
\value{
\code{dvmcos} gives the density  and \code{rvmcos} generates random deviates.
}
\description{
The bivariate von Mises cosine model
}
\details{
The bivariate von Mises cosine model density at the point \eqn{x = (x_1, x_2)} is given by
\deqn{f(x) = C_c (\kappa_1, \kappa_2, \kappa_3) \exp(\kappa_1 \cos(T_1) + \kappa_2 \cos(T_2) + \kappa_3 \cos(T_1 - T_2))}
where
\deqn{T_1 = x_1 - \mu_1;  T_2 = x_2 - \mu_2}
and \eqn{C_c (\kappa_1, \kappa_2, \kappa_3)} denotes the normalizing constant for the cosine model.

Because \eqn{C_c} involves an infinite  alternating series with product of Bessel functions,
if \code{kappa3 < -5} or \code{max(kappa1, kappa2, abs(kappa3)) > 50}, \eqn{C_c} is evaluated
numerically via (quasi) Monte carlo method for
numerical stability. These (quasi) random numbers can be provided through the
argument \code{qrnd}, which must be a two column matrix, with each element being
a  (quasi) random number between 0 and 1. Alternatively, if \code{n_qrnd} is
provided (and \code{qrnd} is missing), a two dimensional sobol sequence of size \code{n_qrnd} is
generated via the function \link[qrng]{sobol} from the R package \code{qrng}. If none of \code{qrnd}
or \code{n_qrnd} is available, a two dimensional sobol sequence of size 1e4 is used. By default Monte
Carlo approximation is used only if \code{kappa3 < -5} or \code{max(kappa1, kappa2, abs(kappa3)) > 50}.
However, a forced Monte Carlo approximation can be made (irrespective of the choice of \code{kappa1, kappa2} and
\code{kappa3}) by setting \code{force_approx_const = TRUE}. See examples.
}
\examples{
kappa1 <- c(1, 2, 3)
kappa2 <- c(1, 6, 5)
kappa3 <- c(0, 1, 2)
mu1 <- c(1, 2, 5)
mu2 <- c(0, 1, 3)
x <- diag(2, 2)
n <- 10

# when x is a bivariate vector and parameters are all scalars,
# dvmcos returns single density
dvmcos(x[1, ], kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when x is a two column matrix and parameters are all scalars,
# dmvsin returns a vector of densities calculated at the rows of
# x with the same parameters
dvmcos(x, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# if x is a bivariate vector and at least one of the parameters is
# a vector, all parameters are recycled to the same length, and
# dvmcos returns a vector with ith element being the density
# evaluated at x with parameter values kappa1[i], kappa2[i],
# kappa3[i], mu1[i] and mu2[i]
dvmcos(x[1, ], kappa1, kappa2, kappa3, mu1, mu2)

# if x is a two column matrix and at least one of the parameters is
# a vector, rows of x and the parameters are recycled to the same
# length, and dvmcos returns a vector with ith element being the
# density evaluated at ith row of x with parameter values kappa1[i],
# kappa2[i], # kappa3[i], mu1[i] and mu2[i]
dvmcos(x, kappa1, kappa2, kappa3, mu1, mu2)

# when parameters are all scalars, number of observations generated
# by rvmcos is n
rvmcos(n, kappa1[1], kappa2[1], kappa3[1], mu1[1], mu2[1])

# when at least one of the parameters is a vector, all parameters are
# recycled to the same length, n is ignored, and the number of
# observations generated by rvmcos is the same as the length of the
# recycled vectors
rvmcos(n, kappa1, kappa2, kappa3, mu1, mu2)


\donttest{
## Visualizing (quasi) Monte Carlo based approximations of
## the normalizing constant through density evaluations.

# "good" setup, where the analytic formula for C_c can be
# calculated without numerical issues
# kappa1 = 1, kappa2 = 1, kappa3 = -2, mu1 = pi, mu2 = pi

n_qrnd <-  (1:500)*20
# analytic
good.a <- dvmcos(c(3,3), 1, 1, -2, pi, pi, log=TRUE)
# using quasi Monte Carlo
good.q <- sapply(n_qrnd,
                 function(j)
                   dvmcos(c(3,3), 1, 1, -2, pi, pi,
                          log=TRUE, n_qrnd = j,
                          force_approx_const = TRUE))
# using ordinary Monte Carlo
set.seed(1)
good.r <- sapply(n_qrnd,
                 function(j)
                   dvmcos(c(3,3), 1, 1, -2, pi, pi,
                          log=TRUE,
                          qrnd = matrix(runif(2*j), ncol = 2),
                          force_approx_const = TRUE))


plot(n_qrnd, good.q, ylim = range(good.a, good.q, good.r),
     col = "orange", type = "l",
     ylab = "",
     main = "dvmcos(c(3,3), 1, 1, -2, pi, pi, log = TRUE)")
points(n_qrnd, good.r, col = "skyblue", type = "l")
abline(h = good.a, lty = 2, col = "grey")
legend("topright",
       legend = c("Sobol", "Random", "Analytic"),
       col = c("orange", "skyblue", "grey"),
       lty = c(1, 1, 2))


# "bad" setup, where the calculating C_c
# numerically using the analytic formula is problematic
# kappa1 = 100, kappa2 = 100, kappa3 = -200, mu1 = pi, mu2 = pi

n_qrnd <-  (1:500)*20

# using quasi Monte Carlo
bad.q <- sapply(n_qrnd,
                function(j)
                  dvmcos(c(3,3), 100, 100, -200, pi, pi,
                         log=TRUE, n_qrnd = j,
                         force_approx_const = TRUE))
# using ordinary Monte Carlo
set.seed(1)
bad.r <- sapply(n_qrnd,
                function(j)
                  dvmcos(c(3,3), 100, 100, -200, pi, pi,
                         log=TRUE,
                         qrnd = matrix(runif(2*j), ncol = 2),
                         force_approx_const = TRUE))


plot(n_qrnd, bad.q, ylim = range(bad.q, bad.r),
     col = "orange", type = "l",
     ylab = "",
     main = "dvmcos(c(3,3), 100, 100, -200, pi, pi, log = TRUE)")
points(n_qrnd, bad.r, col = "skyblue", type = "l")
legend("topright",
       legend = c("Sobol", "Random"),
       col = c("orange", "skyblue"), lty = 1)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{waic.angmcmc}
\alias{waic.angmcmc}
\title{Watanabe-Akaike Information Criterion (WAIC) for angmcmc objects}
\usage{
\method{waic}{angmcmc}(x, ...)
}
\arguments{
\item{x}{angmcmc object.}

\item{...}{additional model specific arguments to be passed to \link[loo]{waic} from loo. For example, \code{int.displ}
specifies integer displacement in wnorm and wnorm2 models. See \link{fit_wnormmix} and
\link{fit_wnorm2mix} for more details.}
}
\value{
Computes the WAIC for a given angmcmc object.
}
\description{
Watanabe-Akaike Information Criterion (WAIC) for angmcmc objects
}
\details{
Given a deviance function \eqn{D(\eta) = -2 \log(p(y|\eta))}, and an estimate
\eqn{\eta* = (\sum \eta_i) / n} of the posterior mean
\eqn{E(\eta|y)}, where \eqn{y = (y_1, ..., y_n)} denote the data, \eqn{\eta} is the unknown
parameter vector of the model, \eqn{\eta_1, ..., \eta_N} are MCMC samples from the posterior
distribution of \eqn{\eta} given \eqn{y} and \eqn{p(y|\eta)} is the likelihood function,
the Watanabe-Akaike Information Criterion (WAIC) is defined as
\deqn{WAIC = LPPD - p_W}
where
\deqn{LPPD  = \sum_{i=1}^n \log (N^{-1} \sum_{s=1}^N p(y_i|\eta_s) )}
and (form 1 of)
\deqn{p_W =  2 \sum_{i=1}^n [ \log (N^{-1} \sum_{s=1}^N p(y_i|\eta_s) ) - N^{-1} \sum_{s=1}^N \log \:p(y_i|\eta_s) ].}
An alternative form (form 2) for \eqn{p_W} is given by
\deqn{p_W = \sum_{i=1}^n \hat{var} \log p(y_i|\eta)}
where for \eqn{i = 1, ..., n}, \eqn{\hat{var} \log p(y_i|\eta)} denotes the estimated variance
of \eqn{\log p(y_i|\eta)} based on the realizations \eqn{\eta_1, ..., \eta_N}.

Note that waic.angmcmc calls \link[loo]{waic} for computation. If the likelihood contribution of each data
point for each MCMC iteration is available in \code{object} (can be returned by setting \code{return_llik_contri = TRUE})
during \link{fit_angmix} call), \code{waic.array} is used; otherwise \code{waic.function} is
called. Computation is much faster if the likelihood contributions are available - however, they are very
memory intensive, and by default not returned in \link{fit_angmix}.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1, return_llik_contri = TRUE)
library(loo)
waic(fit.vmsin.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{latent_allocation}
\alias{latent_allocation}
\title{Finding latent allocation (component indicators) from an angmcmc object}
\usage{
latent_allocation(object, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{...}{passed to \link{pointest} to estimate parameters.}
}
\value{
Returns a vector of length n, where n is the length (if univariate) or
number of rows (if bivariate) of the data used in original fit.
i-th entry of the output vector provides component label for the i-th data point.
}
\description{
Finding latent allocation (component indicators) from an angmcmc object
}
\details{
In order to find the latent component indicators, estimates
of mixing proportions and model parameters are first computed via
pointest. Then, a data point is assigned label j, if the j-th
component gives highest density for that point.
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# now find latent allocation
latent_allocation(fit.vmsin.20)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{loo.angmcmc}
\alias{loo.angmcmc}
\title{Leave-one-out cross-validation (LOO) for angmcmc objects}
\usage{
\method{loo}{angmcmc}(x, ...)
}
\arguments{
\item{x}{angmcmc object.}

\item{...}{additional model specific arguments to be passed to \link[loo]{waic} from loo. For example, \code{int.displ}
specifies integer displacement in wnorm and wnorm2 models. See \link{fit_wnormmix} and
\link{fit_wnorm2mix} for more details.}
}
\description{
Leave-one-out cross-validation (LOO) for angmcmc objects
}
\details{
Note that loo.angmcmc calls \link[loo]{loo} for computation. If the likelihood contribution of each data
point for each MCMC iteration is available in \code{object} (can be returned by setting \code{return_llik_contri = TRUE})
during \link{fit_angmix} call), \code{loo.array} is used; otherwise \code{loo.function} is
called. Computation is much faster if the likelihood contributions are available - however, they are very
memory intensive, and by default not returned in \link{fit_angmix}.
}
\examples{
\donttest{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1, return_llik_contri = TRUE)
library(loo)
loo(fit.vmsin.20)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_wnorm_fns.R
\name{rwnorm}
\alias{rwnorm}
\alias{dwnorm}
\title{The univariate Wrapped Normal distribution}
\usage{
rwnorm(n = 1, kappa = 1, mu = 0)

dwnorm(x, kappa = 1, mu = 0, int.displ, log = FALSE)
}
\arguments{
\item{n}{number of observations. Ignored if at least one of the other parameters have length k > 1, in which
case, all the parameters are recycled to length k to produce k random variates.}

\item{kappa}{vector of concentration (inverse-variance) parameters; \code{kappa} > 0.}

\item{mu}{vector of means.}

\item{x}{vector of angles (in radians) where the densities are to be evaluated.}

\item{int.displ}{integer displacement. If \code{int.displ =} M, then the infinite sum in the
density is approximated by a sum over 2*M + 1 elements. (See Details.) The allowed values are 1, 2, 3, 4 and 5. Default is 3.}

\item{log}{logical. Should the log density be returned instead?}
}
\value{
\code{dwnorm} gives the density  and \code{rwnorm} generates random deviates.
}
\description{
The univariate Wrapped Normal distribution
}
\details{
If \code{mu} and \code{kappa} are not specified they assume the default values of \code{0} and \code{1} respectively.

The univariate wrapped normal distribution has density
\deqn{f(x) = \sqrt(\kappa/(2\pi)) \sum \exp(-\kappa/2 (x - \mu(2\pi\omega))^2)}
where the sum extends over all integers \eqn{\omega},
and is approximated by a sum over \eqn{\omega} in \eqn{\{-M, -M+1, ..., M-1, M \}} if \code{int.displ = } \eqn{M}.
}
\examples{

kappa <- 1:3
mu <- 0:2
x <- 1:10
n <- 10


# when x and both parameters are scalars, dwnorm returns a single density
dwnorm(x[1], kappa[1], mu[1])

# when x is a vector but both the parameters are scalars, dmv returns a vector of
# densities calculated at each entry of x with the same parameters
dwnorm(x, kappa[1], mu[1])

# if x is scalar and at least one of the two paraemters is a vector, both parameters are
# recycled to the same length, and dwnorm returns a vector of with ith element being the
# density evaluated at x with parameter values kappa[i] and mu[i]
dwnorm(x[1], kappa, mu)

# if x and at least one of the two paraemters is a vector, x and the two parameters are
# recycled to the same length, and dwnorm returns a vector of with ith element being the
# density at ith element of the (recycled) x with parameter values kappa[i] and mu[i]
dwnorm(x, kappa, mu)

# when parameters are all scalars, number of observations generated by rwnorm is n
rwnorm(n, kappa[1], mu[1])

# when at least one of the two parameters is a vector, both are recycled to the same length,
# n is ignored, and the number of observations generated by rwnorm is the same as the length
# of the recycled vectors
rwnorm(n, kappa, mu)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{bridge_sampler.angmcmc}
\alias{bridge_sampler.angmcmc}
\title{Log Marginal Likelihood via Bridge Sampling for angmcmc objects}
\usage{
\method{bridge_sampler}{angmcmc}(samples, ..., ave_over_chains = TRUE)
}
\arguments{
\item{samples}{angmcmc object}

\item{...}{additional argument passed to \link[bridgesampling]{bridge_sampler}. Note that default for
the argument \code{method} is \code{"warp3"}, (instead of \code{"normal"} as used in \code{bridgesampling} package)
to account for multi-modality of the posterior density.}

\item{ave_over_chains}{logical. Separately call \link[bridgesampling]{bridge_sampler} on
each chain in the angmcmc object and then take the average? Defaults to \code{TRUE}.
See details.}
}
\description{
Log Marginal Likelihood via Bridge Sampling for angmcmc objects
}
\details{
Marginal likelihood is calculated by first converting the \code{angmcmc} object \code{samples} to an
\code{mcmc.list} object, and then by passing the resulting \code{mcmc.list} object to \link[bridgesampling]{bridge_sampler}.
If variablity across multiple chains (if any) are very different,
then calling \link[bridgesampling]{bridge_sampler} separately for each chain
usually provides more stable results; the final log ML is computed by averaging over
chain specific MLs.
}
\examples{
\donttest{
library(future)
library(parallel)
# plan(multisession, gc = TRUE) # parallelize chains

set.seed(100)
MC.fit <- fit_angmix("vmsin", tim8, ncomp=3, n.iter=5000,
                     n.chains = 3)


library(bridgesampling)
bridge_sampler(MC.fit)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_postprodn_fns.R
\name{summary.angmcmc}
\alias{summary.angmcmc}
\title{Summary statistics for parameters from an angmcmc object}
\usage{
\method{summary}{angmcmc}(object, par.name, comp.label, chain.no, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{...}{additional arguments affecting the summary produced.}
}
\value{
Returns a list with elements \code{estimate, lower, upper, llik} and \code{lpd}. \code{estimate}
is itself a list with three elements: \code{mean}, \code{median} and \code{mode} providing the
sample mean, sample median and (sample) MAP estimates.

Note that \code{summary.angmcmc} has its own print method, providing a table the estimated mean and 95\% credible intervals for each parameter
}
\description{
Summary statistics for parameters from an angmcmc object
}
\details{
Computes (after thinning and discarding burn-in) point estimates with 95\% posterior credible sets for all components and all parameters,
together with the sample averages of log likelihood and log posterior density.
}
\examples{
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
summary(fit.vmsin.20)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/all_diagnostics.R
\name{paramtrace}
\alias{paramtrace}
\title{Trace plot for parameters from an angmcmc object}
\usage{
paramtrace(object, par.name, comp.label, chain.no, ...)
}
\arguments{
\item{object}{angular MCMC object.}

\item{par.name}{vector of names of parameters for which point estimates are to be computed.  If \code{NULL}, results for all parameters are provided.}

\item{comp.label}{vector of component labels (positive integers, e.g., \code{1, 2, ...}) for which point estimates are to be computed.
If \code{NULL}, results for all components are provided.}

\item{chain.no}{vector of chain numbers whose samples are to be be used.
in the estimation. By default all chains are used.}

\item{...}{unused}

\item{par}{parameter for which trace plot is to be created.}
}
\value{
Returns a single plot if a single \code{par} and a single \code{comp.label} is supplied.
Otherwise, a series of plots is produced.
}
\description{
Trace plot for parameters from an angmcmc object
}
\examples{
# first fit a vmsin mixture model
# illustration only - more iterations needed for convergence
fit.vmsin.20 <- fit_vmsinmix(tim8, ncomp = 3, n.iter =  20,
                             n.chains = 1)
# trace plot for kappa1 in component 1
paramtrace(fit.vmsin.20, "kappa1", 1)
# for kappa1 in all components
paramtrace(fit.vmsin.20, "kappa1")
# for all parameters in component 1
paramtrace(fit.vmsin.20, comp.label = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{wind}
\alias{wind}
\title{Saturna Island wind directions}
\format{
A data frame with 239 rows and 2 columns; the column "angle" provides the angular direction (in radian)
and the column day provides the days on which the data points were collected (ranges between 1-10, corresponding to
October 1-10, 2016).
}
\source{
Environment Canada: \url{https://climate.weather.gc.ca/climate_data/data_quality_e.html}.

CBC news on the storm: \url{https://www.cbc.ca/news/canada/british-columbia/storm-bc-1.3795204}.
}
\usage{
data(wind)
}
\description{
A dataset consisting of 239 observations on wind direction in radians (original measurements were
in 10s of degrees), measured at Saturna Island, British Columbia,
Canada during October 1-10, 2016 (obtained from Environment Canada website). There was a severe storm
during October 4-7, which caused significant fluctuations among the wind directions. As a result the
angular data show a clear multimodality.
}
\keyword{datasets}
